/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef treedb_andromeda_avl_offset_string_h_included
#define treedb_andromeda_avl_offset_string_h_included

#include <stdint.h>
#include <string.h>
#include <treedb/avl-template-checkdefs.h>

#define runtime_pack_tree(h) \
	treedb_storage_t_pack_pointer((h)->m_pStorage, h, (h)->m_pTree)

#define runtime_clear(h) AvlOffsetString_clear(h, runtime_pack_tree(h))

#define runtime_insert(h,n) \
	AvlOffsetString_insert(h, runtime_pack_tree(h), n, 0);

#define runtime_pack_node(h, n) \
	treedb_storage_t_pack_pointer((h)->m_pStorage, h, n)
#define runtime_unpack_node(h,n) \
	((AvlOffsetString_node_ptr_t) \
	(treedb_storage_t_unpack_pointer((h)->m_pStorage, h, n)))
#define runtime_remove(h,n) \
	AvlOffsetString_remove(h, runtime_pack_tree(h), n);

#define runtime_count(h) \
	(runtime_unpack_node(h,(h)->m_pTree->m_top)->m_count)
#define runtime_head(h) \
	(runtime_unpack_node(h,(h)->m_pTree->m_head))
#define runtime_top(h) \
	(runtime_unpack_node(h,(h)->m_pTree->m_top))

#define AVL_TMPL_NS(x) AvlOffsetString_##x

#define AVL_TMPL_data_t char
#define AVL_TMPL_item_SUFFIX []

#define AVL_TMPL_key_t const char *
#define AVL_TMPL_input_t const char *

/* for avl-storage.h */
#define AVL_TMPL_data_ref_t char *

#include "andromeda-avl-offset.h"

typedef struct AVL_TMPL_NS(string_storage_t) AVL_TMPL_NS(string_storage_t)
, * AVL_TMPL_NS(string_storage_ptr_t);

#ifdef __cplusplus
struct AVL_TMPL_NS(string_storage_t) : AVL_TMPL_NS(storage_t)
{
	void STDCALL read_lock
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	) {
		(void)runtime;
	}
	void STDCALL write_lock
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	) {
		(void)runtime;
	}
	void STDCALL unlock
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	) {
		(void)runtime;
	}
	AVL_TMPL_NS(node_ptr_t) STDCALL alloc_node
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, AVL_TMPL_input_t d
	) {
		size_t sz = sizeof(AVL_TMPL_NS(node_t)) + strlen(d) + 1;
		runtime->m_pContainer->cur
			= TREEDB_ALIGNED_POINTER(runtime->m_pContainer->cur);
		if(((runtime->m_pContainer->cur + sz) - runtime->m_pContainer->base) > TREEDB_PAGE_SIZE)
			return 0;
		AVL_TMPL_NS(node_ptr_t) ret =
			(AVL_TMPL_NS(node_ptr_t))runtime->m_pContainer->cur;
		runtime->m_pContainer->cur += sz;
		strcpy(ret->m_item, d);
		return ret;
	}
	void STDCALL free_node
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, AVL_TMPL_NS(node_ptr_t) p
	) {
		(void)runtime;
		(void)p;
		//if(p)
		//	free(p);
	}
	void STDCALL free_tree
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, AVL_TMPL_NS(tree_ptr_t) t
	) {
		(void)runtime;
		(void)t;
		// TODO
	}
	void STDCALL free_data
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, AVL_TMPL_data_ref_t p
	) {
		(void)runtime;
		(void)p;
	}
	void * STDCALL unpack_pointer
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, uintptr_t t
	) {
		if(!t)
			return 0;
		void * ret = runtime->m_pContainer->base + t;
		return ret;
	}
	uintptr_t STDCALL pack_pointer
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, const void * t
	) {
		uint16_t u;
		if(!t)
			u = 0;
		else
			u = ((char *)t) - runtime->m_pContainer->base;
		uintptr_t ret = u;
		const void * v
			= unpack_pointer(runtime, ret);
		assert(t == v);
		return ret;
	}
	int STDCALL compare
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, AVL_TMPL_key_t a
	, AVL_TMPL_key_t b
	) {
		(void)runtime;
		return strcmp(a,b);
	}
};

#else /* #ifdef __cplusplus */

typedef struct AVL_TMPL_NS(string_storage_vtbl_t) {
	treedb_storage_t_methods(,,AVL_TMPL_NS(string_storage_t));
} AVL_TMPL_NS(string_storage_vtbl_t), * AVL_TMPL_NS(string_storage_vtbl_ptr_t);

struct AVL_TMPL_NS(string_storage_t) {
	AVL_TMPL_NS(string_storage_vtbl_ptr_t) m_pVtbl;
};

void STDCALL AVL_TMPL_NS(impl_noop_lock)
( AVL_TMPL_NS(string_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
) {
	(void)This;
	(void)runtime;
}

AVL_TMPL_NS(node_ptr_t) STDCALL AVL_TMPL_NS(impl_alloc_node)
( AVL_TMPL_NS(string_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, AVL_TMPL_input_t d
) {
	(void)This;
	size_t sz = sizeof(AVL_TMPL_NS(node_t)) + strlen(d) + 1;
	runtime->m_pContainer->cur
		= TREEDB_ALIGNED_POINTER(runtime->m_pContainer->cur);
	if(((runtime->m_pContainer->cur + sz) - runtime->m_pContainer->base) > TREEDB_PAGE_SIZE)
		return 0;
	AVL_TMPL_NS(node_ptr_t) ret =
		(AVL_TMPL_NS(node_ptr_t))runtime->m_pContainer->cur;
	runtime->m_pContainer->cur += sz;
	strcpy(ret->m_item, d);
	return ret;
}
void STDCALL AVL_TMPL_NS(impl_free_node)
( AVL_TMPL_NS(string_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, AVL_TMPL_NS(node_ptr_t) p
) {
	(void)This;
	(void)runtime;
	(void)p;
	//if(p) free(p);
}
void STDCALL AVL_TMPL_NS(impl_free_tree)
( AVL_TMPL_NS(string_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, AVL_TMPL_NS(tree_ptr_t) p
) {
	(void)This;
	(void)runtime;
	(void)p;
	//if(p) free(p);
}
void STDCALL AVL_TMPL_NS(impl_free_data)
( AVL_TMPL_NS(string_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, AVL_TMPL_data_ref_t p
) {
	(void)This;
	(void)runtime;
	(void)p;
}
void * STDCALL AVL_TMPL_NS(impl_unpack_pointer)
( AVL_TMPL_NS(string_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, uintptr_t t
) {
	(void)This;
	if(!t)
		return 0;
	void * ret = runtime->m_pContainer->base + t;
	return ret;
}
uintptr_t STDCALL AVL_TMPL_NS(impl_pack_pointer)
( AVL_TMPL_NS(string_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, const void * t
) {
	(void)This;
	uintptr_t ret;
	if(!t)
		ret = 0;
	else
		ret = ((char *)t) - runtime->m_pContainer->base;
	const void * v
		= AVL_TMPL_NS(impl_unpack_pointer)(This, runtime, ret);
	assert(t == v);
	return ret;
}
int STDCALL AVL_TMPL_NS(impl_compare)
( AVL_TMPL_NS(string_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, AVL_TMPL_key_t a
, AVL_TMPL_key_t b
) {
	(void)This;
	(void)runtime;
	return strcmp(a,b);
}
#if 0
static void test(AVL_TMPL_NS(string_storage_vtbl_t) * pv)
{
	pv->read_lock = & AVL_TMPL_NS(impl_noop_lock);
	pv->write_lock = & AVL_TMPL_NS(impl_noop_lock);
	pv->unlock = & AVL_TMPL_NS(impl_noop_lock);
	pv->alloc_node = & AVL_TMPL_NS(impl_alloc_node);
	pv->free_node = & AVL_TMPL_NS(impl_free_node);
	pv->free_tree = & AVL_TMPL_NS(impl_free_tree);
	pv->free_data = & AVL_TMPL_NS(impl_free_data);
	pv->unpack_pointer = & AVL_TMPL_NS(impl_unpack_pointer);
	pv->pack_pointer = & AVL_TMPL_NS(impl_pack_pointer);
	pv->compare = & AVL_TMPL_NS(impl_compare);
}
#endif
#define AvlOffsetString_impl_read_lock AvlOffsetString_impl_noop_lock
#define AvlOffsetString_impl_write_lock AvlOffsetString_impl_noop_lock
#define AvlOffsetString_impl_unlock AvlOffsetString_impl_noop_lock
AVL_TMPL_NS(string_storage_vtbl_t) AVL_TMPL_NS(storage_vtbl) = {
	V3C_INVOKE1(treedb_storage_vtbl_t_entries,AVL_TMPL_NS(impl_))
};

#endif /* #ifdef __cplusplus */

#undef AVL_TMPL_NS

#endif /* #ifndef treedb_andromeda_avl_offset_string_h_included */
