/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_andromeda_pointer_int_h_included
#define v3c_andromeda_pointer_int_h_included

#include <stdint.h>
#include <memory.h>
#include <treedb/avl-template-checkdefs.h>

#define AVL_TMPL_NS(x) AvlPointerInt_##x
#define AVL_TMPL_data_t int16_t
#define AVL_TMPL_key_t int16_t
#define AVL_TMPL_input_t int16_t

/* for avl-storage.h */
#define AVL_TMPL_data_ref_t int16_t

#include "andromeda-avl-pointer.h"

typedef struct AVL_TMPL_NS(int_storage_t) AVL_TMPL_NS(int_storage_t)
, * AVL_TMPL_NS(int_storage_ptr_t);

#define runtime_pack_tree(h) \
	treedb_storage_t_pack_pointer((h)->m_pStorage, h, (h)->m_pTree)

#define runtime_clear(h) \
	AvlPointerInt_clear(h, runtime_pack_tree(h))
#define runtime_insert(h, n) \
	AvlPointerInt_insert(h, runtime_pack_tree(h), n, 0);

#define runtime_pack_node(h, n) \
	AvlPointerInt_pack_node_offset \
	(treedb_storage_t_pack_pointer(pack_pointer((h)->m_pStorage, h, n)))
#define runtime_unpack_node(h,n) \
	((AvlPointerInt_node_ptr_t) \
	(treedb_storage_t_unpack_pointer((h)->m_pStorage, h, n)))
#define runtime_count(h) \
	(runtime_unpack_node(h,(h)->m_pTree->m_top)->m_count)
#define runtime_head(h) \
	(runtime_unpack_node(h,(h)->m_pTree->m_head))

#ifdef __cplusplus
struct AVL_TMPL_NS(int_storage_t) : AVL_TMPL_NS(storage_t)
{
	void STDCALL read_lock
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	) {
		(void)runtime;
	}
	void STDCALL write_lock
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	) {
		(void)runtime;
	}
	void STDCALL unlock
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	) {
		(void)runtime;
	}
	AVL_TMPL_NS(node_ptr_t) STDCALL alloc_node
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, AVL_TMPL_input_t d
	) {
		(void)runtime;
		AVL_TMPL_NS(node_ptr_t) ret =
			(AVL_TMPL_NS(node_ptr_t))malloc(sizeof(AVL_TMPL_NS(node_t)));
		if(!ret)
			return 0;
		ret->m_item = d;
		return ret;
	}
	void STDCALL free_node
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, AVL_TMPL_NS(node_ptr_t) p
	) {
		(void)runtime;
		if(p)
			free(p);
	}
	void STDCALL free_tree
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, AVL_TMPL_NS(tree_t) * t
	) {
		(void)runtime;
		(void)t;
		// TODO
	}
	void STDCALL free_data
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, AVL_TMPL_data_ref_t p
	) {
		(void)runtime;
		(void)p;
	}
	void * STDCALL unpack_pointer
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, uintptr_t t
	) {
		(void)runtime;
		return (void *)t;
	}
	uintptr_t STDCALL pack_pointer
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, const void * t
	) {
		(void)runtime;
		return (uintptr_t)t;
	}
	void STDCALL free_data
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, void * p
	) {
		(void)runtime;
		if(p) free(p);
	}
	int STDCALL compare
	( AVL_TMPL_NS(runtime_ptr_t) runtime
	, int16_t a
	, int16_t b
	) {
		(void)runtime;
		return a < b ? -1 : a == b ? 0 : 1;
	}
};

#else /* #ifdef __cplusplus */

typedef struct AVL_TMPL_NS(int_storage_vtbl_t) {
	treedb_storage_t_methods(,,AVL_TMPL_NS(int_storage_t));
} AVL_TMPL_NS(int_storage_vtbl_t), * AVL_TMPL_NS(int_storage_vtbl_ptr_t);

struct AVL_TMPL_NS(int_storage_t) {
	AVL_TMPL_NS(int_storage_vtbl_ptr_t) m_pVtbl;
};

void STDCALL AVL_TMPL_NS(impl_noop_lock)
( AVL_TMPL_NS(int_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
) {
	(void)This;
	(void)runtime;
}
AVL_TMPL_NS(node_ptr_t) STDCALL AVL_TMPL_NS(impl_alloc_node)
( AVL_TMPL_NS(int_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, AVL_TMPL_input_t d
) {
	(void)This;
	(void)runtime;
	AVL_TMPL_NS(node_ptr_t) ret =
		(AVL_TMPL_NS(node_ptr_t))malloc(sizeof(AVL_TMPL_NS(node_t)));
	if(!ret)
		return 0;
	ret->m_item = d;
	return ret;
}
void STDCALL AVL_TMPL_NS(impl_free_node)
( AVL_TMPL_NS(int_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, AVL_TMPL_NS(node_ptr_t) p
) {
	(void)This;
	(void)runtime;
	if(p) free(p);
}
void STDCALL AVL_TMPL_NS(impl_free_tree)
( AVL_TMPL_NS(int_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, AVL_TMPL_NS(tree_ptr_t) p
) {
	(void)This;
	(void)runtime;
	if(p) free(p);
}
void STDCALL AVL_TMPL_NS(impl_free_data)
( AVL_TMPL_NS(int_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, AVL_TMPL_data_ref_t p
) {
	(void)This;
	(void)runtime;
	(void)p;
}
void * STDCALL AVL_TMPL_NS(impl_unpack_pointer)
( AVL_TMPL_NS(int_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, uintptr_t t
) {
	(void)This;
	(void)runtime;
	void * ret = (void *)t;
	/*
	printf("unpack_pointer : %p -> %p\n", ret, (void *)t);
	*/
	return ret;
}
uintptr_t STDCALL AVL_TMPL_NS(impl_pack_pointer)
( AVL_TMPL_NS(int_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, const void * t
) {
	(void)This;
	(void)runtime;
	uintptr_t ret = (uintptr_t)t;
	/*
	printf("pack_pointer : %p -> %p\n", t, (void *)ret);
	*/
	return ret;
}
int STDCALL AVL_TMPL_NS(impl_compare)
( AVL_TMPL_NS(int_storage_ptr_t) This
, AVL_TMPL_NS(runtime_ptr_t) runtime
, int16_t a
, int16_t b
) {
	(void)This;
	(void)runtime;
	return a < b ? -1 : a == b ? 0 : 1;
}
#if 0
static void test(AVL_TMPL_NS(int_storage_vtbl_t) * pv)
{
	pv->read_lock = & AVL_TMPL_NS(impl_noop_lock);
	pv->write_lock = & AVL_TMPL_NS(impl_noop_lock);
	pv->unlock = & AVL_TMPL_NS(impl_noop_lock);
	pv->alloc_node = & AVL_TMPL_NS(impl_alloc_node);
	pv->free_node = & AVL_TMPL_NS(impl_free_node);
	pv->free_tree = & AVL_TMPL_NS(impl_free_tree);
	pv->free_data = & AVL_TMPL_NS(impl_free_data);
	pv->unpack_pointer = & AVL_TMPL_NS(impl_unpack_pointer);
	pv->pack_pointer = & AVL_TMPL_NS(impl_pack_pointer);
	pv->compare = & AVL_TMPL_NS(impl_compare);
}
#endif
#define AvlPointerInt_impl_read_lock AvlPointerInt_impl_noop_lock
#define AvlPointerInt_impl_write_lock AvlPointerInt_impl_noop_lock
#define AvlPointerInt_impl_unlock AvlPointerInt_impl_noop_lock
AVL_TMPL_NS(int_storage_vtbl_t) AVL_TMPL_NS(storage_vtbl) = {
	V3C_INVOKE1(treedb_storage_vtbl_t_entries,AVL_TMPL_NS(impl_))
};

#endif /* #ifdef __cplusplus */

#undef AVL_TMPL_NS

#endif /* #ifndef v3c_andromeda_pointer_int_h_included */
