/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "andromeda-avl-offset-int.h"
#include <string.h>
#include <v3c/utils.h>

#define START_OFFS V3C_ROUND_UP_T2(uint16_t, sizeof(AvlOffsetInt_tree_t), \
	TREEDB_ALIGN)

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	AvlOffsetInt_int_storage_t storage;
#ifndef __cplusplus
	storage.m_pVtbl = & AvlOffsetInt_storage_vtbl;
#endif
	char mem[
		START_OFFS + TREEDB_PAGE_SIZE
	] __attribute__ ((aligned (TREEDB_ALIGN))) = {0};
	// We need a base that can't be used - it's our NULL reference.
	// We also reserve the tree root at the start.
	AvlOffsetInt_container_t container
		= { mem - TREEDB_ALIGN, mem + START_OFFS };
	AvlOffsetInt_tree_t * ptree = (AvlOffsetInt_tree_t *)mem;

	AvlOffsetInt_runtime_t han =
	{ (AvlOffsetInt_storage_ptr_t)(& storage)
	, & container
	, ptree
	, PTHREAD_RWLOCK_INITIALIZER };

	han.m_pStorage = (AvlOffsetInt_storage_ptr_t)(& storage);
	AvlOffsetInt_init(& han, han.m_pTree);

	unsigned int n, count;
	for(n = 0; n < 10u; ++n)
		runtime_insert(& han, n);

	count = runtime_count(& han);
	printf("Inserted %d items.\n", count);

	AvlOffsetInt_node_ptr_t c = runtime_head(& han);

	for(n = 0; c; ++n, c = runtime_unpack_node(&han, c->m_next)) {
		printf("[%u] %d\n", n, c->m_item);
	}
	runtime_clear(& han);

	printf("%zu bytes used, %zu for node, %zu for tree.\n"
		, (size_t)(container.cur - container.base - TREEDB_ALIGN - START_OFFS)
		, sizeof(AvlOffsetInt_node_t)
		, sizeof(AvlOffsetInt_tree_t));

	return count == 10 ? 0 : 1;
}
