/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdlib.h>
#include <iostream>
#include "andromeda-avl-offset-int.h"
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>
#include <v3c/utils.h>

using namespace std;
using namespace v3c;

void dump_tree_node
( TreeStreambuf & sb
, AvlOffsetInt_runtime_ptr_t runtime
, AvlOffsetInt_node_ptr_t node
) {
	if(!node)
		return;
	sb.add(string_inline_printf("Count: %zu, Value : %u"
		, (size_t)node->m_count, node->m_item), true);
	sb.push();
	AvlOffsetInt_node_ptr_t left
		= runtime_unpack_node(runtime, node->m_left);
	AvlOffsetInt_node_ptr_t right
		= runtime_unpack_node(runtime, node->m_right);
	if(left) {
		sb.add("Left", !right);
		sb.push();
		dump_tree_node(sb, runtime, left);
		sb.pop();
	}
	if(right) {
		sb.add("Right", true);
		sb.push();
		dump_tree_node(sb, runtime, right);
		sb.pop();
	}
	sb.pop();
}
void dump_tree(AvlOffsetInt_runtime_ptr_t runtime)
{
	TreeStreambuf sb(cout);
	sb.add("Nodes", false);
	sb.push();
	dump_tree_node(sb, runtime, runtime_unpack_node(runtime, runtime->m_pTree->m_top));
	sb.pop();
}

#define START_OFFS V3C_ROUND_UP_T2(uint16_t, sizeof(AvlOffsetInt_tree_t), \
	TREEDB_ALIGN)

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	AvlOffsetInt_int_storage_t storage;
	char mem[
		START_OFFS + TREEDB_PAGE_SIZE
	] __attribute__ ((aligned (TREEDB_ALIGN))) = {0};
	// We need a base that can't be used - it's our NULL reference.
	// We also reserve the tree root at the start.
	AvlOffsetInt_container_t container
		= { mem - TREEDB_ALIGN, mem + START_OFFS };
	AvlOffsetInt_tree_t * ptree = (AvlOffsetInt_tree_t *)mem;

	AvlOffsetInt_runtime_t han =
	{ (AvlOffsetInt_storage_ptr_t)(& storage)
	, & container
	, ptree
	, PTHREAD_RWLOCK_INITIALIZER };

	AvlOffsetInt_init(& han, han.m_pTree);

	unsigned int n, count;
	for(n = 0; n < 40; ++n)
		runtime_insert(& han, n);

	count = runtime_count(&han);
	printf("Inserted %d items.\n", count);
	int res = (count == 40) ? 0 : 1;

	dump_tree(& han);
	// runtime_clear_tree(& han);
	uint16_t v;
	for(n = 0; n < 40; ++n) {
		v = (uint16_t)rand();
		while(v >= count)
			v >>= 1;
		AvlOffsetInt_node_ptr_t c = runtime_index(& han, v);
		if(!c) {
			printf("get(%d) failed.\n", v);
			return 1;
		}
		if(c->m_item != v) {
			printf("avl_get : wanted %d but returned %d.\n"
				, v, c->m_item);
			return 2;
		}
		printf("got : %d.\n", v);
		AvlOffsetInt_node_ptr_t d = c;
		c = runtime_unpack_node(& han, c->m_next);
		runtime_delete_node(& han, d);
		for(; c; ++v, c = runtime_unpack_node(&han, c->m_next))
			c->m_item = v;
		--count;
		dump_tree(& han);
	}
	printf("%zu bytes used, %zu for node, %zu for tree.\n"
		, (size_t)(container.cur - container.base - TREEDB_ALIGN - START_OFFS)
		, sizeof(AvlOffsetInt_node_t)
		, sizeof(AvlOffsetInt_tree_t));

	return res;
}
