/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "andromeda-avl-offset-string.h"
#include <v3c/utils.h>

const char * items[] =
{ "apple"
, "banana"
, "pear"
, "plum"
, "orange"
, "lemon"
};
const char * sorted_items[] =
{ "apple"
, "banana"
, "lemon"
, "orange"
, "pear"
, "plum"
};

#define START_OFFS V3C_ROUND_UP_T2(uint16_t, sizeof(AvlOffsetString_tree_t), \
	TREEDB_ALIGN)

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	AvlOffsetString_string_storage_t storage;
#ifndef __cplusplus
	storage.m_pVtbl = & AvlOffsetString_storage_vtbl;
#endif
	char mem[
		START_OFFS + TREEDB_PAGE_SIZE
	] __attribute__ ((aligned (TREEDB_ALIGN))) = {0};
	// We need a base that can't be used - it's our NULL reference.
	// We also reserve the tree root at the start.
	AvlOffsetString_container_t container
		= { mem - TREEDB_ALIGN, mem + START_OFFS };
	AvlOffsetString_tree_t * ptree = (AvlOffsetString_tree_t *)mem;

	AvlOffsetString_runtime_t han =
	{ (AvlOffsetString_storage_ptr_t)(& storage)
	, & container
	, ptree
	, PTHREAD_RWLOCK_INITIALIZER };

	AvlOffsetString_init(& han, han.m_pTree);

	unsigned int n;
	for(n = 0; n < sizeof(items)/sizeof(items[0]); ++n)
		runtime_insert(& han, items[n]);

	n = runtime_count(&han);
	int ret = (n != sizeof(items)/sizeof(items[0]));
	printf("Inserted %d items, expected %u.\n", n
		, (unsigned int)(sizeof(items)/sizeof(items[0])));

	AvlOffsetString_node_ptr_t c = runtime_head(&han);

	for(n = 0; c; ++n, c = runtime_unpack_node(&han, c->m_next)) {
		printf("[%u] %s [%s]\n", n, c->m_item, sorted_items[n]);
		if(strcmp(c->m_item, sorted_items[n]))
			ret = 1;
	}
	runtime_clear(& han);

	printf("%zu bytes used, %zu for node, %zu for tree.\n"
		, (size_t)(container.cur - container.base - TREEDB_ALIGN - START_OFFS)
		, sizeof(AvlOffsetString_node_t)
		, sizeof(AvlOffsetString_tree_t));

	return ret;
}
