/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "andromeda-avl-offset-string.h"
#include <iostream>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>
#include <v3c/utils.h>

using namespace std;
using namespace v3c;

const char * items[] =
{ "apple"
, "banana"
, "pear"
, "plum"
, "orange"
, "lemon"
};
const char * sorted_items[] =
{ "apple"
, "banana"
, "lemon"
, "orange"
, "pear"
, "plum"
};

void dump_tree_node
( TreeStreambuf & sb
, AvlOffsetString_runtime_ptr_t runtime
, AvlOffsetString_node_ptr_t node)
{
	if(!node)
		return;
	AvlOffsetString_size_t pp
		= runtime_pack_node(runtime, node);
	uint16_t u = pp/*.node.value*/;
	sb.add( string_inline_printf("<%u> %s", u, node->m_item), true);
	sb.push();
	AvlOffsetString_node_ptr_t left = runtime_unpack_node(runtime, node->m_left);
	AvlOffsetString_node_ptr_t right = runtime_unpack_node(runtime, node->m_left);
	if(left) {
		sb.add("Left", !right);
		sb.push();
		dump_tree_node(sb, runtime, left);
		sb.pop();
	}
	if(right) {
		sb.add("Right", true);
		sb.push();
		dump_tree_node(sb, runtime, right);
		sb.pop();
	}
	sb.pop();
}
void dump_trees(AvlOffsetString_runtime_ptr_t runtime)
{
	TreeStreambuf sb(cout);
	sb.add("Tree", true);
	sb.push();
	dump_tree_node(sb, runtime, runtime_top(runtime));
	sb.pop();
}

#define START_OFFS V3C_ROUND_UP_T2(uint16_t, sizeof(AvlOffsetString_tree_t), \
	TREEDB_ALIGN)

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	AvlOffsetString_string_storage_t storage;
	char mem[
		START_OFFS + TREEDB_PAGE_SIZE
	] __attribute__ ((aligned (TREEDB_ALIGN))) = {0};
	// We need a base that can't be used - it's our NULL offset.
	// We also reserve the tree root at the start.
	AvlOffsetString_container_t container
		= { mem - TREEDB_ALIGN, mem + START_OFFS };
	AvlOffsetString_tree_t * ptree = (AvlOffsetString_tree_t *)mem;

	AvlOffsetString_runtime_t han =
	{ (AvlOffsetString_storage_ptr_t)(& storage)
	, & container
	, ptree
	, PTHREAD_RWLOCK_INITIALIZER };

	AvlOffsetString_init(& han, ptree);

	int ret = 1;
	unsigned int n;
	for(n = 0; n < sizeof(items)/sizeof(items[0]); ++n) {
		runtime_insert(& han, items[n]);
		printf("[%u] inserted %s.\n", n, items[n]);
		dump_trees(&han);
	}
	AvlOffsetString_node_ptr_t pRoot = runtime_top(&han);
	if(pRoot) {
		n = pRoot->m_count;
		printf("Inserted %d items, expected %u.\n", n, (unsigned int)(sizeof(items)/sizeof(items[0])));
		ret = (n != sizeof(items)/sizeof(items[0]));
	}else
		printf("NULL root?\n");
	AvlOffsetString_node_ptr_t c = runtime_head(&han);

	for(n = 0; c; ++n, c = runtime_unpack_node(&han, c->m_next)) {
		printf("[%u] %s [%s]\n", n, c->m_item, sorted_items[n]);
		if(strcmp(c->m_item, sorted_items[n]))
			ret = 1;
	}
	dump_trees(&han);
	for(n = 0; n < sizeof(items)/sizeof(items[0]); ++n) {
		runtime_remove(& han, sorted_items[n]);
		printf("[%u] deleted %s.\n", n, sorted_items[n]);
		dump_trees(&han);
	}

	printf("%u bytes used, %u for node, %u for tree.\n"
		, (unsigned int)(container.cur - container.base - TREEDB_ALIGN)
		, (unsigned int)sizeof(AvlOffsetString_node_t)
		, (unsigned int)sizeof(AvlOffsetString_tree_t));

	return ret;
}
