/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform - passing.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include <v3c/utils.h>
#include "treedb.h"
#include <treedb/ns-types.h>
#define TREEDB_PAYLOAD_FIRST_MEMBER m_dummy
#define TREEDB_PAYLOAD char TREEDB_PAYLOAD_FIRST_MEMBER
#define TREEDB_CHECK_HEAP(runtime) allocator_check_heap()
int allocator_check_heap();

#define AVL_TMPL_omit_pack_node_DECLARE
#include <treedb/storage.h>
#include <treedb/treedb-impl.h>

TREEDB_NS(runtime_ptr_t) runtime = 0;

int heap_check_res = 0;
int allocator_check_heap()
{
	if(heap_check_res)
		return heap_check_res;
	heap_check_res = TREEDB_NS(check_heap)(runtime);
	return heap_check_res;
}

char buf[TREEDB_PAGE_SIZE * 2];
/* Page aligned. */
char * g_buf() { return V3C_ROUND_UP_PTR2(buf,TREEDB_PAGE_SIZE); }

int init(TREEDB_NS(count_t) align, TREEDB_NS(count_t) offs
, TREEDB_NS(count_t) size)
{
	heap_check_res = 0;
	/* Initialise allocator trees. */
	TREEDB_NSF(init)(runtime, TREEDB_ALLOC_free_tree);
	TREEDB_NSU(init)(runtime, TREEDB_ALLOC_used_tree);

	TREEDB_NS(count_t) fno
		= TREEDB_NS_ALIGNED_OFFSET(sizeof(TREEDB_NS(treedb_t)));
	/* Make "used" refer to the adjusted free node offset. */
	fno = V3C_ROUND_UP_T2(TREEDB_NS(count_t), fno + TREEDB_ALLOC_node_size
		, align) + offs - TREEDB_ALLOC_node_size;

	TREEDB_STORAGE_size = TREEDB_PAGE_SIZE;
	/* Manually insert a free node that describes the available storage space in
	the first page. */
	TREEDB_NSA(node_ptr_t) pFirstNode
		= (TREEDB_NSA(node_ptr_t))
			(TREEDB_STORAGE_base + fno);

	TREEDB_ALLOC_node_set_length(pFirstNode, size);
	TREEDB_ALLOC_node_set_prev(pFirstNode, TREEDB_pack_allocator_node(0));
	TREEDB_ALLOC_node_set_next(pFirstNode, TREEDB_pack_allocator_node(0));
	TREEDB_NSF(insert_node)(runtime, TREEDB_ALLOC_free_tree, pFirstNode);
	printf("Init : free node offset %zu, total %zu.\n", (size_t)fno
		, (size_t)fno + TREEDB_ALLOC_node_size + size);
	return fno;
}
/** Test malloc_aligned with one free block that's smaller than the requested
size
*/
int malloc_simple_fail1()
{
	int n, align = 2; /* 4,8 */
	for(n = 0; n < 3; ++n, align <<= 1) {
		init(1, 0, align/2);
		void * p = TREEDB_NS(malloc_aligned)(runtime, align, align, 0);
		if(p) {
			printf("Malloc succeeded! It should have failed.\n");
			return 1;
		}
		if(heap_check_res != 0) {
			printf("Heap check failed.\n");
			return 1;
		}
	}
	return 0;
}
/** Test malloc_aligned with one free block that's equal to the requested size

The block is incorrectly aligned for the request.
*/
int malloc_simple_fail2()
{
	int n, align = sizeof(TREEDB_NS(count_t)) * 2;
	for(n = 0; n < 2; ++n, align <<= 1) {
		init(align, sizeof(TREEDB_NS(count_t)), align);
		void * p = TREEDB_NS(malloc_aligned)(runtime, align, align, 0);
		if(p) {
			printf("Malloc succeeded! It should have failed.\n");
			return 1;
		}
		if(heap_check_res != 0) {
			printf("Heap check failed.\n");
			return 1;
		}
	}
	return 0;
}
/** Test malloc_aligned with one free block that's equal to the requested size

The block is correctly aligned for the request.
*/
int malloc_simple1()
{
	int n, align = 2; /* 4,8,16,32,64 */
	for(n = 0; n < 6; ++n, align <<= 1) {
		init(align, 0, align);
		char * p = (char *)TREEDB_NS(malloc_aligned)(runtime, align, align, 0);
		if(!p) {
			printf("Malloc failed.\n");
			return 1;
		}
		ptrdiff_t diff = p - V3C_ROUND_DOWN_PTR2(p,align);
		if(diff) {
			printf("Returned pointer misaligned by %td bytes.\n", diff);
			return 1;
		}
		if(heap_check_res != 0) {
			printf("Heap check failed.\n");
			return 1;
		}
		if(TREEDB_ALLOCATOR.m_free.m_head/*.node.value*/) {
			printf("Free heap node shouldn't be created.\n");
			return 1;
		}
	}
	return 0;
}
/** Test malloc_aligned with one free block that's greater than the requested
size

The block is correctly aligned for the request, but not large enough to split.
*/
int malloc_greater1()
{
	int n, align = 2; /* 4,8,16,32,64 */
	for(n = 0; n < 6; ++n, align <<= 1) {
		init(align, 0, align + sizeof(TREEDB_NS(count_t)));
		char * p = (char *)TREEDB_NS(malloc_aligned)(runtime, align, align, 0);
		if(!p) {
			printf("Malloc failed.\n");
			return 1;
		}
		ptrdiff_t diff = p - V3C_ROUND_DOWN_PTR2(p,align);
		if(diff) {
			printf("Returned pointer misaligned by %td bytes.\n", diff);
			return 1;
		}
		if(heap_check_res != 0) {
			printf("Heap check failed.\n");
			return 1;
		}
		if(TREEDB_ALLOCATOR.m_free.m_head/*.node.value*/) {
			printf("Free heap node shouldn't be created.\n");
			return 1;
		}
	}
	return 0;
}
/** Test malloc_aligned with one free block that's greater than the requested
size and large enough to split, leaving a zero-length free node
*/
int malloc_greater2()
{
	int n, align = 2; /* 4,8,16,32,64 */
	for(n = 0; n < 6; ++n, align <<= 1) {
		init(align, 0
			, V3C_ROUND_UP_T2(int, align, sizeof(TREEDB_NS(count_t)))
				+ TREEDB_ALLOC_node_size);
		char * p = (char *)TREEDB_NS(malloc_aligned)(runtime, align, align, 0);
		if(!p) {
			printf("Malloc failed.\n");
			return 1;
		}
		ptrdiff_t diff = p - V3C_ROUND_DOWN_PTR2(p,align);
		if(diff) {
			printf("Returned pointer misaligned by %td bytes.\n", diff);
			return 1;
		}
		if(heap_check_res != 0) {
			printf("Heap check failed.\n");
			return 1;
		}
		TREEDB_NSA(node_ptr_t) pfree = TREEDB_ALLOC_free_head;
		if(!pfree) {
			printf("No free heap node.\n");
			return 1;
		}
		size_t fsz = TREEDB_ALLOC_node_length(pfree);
		if(fsz) {
			printf("Free heap node isn't empty (%zu).\n", fsz);
			return 1;
		}
	}
	return 0;
}

typedef struct Entry
{
	int offset, length;
} Entry;

/** Test malloc_aligned with one free block that's greater than the requested
size and large enough to split, leaving a minimal free node

The block is incorrectly aligned for the request, requiring an upper split.

As in \ref malloc_greater2, we need enough space to create a free node of known
size as a confidence factor in our calculations. It can't be zero-length because
of the mis-alignment, but we should know it's length for comparison.

\msc
	a,b,c,d,e,f,g;

	a->g [ label = "total" ] ;
	a->f [ label = "aligned" ] ;
	a->b [ label = "header" ] ,
	b->c [ label = "misalign" ] ,
	c->d [ label = "free node" ] ,
	d->e [ label = "data" ] ,
	e->f [ label = "used node" ] ,
	f->g [ label = "data" ] ;
\endmsc

 - a->b TREEDB_NS_ALIGNED_OFFSET(sizeof(TREEDB_NS(treedb_t)))
 - b->c misalign
 - c->d TREEDB_ALLOC_node_size
 - d->e pad
 - e->f TREEDB_ALLOC_node_size
 - f->g req
*/
int malloc_greater3()
{
	int n, align = sizeof(TREEDB_NS(count_t)) * 2; /* 2x,4x,8x,16x,32x */
	for(n = 0; n < 5; ++n, align <<= 1) {
		int free_data_offset
			= V3C_ROUND_UP_T2(int
				, TREEDB_NS_ALIGNED_OFFSET(sizeof(TREEDB_NS(treedb_t)))
					+ TREEDB_ALLOC_node_size
				, align) + sizeof(TREEDB_NS(count_t));
		int free_node_offset = free_data_offset - TREEDB_ALLOC_node_size;
		int used_data_offset
			= free_data_offset + TREEDB_ALLOC_node_size;
		int aligned_used_data_offset
			= V3C_ROUND_UP_T2(int, used_data_offset, align);

		int pad = aligned_used_data_offset - used_data_offset;

		Entry misalign, freenode, freedata, usednode, useddata;
		misalign.offset = free_node_offset - (int)sizeof(TREEDB_NS(count_t));
		misalign.length = (int)sizeof(TREEDB_NS(count_t));
		freenode.offset = free_node_offset;
		freenode.length = (int)TREEDB_ALLOC_node_size;
		freedata.offset = freenode.offset + freenode.length;
		freedata.length = pad;
		usednode.offset = freedata.offset + freedata.length;
		usednode.length = (int)TREEDB_ALLOC_node_size;
		useddata.offset = usednode.offset + usednode.length;
		useddata.length = align;
		printf("# %d Expected (size) [offset:length]\n"
"\\msc\n"
"	a,b,c,d,e,f,g;\n"
"\n"
"	a->g [ label = \"total(%d)\" ] ;\n"
"	a->f [ label = \"aligned(%d)\" ] ;\n"
"	a->b [ label = \"header(%d)\" ] ,\n"
"	b->c [ label = \"misalign[%d:%d]"  /*0*/ "\" ] ,\n"
"	c->d [ label = \"free node[%d:%d]" /*1*/ "\" ] ,\n"
"	d->e [ label = \"data[%d:%d]"      /*2*/ "\" ] ,\n"
"	e->f [ label = \"used node[%d:%d]" /*3*/ "\" ] ,\n"
"	f->g [ label = \"data[%d:%d]"      /*4*/ "\" ] ;\n"
"\\endmsc\n"
, align
, aligned_used_data_offset + align
, aligned_used_data_offset
, misalign.offset
/*0*/ , misalign.offset, misalign.length
/*1*/ , freenode.offset, freenode.length
/*2*/ , freedata.offset, freedata.length
/*3*/ , usednode.offset, usednode.length
/*4*/ , useddata.offset, useddata.length);

		int fno = init(align, sizeof(TREEDB_NS(count_t))
			, aligned_used_data_offset + align - free_data_offset);

		if((fno + TREEDB_ALLOC_node_size) != (size_t)free_data_offset) {
			printf("Free data offset %zd != expected %d.\n"
				,(size_t)(fno + TREEDB_ALLOC_node_size), free_data_offset);
			return 1;
		}
		char * p = (char *)TREEDB_NS(malloc_aligned)(runtime, align, align, 0);
		if(!p) {
			printf("Malloc failed.\n");
			return 1;
		}
		ptrdiff_t diff = p - V3C_ROUND_DOWN_PTR2(p,align);
		if(diff) {
			printf("Returned pointer misaligned by %td bytes.\n", diff);
			return 1;
		}
		if(heap_check_res != 0) {
			printf("Heap check failed.\n");
			return 1;
		}
		TREEDB_NSA(node_ptr_t) pfree = TREEDB_ALLOC_free_head;
		if(!pfree) {
			printf("No free heap node.\n");
			return 1;
		}

		size_t fsz = TREEDB_ALLOC_node_length(pfree);

		usednode.offset = (int)TREEDB_ALLOCATOR.m_used.m_head/*.node.value*/;
		int aligned_used_data_offset2 = usednode.offset
					+ (int)TREEDB_ALLOC_node_size;
		freenode.offset = (int)TREEDB_pack_pointer(pfree);
		int saw_error = 0;
		if((fsz != (size_t)pad)
				|| (freenode.offset != fno)
				|| (aligned_used_data_offset != aligned_used_data_offset2)) {
			if(fsz != (size_t)pad)
				printf("Free heap node length isn't %d (%zu).\n", pad, fsz);
			if(aligned_used_data_offset != aligned_used_data_offset2)
				printf("Used data offset isn't %d (%d).\n"
					, aligned_used_data_offset, aligned_used_data_offset2);
			if(freenode.offset != fno)
				printf("Free node offset isn't %d (%d).\n"
					, freenode.offset, fno);
			saw_error = 1;
		}
		aligned_used_data_offset2 = aligned_used_data_offset;
		freedata.offset = freenode.offset + (int)TREEDB_ALLOC_node_size;
		freedata.length = fsz;
		useddata.length = (int)TREEDB_ALLOC_node_length(TREEDB_ALLOC_used_head);
		int total = aligned_used_data_offset + useddata.length;
		printf("# %d Actual (size) [offset:length]\n"
"\\msc\n"
"	a,b,c,d,e,f,g;\n"
"\n"
"	a->g [ label = \"total(%d)\" ] ;\n"
"	a->f [ label = \"aligned(%d)\" ] ;\n"
"	a->b [ label = \"header(%d)\" ] ,\n"
"	b->c [ label = \"misalign[%d:%d]"  /*0*/ "\" ] ,\n"
"	c->d [ label = \"free node[%d:%d]" /*1*/ "\" ] ,\n"
"	d->e [ label = \"data[%d:%d]"      /*2*/ "\" ] ,\n"
"	e->f [ label = \"used node[%d:%d]" /*3*/ "\" ] ,\n"
"	f->g [ label = \"data[%d:%d]"      /*4*/ "\" ] ;\n"
"\\endmsc\n"
, align
, total
, aligned_used_data_offset
, misalign.offset
/*0*/ , misalign.offset, misalign.length
/*1*/ , freenode.offset, freenode.length
/*2*/ , freedata.offset, freedata.length
/*3*/ , usednode.offset, usednode.length
/*4*/ , useddata.offset, useddata.length);
		if(saw_error)
			return n + 50;
	}
	return 0;
}

/** Test malloc_aligned with one used block followed by one free block

The free block is greater than the requested size but not large enough to split.

The block is incorrectly aligned for the request (actually
sizeof(TREEDB_NS(count_t)) off), and malloc_aligned solves this by padding the
previous node and moving and shrinking the following free node.

So
\msc
	a,b,c,d,e,f,g;

	a->g [ label = "total" ] ;
	a->f [ label = "aligned" ] ;
	a->b [ label = "header" ] ,
	b->c [ label = "used node" ] ,
	c->d [ label = "data" ] ,
	d->e [ label = "free node" ] ,
	e->f [ label = "misalign" ] ,
	f->g [ label = "data" ] ;
\endmsc
becomes
\msc
	a,b,c,d,e,f,g;

	a->g [ label = "total" ] ;
	a->f [ label = "aligned" ] ;
	a->b [ label = "header" ] ,
	b->c [ label = "used node" ] ,
	c->d [ label = "data" ] ,
	d->e [ label = "misalign" ] ,
	e->f [ label = "used node" ] ,
	f->g [ label = "data" ] ;
\endmsc
*/
int malloc_greater_misaligned()
{
	int n, align = sizeof(TREEDB_NS(count_t)) * 2; /* 2x,4x,8x,16x,32x */
	for(n = 0; n < 5; ++n, align <<= 1) {
		int b, c, d, e, f, g;
		/* Start b off with the miminum size. */
		b = TREEDB_NS_ALIGNED_OFFSET(sizeof(TREEDB_NS(treedb_t)));
		c = V3C_ROUND_UP_T2(int, b + TREEDB_ALLOC_node_size, align);
		b = c - TREEDB_ALLOC_node_size;
		/* Start with node1's data length == 0. */
		d = c;
		e = d + TREEDB_ALLOC_node_size;
		f = V3C_ROUND_UP_T2(int, e + sizeof(TREEDB_NS(count_t)), align);
		g = f + align;
		e = f - sizeof(TREEDB_NS(count_t));
		d = e - TREEDB_ALLOC_node_size;
		printf("# %d Original (size) [offset:length]\n"
"\\msc\n"
"	a,b,c,d,e,f,g;\n"
"\n"
"	a->g [ label = \"total(%d)\" ] ;\n"
"	a->f [ label = \"aligned(%d)\" ] ;\n"
"	a->b [ label = \"header(%d)\" ] ,\n"
"	b->c [ label = \"used node[%d:%d]" /*0*/ "\" ] ,\n"
"	c->d [ label = \"data[%d:%d]"      /*1*/ "\" ] ,\n"
"	d->e [ label = \"free node[%d:%d]" /*2*/ "\" ] ,\n"
"	e->f [ label = \"misalign[%d:%d]"  /*3*/ "\" ] ,\n"
"	f->g [ label = \"data[%d:%d]"      /*4*/ "\" ] ;\n"
"\\endmsc\n"
, align, g, f, b
/*0*/ , b, c - b
/*1*/ , c, d - c
/*2*/ , d, e - d
/*3*/ , e, f - e
/*4*/ , f, g - f
);
	e = d + sizeof(TREEDB_NS(count_t));
	f = e + TREEDB_ALLOC_node_size;
		printf("# %d Expected (size) [offset:length]\n"
"\\msc\n"
"	a,b,c,d,e,f,g;\n"
"\n"
"	a->g [ label = \"total(%d)\" ] ;\n"
"	a->f [ label = \"aligned(%d)\" ] ;\n"
"	a->b [ label = \"header(%d)\" ] ,\n"
"	b->c [ label = \"used node[%d:%d]" /*0*/ "\" ] ,\n"
"	c->d [ label = \"data[%d:%d]"      /*1*/ "\" ] ,\n"
"	d->e [ label = \"misalign[%d:%d]"  /*3*/ "\" ] ,\n"
"	e->f [ label = \"used node[%d:%d]" /*2*/ "\" ] ,\n"
"	f->g [ label = \"data[%d:%d]"      /*4*/ "\" ] ;\n"
"\\endmsc\n"
, align, g, f, b
/*0*/ , b, c - b
/*1*/ , c, d - c
/*2*/ , d, e - d
/*3*/ , e, f - e
/*4*/ , f, g - f
);

		int fno = init(align, 0, g - c);

		if(fno != b) {
			printf("Free node offset %u != expected %u.\n", fno, b);
			return 1;
		}
		char * p0 = (char *)TREEDB_NS(malloc_aligned)(runtime, d - c, 1, 0);
		if(!p0) {
			printf("First allocation failed.\n");
			return 1;
		}
		TREEDB_NSA(node_ptr_t) pn0 = (TREEDB_NSA(node_ptr_t))
			(p0 - TREEDB_ALLOC_node_size);
		if(TREEDB_pack_pointer(p0) != (size_t)c) {
			printf("Free (now used) data offset %u != expected %u.\n"
				, (int)TREEDB_pack_pointer(p0), c);
			return 1;
		}
		/* Misaligned allocation requiring a node move. */
		char * p = (char *)TREEDB_NS(malloc_aligned)(runtime, align, align, 0);
		if(!p) {
			printf("Second allocation failed.\n");
			return 1;
		}
		TREEDB_NSA(node_ptr_t) pn = (TREEDB_NSA(node_ptr_t))
			(p - TREEDB_ALLOC_node_size);

		ptrdiff_t diff = p - V3C_ROUND_DOWN_PTR2(p, align);
		if(diff) {
			printf("Returned pointer misaligned by %td bytes.\n", diff);
			return 1;
		}
		if(heap_check_res != 0) {
			printf("Heap check failed.\n");
			return 1;
		}
		if(TREEDB_ALLOCATOR.m_free.m_head/*.node.value*/) {
			printf("Free heap node is unexpected.\n");
			return 1;
		}
		printf("# %d Actual (size) [offset:length]\n"
"\\msc\n"
"	a,b,c,d,e,f,g;\n"
"\n"
"	a->g [ label = \"total(%d)\" ] ;\n"
"	a->f [ label = \"aligned(%d)\" ] ;\n"
"	a->b [ label = \"header(%d)\" ] ,\n"
"	b->c [ label = \"used node[%d:%d]" /*0*/ "\" ] ,\n"
"	c->d [ label = \"data[%d:%d]"      /*1*/ "\" ] ,\n"
"	d->e [ label = \"misalign[%d:%d]"  /*2*/ "\" ] ,\n"
"	e->f [ label = \"used node[%d:%d]" /*3*/ "\" ] ,\n"
"	f->g [ label = \"data[%d:%d]"      /*4*/ "\" ] ;\n"
"\\endmsc\n"
, align
, /*g*/ (int)TREEDB_pack_pointer(pn) + (int)TREEDB_ALLOC_node_length(pn)
		+ (int)TREEDB_ALLOC_node_size
, (int)TREEDB_pack_pointer(p)
, (int)TREEDB_pack_pointer(pn0)
/*0 b */ , (int)TREEDB_pack_pointer(pn0), (int)TREEDB_ALLOC_node_size
/*1 c */ , (int)TREEDB_pack_pointer(p0)
		, (int)(TREEDB_ALLOC_node_length(pn0) - sizeof(TREEDB_NS(count_t)))
/*2 d */ , (int)(TREEDB_pack_pointer(p0) + TREEDB_ALLOC_node_length(pn0)
			- sizeof(TREEDB_NS(count_t)))
		, (int)sizeof(TREEDB_NS(count_t))
/*3 e */ , (int)TREEDB_pack_pointer(pn), (int)TREEDB_ALLOC_node_size
/*4 f */ , (int)TREEDB_pack_pointer(p), (int)TREEDB_ALLOC_node_length(pn)
);
		if(b != (int)TREEDB_pack_pointer(pn0)) {
			printf("Bad b %d.\n", (int)TREEDB_pack_pointer(pn0));
			return 1;
		}
		if(c != (int)TREEDB_pack_pointer(p0)) {
			printf("Bad c %d.\n", (int)TREEDB_pack_pointer(p0));
			return 1;
		}
		if(d != (int)(TREEDB_pack_pointer(p0)
				+ TREEDB_ALLOC_node_length(pn0)
				- sizeof(TREEDB_NS(count_t)))) {
			printf("Bad d %d.\n"
				, (int)(TREEDB_pack_pointer(p0)
				+ TREEDB_ALLOC_node_length(pn0)
				- sizeof(TREEDB_NS(count_t)))
			);
			return 1;
		}
		if(e != (int)TREEDB_pack_pointer(pn)) {
			printf("Bad e %d.\n", (int)TREEDB_pack_pointer(pn));
			return 1;
		}
		if(f != (int)TREEDB_pack_pointer(p)) {
			printf("Bad f %d.\n", (int)TREEDB_pack_pointer(p));
			return 1;
		}
		if(g != ((int)TREEDB_pack_pointer(pn)
				+ (int)TREEDB_ALLOC_node_length(pn))
				+ (int)TREEDB_ALLOC_node_size) {
			printf("Bad g %d.\n"
				, (int)TREEDB_pack_pointer(pn)
					+ (int)TREEDB_ALLOC_node_length(pn)
					+ (int)TREEDB_ALLOC_node_size
			);
			return 1;
		}
	}
	return 0;
}
/** Test malloc_aligned with one free block that's equal to the requested size

The block is correctly aligned for the request, given the offset.
*/
int malloc_offset()
{
	int n, offs, align = sizeof(TREEDB_NS(count_t)) * 2; /* 2x,4x,8x,16x,32x */
	for(n = 0; n < 5; ++n, align <<= 1) {
		for(offs = 0; offs < align; offs += sizeof(TREEDB_NS(count_t))) {
			init(align, align - offs, align);
			char * p = (char *)TREEDB_NS(malloc_aligned)(runtime, align, align
				, offs);
			if(!p) {
				printf("Malloc failed.\n");
				return 1;
			}
			ptrdiff_t diff = p + offs - V3C_ROUND_DOWN_PTR2(p + offs, align);
			if(diff) {
				printf("Returned pointer misaligned by %td bytes.\n", diff);
				return 1;
			}
			if(heap_check_res != 0) {
				printf("Heap check failed.\n");
				return 1;
			}
			if(TREEDB_ALLOCATOR.m_free.m_head/*.node.value*/) {
				printf("Free heap node shouldn't be created.\n");
				return 1;
			}
		}
	}
	return 0;
}
#define TEST_ENTRY(x) {&x, #x}
typedef int (*PFN)();
typedef struct TestEntry {
	PFN fn;
	const char * name;
} TestEntry;
TestEntry test_entries[] =
{ TEST_ENTRY(malloc_simple_fail1)
, TEST_ENTRY(malloc_simple_fail2)
, TEST_ENTRY(malloc_simple1)
, TEST_ENTRY(malloc_greater1)
, TEST_ENTRY(malloc_greater2)
, TEST_ENTRY(malloc_greater3)
, TEST_ENTRY(malloc_greater_misaligned)
, TEST_ENTRY(malloc_offset)
};
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	TREEDB_NS(runtime_t) ctx;
	ctx.m_pBase = g_buf();
	runtime = & ctx;
	int n;
	for(n = 0; n < (int)(sizeof(test_entries) / sizeof(test_entries[0])); ++n) {
		TestEntry * t = test_entries + n;
		printf("[%d] testing %s.\n", n, t->name);
		int ret = t->fn();
		if(heap_check_res != 0) {
			printf("[%d] %s failed heap check.\n", n, t->name);
			return 1;
		}
		if(ret) {
			printf("[%d] %s failed : %d.\n", n, t->name, ret);
			return 2;
		}
		printf("[%d] %s passed.\n", n, t->name);
	}
	return 0;
}

#endif /* #if TREEDB_NBITS==-1 */
