/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform - passing.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>

#ifdef __cplusplus
#include <stdexcept>
#include <iostream>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;
#endif

#ifdef __cplusplus
#define TREEDB_have_dump_trees
#endif

#ifdef DOXYGEN_processing
{
#endif

#define STRING_TREE_NO_CLEARDEFS
#include <treedb/avl-string.h>


#ifdef DOXYGEN_processing
}
#endif
#include <treedb/comet-guids.h>

const char * items[] =
{ "apple"
, "orange"
, "lemon"
, "pear"
, "plum"
, "banana"
};
const char * sorted_items[] =
{ "apple"
, "banana"
, "lemon"
, "orange"
, "pear"
, "plum"
};

#ifdef __cplusplus
string node_status(TREEDB_NSA(node_ptr_t) pn)
{
	string ret;
	if(pn->m_offset_next/*.node.value*/ || pn->m_offset_prev/*.node.value*/)
		ret += " ";
	if(pn->m_offset_prev/*.node.value*/) {
		ret += string_inline_printf("{<- %zu"
			, (size_t)TREEDB_packed_item_offset(pn->m_offset_prev));
		if(TREEDB_packed_node_isfree(pn->m_offset_prev))
			ret += "[free]}";
		else
			ret += "[used]}";
		if(pn->m_offset_next/*.node.value*/)
			ret += " ";
	}
	if(pn->m_offset_next/*.node.value*/) {
		ret += string_inline_printf("{-> %zu"
			, (size_t)TREEDB_packed_item_offset(pn->m_offset_next));
		if(TREEDB_packed_node_isfree(pn->m_offset_next))
			ret += "[free]}";
		else
			ret += "[used]}";
	}
	return ret;
}
void dump_tree_node(TreeStreambuf & sb, TREEDB_NS(runtime_ptr_t) runtime
, TREEDB_NSA(node_ptr_t) node, bool is_used)
{
	if(!node)
		return;
	if(is_used) {
		const char * p = TREEDB_ALLOC_nodeptr_to_addr(node)
			+ offsetof(STRING_TREE_NS(node_t), m_item);
		sb.add
			( string_inline_printf("<%zu> (%zu) [%s]"
				, (size_t)TREEDB_pack_pointer(node)
				, (size_t)node->m_uLength, p)
			+ node_status(node)
			, true);
	}else
		sb.add
			( string_inline_printf("<%zu> (%zu)"
				, (size_t)TREEDB_pack_pointer(node)
				, (size_t)node->m_uLength)
			+ node_status(node)
			, true);
	sb.push();
	if(node->m_left/*.node.value*/) {
		sb.add("Left", !node->m_right/*.node.value*/);
		sb.push();
		dump_tree_node(sb, runtime
			, TREEDB_NSA_unpack_node(node->m_left)
			, is_used);
		sb.pop();
	}
	if(node->m_right/*.node.value*/) {
		sb.add("Right", true);
		sb.push();
		dump_tree_node(sb, runtime
			, TREEDB_NSA_unpack_node(node->m_right)
			, is_used);
		sb.pop();
	}
	sb.pop();
}
void TREEDB_NS(dump_trees)(TREEDB_NS(runtime_ptr_t) runtime)
{
	TreeStreambuf sb(cout);
	sb.add("Blocks", true);
	sb.push();
	sb.add("Used", false);
	sb.push();
	dump_tree_node(sb, runtime
		, TREEDB_NSA_unpack_node(runtime->m_pTreedb->m_allocator.m_used.m_top)
		, true);
	sb.pop();
	sb.add("Free", true);
	sb.push();
	dump_tree_node(sb, runtime
		, TREEDB_NSA_unpack_node(runtime->m_pTreedb->m_allocator.m_free.m_top)
		, false);
	sb.pop();
	sb.pop();
}
#endif /* #ifdef __cplusplus */

int test(TREEDB_NS(runtime_ptr_t) runtime)
{
	unsigned int n;
	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);

	if(res == eTreedbSearchEmpty)
		return 1;
	size_t size = pFirstNode->m_uLength;

	TREEDB_NS(dump_trees)(runtime);
	for(n = 0; n < sizeof(items)/sizeof(items[0]); ++n) {
		printf("[%u] inserting %s.\n", n, items[n]);
		STRING_TREE_NS(insert)(runtime, PACKED_STRING_TREE, items[n], 0);
		TREEDB_NS(dump_trees)(runtime);
	}

	STRING_TREE_NS(node_ptr_t) pRoot
		= STRING_TREE_NS_unpack_node(runtime->m_pTreedb->m_string_tree.m_top);
	if(pRoot) {
		n = pRoot->m_count;
		printf("Inserted %d items, expected %u.\n", n
			, (unsigned int)(sizeof(items)/sizeof(items[0])));
		if(n != sizeof(items)/sizeof(items[0]))
			return 2;
	}else
		printf("NULL root?\n");
	STRING_TREE_NS(node_ptr_t) c
		= STRING_TREE_NS_unpack_node(runtime->m_pTreedb->m_string_tree.m_head);

	for(n = 0; c; ++n, c = STRING_TREE_NS_unpack_node(c->m_next)) {
		printf("[%u] <%zu> %s [%s]\n", n, (size_t)TREEDB_pack_pointer(c)
			, c->m_item, sorted_items[n]);
		if(strcmp(c->m_item, sorted_items[n]))
			return 3;
	}
	for(n = 0; n < sizeof(sorted_items)/sizeof(sorted_items[0]); ++n) {
		printf("[%u] freeing [%s]\n", n, sorted_items[n]);
		STRING_TREE_NS(remove)(runtime, PACKED_STRING_TREE
			, sorted_items[n]);
		TREEDB_NS(dump_trees)(runtime);
	}
	res = TREEDB_NSF(search_closest)(runtime, TREEDB_ALLOC_free_tree, 0
		, & pFirstNode);
	if(size != pFirstNode->m_uLength) {
		printf("Error : free size %zu != original free size %zu.\n"
			, (size_t)pFirstNode->m_uLength
			, (size_t)size
		);
		return 4;
	}
	return 0;
}

char buf[TREEDB_PAGE_SIZE * 2] __attribute__ ((aligned (TREEDB_ALIGN))) = {0};

char * g_buf() { return buf; }

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	TREEDB_NS(runtime_t) ctx;
	TREEDB_NS(runtime_ptr_t) runtime = & ctx;
	ctx.m_pBase = g_buf();

	int ret = TREEDB_NS(init)(& ctx
				, UUID_treedb_string_tree_schema
				, offsetof(TREEDB_NS(treedb_t), STRING_TREE_MEMBER)
				, sizeof(TREEDB_NS(treedb_t))
				, TREEDB_PAGE_SIZE * 2);
	if(ret) {
		printf("main: treedb init() failed: %d.\n", ret);
		return 1;
	}

	STRING_TREE_NS(init)(runtime, STRING_TREE);
	printf("treedb header size [%u], tree size [%u], allocator node size"
		" [%u]\n"
		, (unsigned int)sizeof(TreedbHeader)
		, (unsigned int)sizeof(STRING_TREE_NS(tree_t))
		, (unsigned int)sizeof(TREEDB_NSA(node_t))
	);
	fflush(stdout);
	return test(runtime);
}
#endif /* #if TREEDB_NBITS==-1 */
