/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform - passing.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>

#ifdef __cplusplus
#include <stdexcept>
#include <iostream>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;
#endif

#undef TREEDB_DEBUG
#define TREEDB_DEBUG 1

#ifdef __cplusplus
#define TREEDB_have_dump_trees
#endif

/* We're not using a file-based treedb. */
#define TREEDB_OMIT_treedb_init
#define TREEDB_OMIT_treedb_init_header
#define TREEDB_PAYLOAD_NS

/* We're defining our own treedb. */
#define OMIT_TREEDB_treedb_t

#ifdef DOXYGEN_processing
{
#endif

// String tree.
#include <treedb/ns-types.h>
#define AVL_omit_treedb_impl
#define AVL_omit_treedb_storage
#define AVL_NO_IMPL
#include <avl-string.h>

// Int tree.
/* AVL - define macros we use. */
#define INT_TREE_NS(x) TREEDB_NS(AvlIntTree_##x)
#define INT_TREE_MEMBER m_int_tree
#define INT_TREE MAKE_AVL_TREE(INT_TREE_MEMBER)
#define PACKED_INT_TREE MAKE_PACKED_AVL_TREE(INT_TREE_NS, INT_TREE_MEMBER)
#define INT_TREE_NS_unpack_tree(x) \
	(INT_TREE_NS(tree_ptr_t))TREEDB_unpack_pointer(x)
#define INT_TREE_NS_unpack_node(x) \
	(INT_TREE_NS(node_ptr_t))TREEDB_unpack_pointer(x)

/* AVL- define macros avl.h uses - it will undefine them. */
#define AVL_NS(x) INT_TREE_NS(x)
#define AVL_TREE_MEMBER INT_TREE_MEMBER
#define AVL_omit_treedb_impl
#define AVL_omit_treedb_storage
#define AVL_NO_IMPL
#include <treedb/avl.h>

#define OMIT_TREEDB_functions
#include <treedb/storage.h>

#ifdef DOXYGEN_processing
}
#endif

struct TREEDB_NS(treedb_t)
{
#if 0
	TreedbHeader m_header;
#else
	uint64_t m_uPad; // To make sure that offset 0 isn't used.
#endif
	TREEDB_NS(allocator_t) m_allocator;
	TREEDB_NS(AvlStringTree_tree_t) m_string_tree;
	TREEDB_NS(AvlIntTree_tree_t) m_int_tree;
};
#define TREEDB_STORAGE_size runtime->m_uSize

/* ...and around again, this time with the implementation. */

#ifdef DOXYGEN_processing
{
#endif

#define AVL_omit_treedb_impl
#define AVL_omit_treedb_storage
#define AVL_NO_TYPES
#define STRING_TREE_NO_CLEARDEFS
#include "avl-string.h"

/* AVL- define macros avl.h uses - it will undefine them. */
#define AVL_NS(x) INT_TREE_NS(x)
#define AVL_TREE_MEMBER INT_TREE_MEMBER
#define AVL_omit_treedb_impl
#define AVL_omit_treedb_storage
#define AVL_NO_TYPES
#include <treedb/avl.h>

#include <treedb/treedb-impl.h>

#ifdef DOXYGEN_processing
}
#endif

const char * items[] =
{ "apple"
, "orange"
, "lemon"
, "pear"
, "plum"
, "banana"
};
const char * sorted_items[] =
{ "apple"
, "banana"
, "lemon"
, "orange"
, "pear"
, "plum"
};

#ifdef __cplusplus
string node_status(TREEDB_NSA(node_ptr_t) pn)
{
	string ret;
	if(pn->m_offset_next || pn->m_offset_prev)
		ret += " ";
	if(pn->m_offset_prev) {
		ret += string_inline_printf("{<- %zu"
			, (size_t)TREEDB_packed_item_offset(pn->m_offset_prev));
		if(TREEDB_packed_node_isfree(pn->m_offset_prev))
			ret += "[free]}";
		else
			ret += "[used]}";
		if(pn->m_offset_next)
			ret += " ";
	}
	if(pn->m_offset_next) {
		ret += string_inline_printf("{-> %zu"
			, (size_t)TREEDB_packed_item_offset(pn->m_offset_next));
		if(TREEDB_packed_node_isfree(pn->m_offset_next))
			ret += "[free]}";
		else
			ret += "[used]}";
	}
	return ret;
}
void dump_tree_node(TreeStreambuf & sb, TREEDB_NS(runtime_ptr_t) runtime
, TREEDB_NSA(node_ptr_t) node, bool is_used)
{
	if(!node)
		return;
	if(is_used) {
		const char * p = TREEDB_ALLOC_nodeptr_to_addr(node)
			+ offsetof(STRING_TREE_NS(node_t), m_item);
		if(STRING_TREE_NS(search_closest)
				(runtime, STRING_TREE, p, 0) == eTreedbSearchEqual)
			sb.add
			( string_inline_printf("<%zu> (%zu) string [%s]"
				, (size_t)TREEDB_pack_pointer(node)
				, (size_t)node->m_uLength, p)
			+ node_status(node)
			, true);
		else{
			p = TREEDB_ALLOC_nodeptr_to_addr(node)
				+ offsetof(INT_TREE_NS(node_t), m_item);
			sb.add
			( string_inline_printf("<%zu> (%zu) uint [%zu]"
				, (size_t)TREEDB_pack_pointer(node)
					, (size_t)node->m_uLength
					, (size_t)(*((TREEDB_NS(count_t) *)p)))
			+ node_status(node)
			, true);
		}
	}else
		sb.add
			( string_inline_printf("<%zu> (%zu)"
				, (size_t)TREEDB_pack_pointer(node)
				, (size_t)node->m_uLength)
			+ node_status(node)
			, true);
	sb.push();
	if(node->m_left) {
		sb.add("Left", !node->m_right);
		sb.push();
		dump_tree_node(sb, runtime, TREEDB_NSA_unpack_node(node->m_left)
			, is_used);
		sb.pop();
	}
	if(node->m_right) {
		sb.add("Right", true);
		sb.push();
		dump_tree_node(sb, runtime, TREEDB_NSA_unpack_node(node->m_right)
			, is_used);
		sb.pop();
	}
	sb.pop();
}
void TREEDB_NS(dump_trees)(TREEDB_NS(runtime_ptr_t) runtime)
{
	TreeStreambuf sb(cout);
	sb.add("Blocks", true);
	sb.push();
	sb.add("Used", false);
	sb.push();
	dump_tree_node(sb, runtime
		, TREEDB_NSA_unpack_node
				(runtime->m_pTreedb->m_allocator.m_used.m_top)
		, true);
	sb.pop();
	sb.add("Free", true);
	sb.push();
	dump_tree_node(sb, runtime
		, TREEDB_NSA_unpack_node
				(runtime->m_pTreedb->m_allocator.m_free.m_top)
		, false);
	sb.pop();
	sb.pop();
}
#endif /* #ifdef __cplusplus */

int test(TREEDB_NS(runtime_ptr_t) runtime)
{
	unsigned int n;
	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res = TREEDB_NSF(search_closest)(runtime
		, TREEDB_ALLOC_free_tree, 0, & pFirstNode);

	if(res == eTreedbSearchEmpty)
		return 1;
	TREEDB_NS(count_t) size = pFirstNode->m_uLength;

	TREEDB_NS(dump_trees)(runtime);
	for(n = 0; n < sizeof(items)/sizeof(items[0]); ++n) {
		printf("[%u] inserting %s.\n", n, items[n]);
		TREEDB_NS(AvlStringTree_insert)(runtime, PACKED_STRING_TREE
			, items[n], 0);
		TREEDB_NS(AvlIntTree_insert)(runtime, PACKED_INT_TREE, n, 0);
		TREEDB_NS(dump_trees)(runtime);
	}

	STRING_TREE_NS(node_ptr_t) pStringRoot
		= STRING_TREE_NS_unpack_node(runtime->m_pTreedb->m_string_tree.m_top);
	if(pStringRoot) {
		n = pStringRoot->m_count;
		printf("Inserted %d items, expected %u.\n", n
			, (unsigned int)(sizeof(items)/sizeof(items[0])));
		if(n != sizeof(items)/sizeof(items[0]))
			return 2;
	}else
		printf("NULL root?\n");

	INT_TREE_NS(node_ptr_t) pIntRoot
		= INT_TREE_NS_unpack_node(runtime->m_pTreedb->m_int_tree.m_top);
	if(pIntRoot) {
		n = pIntRoot->m_count;
		printf("Inserted %d items, expected %u.\n", n
			, (unsigned int)(sizeof(items)/sizeof(items[0])));
		if(n != sizeof(items)/sizeof(items[0]))
			return 2;
	}else
		printf("NULL root?\n");

	STRING_TREE_NS(node_ptr_t) c
		= STRING_TREE_NS_unpack_node(runtime->m_pTreedb->m_string_tree.m_head);

	for(n = 0; c; ++n, c = STRING_TREE_NS_unpack_node(c->m_next)) {
		printf("[%u] <%zu> %s [%s]\n", n, (size_t)TREEDB_pack_pointer(c)
			, c->m_item, sorted_items[n]);
		if(strcmp(c->m_item, sorted_items[n]))
			return 3;
	}
	for(n = 0; n < sizeof(sorted_items)/sizeof(sorted_items[0]); ++n) {
		printf("[%u] freeing [%s]\n", n, sorted_items[n]);
		STRING_TREE_NS(remove)(runtime, PACKED_STRING_TREE
			, sorted_items[n]);
		INT_TREE_NS(remove)(runtime, PACKED_INT_TREE, n);
		TREEDB_NS(dump_trees)(runtime);
	}
	res = TREEDB_NSF(search_closest)(runtime, TREEDB_ALLOC_free_tree, 0
		, & pFirstNode);
	if(size != pFirstNode->m_uLength)
		return 4;
	return 0;
}

// c1ec5655-644b-4898-b117-102b048ed95d
UUID_DEFINE
( UUID_treedb_two_trees_schema
, 0xc1, 0xec, 0x56, 0x55
, 0x64, 0x4b
, 0x48, 0x98
, 0xb1, 0x17
, 0x10, 0x2b, 0x04, 0x8e, 0xd9, 0x5d
);

char buf[TREEDB_PAGE_SIZE * 2];
/* Page aligned. */
char * g_buf() { return V3C_ROUND_UP_PTR2(buf,TREEDB_PAGE_SIZE); }

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	TREEDB_NS(runtime_t) ctx;
	ctx.m_pBase = g_buf();
	TREEDB_NS(runtime_ptr_t) runtime = & ctx;
	TREEDB_STORAGE_size = TREEDB_PAGE_SIZE;
#if 0
	TreedbInitResult ret = TREEDB_NS(init)(& ctx
				, TREEDB_NS_CONCAT(UUID_treedb_,_count_two_trees_schema)
				, offsetof(TREEDB_NS(treedb_t), m_string_tree)
				, sizeof(TREEDB_NS(treedb_t)));
#else
	TreedbInitResult ret = TREEDB_NS(init_memory)(& ctx
				, sizeof(TREEDB_NS(treedb_t)));
#endif
	if(ret != eTreedbInitSuccess) {
		printf("main: treedb init() failed: %s.\n", TreedbInitResultDesc(ret));
		return 1;
	}

	STRING_TREE_NS(init)(runtime, STRING_TREE);
	INT_TREE_NS(init)(runtime, INT_TREE);

	return test(& ctx);
}
#endif /* #if TREEDB_NBITS==-1 */
