/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include "comet-impl.h"

/*#define TREEDB_DEBUG 1
#define TREEDB_MSYNC MS_ASYNC*/

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP

#include <treedb/treedb.h>
#define TREEDB_NS(x) TREEDB_NS_CONCAT(V3C_JOIN3(li_,x,_),)

/* LIST - define macros we use. */
#define INT_LIST_NS(x) TREEDB_NS(IntList_##x)
#define INT_LIST_MEMBER m_int_list
#define INT_LIST MAKE_LIST(INT_LIST_MEMBER)
#define PACKED_INT_LIST MAKE_PACKED_LIST(INT_LIST_NS, INT_LIST_MEMBER)

/* AVL- define macros avl.h uses - it will undefine them. */
#define LIST_NS(x) INT_LIST_NS(x)
#define LIST_MEMBER INT_LIST_MEMBER
#include <treedb/list.h>
#include <treedb/comet-guids.h>

/******************************************************************************/
/* int list                                                                   */
/******************************************************************************/
TreedbHeapStatus TREEDB_NS_CONCAT(int_list,_create)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_int_list_schema
		, filename
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res != eTreedbHeapSuccess)
		return res;
	static INT_LIST_NS(list_t) list_init = LIST_TMPL_init;
	* INT_LIST = list_init;
	TREEDB_NS(heap_unlock)(& han);
	TREEDB_NS(heap_close)(runtime);
	return eTreedbHeapSuccess;
}
TreedbHeapStatus TREEDB_NS_CONCAT(int_list,_open)
( TREEDB_NS(runtime_ptr_t) runtime
, const char * filename
, int flags
) {
	TreedbHeapStatus res = TREEDB_NS(heap_open)
		( runtime
		, UUID_treedb_int_list_schema
		, filename
		, flags
		, (flags & eHeapOpenWrite ? eHeapAccessWrite : 0)
			| eHeapAccessRead | eHeapAccessShared
		, 0
		);
	return res;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_dump)(const char * filename)
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_list,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	INT_LIST_NS(node_ptr_t) node
		= (INT_LIST_NS(node_ptr_t))
			TREEDB_NS(impl_unpack_pointer)(runtime, INT_LIST->m_head/*.node*/);
	for(; node; node
			= (INT_LIST_NS(node_ptr_t))
				TREEDB_NS(impl_unpack_pointer)(runtime, node->m_next)) {
		TREEDB_printf2("[%zu] "
			, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, &node->m_item));
		printf("%d\n", (int)node->m_item);
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_count)
	(const char * filename)
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_list,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	unsigned int ret = INT_LIST_NS(count)(runtime, INT_LIST);
	printf("%u\n", ret);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_reverse)(const char * filename)
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_list,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	INT_LIST_NS(reverse)(runtime, INT_LIST);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_push_back)
	( const char * filename
	, const char * szitem)
{
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_list,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	CometCommandStatus ret
		= INT_LIST_NS(push_back)(runtime, PACKED_INT_LIST, item)
		? eCometCommandSuccess : eCometCommandOutOfResources;
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_push_front)
	( const char * filename
	, const char * szitem)
{
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_list,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	CometCommandStatus ret
		= INT_LIST_NS(push_front)(runtime, PACKED_INT_LIST, item)
		? eCometCommandSuccess : eCometCommandOutOfResources;
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret == 1 ? eCometCommandSuccess : eCometCommandOutOfResources;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_pop_back)(const char * filename)
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_list,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	CometCommandStatus ret
		= INT_LIST_NS(pop_back)(runtime, PACKED_INT_LIST)
		? eCometCommandFailure : eCometCommandSuccess;
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_pop_front)(const char * filename)
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_list,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	CometCommandStatus ret
		= INT_LIST_NS(pop_front)(runtime, PACKED_INT_LIST)
		? eCometCommandFailure : eCometCommandSuccess;
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_stream)(const char * filename)
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_list,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	int n;
	CometCommandStatus ret = eCometCommandSuccess;
	int c;
	for(;;) {
		c = scanf("%d", &n);
		if(c != 1)
			break;
		ret = INT_LIST_NS(push_back)(runtime, PACKED_INT_LIST, n)
			? eCometCommandSuccess : eCometCommandOutOfResources;
		if(ret == eCometCommandOutOfResources)
			break;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_type)
	(const char * filename)
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_list,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	/*int bad = TREEDB_NS(check_heap)(runtime);*/
	int bad = 0;
	char u[37];
	uuid_unparse_lower(runtime->m_pTreedb->m_header.m_uSchema, u);
	if(bad) {
		printf(
			"%s is a " TREEDB_NBITS_STRING " bit int list treedb [%s] "
			"[check failed - %d].\n", filename, u, bad);
	}else{
		size_t count = INT_LIST_NS(count)(runtime, INT_LIST);
		printf(
			"%s is a " TREEDB_NBITS_STRING " bit int list treedb [%s] "
			"with %zu entries.\n", filename, u, count);
	}
	TREEDB_NS_CONCAT(allocator,_schema)();

	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(int_list,_process)
( int argc, char *argv[]
) {
	switch(argc) {
	case 3: {
		/* prg dump <filename> */
		int is_dump      = !strcmp(argv[1], "dump");
		int is_count     = !strcmp(argv[1], "count");
		int is_check     = !strcmp(argv[1], "check");
		int is_stream    = !strcmp(argv[1], "stream");
		int is_type      = !strcmp(argv[1], "type");
		if(is_dump)
			return TREEDB_NS_CONCAT(int_list,_dump)(argv[2]);
		if(is_count)
			return TREEDB_NS_CONCAT(int_list,_count)(argv[2]);
		if(is_check)
			return TREEDB_NS_CONCAT(heap,_check)(argv[2]);
		if(is_stream)
			return TREEDB_NS_CONCAT(int_list,_stream)(argv[2]);
		if(is_type)
			return TREEDB_NS_CONCAT(int_list,_type)(argv[2]);
		return eCometCommandUnrecognisedCommand; }
	case 5: {
		/* prg set string <op> <filename> */
		int is_dump      = !strcmp(argv[3], "dump");
		int is_create    = !strcmp(argv[3], "create") || !strcmp(argv[3], "init");
		int is_pop_back  = !strcmp(argv[3], "pop_back");
		int is_pop_front = !strcmp(argv[3], "pop_front");
		int is_count     = !strcmp(argv[3], "count");
		int is_reverse   = !strcmp(argv[3], "reverse");
		int is_stream    = !strcmp(argv[3], "stream");
		if(is_dump)
			return TREEDB_NS_CONCAT(int_list,_dump)(argv[4]);
		if(is_create)
			return TREEDB_NS_CONCAT(int_list,_create)(argv[4])
				== eTreedbHeapSuccess ? eCometCommandSuccess
					: eCometCommandFailure;
		if(is_pop_back)
			return TREEDB_NS_CONCAT(int_list,_pop_back)(argv[4]);
		if(is_pop_front)
			return TREEDB_NS_CONCAT(int_list,_pop_front)(argv[4]);
		if(is_count)
			return TREEDB_NS_CONCAT(int_list,_count)(argv[4]);
		if(is_reverse)
			return TREEDB_NS_CONCAT(int_list,_reverse)(argv[4]);
		if(is_stream)
			return TREEDB_NS_CONCAT(int_list,_stream)(argv[4]);
		return eCometCommandUnrecognisedCommand; }
	case 6: {
		/* prg set int create <filename> 16 */
		/* prg set int <op> <filename> arg */
		int is_create     = !strcmp(argv[3], "create") || !strcmp(argv[3], "init");
		int is_push_back  = !strcmp(argv[3], "push_back");
		int is_push_front = !strcmp(argv[3], "push_front");
		if(is_create) {
			if(strcmp(argv[5], TREEDB_NBITS_STRING))
				return eCometCommandUnknownSchema;
			return TREEDB_NS_CONCAT(int_list,_create)(argv[4])
				== eTreedbHeapSuccess ? eCometCommandSuccess
					: eCometCommandFailure;
		}
		if(is_push_back)
			return TREEDB_NS_CONCAT(int_list,_push_back)(argv[4], argv[5]);
		if(is_push_front)
			return TREEDB_NS_CONCAT(int_list,_push_front)(argv[4], argv[5]);
		return eCometCommandUnrecognisedCommand; }
	}
	return eCometCommandUnrecognisedCommand;
}
