/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include "comet-impl.h"

/*#define TREEDB_DEBUG 1
#define TREEDB_MSYNC MS_ASYNC*/

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP

#include <treedb/treedb.h>
#define TREEDB_NS(x) TREEDB_NS_CONCAT(V3C_JOIN3(ti_,x,_),)

/* AVL - define macros we use. */
#define INT_TREE_NS(x) TREEDB_NS(AvlIntTree_##x)
#define INT_TREE_MEMBER m_int_tree
#define INT_TREE MAKE_AVL_TREE(INT_TREE_MEMBER)
#define PACKED_INT_TREE MAKE_PACKED_AVL_TREE(INT_TREE_NS, INT_TREE_MEMBER)

/* AVL- define macros avl.h uses - it will undefine them. */
#define AVL_NS(x) INT_TREE_NS(x)
#define AVL_TREE_MEMBER INT_TREE_MEMBER
#include <treedb/avl.h>
#include <treedb/comet-guids.h>

/******************************************************************************/
/* int tree                                                                   */
/******************************************************************************/
TreedbHeapStatus TREEDB_NS_CONCAT(int_tree,_create)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_int_tree_schema
		, filename
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res != eTreedbHeapSuccess)
		return res;
	INT_TREE_NS(tree_t) tree_init = AVL_TMPL_init;
	* INT_TREE = tree_init;
	TREEDB_NS(heap_unlock)(& han);
	TREEDB_NS(heap_close)(runtime);
	return eTreedbHeapSuccess;
}
TreedbHeapStatus TREEDB_NS_CONCAT(int_tree,_open)
( TREEDB_NS(runtime_ptr_t) runtime
, const char * filename
, int flags
) {
	TreedbHeapStatus res = TREEDB_NS(heap_open)
		( runtime
		, UUID_treedb_int_tree_schema
		, filename
		, flags
		, (flags & eHeapOpenWrite ? eHeapAccessWrite : 0)
			| eHeapAccessRead | eHeapAccessShared
		, 0
		);
	return res;
}
CometCommandStatus TREEDB_NS_CONCAT(int_tree,_dump)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_tree,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	INT_TREE_NS(node_ptr_t) node
		= INT_TREE_NS(unpack_node)(runtime
		, runtime->m_pTreedb->m_int_tree.m_head/*.node*/);
	/*printf("%u nodes.\n", TREE_STRING(count)(runtime));*/
	for(; node; node
			= INT_TREE_NS(unpack_node)(runtime, node->m_next/*.node*/)) {
		TREEDB_printf2("[%zu] "
			, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, &node->m_item));
		printf("%d\n", (int)node->m_item);
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(int_tree,_count)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_tree,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	unsigned int ret = INT_TREE_NS(count)(runtime, INT_TREE);
	printf("%u\n", ret);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(int_tree,_insert)
( const char * filename
, const char * szitem
) {
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_tree,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	TreedbInsertResult ret = INT_TREE_NS(insert)(runtime, PACKED_INT_TREE, item
		, 0);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	switch(ret) {
	case eTreedbInsertSuccess:
		return eCometCommandSuccess;
	case eTreedbInsertExists:
		return eCometCommandFailure;
	default:
	case eTreedbInsertNoMem:
		return eCometCommandOutOfResources;
	}
}
CometCommandStatus TREEDB_NS_CONCAT(int_tree,_remove)
( const char * filename
, const char * szitem
) {
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_tree,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	int ret = INT_TREE_NS(remove)(runtime, PACKED_INT_TREE, item);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret == 0 ? eCometCommandSuccess : eCometCommandFailure;
}
CometCommandStatus TREEDB_NS_CONCAT(int_tree,_contains)
( const char * filename
, const char * szitem
) {
	char * tailptr;
	int item = (int)strtol(szitem, & tailptr, 10);
	if(!tailptr || *tailptr) {
		  fprintf(stderr, "Error \"%s\" isn't a number.\n", szitem);
		  return eCometCommandBadCommand;
	}
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_tree,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	TREEDB_printf2("Node top: %zu, item offset: %zu.\n"
		, (size_t)INT_TREE->m_top, (size_t)offsetof(INT_TREE_NS(node_t), m_item)
	);
	int ret = INT_TREE_NS(search)(runtime, INT_TREE, item) ? 1 : 0;
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret == 1 ? eCometCommandSuccess : eCometCommandFailure;
}
CometCommandStatus TREEDB_NS_CONCAT(int_tree,_stream)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_tree,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	int n;
	CometCommandStatus ret = eCometCommandSuccess;
	int c;
	for(;;) {
		c = scanf("%d", &n);
		if(c != 1)
			break;
		ret = (INT_TREE_NS(insert)(runtime, PACKED_INT_TREE, n, 0)
				== eTreedbInsertNoMem)
			? eCometCommandOutOfResources : eCometCommandSuccess;
		if(ret == eCometCommandOutOfResources)
			break;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret;
}
CometCommandStatus TREEDB_NS_CONCAT(int_tree,_type)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(int_tree,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	/*int bad = TREEDB_NS(check_heap)(runtime);*/
	int bad = 0;
	char u[37];
	uuid_unparse_lower(runtime->m_pTreedb->m_header.m_uSchema, u);
	if(bad) {
		printf(
			"%s is a " TREEDB_NBITS_STRING " bit int set treedb [%s] "
			"[check failed - %d].\n", filename, u, bad);
	}else{
		size_t count = INT_TREE_NS(count)(runtime, INT_TREE);
		printf(
			"%s is a " TREEDB_NBITS_STRING " bit int set treedb [%s] "
			"with %zu entries.\n", filename, u, count);
	}
	TREEDB_NS_CONCAT(allocator,_schema)();

	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(int_tree,_process)
( int argc
, char *argv[]
) {
	switch(argc) {
	case 3: {
		/* prg dump <filename> */
		int is_dump     = !strcmp(argv[1], "dump");
		int is_count    = !strcmp(argv[1], "count");
		int is_check    = !strcmp(argv[1], "check");
		int is_stream   = !strcmp(argv[1], "stream");
		int is_type     = !strcmp(argv[1], "type");
		if(is_dump)
			return TREEDB_NS_CONCAT(int_tree,_dump)(argv[2]);
		if(is_count)
			return TREEDB_NS_CONCAT(int_tree,_count)(argv[2]);
		if(is_check)
			return TREEDB_NS_CONCAT(heap,_check)(argv[2]);
		if(is_stream)
			return TREEDB_NS_CONCAT(int_tree,_stream)(argv[2]);
		if(is_type)
			return TREEDB_NS_CONCAT(int_tree,_type)(argv[2]);
		return eCometCommandUnrecognisedCommand; }
	case 5: {
		/* prg set string <op> <filename> */
		int is_dump     = !strcmp(argv[3], "dump");
		int is_create   = !strcmp(argv[3], "create") || !strcmp(argv[3], "init");
		int is_count    = !strcmp(argv[3], "count");
		int is_stream   = !strcmp(argv[3], "stream");
		if(is_dump)
			return TREEDB_NS_CONCAT(int_tree,_dump)(argv[4]);
		if(is_create)
			return TREEDB_NS_CONCAT(int_tree,_create)(argv[4])
				== eTreedbHeapSuccess ? eCometCommandSuccess
					: eCometCommandFailure;
		if(is_count)
			return TREEDB_NS_CONCAT(int_tree,_count)(argv[4]);
		if(is_stream)
			return TREEDB_NS_CONCAT(int_tree,_stream)(argv[4]);
		return eCometCommandUnrecognisedCommand; }
	case 6: {
		/* prg set string create <filename> 16 */
		/* prg set string <op> <filename> arg */
		int is_create   = !strcmp(argv[3], "create") || !strcmp(argv[3], "init");
		int is_insert   = !strcmp(argv[3], "insert");
		int is_remove   = !strcmp(argv[3], "remove");
		int is_contains = !strcmp(argv[3], "contains");
		if(is_create) {
			if(strcmp(argv[5], TREEDB_NBITS_STRING))
				return eCometCommandUnknownSchema;
			return TREEDB_NS_CONCAT(int_tree,_create)(argv[4])
				== eTreedbHeapSuccess ? eCometCommandSuccess
					: eCometCommandFailure;
		}
		if(is_insert)
			return TREEDB_NS_CONCAT(int_tree,_insert)(argv[4], argv[5]);
		if(is_remove)
			return TREEDB_NS_CONCAT(int_tree,_remove)(argv[4], argv[5]);
		if(is_contains)
			return TREEDB_NS_CONCAT(int_tree,_contains)(argv[4], argv[5]);
		return eCometCommandUnrecognisedCommand; }
	}
	return eCometCommandUnrecognisedCommand;
}
