/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include <malloc.h>
#include "comet-impl.h"

#define HOWTO_DEBUG_STREAM

#ifdef HOWTO_DEBUG_STREAM
#include <v3c/time.hpp>
#endif

/*#define TREEDB_DEBUG 1
#define TREEDB_MSYNC MS_ASYNC*/

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP

#include <treedb/treedb.h>
#define TREEDB_NS(x) TREEDB_NS_CONCAT(V3C_JOIN3(ts_,x,_),)

#define STRING_TREE_NO_CLEARDEFS
#include <treedb/avl-string.h>
#include <treedb/comet-guids.h>

/******************************************************************************/
/* check                                                                      */
/******************************************************************************/
CometCommandStatus TREEDB_NS_CONCAT(heap,_check)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS(heap_open)
		( runtime
		, NULL
		, filename
		, eHeapOpenRead
		, eHeapAccessRead | eHeapAccessShared
		, 0
		);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	int ret = TREEDB_NS(check_heap)(runtime);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	if(ret)
		return eCometCommandHeapError;
	return eCometCommandSuccess;
}
#define ALLOC_OFFS(mem) \
(unsigned int)offsetof(TREEDB_NSA(node_t), mem)
void TREEDB_NS_CONCAT(allocator,_schema)
(
) {
#ifdef EXAMPLE
Header: 56, Allocator: 24, Node: 28, Item 12, Node total 40.

Allocator Node     Allocator Item
Of Name     Sz     Of Name          Sz     flags
-- -------- --     -- ------------- --     -----
 0 m_next    4      0 m_offset_prev  4     count
 4 m_prev    4      4 m_offset_next  4     depth
 8 m_parent  4      8 m_uLength      4
12 m_left    4
16 m_right   4
20 m_count   4
24 m_depth   1
--------------------------
40 Aligned total node size
#endif
	unsigned int sz = (unsigned int)sizeof(TREEDB_NS(count_t));
	unsigned int nsz
		= (unsigned int)offsetof(TREEDB_NSA(node_t), m_offset_prev);
	unsigned int totalsz = V3C_ROUND_UP_T2(unsigned int
		, sizeof(TREEDB_NSA(node_t)), sizeof(TREEDB_NS(count_t)));
	fflush(stdout);
	printf(
	/* 1*/  "Header: %u, Allocator: %u, Node: %u, Item %u, Node total %u.\n\n"
	        "Allocator Node     Allocator Item\n"
	        "Of Name     Sz     Of Name          Sz     flags\n"
	        "-- -------- --     -- ------------- --     -----\n"
	/* 2*/  " 0 m_next   %2u     %2u m_offset_prev %2u     count\n"
	/* 3*/ "%2u m_prev   %2u     %2u m_offset_next %2u     depth\n"
	/* 4*/ "%2u m_parent %2u     %2u m_uLength     %2u\n"
	/* 5*/ "%2u m_left   %2u\n"
	/* 6*/ "%2u m_right  %2u\n"
	/* 7*/ "%2u m_count  %2u\n"
	/* 8*/ "%2u m_depth   1\n"
	"--------------------------\n"
	/* 9*/ "%2u Aligned total node size\n\n"
	, /* 1*/ (unsigned int)sizeof(TreedbHeader)
	, (unsigned int)sizeof(TREEDB_NS(allocator_t))
	, (unsigned int)offsetof(TREEDB_NSA(node_t), m_depth) + 1
	, (unsigned int)sizeof(TREEDB_NSA(node_t))
		- (unsigned int)offsetof(TREEDB_NSA(node_t), m_offset_prev)
	, (unsigned int)sizeof(TREEDB_NSA(node_t))
	, /* 2*/ sz, nsz, sz
	, /* 3*/ ALLOC_OFFS(m_prev), sz, ALLOC_OFFS(m_offset_next), sz
	, /* 4*/ ALLOC_OFFS(m_parent), sz, ALLOC_OFFS(m_uLength), sz
	, /* 5*/ ALLOC_OFFS(m_left), sz
	, /* 6*/ ALLOC_OFFS(m_right), sz
	, /* 7*/ ALLOC_OFFS(m_count), sz
	, /* 8*/ ALLOC_OFFS(m_depth)
	, /* 9*/ totalsz
	);
	fflush(stdout);
}
#undef ALLOC_OFFS
/******************************************************************************/
/* string tree                                                                */
/******************************************************************************/
TreedbHeapStatus TREEDB_NS_CONCAT(string_tree,_create)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_string_tree_schema
		, filename
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res != eTreedbHeapSuccess)
		return res;
	STRING_TREE_NS(tree_t) tree_init = AVL_TMPL_init;
	* STRING_TREE = tree_init;
	TREEDB_printf2("Node size %zu.\n"
		, (size_t)offsetof(STRING_TREE_NS(node_t), m_item));
	CometCommandStatus ret = TREEDB_NS_CONCAT(heap,_check)(filename);
	TREEDB_NS(heap_unlock)(& han);
	TREEDB_NS(heap_close)(runtime);
	return ret == eCometCommandSuccess ? eTreedbHeapSuccess : eTreedbHeapBad;
}
TreedbHeapStatus TREEDB_NS_CONCAT(string_tree,_open)
( TREEDB_NS(runtime_ptr_t) runtime
, const char * filename
, int flags
) {
	TreedbHeapStatus res = TREEDB_NS(heap_open)
		( runtime
		, UUID_treedb_string_tree_schema
		, filename
		, flags
		, (flags & eHeapOpenWrite ? eHeapAccessWrite : 0)
			| eHeapAccessRead | eHeapAccessShared
		, 0
		);
	return res;
}
CometCommandStatus TREEDB_NS_CONCAT(string_tree,_dump)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(string_tree,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	STRING_TREE_NS(node_ptr_t) node
		= STRING_TREE_NS(unpack_node)(runtime, STRING_TREE->m_head/*.node*/);
	/*printf("%u nodes.\n", STRING_TREE_NS(count)(runtime));*/
	for(; node; node
			= STRING_TREE_NS(unpack_node)(runtime, node->m_next/*.node*/))
		printf("%s\n", node->m_item);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(string_tree,_count)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(string_tree,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	unsigned int ret = STRING_TREE_NS(count)(runtime, STRING_TREE);
	printf("%u\n", ret);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(string_tree,_insert)
( const char * filename
, const char * item
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(string_tree,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	TreedbInsertResult ret = STRING_TREE_NS(insert)(runtime
		, PACKED_STRING_TREE, item, 0);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	switch(ret) {
	case eTreedbInsertSuccess:
		return eCometCommandSuccess;
	case eTreedbInsertExists:
		return eCometCommandFailure;
	default:
	case eTreedbInsertNoMem:
		return eCometCommandOutOfResources;
	}
}
CometCommandStatus TREEDB_NS_CONCAT(string_tree,_remove)
( const char * filename
, const char * item
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(string_tree,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	int ret = STRING_TREE_NS(remove)(runtime, PACKED_STRING_TREE, item);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret == 0 ? eCometCommandSuccess : eCometCommandFailure;
}
CometCommandStatus TREEDB_NS_CONCAT(string_tree,_contains)
( const char * filename
, const char * item
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(string_tree,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	int ret = STRING_TREE_NS(search)(runtime, STRING_TREE, item) ? 1 : 0;
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret == 1 ? eCometCommandSuccess : eCometCommandFailure;
}
CometCommandStatus TREEDB_NS_CONCAT(string_tree,_stream)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(string_tree,_open)
		(runtime, filename, eHeapOpenRead | eHeapOpenWrite);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_unlock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to unlock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	res = TREEDB_NS(heap_write_lock)(runtime);
	if(res != eTreedbHeapSuccess) {
		fprintf(stderr, "Failed to write-lock \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		return eCometCommandHeapError;
	}
	char * buf;
	CometCommandStatus ret = eCometCommandSuccess;
	int c;
#ifdef HOWTO_DEBUG_STREAM
	int fillerup = 0;
	size_t fillerup_count = 0;
#endif
	for(;;) {
#ifdef HOWTO_DEBUG_STREAM
		if(!fillerup) {
#endif
			c = scanf("%ms", &buf);
			if(c != 1)
				break;
#ifdef HOWTO_DEBUG_STREAM
		}else{
			++fillerup_count;
			if(-1 == asprintf(& buf, "breakpoint-%zu", fillerup_count))
				return eCometCommandOutOfResources;
		}
		if(!fillerup)
			if(!strcmp(buf, "breakpoint")) {
				fillerup = 1;
				// Sleep for 5 minutes. Modify n in the debugger.
				int n, count = 10*60*5;
				for(n = 0; n < count; ++n)
					deci_sleep(1);
			}
#endif
		TreedbInsertResult res
			= STRING_TREE_NS(insert)(runtime, PACKED_STRING_TREE, buf, 0);
		free(buf);
		switch(res) {
		case eTreedbInsertNoMem:
			ret = eCometCommandOutOfResources;
			break;
		default:
			continue;
		}
		break;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return ret;
}
CometCommandStatus TREEDB_NS_CONCAT(string_tree,_type)
( const char * filename
) {
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = & han;
	TreedbHeapStatus res = TREEDB_NS_CONCAT(string_tree,_open)
		(runtime, filename, eHeapOpenRead);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	/*int bad = TREEDB_NS(check_heap)(runtime);*/
	int bad = 0;
	char u[37];
	uuid_unparse_lower(runtime->m_pTreedb->m_header.m_uSchema, u);
	if(bad) {
		printf(
			"%s is a " TREEDB_NBITS_STRING " bit string set treedb [%s] "
			"[check failed - %d].\n", filename, u, bad);
	}else{
		size_t count = STRING_TREE_NS(count)(runtime, STRING_TREE);
		printf(
			"%s is a " TREEDB_NBITS_STRING " bit string set treedb [%s] "
			"with %zu entries.\n", filename, u, count);
	}
	TREEDB_NS_CONCAT(allocator,_schema)();

	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return eCometCommandSuccess;
}
CometCommandStatus TREEDB_NS_CONCAT(string_tree,_process)
( int argc
, char *argv[]
) {
	switch(argc) {
	case 3: {
		/* prg dump <filename> */
		int is_dump     = !strcmp(argv[1], "dump");
		int is_count    = !strcmp(argv[1], "count");
		int is_check    = !strcmp(argv[1], "check");
		int is_stream   = !strcmp(argv[1], "stream");
		int is_type       = !strcmp(argv[1], "type");
		if(is_dump)
			return TREEDB_NS_CONCAT(string_tree,_dump)(argv[2]);
		if(is_count)
			return TREEDB_NS_CONCAT(string_tree,_count)(argv[2]);
		if(is_check)
			return TREEDB_NS_CONCAT(heap,_check)(argv[2]);
		if(is_stream)
			return TREEDB_NS_CONCAT(string_tree,_stream)(argv[2]);
		if(is_type)
			return TREEDB_NS_CONCAT(string_tree,_type)(argv[2]);
		return eCometCommandUnrecognisedCommand; }
	case 5: {
		/* prg set string <op> <filename> */
		int is_dump     = !strcmp(argv[3], "dump");
		int is_create   = !strcmp(argv[3], "create") || !strcmp(argv[3], "init");
		int is_count    = !strcmp(argv[3], "count");
		int is_stream   = !strcmp(argv[3], "stream");
		if(is_dump)
			return TREEDB_NS_CONCAT(string_tree,_dump)(argv[4]);
		if(is_create)
			return TREEDB_NS_CONCAT(string_tree,_create)(argv[4])
				== eTreedbHeapSuccess ? eCometCommandSuccess
					: eCometCommandFailure;
		if(is_count)
			return TREEDB_NS_CONCAT(string_tree,_count)(argv[4]);
		if(is_stream)
			return TREEDB_NS_CONCAT(string_tree,_stream)(argv[4]);
		return eCometCommandUnrecognisedCommand; }
	case 6: {
		/* prg set string create <filename> 16 */
		/* prg set string <op> <filename> arg */
		int is_create   = !strcmp(argv[3], "create") || !strcmp(argv[3], "init");
		int is_insert   = !strcmp(argv[3], "insert");
		int is_remove   = !strcmp(argv[3], "remove");
		int is_contains = !strcmp(argv[3], "contains");
		if(is_create) {
			if(strcmp(argv[5], TREEDB_NBITS_STRING))
				return eCometCommandUnknownSchema;
			return TREEDB_NS_CONCAT(string_tree,_create)(argv[4])
				== eTreedbHeapSuccess ? eCometCommandSuccess
					: eCometCommandFailure;
		}
		if(is_insert)
			return TREEDB_NS_CONCAT(string_tree,_insert)(argv[4], argv[5]);
		if(is_remove)
			return TREEDB_NS_CONCAT(string_tree,_remove)(argv[4], argv[5]);
		if(is_contains)
			return TREEDB_NS_CONCAT(string_tree,_contains)(argv[4], argv[5]);
		return eCometCommandUnrecognisedCommand; }
	}
	return eCometCommandUnrecognisedCommand;
}
