/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include "comet.h"

extern CometCommandStatus string_tree_process(int argc, char *argv[]);
extern CometCommandStatus int_tree_process(int argc, char *argv[]);
extern CometCommandStatus string_list_process(int argc, char *argv[]);
extern CometCommandStatus int_list_process(int argc, char *argv[]);

int is_readable(const char * filename)
{
	return access(filename, R_OK);
}
int is_writable(const char * filename)
{
	(void)filename;
	return 0;
}
const char * program = 0;
const char * program_name(const char *p)
{
	const char * name = strrchr(p,'/');
	if(!name)
		return p;
	if(name != (p + strlen(p) - 1))
		name++;
	return name;
}
void usage()
{
	printf("Usage:"
		"\tv3c-comet <string|int> <list|set> <op> <filename> [arg]\n"
		"\tv3c-comet <string|int> <list|set> <op> <filename>\n"
		"\tv3c-comet <op> <filename> [arg]\n\n"
		"Common operations are type, init|create, count stream, check and dump.\n"
		"For sets, the extra operations are insert, remove and contains.\n"
		"For lists, the valid operations are reverse, push_back, push_front, pop_back, and pop_front.\n"
		"To create a default sandbox, use\n"
		"\tv3c-comet <string|int> <list|set> <create|init> <filename> [bits],\n"
		"where bits is one of 16,32,64. If omitted it uses the default for\n"
		"your platform - 32 or 64 bits.\n"
		"Stream is the slip-stream way of populating a treedb from a pipe, as in\n"
		"\tfind . | v3c-comet stream <filename>\n"
		"You cannot create or read a 64 bit treedb on a 32 bit platform.\n"
		);
}
/* v3c-comet <type> <container> <op> <file-name> [arg] */
/* v3c-comet int list create <file-name> */
/* v3c-comet string set create <file-name> */
/* v3c-comet string set insert <file-name> "abc" */
/* v3c-comet int list push_back <file-name> 123 */
/* v3c-comet dump <file-name> */
int main(int argc, char *argv[])
{
	program = program_name(argv[0]);
	switch(argc) {
	case 2:
		if(!strcmp(argv[1], "-?")) {
			usage();
			return 0;
		}else if(!strcmp(argv[1], "-h")) {
			usage();
			return 0;
		}else if(!strcmp(argv[1], "-v")) {
			printf(PACKAGE_VERSION "\n");
			return 0;
		}else if(!strcmp(argv[1], "--help")) {
			usage();
			return 0;
		}else if(!strcmp(argv[1], "--version")) {
			printf(PACKAGE_VERSION "\n");
			return 0;
		}else if(!strcmp(argv[1], "man")) {
			printf(
			".\\\"                                      Hey, EMACS: -*- nroff -*-\n"
			".\\\" First parameter, NAME, should be all caps\n"
			".\\\" Second parameter, SECTION, should be 1-8, maybe w/ subsection\n"
			".\\\" other parameters are allowed: see man(7), man(1)\n"
			".TH treedb 1 \"" PACKAGE_SHORT_BUILD_TIME "\"\n"
			".\\\" Please adjust this date whenever revising the manpage.\n"
			".\\\"\n"
			".\\\" Some roff macros, for reference:\n"
			".\\\" .nh        disable hyphenation\n"
			".\\\" .hy        enable hyphenation\n"
			".\\\" .ad l      left justify\n"
			".\\\" .ad b      justify to both left and right margins\n"
			".\\\" .nf        disable filling\n"
			".\\\" .fi        enable filling\n"
			".\\\" .br        insert line break\n"
			".\\\" .sp <n>    insert n+1 empty lines\n"
			".\\\" for manpage-specific macros, see man(7)\n"
			".SH NAME\n"
			"v3c\\-comet \\- a treedb demo program\n"
			".SH SYNOPSIS\n"
			".B v3c\\-comet\n"
			".RI [ command ] \" options\" ...\n"
			".SH DESCRIPTION\n"
			"This manual page documents briefly the\n"
			".B v3c\\-comet\n"
			"command.\n"
			".PP\n"
			".\\\" TeX users may be more comfortable with the \\fB<whatever>\\fP and\n"
			".\\\" \\fI<whatever>\\fP escape sequences to invode bold face and italics,\n"
			".\\\" respectively.\n"
			"\\fBv3c\\-comet\\fP is a treedb demo program with the following commands\n"
			".br\n");
			char * rp = realpath(argv[0], NULL);
			if(!rp)
				return 2;
			char cmd[4096];
			sprintf(cmd, "%s --help | sed -e 's|-|\\\\-|g' "
				"-e 's|\\t||g' -e 's|$|\\n\\.br|g' -re 's|(.+) :|\\\\fB\\1\\\\fP :|g'", rp);
			free(rp);
			printf(".\\\" %s\n", cmd);
			fflush(stdout);
			return system(cmd);
		}
		usage();
		return 1;
	case 3:
		/* prg dump <filename> */
		if(is_readable(argv[2]))
			return eCometCommandFileOpenFailure;
		do {
			CometCommandStatus ret;
			ret = string_tree_process(argc, argv);
			if(ret != eCometCommandUnknownSchema)
				return ret;
			ret = int_tree_process(argc, argv);
			if(ret != eCometCommandUnknownSchema)
				return ret;
			ret = string_list_process(argc, argv);
			if(ret != eCometCommandUnknownSchema)
				return ret;
			ret = int_list_process(argc, argv);
			if(ret != eCometCommandUnknownSchema)
				return ret;
			if(ret == eCometCommandUnknownSchema)
				fprintf(stderr, "Unrecognised file schema.\n");
			else if(ret == eCometCommandUnknownAllocatorAlgorithm)
				fprintf(stderr, "Unrecognised file allocator algorithm.\n");
			return ret;
		}while(0);
	case 4:
		/* prg string set ? */
		usage();
		return 1;
	case 5: {
		/* prg string set <op> <filename> */
		int is_int    = !strcmp(argv[1], "int");
		int is_string = !strcmp(argv[1], "string");
		if(!(is_int || is_string)) {
			usage();
			return eCometCommandUnrecognisedCommand;
		}
		int is_set  = !strcmp(argv[2], "set");
		int is_list = !strcmp(argv[2], "list");
		if(!(is_set || is_list)) {
			usage();
			return eCometCommandUnrecognisedCommand;
		}
		if(is_set && is_int)
			return int_tree_process(argc, argv);
		if(is_set && is_string)
			return string_tree_process(argc, argv);
		if(is_list && is_int)
			return int_list_process(argc, argv);
		if(is_list && is_string)
			return string_list_process(argc, argv);
		return eCometCommandUnrecognisedCommand; }
	case 6: {
		/* prg int list <op> <filename> <arg> */
		int is_int    = !strcmp(argv[1], "int");
		int is_string = !strcmp(argv[1], "string");
		if(!(is_int || is_string)) {
			usage();
			return eCometCommandUnrecognisedCommand;
		}
		int is_set  = !strcmp(argv[2], "set");
		int is_list = !strcmp(argv[2], "list");
		if(!(is_set || is_list)) {
			usage();
			return eCometCommandUnrecognisedCommand;
		}
		if(is_writable(argv[4])) {
			fprintf(stderr, "file \"%s\" isn't writable.\n", argv[4]);
			return eCometCommandFileOpenFailure;
		}
		if(is_set && is_int)
			return int_tree_process(argc, argv);
		if(is_set && is_string)
			return string_tree_process(argc, argv);
		if(is_list && is_int)
			return int_list_process(argc, argv);
		if(is_list && is_string)
			return string_list_process(argc, argv);
		return eCometCommandUnrecognisedCommand; }
	default:
		usage();
		return eCometCommandUnrecognisedCommand;
	}
}
