/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#define HAVE_TREEDB_DUMP_TREES

string node_status(TREEDB_NSA(node_ptr_t) pn)
{
	string ret;
	if(pn->m_offset_next || pn->m_offset_prev)
		ret += " ";
	if(pn->m_offset_prev) {
		ret += string_inline_printf("{<- %zu"
			, (size_t)TREEDB_packed_item_offset(pn->m_offset_prev));
		if(TREEDB_packed_node_isfree(pn->m_offset_prev))
			ret += "[free]}";
		else
			ret += "[used]}";
		if(pn->m_offset_next)
			ret += " ";
	}
	if(pn->m_offset_next) {
		ret += string_inline_printf("{-> %zu"
			, (size_t)TREEDB_packed_item_offset(pn->m_offset_next));
		if(TREEDB_packed_node_isfree(pn->m_offset_next))
			ret += "[free]}";
		else
			ret += "[used]}";
	}
	return ret;
}
void dump_tree_node(TreeStreambuf & sb
, TREEDB_NS(runtime_ptr_t) runtime
, TREEDB_NSA(node_ptr_t) node
, bool is_used)
{
	if(!node)
		return;
	if(is_used) {
		const char * p = TREEDB_ALLOC_nodeptr_to_addr(node)
			+ offsetof(TREEDB_PAYLOAD_NS(node_t), m_item);
		sb.add
			( string_inline_printf("[%zu] payload offs[%zu] + text[%zu] = "
				"%zu [%s]"
				, (size_t)TREEDB_pack_pointer(node)
				, offsetof(TREEDB_PAYLOAD_NS(node_t), m_item)
				, (size_t)(node->m_uLength)
					- offsetof(TREEDB_PAYLOAD_NS(node_t), m_item)
				, (size_t)(node->m_uLength)
				, p)
			+ node_status(node)
			, true);
	}else
		sb.add
			( string_inline_printf("[%zu], length = %zu"
				, (size_t)TREEDB_pack_pointer(node)
				, (size_t)(node->m_uLength)
				)
			+ node_status(node)
			, true);
	sb.push();
	if(node->m_left) {
		sb.add("Left", !node->m_right);
		sb.push();
		dump_tree_node(sb, runtime
			, TREEDB_NSA_unpack_node(node->m_left)
			, is_used);
		sb.pop();
	}
	if(node->m_right) {
		sb.add("Right", true);
		sb.push();
		dump_tree_node(sb, runtime
			, TREEDB_NSA_unpack_node(node->m_right)
			, is_used);
		sb.pop();
	}
	sb.pop();
}
string payload_node_status(TREEDB_PAYLOAD_NS(node_ptr_t) pp)
{
	string ret = string_inline_printf("# %zu", (size_t)pp->m_count);
	return ret;
}
void dump_payload_node(TreeStreambuf & sb, TREEDB_NS(runtime_ptr_t) runtime, STRING_TREE_NS(node_ptr_t) node)
{
	if(!node)
		return;
	sb.add
		( string_inline_printf("<%zu> = \"%s\" "
			, (size_t)TREEDB_pack_pointer(node), node->m_item)
		+ payload_node_status(node)
		, true);
	sb.push();
	if(node->m_left) {
		sb.add("Left", !node->m_right);
		sb.push();
		dump_payload_node
			(sb, runtime, STRING_TREE_NS_unpack_node(node->m_left));
		sb.pop();
	}
	if(node->m_right) {
		sb.add("Right", true);
		sb.push();
		dump_payload_node(sb, runtime
			, STRING_TREE_NS_unpack_node(node->m_right)
		);
		sb.pop();
	}
	sb.pop();
}
void TREEDB_NS(dump_trees)(TREEDB_NS(runtime_ptr_t) runtime)
{
	{
		TreeStreambuf sb(cout);
		sb.add("Blocks", true);
		sb.push();
		sb.add("Used", false);
		sb.push();
		dump_tree_node(sb, runtime
			, TREEDB_NSA_unpack_node
					(runtime->m_pTreedb->m_allocator.m_used.m_top)
			, true);
		sb.pop();
		sb.add("Free", false);
		sb.push();
		dump_tree_node(sb, runtime
			, TREEDB_NSA_unpack_node
				(runtime->m_pTreedb->m_allocator.m_free.m_top)
			, false);
		sb.pop();
		sb.add("Items", true);
		sb.push();
		dump_payload_node(sb, runtime
			, STRING_TREE_NS_unpack_node
				(runtime->m_pTreedb->m_string_tree.m_top));

		sb.pop();
		sb.pop();
	}
	cout << flush;
	fflush(stdout);
}
