/* Copyright (C) 2011  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <sys/mman.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

int main(int argc, char * argv[])
{
	(void)argc;
	(void)argv;
	int zfd = open("/dev/zero", O_RDONLY | O_CLOEXEC);
	if(!zfd) {
		fprintf(stderr, "opening /dev/zero failed.\n");
		return 1;
	}
	/* Make a private allocation. */
	fprintf(stderr, "Creating anonymous read-write mapping.\n");
	char * p = (char *)mmap(0, 8192, PROT_READ | PROT_WRITE
		, MAP_SHARED | MAP_ANON, 0, 0);
	if(p == (char *)-1) {
		fprintf(stderr, "Creating anonymous read-write mapping failed : %s\n"
			, strerror(errno));
		close(zfd);
		return 1;
	}
	fprintf(stderr, "Creating fixed read-only mapping onto /dev/zero.\n");
	char * q = (char *)mmap(p + 4096, 4096, PROT_READ
		, MAP_SHARED | MAP_FIXED, zfd, 0);
	if(q == (char *)-1) {
		fprintf(stderr, "Creating fixed read-only mapping failed : %s\n"
			, strerror(errno));
		close(zfd);
		munmap(p, 8192);
		return 1;
	}
	close(zfd); /* mmap makes its own copy. */

	fprintf(stderr, "Writing.\n");
	*p = 1;
	fprintf(stderr, "Reading 1.\n");
	char c = *p;
	int res1 = (c != 1);
	fprintf(stderr, "Reading 2.\n");
	c = *q;
	int res2 = (c != 0);

	fprintf(stderr, "Unmapping.\n");
	munmap(p, 8192);

	if(res1)
		fprintf(stderr, "Reading p != 1.\n");
	if(res2)
		fprintf(stderr, "Reading q != 0.\n");
	return res1 + res2;
}
