/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform - passing.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include <v3c/time.hpp>
#include <sys/wait.h>

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP

// Int tree.
/* AVL - define macros we use. */
#define INT_TREE_NS(x) TREEDB_NS(AvlIntTree_##x)
#define INT_TREE_MEMBER m_int_tree
#define INT_TREE MAKE_AVL_TREE(INT_TREE_MEMBER)
#define PACKED_INT_TREE MAKE_PACKED_AVL_TREE(INT_TREE_NS, INT_TREE_MEMBER)
#define INT_TREE_NS_unpack_tree(x) \
	(INT_TREE_NS(tree_ptr_t))TREEDB_unpack_pointer(x)
#define INT_TREE_NS_unpack_node(x) \
	(INT_TREE_NS(node_ptr_t))TREEDB_unpack_pointer(x)

/* AVL- define macros avl.h uses - it will undefine them. */
#define AVL_NS(x) INT_TREE_NS(x)
#define AVL_TREE_MEMBER INT_TREE_MEMBER

#include <treedb/avl.h>

#include <treedb/comet-guids.h>

const char * fork_file = "fork-test.3db";
TREEDB_NS(count_t) free_offs, free_size, packed_pointer;

/* Setting this value too high with TREEDB_NBITS==16 runs out of addresses (64K),
making this a good test for graceful insertion/malloc failure. */
int fork_length = 100;
int process_id = 0, treedb_fd;
pid_t pid;

int create_treedb();
int process_treedb();

FileLockState second_byte = FILE_LOCK_INIT;

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Root : 1. Create fork-test.3db.\n");
	int res = create_treedb();
	if(res)
		return res;
	printf("Root : 2. write-lock the second byte of fork-test.3db.\n");
	treedb_fd = open(fork_file, O_RDWR);
	if(treedb_fd == -1) {
		printf("Root : 2. couldn't open fork-test.3db?\n");
		unlink(fork_file);
		return 1;
	}
	res = file_lock_state_init(treedb_fd, & second_byte, 2, 1
		, eFileLockWriteLocked);
	if(res) {
		printf("Root : 2. couldn't write-lock fork-test.3db?\n");
		close(treedb_fd);
		unlink(fork_file);
		return 1;
	}

	res = process_treedb();
	if(!process_id) {
		close(treedb_fd); /* Loses all locks. */
#if TREEDB_DEBUG
		if(!res) {
			printf("Root : 13. Deleting the treedb.\n");
			unlink(fork_file);
			int n;
			char filename[30];
			for(n = 1; n < fork_length; ++n) {
				sprintf(filename, "fork_test-%.2d.txt", n);
				if(n == 1) {
					printf("Root : 13. Dumping output of sub-process 1.\n\n\n");
					fflush(stdout);
					char cmd[40];
					sprintf(cmd, "cat %s", filename);
					(void)system(cmd);
				}
				unlink(filename);
			}
		}else{
			printf("Root : 13. (fail) leaving treedb and trace files.\n");
		}
#else
		printf("Root : 13. Deleting the treedb.\n");
		unlink(fork_file);
#endif
	}
	fflush(stdout);
	return res;
}
#define WRITE_LOCK_TREEDB(step) \
	printf("Process %d: " #step ". Write-locking.\n", process_id); \
	ret = TREEDB_NS(heap_write_lock)(runtime); \
	if(ret != eTreedbHeapSuccess) { \
		printf("Process %d: " #step ". TREEDB_NS(heap_write_lock)() "\
			"failed : %s.\n", process_id, TreedbHeapStatusDesc(ret)); \
		TREEDB_NS(heap_close)(&han); \
		return 1; \
	}

#define READ_LOCK_TREEDB(step) \
	printf("Process %d: " #step ". Read-locking.\n", process_id); \
	ret = TREEDB_NS(heap_read_lock)(runtime); \
	if(ret != eTreedbHeapSuccess) { \
		printf("Process %d: " #step ". TREEDB_NS(heap_read_lock)() " \
			"failed : %s.\n", process_id, TreedbHeapStatusDesc(ret)); \
		TREEDB_NS(heap_close)(&han); \
		return 1; \
	}
#define UNLOCK_TREEDB(step) \
	printf("Process %d: " #step ". Unlocking.\n", process_id); \
	TREEDB_NS(heap_unlock)(runtime)
#define CLOSE_TREEDB \
	TREEDB_NS(heap_close)(runtime)
void wait_child_exit(int num, int fail)
{
	if(pid) {
		printf("Process %d : %d. %sWaiting for child to exit.\n"
			, process_id, num, fail ? "(fail) " : "");
		waitpid(pid, 0, 0);
		printf("Process %d : %d. %sChild exited.\n"
			, process_id, num, fail ? "(fail) " : "");
	}
}
int process_treedb()
{
	printf("Process %d : 3. Spawn child process before opening the "
		"treedb.\n", process_id);
	while(process_id < (fork_length-1)) {
		fflush(stdout);
		pid = fork();
		if(pid == (pid_t)-1) {
			printf("Process %d: 3. fork() failed.\n", process_id);
			return 3;
		}else if(!pid) {
			/* The child. */
			close(treedb_fd);
			++process_id;
#if TREEDB_DEBUG
			char filename[20];
			sprintf(filename, "fork_test-%.2d.txt", process_id);
			int file = TEMP_FAILURE_RETRY (open(filename
				, O_RDWR | O_CREAT | O_TRUNC, 0644));
			dup2(file, STDOUT_FILENO);
			(void)TEMP_FAILURE_RETRY(close(file));
#endif
			printf("Process %d : 3. Child process started.\n", process_id);
		}else{
			/* The parent continues. */
			break;
		}
	}
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = &han;
	TreedbHeapStatus ret = TREEDB_NS(heap_open)
		( runtime
		, UUID_treedb_int_tree_schema
		, fork_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(ret != eTreedbHeapSuccess) {
		printf("Process %d: 3. TREEDB_NS(heap_open)() failed : %s.\n"
			, process_id, TreedbHeapStatusDesc(ret));
		wait_child_exit(2, 1);
		return 3;
	}
	UNLOCK_TREEDB(3);

	if(process_id) {
		printf("Process %d: 4. read-locking the second byte.\n", process_id);
		second_byte.m_fd = runtime->m_fls.m_fd;
		if(file_lock_state_update(& second_byte, eFileLockReadLocked)) {
			printf("Process %d: 4. read-locking the second byte failed.\n"
				, process_id);
			CLOSE_TREEDB;
			wait_child_exit(4, 1);
			return 4;
		}
	}else{
		printf("Root : 5. Unlocking the second byte.\n");
		centi_sleep(40);
		if(file_lock_state_update(& second_byte, eFileLockUnlocked)) {
			printf("Root : 5. Unlocking second byte failed.\n");
			CLOSE_TREEDB;
			wait_child_exit(5, 1);
			return 5;
		}
	}

	if(process_id) {
		WRITE_LOCK_TREEDB(6);
		printf("Process %d : 6. Inserting %d.\n", process_id, process_id);
		TreedbInsertResult ret
			= INT_TREE_NS(insert)(runtime, PACKED_INT_TREE, process_id, 0);
		if(ret != eTreedbInsertSuccess) {
			printf("Process %d: 6. INT_TREE_NS(insert)() failed : "
				"%s.\n", process_id, TreedbInsertResultDesc(ret));
			UNLOCK_TREEDB(6);
			CLOSE_TREEDB;
			wait_child_exit(6, 1);
			return 6;
		}
		UNLOCK_TREEDB(6);
	}
	if(process_id < (fork_length-1)) {
		printf("Process %d : 7. Waiting for %d to be inserted.\n"
			, process_id, process_id + 1);
		/* Macro madness. */
		int to_remove = process_id + 1;
		for(;;) {
			READ_LOCK_TREEDB(7);
			int res
				= INT_TREE_NS(search)(runtime, INT_TREE, process_id + 1) != 0;
			UNLOCK_TREEDB(7);
			if(res) {
				printf("Process %d : 7. %d was inserted.\n"
					, process_id, to_remove);
				break;
			}
			int wres = waitpid(pid, 0, WNOHANG);
			if((wres == -1) || (wres == pid)) {
				printf("Process %d : 7. waitpid(%d) returned %d.\n"
					, process_id, pid, wres);
				break;
			}
			/* It seems that omitting the following line leaves too little time
			for other processes to lock the treedb, causing this process to
			hog the treedb. */
			centi_sleep(1);
		}
		WRITE_LOCK_TREEDB(8);
		int res
			= INT_TREE_NS(remove)(runtime, PACKED_INT_TREE, process_id + 1);
		if(res) {
			printf("Process %d: 8. TREEDB_NS(remove)(%d) failed.\n"
				, process_id, process_id + 1);
			if(process_id) {
				void * p = TREEDB_unpack_pointer(packed_pointer);
				TREEDB_NS(free)(runtime, p);
			}
			UNLOCK_TREEDB(8);
			CLOSE_TREEDB;
			wait_child_exit(8, 1);
			return 8;
		}
		UNLOCK_TREEDB(8);
		printf("Process %d : 8. Removed %d.\n", process_id, process_id + 1);
	}
	if(process_id) {
		WRITE_LOCK_TREEDB(9);
		printf("Process %d : 9. Allocating chunk.\n", process_id);
		void * p = TREEDB_NS(malloc)(runtime
			, TREEDB_NBITS == 16 ? 1000 : 8000);
		if(!p) {
			printf("Process %d: 9. TREEDB_NS(malloc)() failed.\n"
				, process_id);
			UNLOCK_TREEDB(9);
			if(process_id < (fork_length-1)) {
				wait_child_exit(9, 1);
			}
			CLOSE_TREEDB;
			return 9;
		}
		packed_pointer = TREEDB_pack_pointer(p);
		UNLOCK_TREEDB(9);
	}
	if(process_id < (fork_length-1)) {
		wait_child_exit(10, 0);
	}
	if(process_id) {
		WRITE_LOCK_TREEDB(11);
		printf("Process %d : 11. Free'ing chunk.\n", process_id);
		void * p = TREEDB_unpack_pointer(packed_pointer);
		TreedbFreeResult ret = TREEDB_NS(free)(runtime, p);
		UNLOCK_TREEDB(11);
		if(ret != eTreedbFreeSuccess) {
			printf("Process %d : 11. TREEDB_NS(free)() failed : %s.\n"
				, process_id, TreedbFreeResultDesc(ret));
			CLOSE_TREEDB;
			return 11;
		}
	}
	int retval = 0;
	if(!process_id) {
		READ_LOCK_TREEDB(12);
		printf("Root : 12. Checking size.\n");
		retval = TREEDB_NS(check_heap)(runtime);
		TREEDB_NSA(node_ptr_t) pFirstNode = 0;
		TREEDB_NSF(search_closest)(runtime, TREEDB_ALLOC_free_tree, 0
			, & pFirstNode);
		if(!pFirstNode) {
			printf("Root : 12. no free node!\n");
			UNLOCK_TREEDB(12);
			CLOSE_TREEDB;
			return 12;
		}
		TREEDB_NS(count_t) free_size2 = pFirstNode->m_uLength;
		TREEDB_NS(count_t) free_offs2 = TREEDB_pack_pointer(pFirstNode);
		if((free_offs != free_offs2) || (free_size != free_size2)) {
			printf("Root : 12. free size mismatch : free offset(%zu) "
				", expected %zu, free_size(%zu), expected %zu.\n"
				, (size_t)free_offs2, (size_t)free_offs, (size_t)free_size2
				, (size_t)free_size
			);
			UNLOCK_TREEDB(12);
			CLOSE_TREEDB;
			return 12;
		}
		UNLOCK_TREEDB(12);
	}
	CLOSE_TREEDB;
	return retval;
}
int create_treedb()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = &han;
	TreedbHeapStatus ret = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_int_tree_schema
		, fork_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite
			| eHeapCreateGroupRead | eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(ret) {
		printf("Root: TREEDB_NS(heap_create)() failed: %s.\n"
			, TreedbHeapStatusDesc(ret));
		return 1;
	}
	INT_TREE_NS(init)(runtime, INT_TREE);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);

	if(res2 == eTreedbSearchEmpty) {
		printf("Root: TREEDB_NS(heap_create)() no root node?\n");
		return 1;
	}
	free_size = pFirstNode->m_uLength;
	free_offs = TREEDB_pack_pointer(pFirstNode);
	UNLOCK_TREEDB(1);
	CLOSE_TREEDB;
	return 0;
}
#endif /* #if TREEDB_NBITS==-1 */
