/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */
/*
#undef TREEDB_DEBUG
#define TREEDB_DEBUG 1
*/
#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;

extern void impl_memmove(void * to, const void * from, size_t size);

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define TREEDB_ARRAY_MEMMOVE
#define ARRAY_NO_CLEARDEFS

#include <treedb/handle-array.h>

// 1b637403-525b-4335-ae3a-6b80a2bbfb47
UUID_DEFINE
( UUID_treedb_handle_array_int_schema
, 0x1b, 0x63, 0x74, 0x03
, 0x52, 0x5b
, 0x43, 0x35
, 0xae, 0x3a
, 0x6b, 0x80, 0xa2, 0xbb, 0xfb, 0x47
);
const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;
TREEDB_NS(runtime_ptr_t) runtime;
int test_num = 0;

void impl_memmove(void * to, const void * from, size_t size)
{
	printf("### memmove(%zu, %zu, %zu) ###\n"
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, to)
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, from)
		, size
	);
	memmove(to, from, size);
}

void dump()
{
	int n;
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	for(n = 0; n < (int)pa->m_count; ++n)
		cout << '[' << n << "] " << pa->m_items[n] << '\n';
	cout << flush;
}

typedef int (*PFN)();

struct TestEntry
{
	PFN fn;
	const char * name;
};
int clear_fn()
{
	cerr << "PACKED_ARRAY: " << PACKED_ARRAY << ", PACKED2_ARRAY: "
		<< PACKED2_ARRAY << '.' << endl;
	TREEDB_PAYLOAD_NS(clear)(runtime, PACKED2_ARRAY);
	return 0;
}
int push_front_many()
{
	TREEDB_PAYLOAD_NS(push_front_many)(runtime, PACKED2_ARRAY, 3);
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	uint16_t n;
	for(n = 0; n < 3; ++n)
		pa->m_items[n] = n - 3;
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != (n - 3)) {
			fprintf(stderr, "[%u] %d != %d\n"
				, n, pa->m_items[n], (n - 3));
			return 1;
		}
	return 0;
}
int push_back_many()
{
	TREEDB_PAYLOAD_NS(push_back_many)(runtime, PACKED2_ARRAY, 3);
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	uint16_t n;
	for(n = pa->m_count - 3; n < pa->m_count; ++n)
		pa->m_items[n] = n;
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != n)
			return 1;
	return 0;
}
int insert_many()
{
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	uint16_t n;
	for(n = 4; n < pa->m_count; ++n)
		pa->m_items[n] = n + 3;
	TREEDB_PAYLOAD_NS(insert_many)(runtime, PACKED2_ARRAY, 4, 3);
	pa = ARRAY;
	for(n = 4; n < 7; ++n)
		pa->m_items[n] = n;
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != n)
			return 1;
	return 0;
}
int pop_front_many()
{
	TREEDB_PAYLOAD_NS(pop_front_many)(runtime, PACKED2_ARRAY, 4);
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	uint16_t n;
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != (n + 4)) {
			fprintf(stderr, "[%u] %d != %d\n"
				, n, pa->m_items[n], (n + 4));
			return 1;
		}
	return 0;
}
int pop_back_many()
{
	TREEDB_PAYLOAD_NS(pop_back_many)(runtime, PACKED2_ARRAY, 4);
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	if(pa->m_count != 6) {
		fprintf(stderr, "Count %d != 6.\n", (int)pa->m_count);
		return 1;
	}
	uint16_t n;
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != n) {
			fprintf(stderr, "[%u] %d != %d\n"
				, n, pa->m_items[n], n);
			return 2;
		}
	return 0;
}
int remove_many()
{
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	uint16_t n;
	for(n = 7; n < (int)pa->m_count; ++n)
		pa->m_items[n] -= 3;
	TREEDB_PAYLOAD_NS(remove_many)(runtime, PACKED2_ARRAY, 4, 3);
	pa = ARRAY;
	if(pa->m_count != 7) {
		fprintf(stderr, "Count %d != 7.\n", (int)pa->m_count);
		return 1;
	}
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != n) {
			fprintf(stderr, "[%u] %d != %d\n"
				, n, pa->m_items[n], n);
			return 2;
		}
	return 0;
}
int remove_all()
{
	TREEDB_PAYLOAD_NS(clear)(runtime, PACKED2_ARRAY);
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	if(pa->m_count != 0)
		return 1;
	return 0;
}
int reverse_some_start()
{
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	uint16_t n;
	for(n = 0; n < 3; ++n)
		pa->m_items[n] = (2 - n);
	if(TREEDB_PAYLOAD_NS(reverse_some)(runtime, pa, 0, 3)) {
		fprintf(stderr, "reverse_some() failed.\n");
		return 1;
	}
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != n)
			return 1;
	return 0;
}
int reverse_some_middle()
{
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	uint16_t n;
	for(n = 4; n < 8; ++n)
		pa->m_items[n] = (11 - n);
	TREEDB_PAYLOAD_NS(reverse_some)(runtime, pa, 4, 4);
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != n)
			return 1;
	return 0;
}
int reverse_some_end()
{
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	uint16_t n;
	for(n = 6; n < pa->m_count; ++n)
		pa->m_items[n] = ((int)pa->m_count + 6 - n - 1);
	//dump();
	if(TREEDB_PAYLOAD_NS(reverse_some)(runtime, pa, 6, pa->m_count - 6)) {
		 cerr << "reverse_some)(runtime, pa, 6, " << (pa->m_count - 6)
			<< ") failed." << endl;
		 return 1;
	}
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != n)
			return 1;
	return 0;
}
int reverse_all()
{
	TREEDB_PAYLOAD_NS(reverse)(runtime, ARRAY);
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	uint16_t n;
	for(n = 0; n < pa->m_count; ++n)
		if(pa->m_items[n] != (int)(pa->m_count - n - 1))
			return 1;
	return 0;
}
#define TEST_ENTRY(name) {& name, #name}
TestEntry tests[] =
{
	TEST_ENTRY(clear_fn),
	TEST_ENTRY(push_front_many),
	TEST_ENTRY(push_back_many),
	TEST_ENTRY(insert_many),
	TEST_ENTRY(pop_front_many),
	TEST_ENTRY(pop_back_many),
	TEST_ENTRY(remove_many),
	TEST_ENTRY(remove_all),
	TEST_ENTRY(reverse_some_start),
	TEST_ENTRY(reverse_some_middle),
	TEST_ENTRY(reverse_some_end),
	TEST_ENTRY(reverse_all)
};

int setup_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	runtime = &han;
	int res1 = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_handle_array_int_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(res2 == eTreedbSearchEmpty)
		return 1;

	free_size = pFirstNode->m_uLength;

	if(TREEDB_PAYLOAD_NS(new)(runtime, PACKED2_ARRAY, 8, 4)) {
		fprintf(stderr, "Array creation failed\n");
		return 1;
	}

	unsigned int n;
	TREEDB_NS(dump_trees)(runtime);
	for(n = 0; n < 11; ++n)
		TREEDB_PAYLOAD_NS(push_back)(runtime, PACKED2_ARRAY, n);
	TREEDB_PAYLOAD_NS(ptr_t) pa = ARRAY;
	if(pa->m_count != 11) {
		cerr << "pa-m_count(" << pa->m_count << ") != 11." << endl;
		return 1;
	}

	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}
int run_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_handle_array_int_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("run_test: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	runtime = &han;

	if(!TREEDB_PAYLOAD_NS(remove)(runtime, PACKED2_ARRAY, 11)) {
		printf("run_test: remove(11) XPASS.\n");
		return 2;
	}
	if(TREEDB_PAYLOAD_NS(pop_back)(runtime, PACKED2_ARRAY)) {
		printf("run_test: pop_back() FAIL.\n");
		return 2;
	}

	printf("Testing %s\n", tests[test_num].name);
	if(tests[test_num].fn()) {
		fprintf(stderr, "Test \"%s\" failed.\n", tests[test_num].name);
		dump();
		return 3;
	}
	TREEDB_PAYLOAD_NS(free)(runtime, PACKED_ARRAY);

	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_read_lock)(runtime);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		fprintf(stderr, "Free size(%zu) != %zu.\n", (size_t)free_size
			, (size_t)pFirstNode->m_uLength);
		return 4;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

int test()
{
	int res = setup_test();
	if(res) {
		printf("setup_test failed.\n");
		return res;
	}
	res = run_test();
	if(res) {
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res;
	for(test_num = 0; test_num < (int)(sizeof(tests)/sizeof(tests[0]));
			++test_num) {
		res = test();
		unlink(test_file);
		if(res)
			break;
	}
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
