/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform - passing.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <string>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>
#include <v3c/thread.hpp>
#include <v3c/smart_ptr.hpp>

#include <v3c/file-lock.h>
#include <treedb/ns-types.h>

//#define V3C_AVL_DEBUG_AVL_GET
bool g_debug_treedb = false;
#undef TREEDB_DEBUG
#define TREEDB_DEBUG g_debug_treedb
/*#define TREEDB_MSYNC MS_ASYNC*/
#define NUM_ITEMS 100
#define NUM_ITERS 32

using namespace std;
using namespace v3c;

uint16_t int_pad_size = 0;

const unsigned int uCpus = sysconf(_SC_NPROCESSORS_ONLN);
const unsigned int g_uThreads = uCpus;
#define TREEDB_have_dump_trees

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP

#define TREEDB_OFFSET_FIXED_PAD(ctx, d) int_pad_size

// Int tree.
/* AVL - define macros we use. */
#define INT_TREE_NS(x) TREEDB_NS(AvlIntTree_##x)
#define INT_TREE_MEMBER m_int_tree
#define INT_TREE MAKE_AVL_TREE(INT_TREE_MEMBER)
#define PACKED_INT_TREE MAKE_PACKED_AVL_TREE(INT_TREE_NS, INT_TREE_MEMBER)
#define INT_TREE_NS_unpack_tree(x) \
	(INT_TREE_NS(tree_ptr_t))TREEDB_unpack_pointer(x)
#define INT_TREE_NS_unpack_node(x) \
	(INT_TREE_NS(node_ptr_t))TREEDB_unpack_pointer(x)

/* AVL- define macros avl.h uses - it will undefine them. */
#define AVL_NS(x) INT_TREE_NS(x)
#define AVL_TREE_MEMBER INT_TREE_MEMBER

#include <treedb/avl.h>
#include <treedb/comet-guids.h>

string node_status(TREEDB_NSA(node_ptr_t) pn)
{
	string ret;
	if(pn->m_offset_next || pn->m_offset_prev)
		ret = " ";
	if(pn->m_offset_prev) {
		ret += string_inline_printf("{<- %zu"
			, (size_t)TREEDB_packed_item_offset(pn->m_offset_prev));
		if(TREEDB_packed_node_isfree(pn->m_offset_prev))
			ret += "[free]}";
		else
			ret += "[used]}";
		if(pn->m_offset_next)
			ret += " ";
	}
	if(pn->m_offset_next) {
		ret += string_inline_printf("{-> %zu"
			, (size_t)TREEDB_packed_item_offset(pn->m_offset_next));
		if(TREEDB_packed_node_isfree(pn->m_offset_next))
			ret += "[free]}";
		else
			ret += "[used]}";
	}
	return ret;
}
string payload_node_status(INT_TREE_NS(node_ptr_t) pp)
{
	string ret = string_inline_printf("# %zu", (size_t)pp->m_count);
	return ret;
}
void dump_payload_node(TreeStreambuf & sb, TREEDB_NS(runtime_ptr_t) runtime
, INT_TREE_NS(node_ptr_t) node)
{
	if(!node)
		return;
	sb.add
		( string_inline_printf("<%zu> = %zu "
			, (size_t)TREEDB_pack_pointer(node), (size_t)node->m_item)
		+ payload_node_status(node)
		, true);
	sb.push();
	if(node->m_left) {
		sb.add("Left", !node->m_right);
		sb.push();
		dump_payload_node(sb, runtime
			, INT_TREE_NS_unpack_node(node->m_left));
		sb.pop();
	}
	if(node->m_right) {
		sb.add("Right", true);
		sb.push();
		dump_payload_node(sb, runtime
			, INT_TREE_NS_unpack_node(node->m_right));
		sb.pop();
	}
	sb.pop();
}
void dump_tree_node(TreeStreambuf & sb, TREEDB_NS(runtime_ptr_t) runtime
, TREEDB_NSA(node_ptr_t) node, bool is_used)
{
	if(!node)
		return;
	if(is_used) {
		INT_TREE_NS(node_ptr_t) pp = ((INT_TREE_NS(node_ptr_t))
			TREEDB_ALLOC_nodeptr_to_addr(node));
		sb.add
			( string_inline_printf("%zu : (%zu) = %zu "
				, (size_t)TREEDB_pack_pointer(node)
				, (size_t)(node->m_uLength), (size_t)pp->m_item)
			+ node_status(node)
			, true);
	}else
		sb.add
			( string_inline_printf("%zu : (%zu)"
				, (size_t)TREEDB_pack_pointer(node)
				, (size_t)(node->m_uLength))
			+ node_status(node)
			, true);
	sb.push();
	if(node->m_left) {
		sb.add("Left", !node->m_right);
		sb.push();
		dump_tree_node(sb, runtime
			, TREEDB_NSA_unpack_node(node->m_left)
			, is_used);
		sb.pop();
	}
	if(node->m_right) {
		sb.add("Right", true);
		sb.push();
		dump_tree_node(sb, runtime
			, TREEDB_NSA_unpack_node(node->m_right), is_used);
		sb.pop();
	}
	sb.pop();
}
void TREEDB_NS(dump_trees)(TREEDB_NS(runtime_ptr_t) runtime)
{
	if(!g_debug_treedb)
		return;
	TreeStreambuf sb(cout);
	sb.add("Blocks", true);
	sb.push();
	sb.add("Used", false);
	sb.push();
	dump_tree_node(sb, runtime
		, TREEDB_NSA_unpack_node(runtime->m_pTreedb->m_allocator.m_used.m_top)
		, true);
	sb.pop();
	sb.add("Free", false);
	sb.push();
	dump_tree_node(sb, runtime
		, TREEDB_NSA_unpack_node(runtime->m_pTreedb->m_allocator.m_free.m_top)
		, false);
	sb.pop();
	sb.add("Items", true);
	sb.push();
	dump_payload_node(sb, runtime
		, INT_TREE_NS_unpack_node
			(runtime->m_pTreedb->m_int_tree.m_top)
		);
	sb.pop();
	sb.pop();
}

#define STACKSIZE 0x800000 // 8Mb
class MyThread : public ThreadT< MyThread, unsigned int >
{
	friend class ThreadReturnValueTraits< MyThread, unsigned int >;
public:
	inline MyThread(char c, unsigned int uHop, void * pStack)
	: m_name(string("Worker ") + c)
	, m_test_file(string("test" TREEDB_NBITS_STRING "-") + c + ".3db")
	, m_uHop(uHop)
	, m_pStack(pStack)
	, m_free_size(0)
	, m_iter(0)
	{
		//start(); <- more trouble than it's worth calling from here.
	}
	virtual ~MyThread()
	{
		munmap(m_pStack, STACKSIZE);
	}
protected:
	friend class ThreadReturnValueTraits< MyThread, int >;
	inline unsigned int work();
	inline unsigned int test();
	inline unsigned int fill_test();
	inline unsigned int empty_test();
public:
	string m_name, m_test_file;
	unsigned int m_uHop;
	void * m_pStack;
	TREEDB_NS(count_t) m_free_size;
	unsigned int m_iter;
	static unsigned int m_random[NUM_ITEMS];
private:
	MyThread(const MyThread &);
	MyThread & operator =(const MyThread &);
};

unsigned int MyThread::m_random[NUM_ITEMS];

inline unsigned int MyThread::fill_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = &han;
	int res1 = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_int_tree_schema
		, m_test_file.c_str()
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		TREEDB_cout << m_name << " : TREEDB_NS(heap_create)() failed: " << res1
			<< endl;
		return 1;
	}
	INT_TREE_NS(init)(runtime, INT_TREE);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);

	if(res2 == eTreedbSearchEmpty)
		return 1;
	m_free_size = pFirstNode->m_uLength;

	int n;
	for(n = 0; n < NUM_ITEMS; ++n) {
		TREEDB_NS(dump_trees)(runtime);
		if(eTreedbInsertSuccess != INT_TREE_NS(insert)
				(runtime, PACKED_INT_TREE, n, 0))
			break;
	}
	TREEDB_printf("Allocated %d nodes.\n", (n == NUM_ITEMS) ? NUM_ITEMS : n + 1);
	if(n == NUM_ITEMS)
		TREEDB_NS(dump_trees)(runtime);
	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

inline unsigned int MyThread::empty_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_int_tree_schema
		, m_test_file.c_str()
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		TREEDB_cout << m_name << " : TREEDB_NS(heap_open)() failed : " << res
			<< '.' << endl;
		return 1;
	}
	TREEDB_NS(runtime_ptr_t) runtime = &han;
	TREEDB_NS(heap_unlock)(runtime);

	INT_TREE_NS(node_t) * p1;
	int n;
	unsigned int v, ret = 0;
	while(true) {
		TREEDB_NS(heap_write_lock)(runtime);
		TREEDB_NS(dump_trees)(runtime);
		n = (int)INT_TREE_NS(count)(runtime, INT_TREE);
		if(!n) {
			TREEDB_NS(heap_unlock)(runtime);
			break;
		}
		v = m_random[(++m_iter) % NUM_ITEMS] % n;
		// Look up the item at index v.
		p1 = INT_TREE_NS(index)(runtime, INT_TREE, v);
		if(!p1) {
			TREEDB_cout << m_name << " : get failed." << endl;
			TREEDB_NS(heap_unlock)(runtime);
			break;
		}
		if(p1->m_item != (int)v) {
			TREEDB_cout << m_name << " : get wanted "
				<< v << " but returned " << p1->m_item << endl;
			TREEDB_NS(heap_unlock)(runtime);
			ret = 1;
			break;
		}
		INT_TREE_NS(node_t) * d1 = p1;
		p1 = INT_TREE_NS_unpack_node(p1->m_next);
		INT_TREE_NS(delete_node)(runtime, PACKED_INT_TREE, d1);
		for(; p1; ++v, p1 = INT_TREE_NS_unpack_node(p1->m_next))
			p1->m_item = v;
		--n;

		TREEDB_NS(heap_unlock)(runtime);
	}
	if(!ret) {
		TREEDB_NS(heap_read_lock)(runtime);
		TREEDB_NS(dump_trees)(runtime);
		TREEDB_NSA(node_ptr_t) pFirstNode = 0;
		TREEDB_NSF(search_closest)
			(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
		if(!pFirstNode) {
			TREEDB_cout << m_name << " : no free nodes!" << endl;
			ret = 2;
		}else if(m_free_size != pFirstNode->m_uLength) {
			TREEDB_cout << m_name << " : wanted free_size == " << m_free_size
				<< " but got " << pFirstNode->m_uLength << " instead." << endl;
			ret = 3;
		}else if(pFirstNode->m_offset_next) {
			TREEDB_cout << m_name << " : more than one free node." << endl;
			ret = 4;
		}
		TREEDB_NS(heap_unlock)(runtime);
	}
	TREEDB_NS(heap_close)(&han);
	return ret;
}

inline unsigned int MyThread::test()
{
	srand(1);
	int res = 0;
	do {
		res = fill_test();
		if(res) {
			printf("fill test failed.\n");
			break;
		}
		res = empty_test();
		if(res) {
			printf("empty test failed.\n");
			break;
		}
	} while(0);
	unlink(m_test_file.c_str());
	return res;
}
inline unsigned int MyThread::work()
{
	uint16_t n, count = 0;
	cout << m_name << " running on CPU " << sched_getcpu() << '.' << endl;
	for(n = m_uHop; n < NUM_ITERS; n += g_uThreads) {
		int_pad_size = n;

		int res = test();
		if(res) {
			cout << "Pad " << n << " FAILED : running with dumps enabled.\n";
			cout << "#########################################################"
				"###########" << endl;
			g_debug_treedb = true;
			test();
			cout << "#########################################################"
				"###########" << endl;
			if(n) {
				int_pad_size = n-1;
				cout << "And here's pad " << int_pad_size
					<< " that worked, for comparison : "
					"running with dumps enabled." << endl;
				test();
			}
			return res;
		}
		++count;
	}
	return count;
}

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	printf("###########################################################"
		"#########\n");
	cout << "Running " TREEDB_NBITS_STRING " bit tests with " << g_uThreads
		<< " threads on " << uCpus << " cpus." << endl;

	unsigned int n;

	// clang doesn't like this: variable length array of non-POD element
	//smart_ptr< MyThread > threads[g_uThreads];
	void * vthreads[g_uThreads];
	for(n = 0; n < g_uThreads; ++n)
		vthreads[n] = 0;
	smart_ptr< MyThread > *threads = (smart_ptr< MyThread > *)vthreads;

	char c = 'A';
	for(n = 0; n < g_uThreads; ++n, ++c) {
		// Apparently, if we want processor affinity, we need a thread
		// attribute. If we use a thread attribute, we need to set up our own
		// stack (?)
		void * stk = mmap(NULL, STACKSIZE,
						PROT_READ | PROT_WRITE | PROT_EXEC,
						MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
		if(stk == MAP_FAILED) {
			cerr << '[' << n << "] mmap failed." << endl;
			break;
		}
		MyThread * p = new MyThread(c, n, stk);
		if(!p) {
			cerr << '[' << n << "] new failed." << endl;
			break;
		}
		threads[n] = p;
	}
	// Set their thread attributes and start them.
	Thread::Attr::CpuBitSet cpu_set;
	cpu_set[0] = true;
	Thread::Attr a;
	Thread::Attr::StackInfo si;
	si.size = STACKSIZE;

	for(n = 0; n < NUM_ITEMS; ++n)
		MyThread::m_random[n] = rand();

	for(n = 0; n < g_uThreads; ++n) {
		MyThread & t = *threads[n];
		cpu_set[n % uCpus] = true;
		a.setAffinity(cpu_set);
		si.address = t.m_pStack;
		a.setStackInfo(si);
		// According to the pthread docs, if we set up our own stack, the guard
		// size is ignored (?).
		a.setGuardSize(getpagesize());
		threads[n]->start(&a.handle());
		cpu_set[n % uCpus] = false;
	}
	int res = 0;
	for(n = 0; n < g_uThreads; ++n) {
		threads[n]->join();
		cout << "Thread \"" << threads[n]->m_name << "\" returned "
			<< threads[n]->value() << '.' << endl;
		if(!res)
			res = (threads[n]->value() != 0) ? 0 : 1;
	}
	if(res)
		return res;
	unsigned int count = 0;
	for(n = 0; n < g_uThreads; ++n) {
		if(n)
			cout << " + ";
		cout << threads[n]->value();
		count += threads[n]->value();
	}
	if(count == NUM_ITERS) {
		cout << " == " << NUM_ITERS << '.' << endl;
		return 0;
	}else{
		cout << " != " << NUM_ITERS << '.' << endl;
		return 1;
	}
}
#endif /* #if #if TREEDB_NBITS==-1 */
