/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform.\n");
	return 0;
}
#else
//#undef TREEDB_DEBUG
//#define TREEDB_DEBUG 4

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <iomanip>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>
#include <treedb/comet-guids.h>
#include <signal.h> // signal()

using namespace std;
using namespace v3c;

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP

// Int tree.
/* AVL - define macros we use. */
#define INT_TREE_NS(x) TREEDB_NS(AvlIntTree_##x)
#define INT_TREE_MEMBER m_int_tree
#define INT_TREE MAKE_AVL_TREE(INT_TREE_MEMBER)
#define PACKED_INT_TREE MAKE_PACKED_AVL_TREE(INT_TREE_NS, INT_TREE_MEMBER)
#define INT_TREE_NS_unpack_tree(x) \
	(INT_TREE_NS(tree_ptr_t))TREEDB_unpack_pointer(x)
#define INT_TREE_NS_unpack_node(x) \
	(INT_TREE_NS(node_ptr_t))TREEDB_unpack_pointer(x)
#define INT_TREE_NS_count_t TREEDB_NS_CONCAT(uint,_t)

/* AVL- define macros avl.h uses - it will undefine them. */
#define AVL_NS(x) INT_TREE_NS(x)
#define AVL_TREE_MEMBER INT_TREE_MEMBER
#define AVL_count_t INT_TREE_NS_count_t

#include <treedb/avl.h>

const char * test_file = "test.3db";
INT_TREE_NS_count_t free_size;

TREEDB_NS(runtime_t) default_treedb_runtime = TREEDB_STORAGE_RUNTIME_INIT;
TREEDB_NS(runtime_t) treedb_runtime;
TREEDB_NS(runtime_ptr_t) runtime;
int test_num = 0;

#define ITEMS_COUNT 21

INT_TREE_NS_count_t lk = 0, expected_lower_count = 0, lower_count = 0;
INT_TREE_NS_count_t uk = 0, expected_upper_count = 0, upper_count = 0;
INT_TREE_NS_count_t removed = 0;
#if TREEDB_DEBUG
string int_tree_node_status(INT_TREE_NS(node_ptr_t) pp)
{
	string ret = string_inline_printf("(%lu)", (size_t)pp->m_count);
	return ret;
}
void dump_int_tree_node(TreeStreambuf & sb, TREEDB_NS(runtime_ptr_t) runtime
, INT_TREE_NS(node_ptr_t) node)
{
	if(!node)
		return;
	sb.add
		( string_inline_printf("<%lu> = %lu = "
			, (size_t)TREEDB_pack_pointer(node), (size_t)node->m_item)
		+ int_tree_node_status(node)
		, true);
	sb.push();
	if(node->m_left) {
		sb.add("Left", !node->m_right);
		sb.push();
		dump_int_tree_node(sb, runtime
			, INT_TREE_NS_unpack_node(node->m_left));
		sb.pop();
	}
	if(node->m_right) {
		sb.add("Right", true);
		sb.push();
		dump_int_tree_node(sb, runtime
			, INT_TREE_NS_unpack_node(node->m_right));
		sb.pop();
	}
	sb.pop();
}
#endif /* #if TREEDB_DEBUG */
void dump()
{
	unsigned int n, count = INT_TREE_NS(count)(runtime, INT_TREE);
	INT_TREE_NS(node_ptr_t) pn;
	cout << "dump :" << setfill('0');
	for(n = 0; n < count; ++n) {
		if(n && (n == lower_count))
			cout << ' ';
		cout << ' ' << setw(2) << n;
	}
	cout << setfill(' ') << "\n      ";
	for(n = 0; n < count; ++n) {
		if(n && (n == lower_count))
			cout << ' ';
		pn = INT_TREE_NS(index)(runtime, INT_TREE, n);
		cout << ' ' << setw(2) << pn->m_item;
	}
	cout << endl;
#if TREEDB_DEBUG
	TreeStreambuf sb(cout);
	sb.add("Items", true);
	sb.push();
	dump_int_tree_node(sb, runtime
		, INT_TREE_NS_unpack_node
			(runtime->m_pTreedb->m_int_tree.m_top)
		);
	sb.pop();
#endif /* #if TREEDB_DEBUG */
}

int fill()
{
	unlink(test_file);
	treedb_runtime = default_treedb_runtime;
	runtime = & treedb_runtime;
	int res1 = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_int_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(res2 == eTreedbSearchEmpty)
		return 1;

	free_size = pFirstNode->m_uLength;

	INT_TREE_NS(tree_t) tree_init = AVL_TMPL_init;
	* INT_TREE = tree_init;

	unsigned int n;
	TREEDB_NS(dump_trees)(runtime);
	for(n = 0; n < ITEMS_COUNT; ++n) {
		INT_TREE_NS(insert)(runtime, PACKED_INT_TREE
			, n, 0);
	}
	TreedbHeapStatus hr = TREEDB_NS(heap_unlock)(runtime);
	if(hr != eTreedbHeapSuccess) {
		cout << "heap_unlock failed : " << TreedbHeapStatusDesc(hr)
			<< '.' << endl;
	}
	TreedbHeapStatus hr2 = TREEDB_NS(heap_close)(runtime);
	if(hr2 != eTreedbHeapSuccess) {
		cout << "heap_unlock failed : " << TreedbHeapStatusDesc(hr2)
			<< '.' << endl;
	}
	runtime = 0;
	return hr || hr2;
}
int test_prefix()
{
	int res = fill();
	if(res) {
		cout << "fill failed." << endl;
		return res;
	}
	treedb_runtime = default_treedb_runtime;
	runtime = & treedb_runtime;
	res = TREEDB_NS(heap_open)
		( runtime
		, UUID_treedb_int_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		cout << "empty_test: TREEDB_NS(heap_open)() failed: " << res << '.'
			<< endl;
		return 1;
	}
	if(INT_TREE_NS(index)(runtime, INT_TREE, ITEMS_COUNT)) {
		cout << "empty_test: get(ITEMS_COUNT) XPASS." << endl;
		return 2;
	}
	return 0;
}
int test_suffix()
{
	if(!runtime)
		return 1;
	INT_TREE_NS(clear)(runtime, PACKED_INT_TREE);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(!pFirstNode) {
		cout << "No free node." << endl;
		return 4;
	}
	if(free_size != pFirstNode->m_uLength) {
		cout << "Free size(" << free_size << ") != "
			<< pFirstNode->m_uLength << '.' << endl;
		return 4;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	unlink(test_file);
	runtime = 0;
	return 0;
}

enum TestResult
{ ePass
, eFail
, eFailCleanup
};

TestResult clear_test()
{
	if(test_prefix())
		return eFail;
	INT_TREE_NS(clear)(runtime, PACKED_INT_TREE);
	if(test_suffix())
		return eFail;
	return ePass;
}
int make_all_even()
{
	unsigned int n;
	INT_TREE_NS(node_ptr_t) pn
		= INT_TREE_NS(unpack_node)(runtime, INT_TREE->m_head);
	for(n = 0; pn && (n < ITEMS_COUNT); ++n) {
		pn->m_item = n * 2;
		pn = INT_TREE_NS(unpack_node)(runtime, pn->m_next);
	}
	if(n != ITEMS_COUNT)
		return 1;
	return 0;
}
static const TreedbKeyDistanceFlags kd_include
	= MakeKeyDistanceFlags(eKeyDistanceIncludeLower | eKeyDistanceIncludeUpper);
static const TreedbKeyDistanceFlags kd_exclude = MakeKeyDistanceFlags(0);
#define CALL_KEY_DISTANCE(name, kl, ku, flags, kd) \
TREEDB_cout4 << "[" name "] Calling key_distance(lk " << (kl) << ", uk " << ku \
	<< ", [" << TreedbKeyDistanceFlagsDesc(flags) << "])\n"; \
res = INT_TREE_NS(key_distance)(runtime, INT_TREE, kl, ku \
	, flags, & kd); \
if(res) { \
	TREEDB_cout4 << '[' << u.n << "] key_distance" \
		<< "( lk " << (kl) << ", uk " << (ku) \
		<< ", [" \
		<< TreedbKeyDistanceFlagsDesc(flags) \
		<< "]) failed : " << res << '.' << endl; \
	break; \
} \
if(kd < 0) { \
	TREEDB_cout4 << '[' << u.n << "] key_distance" \
		<< "( lk " << (kl) << ", uk " << (ku) \
		<< ", [" \
		<< TreedbKeyDistanceFlagsDesc(flags) \
		<< "]) returned : " << kd << '.' << endl; \
	break; \
} \
TREEDB_cout4 << "[" name "] key_distance(lk " << (kl) << ", uk " << ku \
	<< ", [" << TreedbKeyDistanceFlagsDesc(flags) << "]) returned " << kd \
	<< ".\n"

typedef union {
	unsigned int n;
	struct {
		unsigned int start:2;
		unsigned int single:1;
	};
} remove_by_get_U;
enum { remove_by_get_TOTAL = 2<<3 };
inline ostream & operator <<(ostream & s, const remove_by_get_U & u)
{
	s << "start " << u.start
		<< ", single " << u.single
	;
	return s;
}
/* Removes the items inclusively. */
int remove_by_get_impl(INT_TREE_NS_count_t lk, INT_TREE_NS_count_t uk)
{
	ssize_t n, count = uk - lk + 1;
	for(n = 0; n < count; ++n) {
		INT_TREE_NS(node_ptr_t) pn = INT_TREE_NS(index)(runtime
			, INT_TREE, lk);
		if(!pn)
			return 1;
		INT_TREE_NS(delete_node)(runtime, PACKED_INT_TREE, pn);
	}
	return 0;
}
TestResult remove_by_get_test()
{
	remove_by_get_U u;
	for(u.n = 15; u.n < remove_by_get_TOTAL; ++u.n) {
		if(u.start == 3)
			continue;
		if(test_prefix())
			return eFail;
		int res;
		ssize_t kd;
		bool good = false;
		do {
			lk = (ITEMS_COUNT / 3) * u.start;
			uk = lk + (ITEMS_COUNT / 3);
			if(uk == ITEMS_COUNT)
				--uk;
			if(u.single)
				lk = uk;
			expected_lower_count = lk;
			CALL_KEY_DISTANCE("remove_by_get - remove", lk, uk, kd_include, kd);
			removed = (INT_TREE_NS_count_t)kd;
			expected_upper_count = ITEMS_COUNT - expected_lower_count - removed;
			/*dump();*/
			res = remove_by_get_impl(lk, uk);
			if(res) {
				cout << '[' << u.n << "] remove_by_get"
					<< "( lk " << lk << ", uk " << uk
					<< ", " << u << ") failed : " << res << '.' << endl;
				return eFailCleanup;
			}
			if(lk) {
				CALL_KEY_DISTANCE("remove_by_get - lower_count", 0, lk
					, kd_include, kd);
				lower_count = (INT_TREE_NS_count_t)kd;
			}else
				lower_count = 0;
			upper_count = ITEMS_COUNT - lower_count - removed;

			if((INT_TREE_NS_count_t)lower_count != expected_lower_count) {
				cout << "remove_by_get failed : lower count mismatch." << endl;
					break;
			}
			if((INT_TREE_NS_count_t)upper_count != expected_upper_count) {
				cout << "remove_by_get failed : upper count mismatch." << endl;
					break;
			}
			INT_TREE_NS_count_t n;
			INT_TREE_NS(node_ptr_t) pn;
			for(n = 0; n < lower_count; ++n) {
				pn = INT_TREE_NS(search)(runtime, INT_TREE, n);
				if(!pn) {
					cout << "remove_by_get lower match failed("
						<< n << ")." << endl;
					break;
				}
			}
			if(n != lower_count)
				break;
			for(; n < (lower_count + removed); ++n) {
				pn = INT_TREE_NS(search)(runtime, INT_TREE, n);
				if(pn) {
					cout << "remove_by_get middle match succeeded("
						<< n << ")." << endl;
					break;
				}
			}
			if(n != (lower_count + removed))
				break;
			for(; n < ITEMS_COUNT; ++n) {
				pn = INT_TREE_NS(search)(runtime, INT_TREE, n);
				if(!pn) {
					cout << "remove_by_get upper match failed("
						<< n << ")." << endl;
					break;
				}
			}
			if(n != ITEMS_COUNT)
				break;
			good = true;
		} while(false);
		if(!good) {
			cout << "[" << u.n << "] remove_by_get : tried "
				<< "(lk " << lk << ", uk " << uk
				<< ", " << u << ", lower_count " << expected_lower_count
				<< ", removed " << removed
				<< ", upper_count " << expected_upper_count << ')' << endl;
			dump();
			cout << "remove_by_get : got "
				<< "(lk " << lk << ", uk " << uk
				<< ", " << u << ", lower_count " << lower_count
				<< ", removed " << removed
				<< ", upper_count " << upper_count << ')' << endl;
			return eFailCleanup;
		}
#ifdef TREEDB_DEBUG
		dump();
#endif
		if(test_suffix())
			return eFail;
	}
	return ePass;
}
typedef union {
	unsigned int n;
	struct {
		unsigned int include_lower:1;
		unsigned int include_upper:1;
		unsigned int exact_lower:1;
		unsigned int exact_upper:1;
		unsigned int start:2;
		unsigned int single:1;
		unsigned int fail_exact_lower:1;
		unsigned int fail_exact_upper:1;
	};
	TreedbKeyDistanceFlags key_distance;
} remove_many_U;
enum { remove_many_TOTAL = 2<<9 };
inline ostream & operator <<(ostream & s, const remove_many_U & u)
{
	s
		<< "exact_lower " << u.exact_lower
		<< ", include_lower " << u.include_lower
		<< ", exact_upper " << u.exact_upper
		<< ", include_upper " << u.include_upper
		<< ", fail_exact_lower " << u.fail_exact_lower
		<< ", fail_exact_upper " << u.fail_exact_upper
	;
	return s;
}
TestResult remove_many_test()
{
	remove_many_U u, u2;
	bool should_fail;
	for(u2.n = 0; u2.n < remove_many_TOTAL; ++u2.n) {
		u = u2;
		// Eliminate tests that don't make sense. Earlier tests simplify
		// succeeding tests.
		if(u.start == 3)
			continue; // We want begin, middle, end.
		if(u.fail_exact_lower && (! u.exact_lower))
			continue;
		if(u.fail_exact_upper && (! u.exact_upper))
			continue;
		if((!u.exact_lower || u.fail_exact_lower) && (!u.start))
			continue;
		if((!u.exact_upper) && (u.start == 2))
			continue;
		if(!u.exact_lower && u.include_lower)
			continue;
		if(!u.exact_upper && u.include_upper)
			continue;

		should_fail = (u.fail_exact_lower || u.fail_exact_upper);
		lk = (ITEMS_COUNT / 3) * u.start * 2;
		uk = lk + (ITEMS_COUNT / 3) * 2;
		if(uk == ITEMS_COUNT*2)
			uk -= 2;
		if(u.single)
			uk = lk;
		if((!u.exact_lower) || u.fail_exact_lower)
			--lk;
		if((!u.exact_upper) || u.fail_exact_upper)
			++uk;
		if(test_prefix())
			return eFail;
		if(make_all_even())
			return eFailCleanup;
		int res;
		ssize_t kd;
		bool good = false;
		do {
			if(!should_fail) {
				if((!lk) && u.include_lower)
					expected_lower_count = 0;
				else{
					CALL_KEY_DISTANCE("remove_many - expected_lower_count", 0,lk
						, MakeKeyDistanceFlags(eKeyDistanceIncludeLower
							| eKeyDistanceExactLower
							| (eKeyDistanceIncludeUpper
								* (!u.include_lower
									&& !((lk == uk) && u.include_upper)))
							| (eKeyDistanceExactUpper * u.exact_lower)
							)
						, kd);
					expected_lower_count = (INT_TREE_NS_count_t)kd;
				}
				CALL_KEY_DISTANCE("remove_many - removed", lk, uk
					, u.key_distance, kd);
				removed = (INT_TREE_NS_count_t)kd;
				expected_upper_count = ITEMS_COUNT - expected_lower_count
					- removed;
			}
			res = INT_TREE_NS(remove_many)(runtime, PACKED_INT_TREE, lk, uk
				, u.key_distance);
			if(should_fail && !res) {
				cout << '[' << setw(2) << u.n << "] remove_many : XPASS."
					<< endl;
				break;
			}
			if(should_fail) {
				good = true;
				break;
			}
			if(res) {
				cout << "remove_many failed(" << res << ")." << endl;
				break;
			}
			if((!lk) && u.include_lower)
				lower_count = 0;
			else{
				CALL_KEY_DISTANCE("remove_many - lower_count", 0, lk
					, MakeKeyDistanceFlags(eKeyDistanceIncludeLower
						| (eKeyDistanceIncludeUpper
							* (!u.include_lower
								&& !((lk == uk) && u.include_upper)))
						)
					, kd);
				lower_count = (INT_TREE_NS_count_t)kd;
			}
			upper_count = ITEMS_COUNT - lower_count - removed;

			if((INT_TREE_NS_count_t)expected_lower_count != lower_count) {
				cout << "remove_many failed : bad lower count." << endl;
				break;
			}
			if((INT_TREE_NS_count_t)expected_upper_count != upper_count) {
				cout << "remove_many failed : bad upper count." << endl;
				break;
			}
			INT_TREE_NS_count_t n;
			INT_TREE_NS(node_ptr_t) pn;
			for(n = 0; n < lower_count; ++n) {
				pn = INT_TREE_NS(search)(runtime, INT_TREE, n*2);
				if(!pn) {
					cout << "remove_many lower match failed("
						<< n*2 << ")." << endl;
					break;
				}
			}
			if(n != lower_count)
				break;
			for(; n < (lower_count + removed); ++n) {
				pn = INT_TREE_NS(search)(runtime, INT_TREE, n*2);
				if(pn) {
					cout << "remove_many middle match succeeded("
						<< n*2 << ")." << endl;
					break;
				}
			}
			if(n != (lower_count + removed))
				break;
			for(; n < ITEMS_COUNT; ++n) {
				pn = INT_TREE_NS(search)(runtime, INT_TREE, n*2);
				if(!pn) {
					cout << "remove_many upper match failed("
						<< n*2 << ")." << endl;
					break;
				}
			}
			if(n != ITEMS_COUNT)
				break;
			good = true;
		} while(false);
		if(!good) {
			cout << "[" << u.n << "] remove_many: tried "
				<< "(lk " << lk << ", uk " << uk
				<< ", " << u << ", lower_count " << expected_lower_count
				<< ", removed " << removed
				<< ", upper_count " << expected_upper_count << ')' << endl;
			dump();
			cout << '[' << u.n << "] remove_many : got "
				<< "(lk " << lk << ", uk " << uk
				<< ", " << u << ", lower_count " << lower_count
				<< ", removed " << removed
				<< ", upper_count " << upper_count << ')' << endl;
			return eFailCleanup;
		}
		if(test_suffix())
			return eFail;
	}
	return ePass;
}
typedef union {
	unsigned int n;
	struct {
		unsigned int include_lower:1;
		unsigned int include_upper:1;
		unsigned int start:2;
		unsigned int single:1;
		unsigned int none:1;
	};
	TreedbNodeDistanceFlags node_distance;
} remove_between_U;
enum { remove_between_TOTAL = 2<<6 };
inline ostream & operator <<(ostream & s, const remove_between_U & u)
{
	s << "start " << u.start
		<< ", single " << u.single
		<< ", none " << u.none
		<< ", include_lower " << u.include_lower
		<< ", include_upper " << u.include_upper
	;
	return s;
}
TestResult remove_between_test()
{
	remove_between_U u;
	for(u.n = 0; u.n < remove_between_TOTAL; ++u.n) {
		if(u.start == 3)
			continue; // We want begin, middle, end.
		if((!u.include_lower) && (u.start == 0))
			continue;
		if((!u.include_upper) && u.single && (u.start == 2))
			continue;
		if(u.single && u.none)
			continue;
		if(u.include_lower && u.include_upper && u.none)
			continue;
		if(test_prefix())
			return eFail;
		int res;
		ssize_t kd;
		bool good = false;
		do {
			lk = (ITEMS_COUNT / 3) * u.start;
			uk = lk + (ITEMS_COUNT / 3);
			if(uk == ITEMS_COUNT)
				uk -= 1;
			if(u.single || u.none)
				uk = lk;
			if(lk) {
				CALL_KEY_DISTANCE("remove_between - expected_lower_count", 0, lk
					, MakeKeyDistanceFlags(eNodeDistanceIncludeLower
						| eKeyDistanceExactLower
						| (eKeyDistanceIncludeUpper * !u.include_lower)
						)
					, kd);
				expected_lower_count = (INT_TREE_NS_count_t)kd;
			}else
				expected_lower_count = 0;
			CALL_KEY_DISTANCE("remove_between - removed", lk, uk
				, MakeKeyDistanceFlags(
					  (eKeyDistanceIncludeLower * u.include_lower)
					| (eKeyDistanceIncludeUpper * u.include_upper)
					)
				, kd);
			removed = (INT_TREE_NS_count_t)kd;
			expected_upper_count = ITEMS_COUNT - expected_lower_count
				- removed;

			INT_TREE_NS(node_ptr_t) pnlk
				= INT_TREE_NS(search)(runtime, INT_TREE, lk);
			if(!pnlk)
				return eFailCleanup;
			INT_TREE_NS(node_ptr_t) pnuk
				= INT_TREE_NS(search)(runtime, INT_TREE, uk);
			if(!pnuk) {
				cout << "remove_between failed : pnuk is NULL!" << endl;
				break;
			}
			res = INT_TREE_NS(remove_between)(runtime, PACKED_INT_TREE
				, TREEDB_pack_pointer(pnlk), TREEDB_pack_pointer(pnuk)
				, u.node_distance);
			if(res) {
				cout << "remove_between failed(" << res << ")." << endl;
				break;
			}
			if(lk) {
				CALL_KEY_DISTANCE("remove_between - lower_count", 0, lk
					, MakeKeyDistanceFlags(eNodeDistanceIncludeLower
						| eKeyDistanceExactLower
						| (eKeyDistanceIncludeUpper * !u.include_lower)
						)
					, kd);
				lower_count = (INT_TREE_NS_count_t)kd;
			}else
				lower_count = 0;
			upper_count = ITEMS_COUNT - lower_count - removed;

			if((INT_TREE_NS_count_t)lower_count != expected_lower_count) {
				cout << "remove_between failed : bad lower count." << endl;
				break;
			}
			if((INT_TREE_NS_count_t)upper_count != expected_upper_count) {
				cout << "remove_between failed : bad upper count." << endl;
				break;
			}
			INT_TREE_NS_count_t n;
			INT_TREE_NS(node_ptr_t) pn;
			for(n = 0; n < lower_count; ++n) {
				pn = INT_TREE_NS(search)(runtime, INT_TREE, n);
				if(!pn) {
					cout << "remove_between lower match failed("
						<< n << ")." << endl;
					break;
				}
			}
			if(n != lower_count)
				break;
			for(; n < (lower_count + removed); ++n) {
				pn = INT_TREE_NS(search)(runtime, INT_TREE, n);
				if(pn) {
					cout << "remove_between middle match succeeded("
						<< n << ")." << endl;
					break;
				}
			}
			if(n != (lower_count + removed))
				break;
			for(; n < ITEMS_COUNT; ++n) {
				pn = INT_TREE_NS(search)(runtime, INT_TREE, n);
				if(!pn) {
					cout << "remove_between upper match failed("
						<< n << ")." << endl;
					break;
				}
			}
			if(n != ITEMS_COUNT)
				break;
			good = true;
		} while(false);
		if(!good) {
			cout << "[" << u.n << "] remove_between : tried "
				<< "(lk " << lk << ", uk " << uk
				<< ", " << u << ", lower_count " << expected_lower_count
				<< ", removed " << removed
				<< ", upper_count " << expected_upper_count << ')' << endl;
			dump();
			cout << '[' << u.n << "] remove_between : got "
				<< "(lk " << lk << ", uk " << uk
				<< ", " << u << ", lower_count " << lower_count
				<< ", removed " << removed
				<< ", upper_count " << upper_count << ')' << endl;
			return eFailCleanup;
		}
		if(test_suffix())
			return eFail;
	}
	return ePass;
}
typedef TestResult (*PFN)();
struct TestEntry
{
	PFN fn;
	const char * name;
};
#define TEST_ENTRY(name) {& name, #name}
TestEntry tests[] =
{
	TEST_ENTRY(clear_test),
	TEST_ENTRY(remove_by_get_test),
	TEST_ENTRY(remove_many_test),
	TEST_ENTRY(remove_between_test),
};

void fatal_repo_error(const char * msg)
{
	fflush(stdout);
	fflush(stderr);
	cout << flush;
	cerr << flush;
	cout << "Terminating because of " << msg << '.' << endl;
	unlink(test_file);

	exit(1);
}

void sigsegv_handler(int signum)
{
	fatal_repo_error("sigsegv");
	signal(signum, SIG_DFL);
	raise(signum);
}

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	signal(SIGSEGV, & sigsegv_handler);

	TestResult res;
	for(test_num = 0; test_num < (int)(sizeof(tests)/sizeof(tests[0]));
			++test_num) {
		cout << "Testing " << tests[test_num].name << '.' << endl;
		res = tests[test_num].fn();
		if(res != ePass) {
			cout << '\t' << tests[test_num].name << " failed." << endl;
			if(res == eFailCleanup) {
				test_suffix();
			}
			return 1;
		}
		cout << '\t' << tests[test_num].name << " passed." << endl;
	}
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
