#!/bin/sh
# Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)
# License: LGPLv3.  See LICENSE.txt for the full license.

program="$1"

if test "$2" = "debug" ; then
	debug="-d"
else
	debug=""
fi
if test -n "$3" ; then
	size="$3"
else
	size="8K"
fi
echo "Starting..."

if test -d fuse; then
	# Try to recover from previous interrupted or failed runs.
	fusermount -u fuse 2>/dev/null || true
	rm -fr fuse 2>/dev/null || true
fi
mkdir fuse || {
	echo "fuse directory creation failed - aborting."
	exit 1
}

daemon="../treedb-malloc-daemon$debug"
TREEDB_FUSE_TMPDIR="$(pwd)" $daemon fuse

while test ! -f fuse/control ; do
	echo "No fuse/control yet..."
	sleep 1
done
echo "Got fuse/control."

# [    9.227322] ACPI Warning: 0x000000000000e040-0x000000000000e05f SystemIO conflicts with Region \_SB_.PCI0.SBUS.SMBI 1 (20120320/utaddress-251)
export V3C_TREEDB_MALLOC_FUSE_DIR="$PWD/fuse"
export LD_LIBRAY_PATH="./libs:../.libs"
./$program $size 1>&2
res=$?
if test $res -ne 0 ; then
	echo "$program failed ($res)."
	if ! test -f fuse/control; then
		echo "treedb-malloc-daemon$debug STOPPED."
	fi
fi
fusermount -u -z fuse
while test -f fuse/control ; do
	fusermount -u -z fuse
	echo "Waiting for fuse/control to go away..."
	sleep 1
done
if test $res -ne 0 ; then
	# Cause any subsequent fuse tests to refuse to run by not deleting the fuse
	# directory.
	exit $res
fi
rmdir fuse
test -f tmd-trace.txt && rm tmd-trace.txt
#mv tmd-trace.txt tmd-trace-$program.txt
echo "Success!"

################################################################################
# So you want to try this as an alternative to libc's malloc? Here's how.
################################################################################
# Firstly, ld segfaults with this malloc replacement trick - it's TODO.

# Anyway, here are the steps.

# 1. mkdir fuse
# 2. start the daemon
#    TREEDB_FUSE_TMPDIR="$(pwd)" treedb-malloc-daemon "$(pwd)/fuse"
# 3. run the program
#    V3C_TREEDB_MALLOC_FUSE_DIR="$(pwd)/fuse" LD_PRELOAD=/usr/lib/libtreedb-malloc-1.2.so <program>
