/* Copyright (C) 2011  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <malloc.h>
#include <stdio.h>
#include <mcheck.h>

#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h> /* umask */
#include <sys/stat.h> /* umask */
#include <ctype.h> /* isprint */
#include <sys/wait.h> // signal

#define KIBIBYTE ((size_t)1024)
#define MEBIBYTE (KIBIBYTE * KIBIBYTE)
#define GIBIBYTE (KIBIBYTE * MEBIBYTE)


#include <dlfcn.h>
#include <v3c/smart_ptr.hpp>
#include <v3c/trace_scope.hpp>
#include <v3c/filesystem.hpp>

using namespace v3c;
using namespace std;

void sigsegv_handler(int signum)
{
	fprintf(stderr, "segfault\n");
	signal(signum, SIG_DFL);
	raise(signum);
}

typedef int (* DIDALLOC_FN)(void */*__ptr*/);

DIDALLOC_FN didalloc_fn = 0;

int default_didalloc(void *__ptr)
{
	(void)__ptr;
	return 1; // Assume yes.
}

void my_free(void *__ptr)
{
	if((*didalloc_fn)(__ptr)) {
		free(__ptr);
		return;
	}
	abort();
}
void dump_proc_maps()
{
	return;
	char cmd[25];
	snprintf(cmd, 25, "cat /proc/%d/maps", (int)getpid());
	system(cmd);
}
int using_treedb = 0;
#define do_write(msg) (void)write(STDERR_FILENO, msg, strlen(msg))
int main(int argc, char * argv[])
{
	dump_proc_maps();
	const char * fuse_dir = getenv("V3C_TREEDB_MALLOC_FUSE_DIR");
	if(!fuse_dir)
		fuse_dir = "(null)";
	do_write("V3C_TREEDB_MALLOC_FUSE_DIR = ");
	do_write(fuse_dir);
	do_write(".\n");

	signal(SIGSEGV, & sigsegv_handler);

	void * h = dlopen(0, RTLD_NOW);
	didalloc_fn = (DIDALLOC_FN)dlsym(h, "didalloc");
	dlclose(h);
	if(!didalloc_fn) {
		didalloc_fn = default_didalloc;
		fprintf(stderr, "Using regular malloc\n");
	}else{
		fprintf(stderr, "Using treedb malloc\n");
		using_treedb = 1;
	}
	mcheck(0);
	CoutTraceWriter ctw("");

	smart_ptr< FileTraceWriter > ftw;
	if(using_treedb && v3c::fs::file_type("fuse/trace")) {
		ftw = new FileTraceWriter();
		ftw->open("fuse/trace", ios_base::out | ios_base::app);
	}

	TraceWrap("main");

	size_t sz;
	if(argc < 2)
		sz = 1; /* not a stress test */
	else{
		int len = strlen(argv[1]);
		errno = 0;
		char * tailptr;
		sz = strtoull(argv[1], & tailptr, 10);
		if(errno) {
			fprintf(stderr, "Error : %s isn't a number.\n", argv[1]);
			return 1;
		}
		int diff = tailptr - argv[1];
		size_t mult = 1;
		if(diff == len)
			; /* The whole number was parsed */
		else if(diff == (len - 1)) {
			switch(*tailptr) {
			case 'K':
				mult = KIBIBYTE;
				break;
			case 'M':
				mult = MEBIBYTE;
				break;
			case 'G':
				mult = GIBIBYTE;
				break;
			default:
				fprintf(stderr, "Error : %s isn't a number.\n", argv[1]);
				return 1;
			}
			if(sz * mult < sz) { /* overflow */
				fprintf(stderr, "Error : %s is too big.\n", argv[1]);
				return 1;
			}
			sz *= mult;
		}else{
			fprintf(stderr, "Error : %s isn't a number.\n", argv[1]);
			return 1;
		}
	}
	mcheck(0);
	char * p;
	trace << "<dummy> malloc(1)." << endl;
	p = (char *)malloc(1);
	free(p);
	trace << "<dummy> free(1) worked." << endl;
	trace << getpid() << ": will malloc " << sz << " bytes." << endl;
	p = (char *)malloc(sz);
	if(!p) {
		trace_cerr << "malloc failed." << endl;
		return 1;
	}
	int res = 0;
#if 1
	size_t n;
	for(n = 0; n < sz; n += 4096)
		p[n] = 'A';
	trace << "Will write " << (sz / 4096) << " 'A's in page steps." << endl;
	for(n = 0; n < sz; n += 4096)
		if(p[n] != 'A') {
			char c = '.';
			if(isprint(p[n]))
				c = p[n];
				trace_cerr << "Comparison failed at " << n << " : "
					"wanted 'A'(" << (int)'A' << "), got '" << c << "'("
					<< (int)c << ")" << endl;
			res = 1;
			break;
		}
#endif
	mcheck(0);
	trace << "Calling mcheck() before free." << endl;
	mcheck(0);
	trace << "free'ing memory." << endl;
	free(p);
	trace << "Calling mcheck()." << endl;
	mcheck(0);
	trace << (res ? "Failed." : "Success.") << endl;
	dump_proc_maps();
	return res;
}
