/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform - passing.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <pthread.h>
#include <stddef.h>
#include <unistd.h>
#include <error.h>
#include <stdlib.h> /* rand() */

#ifdef __cplusplus
#include <stdexcept>
#include <iostream>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

//#define TREEDB_DEBUG 1

using namespace std;
using namespace v3c;
#endif

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define STRING_TREE_NO_CLEARDEFS
#define AVL_NO_CLEARDEFS
#ifdef __cplusplus
  #define TREEDB_have_dump_trees
  #include <treedb/avl-string.h>
  #include "dump-string-trees.h"
#else
  #include <treedb/avl-string.h>
#endif
#include <treedb/comet-guids.h>

TREEDB_NS(runtime_ptr_t) runtime = 0;
int to_compare_initialised = 0;
TREEDB_NSA(node_t) to_compare;

const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;
#define NUM_STRINGS 100

#define NUM_CHARS 26
#define DATA_END (NUM_CHARS+1)
char data[DATA_END] = "abcdefghijklmnopqrstuvwxyz";

int my_rand(unsigned int top)
{
	unsigned int n = rand() % top;
	TREEDB_printf("rand(%u) returned %u.\n", top, n);
	return n;
}

int fill()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	runtime = &han;
	int res1 = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_string_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}
	if(!STRING_TREE_NS(init)(runtime, STRING_TREE))
		return 2;
	/* The heap is now write-locked. */
	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);

	if(res2 == eTreedbSearchEmpty)
		return 3;
	free_size = pFirstNode->m_uLength;

	unsigned int n;
	for(n = 0; n < NUM_STRINGS; ++n) {
		unsigned int start, end;
		start = my_rand(NUM_CHARS/2);
		end = my_rand(DATA_END);
		if(start == end) {
			if(!start)
				++end;
			else
				--start;
		}
		char save = data[end];
		data[end] = '\0';
		TREEDB_printf("Inserting \"%s\".\n", data + start);
		STRING_TREE_NS(insert)(runtime, PACKED_STRING_TREE, data + start, 0);
		data[end] = save;
	}
	TREEDB_NS(check_heap)(runtime);
	TREEDB_NS(dump_trees)(runtime);
	runtime = 0;
	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

int resize_node(TREEDB_NS(runtime_ptr_t) runtime
, STRING_TREE_NS(node_ptr_t) node)
{
	if(!node)
		return 0;
	unsigned int n;
	for(n = 0; node; ++n) {
		TREEDB_NSA(node_ptr_t) p = TREEDB_ALLOC_addr_to_nodeptr(node);
		(void)p; /* -Wunused-but-set-variable */
		TREEDB_printf("[%u] resize_node [%zu : %zu] [%zu : %s]\n"
			, n
			, (size_t)TREEDB_pack_pointer(p)
			, (size_t)(p->m_uLength)
			, (size_t)TREEDB_pack_pointer(node)
			, node->m_item);
		if(TREEDB_NS(check_heap)(runtime))
			return 1;
		unsigned int start, end;
		start = my_rand(DATA_END);
		end = my_rand(DATA_END);
		int shrink = end - start;
		if(!shrink) {
			node = STRING_TREE_NS_unpack_node(node->m_next);
			continue;
		}
		/* The node's storage may be freed so unlink it just in case. */
		STRING_TREE_NS(unlink_node)(runtime, STRING_TREE, node);
		STRING_TREE_NS(node_ptr_t) node2;

		size_t len = strlen(node->m_item);
		size_t space = len;
		if(shrink > (int)len) {
			TREEDB_printf("[%u] resize_node : Shrinking \"%s\" out of existance.\n"
				, n, node->m_item);
			node2 = node;
			node = STRING_TREE_NS_unpack_node(node->m_next);
			TREEDB_NS(free)(runtime, node2);
			if(TREEDB_NS(check_heap)(runtime))
				return 2;
			continue;
		}
		if(shrink > 0) {
			TREEDB_printf("[%u] resize_node : shrinking \"%s\" by %d bytes.\n"
				, n, node->m_item, shrink);
			len = space = len - shrink;
		}else{
			TREEDB_printf("[%u] resize_node : enlarging \"%s\" by %d bytes.\n"
				, n, node->m_item, -shrink);
			space += -shrink;
			unsigned int tmp = start;
			start = end;
			end = tmp;
		}
		space += 1 + offsetof(STRING_TREE_NS(node_t), m_item);
		node2 = (STRING_TREE_NS(node_ptr_t))TREEDB_NS(realloc)
			(runtime, node, space);

		p = TREEDB_ALLOC_addr_to_nodeptr(node2);
		TREEDB_printf("[%u] resize_node : after realloc : node[%zu] -> %zu "
			"[%zu : %zu].\n"
			, n
			, (size_t)TREEDB_pack_pointer(node)
			, (size_t)TREEDB_pack_pointer(node2)
			, (size_t)TREEDB_pack_pointer(p)
			, (size_t)(p->m_uLength));

		/* Don't use node from here on, in case the address space moved. */
		if(!node2) {
			TREEDB_printf("[%u] resize_node : realloc failed.\n", n);
			/* We now know that no address space move occurred. */
			STRING_TREE_NS(insert_node)(runtime, STRING_TREE, node);
			//TREEDB_NS(dump_trees)(runtime);
			node = STRING_TREE_NS_unpack_node(node->m_next);
			if(TREEDB_NS(check_heap)(runtime))
				return 2;
			continue;
		}
		if(shrink > 0)
			node2->m_item[len] = '\0';
		else{
			char save = data[end];
			data[end] = '\0';
			len = strlen(node2->m_item);
			strcpy(node2->m_item + len, data + start);
			node2->m_item[len - shrink] = '\0';
			data[end] = save;
			TREEDB_printf("[%u] resize_node : %s\n", n, node2->m_item);
			if(TREEDB_NS(check_heap)(runtime))
				return 2;
		}
		if(STRING_TREE_NS(insert_node)(runtime, STRING_TREE, node2))
			node = STRING_TREE_NS_unpack_node(node2->m_next);
		else{
			/* A duplicate. */
			TREEDB_printf("[%u] resize_node : \"%s\" became a duplicate. Zoink.\n"
				, n, node->m_item);
			node = STRING_TREE_NS_unpack_node(node2->m_next);
			TREEDB_NS(free)(runtime, node2);
		}
	}
	return 0;
}
int resize()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_string_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("main: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	TREEDB_NS(heap_unlock)(& han);

	unsigned int n = 0, count = 0, total = 0;
	int v;
	TREEDB_NSA(node_ptr_t) pFirstNode = 0;

	STRING_TREE_NS(node_ptr_t) p;
	for(n = 0; 1; ++n) {
		TREEDB_NS(heap_write_lock)(& han);
		runtime = &han;
		count = STRING_TREE_NS(count)(runtime, STRING_TREE);
		if(!total)
			total = count;
		if(!count) {
			runtime = 0;
			TREEDB_NS(heap_unlock)(& han);
			break;
		}
		v = my_rand(count);
		if(v < 0)
			v = -v;
		while((unsigned int)v >= count)
			v >>= 1;
		// Look up the item at index v.
		p = STRING_TREE_NS(index)(runtime, STRING_TREE, v);
		if(!p) {
			printf("get failed.\n");
			runtime = 0;
			TREEDB_NS(heap_unlock)(& han);
			break;
		}
		TREEDB_printf("Remove : %u/%u %s\n", n + 1, total, p->m_item);
		STRING_TREE_NS(delete_node)(runtime, PACKED_STRING_TREE, p);
		TREEDB_printf("Remove : after.\n");
		TREEDB_NS(dump_trees)(runtime);
		res = resize_node(runtime
			, STRING_TREE_NS_unpack_node
				(runtime->m_pTreedb->m_string_tree.m_head));
		runtime = 0;
		TREEDB_NS(heap_unlock)(& han);
		if(res)
			return 2;
	}
	if(TREEDB_NS(heap_read_lock)(& han))
		return 3;
	runtime = &han;
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		printf("Error : wanted free_size == %zu, got %zu instead.\n"
			, (size_t)(free_size), (size_t)(pFirstNode->m_uLength));
		runtime = 0;
		TREEDB_NS(heap_unlock)(& han);
		TREEDB_NS(heap_close)(&han);
		return 4;
	}
	if(pFirstNode->m_offset_next/*.node.value*/) {
		printf("Error : more than 1 free node.\n");
		runtime = 0;
		TREEDB_NS(heap_unlock)(& han);
		TREEDB_NS(heap_close)(&han);
		return 5;
	}
	runtime = 0;
	TREEDB_NS(heap_unlock)(& han);
	TREEDB_NS(heap_close)(& han);
	return 0;
}

int test()
{
	int res = fill();
	if(res) {
		printf("fill failed : %d.\n", res);
		return res;
	}
	res = resize();
	if(res) {
		printf("resize failed : %d.\n", res);
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res = test();
	unlink(test_file);
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
