/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */
/*
#undef TREEDB_DEBUG
#define TREEDB_DEBUG 1
*/
#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define STRING_VALUE_TREE_NO_CLEARDEFS
#include "avl-string-value.h"

// 3dbba997-0720-4216-b23f-f322176dd6ff
UUID_DEFINE
( UUID_treedb_string_value_tree_schema
, 0x3d, 0xbb, 0xa9, 0x97
, 0x07, 0x20
, 0x42, 0x16
, 0xb2, 0x3f
, 0xf3, 0x22, 0x17, 0x6d, 0xd6, 0xff
);
const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;
TREEDB_NS(runtime_ptr_t) runtime;
int test_num = 0;

const char * sorted_items[] =
{ /*0*/"apple"
, /*1*/"banana"
, /*2*/"lemon"
, /*3*/"orange"
, /*4*/"pear"
, /*5*/"plum"
};
STRING_VALUE_TREE_NS(input) items[] =
{ { 0, "apple" }
, { 3, "orange" }
, { 2, "lemon" }
, { 4, "pear" }
, { 5, "plum" }
, { 1, "banana" }
};
#define ITEMS_COUNT (sizeof(items) / sizeof(items[0]))
int fill_test()
{ 
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	runtime = &han;
	int res1 = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_string_value_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(res2 == eTreedbSearchEmpty)
		return 1;

	free_size = pFirstNode->m_uLength;

	STRING_VALUE_TREE_NS(tree_t) tree_init = AVL_TMPL_init;
	* STRING_VALUE_TREE = tree_init;

	unsigned int n;
	TREEDB_NS(dump_trees)(runtime);
	for(n = 0; n < ITEMS_COUNT; ++n)
		STRING_VALUE_TREE_NS(insert)(runtime, PACKED_STRING_VALUE_TREE, items + n
			, 0);
	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

void dump()
{
	int n, count = STRING_VALUE_TREE_NS(count)(runtime, STRING_VALUE_TREE);
	STRING_VALUE_TREE_NS(node_ptr_t) pn;
	for(n = 0; n < count; ++n) {
		pn = STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, n);
		cout << pn->m_item.m_key << ", "
			<< pn->m_item.m_value << '\n';
	}
	cout << flush;
}

typedef int (*PFN)();

struct TestEntry
{
	PFN fn;
	const char * name;
};
int clear_fn()
{
	STRING_VALUE_TREE_NS(clear)(runtime, PACKED_STRING_VALUE_TREE);
	return 0;
}
int remove_first()
{
	STRING_VALUE_TREE_NS(node_ptr_t) pn = STRING_VALUE_TREE_NS(index)(runtime
		, STRING_VALUE_TREE, 0);
	if(!pn)
		return 1;
	STRING_VALUE_TREE_NS(delete_node)(runtime, PACKED_STRING_VALUE_TREE, pn);
	if(STRING_VALUE_TREE_NS(count)(runtime, STRING_VALUE_TREE) != (ITEMS_COUNT - 1))
		return 1;
	int n;
	for(n = 0; n < (int)(ITEMS_COUNT - 2); ++n) {
		pn = STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, n);
		if(strcmp(pn->m_item.m_value, sorted_items[n + 1])) {
			cerr << "pn->m_item.m_value(" << pn->m_item.m_value
				<< ") != sorted_items[n + 1]("
				<< sorted_items[n + 1]
				<< ")." << endl;
			return 1;
		}
	}
	return 0;
}
int remove_first_many()
{
	int n;
	STRING_VALUE_TREE_NS(node_ptr_t) pn;
	for(n = 0; n < (int)(ITEMS_COUNT - 1); ++n) {
		pn = STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, 0);
		if(!pn)
			return 1;
		STRING_VALUE_TREE_NS(delete_node)(runtime, PACKED_STRING_VALUE_TREE, pn);
	}
	if(STRING_VALUE_TREE_NS(count)(runtime, STRING_VALUE_TREE) != 1)
		return 1;
	pn = STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, 0);
	if(strcmp(pn->m_item.m_value, sorted_items[n])) {
		cerr << "pn->m_item.m_value(" << pn->m_item.m_value
			<< ") != sorted_items[n]("
			<< sorted_items[n]
			<< ")." << endl;
		return 1;
	}
	return 0;
}
/*
0, "apple"
1, "banana"
2, "lemon"
4, "pear"
5, "plum"
*/
int remove_middle()
{
	int n;
	STRING_VALUE_TREE_NS(node_ptr_t) pn
		= STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, ITEMS_COUNT/2);
	if(!pn)
		return 1;
	STRING_VALUE_TREE_NS(delete_node)(runtime, PACKED_STRING_VALUE_TREE, pn);
	if(STRING_VALUE_TREE_NS(count)(runtime, STRING_VALUE_TREE) != (ITEMS_COUNT - 1))
		return 1;
	for(n = 0; n < 3; ++n) {
		pn = STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, n);
		if(strcmp(pn->m_item.m_value, sorted_items[n])) {
			cerr << "pn->m_item.m_value(" << pn->m_item.m_value
				<< ") != sorted_items[n]("
				<< sorted_items[n]
				<< ")." << endl;
			return 1;
		}
		if(n != pn->m_item.m_key) {
			cerr << "n(" << n
				<< ") != pn->m_item.m_key("
				<< pn->m_item.m_key
				<< ")." << endl;
			return 1;
		}
	}
	for(; n < (int)(ITEMS_COUNT - 1); ++n) {
		pn = STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, n);
		if(strcmp(pn->m_item.m_value, sorted_items[n + 1])) {
			cerr << "pn->m_item.m_value(" << pn->m_item.m_value
				<< ") != sorted_items[n + 1]("
				<< sorted_items[n + 1]
				<< ")." << endl;
			return 1;
		}
		if(n != (pn->m_item.m_key - 1)) {
			cerr << "n(" << n
				<< ") != pn->m_item.m_key("
				<< pn->m_item.m_key
				<< ") - 1." << endl;
			return 1;
		}
	}
	return 0;
}
int remove_middle_many()
{
	int n;
	for(n = 0; n < (int)(ITEMS_COUNT - 1); ++n) {
		STRING_VALUE_TREE_NS(node_ptr_t) pn
			= STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE
				, ITEMS_COUNT - 1 - n);
		if(!pn)
			return 1;
		STRING_VALUE_TREE_NS(delete_node)(runtime, PACKED_STRING_VALUE_TREE, pn);
	}
	if(STRING_VALUE_TREE_NS(count)(runtime, STRING_VALUE_TREE) != 1)
		return 1;
	return 0;
}
int remove_last()
{
	STRING_VALUE_TREE_NS(node_ptr_t) pn
		= STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, ITEMS_COUNT - 1);
	if(!pn)
		return 1;
	STRING_VALUE_TREE_NS(delete_node)(runtime, PACKED_STRING_VALUE_TREE, pn);
	if(STRING_VALUE_TREE_NS(count)(runtime, STRING_VALUE_TREE) != (ITEMS_COUNT - 1))
		return 1;
	return 0;
}
int remove_last_many()
{
	int n;
	for(n = 0; n < (int)(ITEMS_COUNT - 1); ++n) {
		STRING_VALUE_TREE_NS(node_ptr_t) pn
			= STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, 0);
		if(!pn)
			return 1;
		STRING_VALUE_TREE_NS(delete_node)(runtime, PACKED_STRING_VALUE_TREE, pn);
	}
	if(STRING_VALUE_TREE_NS(count)(runtime, STRING_VALUE_TREE) != 1)
		return 1;
	return 0;
}
int remove_all()
{
	STRING_VALUE_TREE_NS(clear)(runtime, PACKED_STRING_VALUE_TREE);
	if(STRING_VALUE_TREE_NS(count)(runtime, STRING_VALUE_TREE) != 0)
		return 1;
	return 0;
}
#define TEST_ENTRY(name) {& name, #name}
TestEntry tests[] =
{
	TEST_ENTRY(clear_fn),
	TEST_ENTRY(remove_first),
	TEST_ENTRY(remove_first_many),
	TEST_ENTRY(remove_middle),
	TEST_ENTRY(remove_middle_many),
	TEST_ENTRY(remove_last),
	TEST_ENTRY(remove_last_many),
	TEST_ENTRY(remove_all)
};

int empty_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_string_value_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("empty_test: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	runtime = &han;
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_write_lock)(runtime);

	if(STRING_VALUE_TREE_NS(index)(runtime, STRING_VALUE_TREE, ITEMS_COUNT)) {
		printf("empty_test: node_get(ITEMS_COUNT) XPASS.\n");
		return 2;
	}
	if(tests[test_num].fn()) {
		fprintf(stderr, "Test \"%s\" failed.\n", tests[test_num].name);
		dump();
		return 1;
	}
	STRING_VALUE_TREE_NS(clear)(runtime, PACKED_STRING_VALUE_TREE);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_read_lock)(runtime);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		fprintf(stderr, "Free size(%zu) != %zu.\n", (size_t)free_size
			, (size_t)pFirstNode->m_uLength);
		return 4;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

int test()
{
	int res = fill_test();
	if(res) {
		printf("fill test failed.\n");
		return res;
	}
	res = empty_test();
	if(res) {
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res;
	for(test_num = 0; test_num < (int)(sizeof(tests)/sizeof(tests[0]));
			++test_num) {
		res = test();
		unlink(test_file);
		if(res)
			break;
	}
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
