/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */
/*
#undef TREEDB_DEBUG
#define TREEDB_DEBUG 1
*/
#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define STRUCT_LIST_NO_CLEARDEFS
#include <treedb/struct-list.h>
#include <signal.h> // signal()

// 6af2c199-4ae0-4b81-a071-82af398695f4
UUID_DEFINE
( UUID_treedb_struct_list_schema
, 0x6a, 0xf2, 0xc1, 0x99
, 0x4a, 0xe0
, 0x4b, 0x81
, 0xa0, 0x71
, 0x82, 0xaf, 0x39, 0x86, 0x95, 0xf4
);const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;
TREEDB_NS(runtime_ptr_t) runtime;
int test_num = 0;

STRUCT_LIST_NS(input_t) items[] =
{ { "apple" , 0 }
, { "orange", 1 }
, { "lemon" , 2 }
, { "pear"  , 3 }
, { "plum"  , 4 }
, { "banana", 5 }
};
#define ITEMS_COUNT (sizeof(items) / sizeof(items[0]))
int fill_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res1 = TREEDB_NS(heap_create)
		(& han
		, UUID_treedb_struct_list_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}
	runtime = &han;

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(res2 == eTreedbSearchEmpty)
		return 1;

	free_size = pFirstNode->m_uLength;

	STRUCT_LIST_NS(list_t) list_init = LIST_TMPL_init;
	* STRUCT_LIST = list_init;

	unsigned int n;
	TREEDB_NS(dump_trees)(runtime);
	for(n = 0; n < ITEMS_COUNT; ++n) {
		TREEDB_PAYLOAD_NS(push_back)(runtime, PACKED_STRUCT_LIST, items + n);
	}
	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

void dump()
{
	int n, count = TREEDB_PAYLOAD_NS(count)(runtime, STRUCT_LIST);
	TREEDB_PAYLOAD_NS(node_ptr_t) pn
		= STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_head);
	for(n = 0; n < count; ++n) {
		cout << pn->m_item.m_string << ", "
			<< pn->m_item.m_int << '\n';
		pn = STRUCT_LIST_NS_unpack_node(pn->m_next);
	}
}
typedef int (*PFN)();

struct TestEntry
{
	PFN fn;
	const char * name;
};
int clear_fn()
{
	TREEDB_PAYLOAD_NS(clear)(runtime, PACKED_STRUCT_LIST);
	return 0;
}
int swap_first2()
{
	TREEDB_PAYLOAD_NS(node_ptr_t) pn
		= STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_head);
	TREEDB_PAYLOAD_NS(unlink_node)(runtime, STRUCT_LIST, pn);
	TREEDB_PAYLOAD_NS(insert_node_after)(runtime, STRUCT_LIST
		, STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_head), pn);
	if(TREEDB_PAYLOAD_NS(count)(runtime, STRUCT_LIST) != ITEMS_COUNT)
		return 1;
	unsigned int n;
	pn = STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_head);
	for(n = 0; n < 2; ++n) {
		if(pn->m_item.m_int != items[1 - n].m_int) {
			cerr << "pn->m_item.m_int(" << pn->m_item.m_int
				<< ") != items[1 - n].m_int("
				<< items[1 - n].m_int
				<< ")." << endl;
			return 1;
		}
		pn = STRUCT_LIST_NS_unpack_node(pn->m_next);
	}
	for(; n < ITEMS_COUNT; ++n) {
		if(pn->m_item.m_int != items[n].m_int) {
			cerr << "pn->m_item.m_int(" << pn->m_item.m_int
				<< ") != items[n].m_int("
				<< items[n].m_int
				<< ")." << endl;
			return 1;
		}
		pn = STRUCT_LIST_NS_unpack_node(pn->m_next);
	}
	return 0;
}
int swap_last2()
{
	TREEDB_PAYLOAD_NS(node_ptr_t) pn
		= STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_tail);
	TREEDB_PAYLOAD_NS(unlink_node)(runtime, STRUCT_LIST, pn);
	TREEDB_PAYLOAD_NS(insert_node_before)(runtime, STRUCT_LIST
		, STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_tail), pn);
	if(TREEDB_PAYLOAD_NS(count)(runtime, STRUCT_LIST) != ITEMS_COUNT)
		return 1;
	unsigned int n;
	pn = STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_head);
	for(n = 0; n < (ITEMS_COUNT - 2); ++n) {
		if(pn->m_item.m_int != items[n].m_int) {
			cerr << "pn->m_item.m_int(" << pn->m_item.m_int
				<< ") != items[n].m_int("
				<< items[n].m_int
				<< ")." << endl;
			return 1;
		}
		pn = STRUCT_LIST_NS_unpack_node(pn->m_next);
	}
	for(; n < ITEMS_COUNT; ++n) {
		if(pn->m_item.m_int != items[(ITEMS_COUNT - 2)*2 - n + 1].m_int) {
			cerr << "pn->m_item.m_int(" << pn->m_item.m_int
				<< ") != items[(ITEMS_COUNT - 2)*2 - n + 1].m_int("
				<< items[(ITEMS_COUNT - 2)*2 - n + 1].m_int
				<< ")." << endl;
			return 1;
		}
		pn = STRUCT_LIST_NS_unpack_node(pn->m_next);
	}
	return 0;
}
int remove_first()
{
	int res = TREEDB_PAYLOAD_NS(pop_front)(runtime, PACKED_STRUCT_LIST);
	if(res)
		return 1;
	if(TREEDB_PAYLOAD_NS(count)(runtime, STRUCT_LIST) != (ITEMS_COUNT - 1))
		return 1;
	int n;
	TREEDB_PAYLOAD_NS(node_ptr_t) pn
		= STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_head);
	for(n = 0; n < (int)(ITEMS_COUNT - 2); ++n) {
		if(pn->m_item.m_int != items[n + 1].m_int) {
			cerr << "pn->m_item.m_int(" << pn->m_item.m_int
				<< ") != items[n + 1].m_int("
				<< items[n + 1].m_int
				<< ")." << endl;
			return 1;
		}
		pn = STRUCT_LIST_NS_unpack_node(pn->m_next);
	}
	return 0;
}
int remove_first_many()
{
	int n, res;
	for(n = 0; n < (int)(ITEMS_COUNT - 1); ++n) {
		res = TREEDB_PAYLOAD_NS(pop_front)(runtime, PACKED_STRUCT_LIST);
		if(res)
			return 1;
	}
	if(TREEDB_PAYLOAD_NS(count)(runtime, STRUCT_LIST) != 1)
		return 1;
	TREEDB_PAYLOAD_NS(node_ptr_t) pn
		= STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_head);
	if(pn->m_item.m_int != items[n].m_int) {
		cerr << "pn->m_item.m_int(" << pn->m_item.m_int
			<< ") != items[n].m_int("
			<< items[n].m_int
			<< ")." << endl;
		return 1;
	}
	return 0;
}
int remove_last()
{
	int res = TREEDB_PAYLOAD_NS(pop_back)(runtime, PACKED_STRUCT_LIST);
	if(res)
		return 1;
	if(TREEDB_PAYLOAD_NS(count)(runtime, STRUCT_LIST) != (ITEMS_COUNT - 1))
		return 1;
	int n;
	TREEDB_PAYLOAD_NS(node_ptr_t) pn
		= STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_head);
	for(n = 0; n < (int)(ITEMS_COUNT - 1); ++n) {
		if(pn->m_item.m_int != items[n].m_int) {
			cerr << "pn->m_item.m_int(" << pn->m_item.m_int
				<< ") != items[n].m_int("
				<< items[n].m_int
				<< ")." << endl;
			return 1;
		}
		pn = STRUCT_LIST_NS_unpack_node(pn->m_next);
	}
	return 0;
}
int remove_last_many()
{
	int n, res;
	for(n = 0; n < (int)(ITEMS_COUNT - 1); ++n) {
		res = TREEDB_PAYLOAD_NS(pop_back)(runtime, PACKED_STRUCT_LIST);
		if(res)
			return 1;
	}
	if(TREEDB_PAYLOAD_NS(count)(runtime, STRUCT_LIST) != 1)
		return 1;
	TREEDB_PAYLOAD_NS(node_ptr_t) pn
		= STRUCT_LIST_NS_unpack_node(STRUCT_LIST->m_head);
	if(pn->m_item.m_int != items[0].m_int) {
		cerr << "pn->m_item.m_int(" << pn->m_item.m_int
			<< ") != items[0].m_int("
			<< items[0].m_int
			<< ")." << endl;
		return 1;
	}
	return 0;
}
int remove_all()
{
	TREEDB_PAYLOAD_NS(clear)(runtime, PACKED_STRUCT_LIST);
	if(TREEDB_PAYLOAD_NS(count)(runtime, STRUCT_LIST) != 0)
		return 1;
	return 0;
}
#define TEST_ENTRY(name) {& name, #name}
TestEntry tests[] =
{
	TEST_ENTRY(clear_fn),
	TEST_ENTRY(swap_first2),
	TEST_ENTRY(swap_last2),
	TEST_ENTRY(remove_first),
	TEST_ENTRY(remove_first_many),
	TEST_ENTRY(remove_last),
	TEST_ENTRY(remove_last_many),
	TEST_ENTRY(remove_all)
};

int empty_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_struct_list_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("empty_test: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	runtime = &han;
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_write_lock)(runtime);

	if(tests[test_num].fn()) {
		fprintf(stderr, "Test \"%s\" failed.\n", tests[test_num].name);
		dump();
		return 1;
	}
	TREEDB_PAYLOAD_NS(clear)(runtime, PACKED_STRUCT_LIST);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_read_lock)(runtime);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		fprintf(stderr, "Free size(%zu) != %zu.\n", (size_t)free_size
			, (size_t)pFirstNode->m_uLength);
		return 4;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

void fatal_repo_error(const char * msg)
{
	fflush(stdout);
	fflush(stderr);
	cout << flush;
	cerr << flush;
	cout << "Terminating because of " << msg << '.' << endl;
	unlink(test_file);

	exit(1);
}

void sigsegv_handler(int signum)
{
	fatal_repo_error("sigsegv");
	signal(signum, SIG_DFL);
	raise(signum);
}

int test()
{
	int res = fill_test();
	if(res) {
		printf("fill test failed.\n");
		return res;
	}
	res = empty_test();
	if(res) {
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	signal(SIGSEGV, & sigsegv_handler);

	int res;
	for(test_num = 0; test_num < (int)(sizeof(tests)/sizeof(tests[0]));
			++test_num) {
		res = test();
		unlink(test_file);
		if(res)
			break;
	}
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
