/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */
/*
#undef TREEDB_DEBUG
#define TREEDB_DEBUG 1
*/
#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;

extern void impl_memmove(void * to, const void * from, size_t size);

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define TREEDB_ARRAY_MEMMOVE
#define ARRAY_NO_CLEARDEFS

#include <treedb/ns-types.h>

#define TREEDB_PAYLOAD \
	TREEDB_PAYLOAD_NS(array_t) a, b
#define ARRAY_MEMBER a
#include <treedb/array.h>

#undef ARRAY
#undef PACKED_ARRAY
#define A (& runtime->m_pTreedb->a)
#define PACKED_A TREEDB_NS(impl_pack_pointer)(runtime, A)
#define B (& runtime->m_pTreedb->b)
#define PACKED_B TREEDB_NS(impl_pack_pointer)(runtime, B)

#define TEST_A PACKED_A
#define TEST_B PACKED_B

// 220c16f0-4a8b-4c50-9c9d-402700848c8f
UUID_DEFINE
( UUID_treedb_two_array_int_schema
, 0x22, 0x0c, 0x16, 0xf0
, 0x4a, 0x8b
, 0x4c, 0x50
, 0x9c, 0x9d
, 0x40, 0x27, 0x00, 0x84, 0x8c, 0x8f
);

const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;
TREEDB_NS(runtime_ptr_t) runtime;
int test_num = 0;

void impl_memmove(void * to, const void * from, size_t size)
{
	printf("### memmove(%zu, %zu, %zu) ###\n"
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, to)
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, from)
		, size
	);
	memmove(to, from, size);
}

void dumpA()
{
	cout << "A\n===\n";
	int n;
	TREEDB_PAYLOAD_NS(array_ptr_t) pa = A;
	int32_t * pd
		= TREEDB_PAYLOAD_NS(unpack_data)(runtime, pa->m_items);
	for(n = 0; n < (int)pa->m_count; ++n)
		cout << '[' << n << "] " << pd[n] << '\n';
	cout << flush;
}
void dumpB()
{
	cout << "B\n===\n";
	int n;
	TREEDB_PAYLOAD_NS(array_ptr_t) pa = B;
	int32_t * pd
		= TREEDB_PAYLOAD_NS(unpack_data)(runtime, pa->m_items);
	for(n = 0; n < (int)pa->m_count; ++n)
		cout << '[' << n << "] " << pd[n] << '\n';
	cout << flush;
}
void dump()
{
	dumpA();
	dumpB();
}

int check_expected(const char * s, TREEDB_PAYLOAD_NS(array_ptr_t) pa)
{
	uint16_t count = strlen(s);
	if(count != pa->m_count) {
		cerr << "check_expected failed(" << (pa == A ? 'A' : 'B' )
			<< ") count(" << pa->m_count << ") != expected("
			<< count << ")." << endl;
		return 1;
	}
	int32_t * pd
		= TREEDB_PAYLOAD_NS(unpack_data)(runtime, pa->m_items);
	uint16_t n;
	int err = 0;
	for(n = 0; n < count; ++n) {
		int16_t u = s[n] - '0';
		if(pd[n] != u) {
			cerr << "check_expected failed(" << (pa == A ? 'A' : 'B' )
				<< ") entry[" << n << "] "
				<< pd[n]
				<< " != expected "
				<< u
				<< '.' << endl;
			err = 1;
		}
	}
	return err;
}

#include "two-test-common.h"

int setup_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	runtime = &han;
	int res1 = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_two_array_int_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(res2 == eTreedbSearchEmpty)
		return 1;

	free_size = pFirstNode->m_uLength;

	if(TREEDB_PAYLOAD_NS(init)(runtime, PACKED_A, 8, 4)) {
		fprintf(stderr, "Array initialisation failed\n");
		return 1;
	}
	if(TREEDB_PAYLOAD_NS(init)(runtime, PACKED_B, 4, 2)) {
		fprintf(stderr, "Array initialisation failed\n");
		return 1;
	}

	unsigned int n;
	TREEDB_NS(dump_trees)(runtime);

	for(n = 0; n < 10; ++n)
		TREEDB_PAYLOAD_NS(push_back)(runtime, PACKED_A, n);
	for(n = 0; n < 10; ++n)
		TREEDB_PAYLOAD_NS(push_back)(runtime, PACKED_B, n);

	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}
int run_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_two_array_int_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("run_test: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	runtime = &han;

	fprintf(stderr, "Testing %s\n", tests[test_num].name);
	res = do_the_test(tests[test_num]);
	if(res) {
		fprintf(stderr, "Test \"%s\" failed - %d.\n", tests[test_num].name, res);
		dump();
		TREEDB_NS(heap_unlock)(runtime);
		TREEDB_NS(heap_close)(&han);
		return 3;
	}
	TREEDB_PAYLOAD_NS(reset)(runtime, PACKED_A);
	TREEDB_PAYLOAD_NS(reset)(runtime, PACKED_B);

	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_read_lock)(runtime);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		fprintf(stderr, "Free size(%zu) != %zu.\n", (size_t)free_size
			, (size_t)pFirstNode->m_uLength);
		return 4;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

int test()
{
	int res = setup_test();
	if(res) {
		printf("setup_test failed.\n");
		return res;
	}
	res = run_test();
	if(res) {
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res;
	for(test_num = 0; test_num < (int)(sizeof(tests)/sizeof(tests[0]));
			++test_num) {
		res = test();
		unlink(test_file);
		if(res)
			break;
	}
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
