/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */
/*
#undef TREEDB_DEBUG
#define TREEDB_DEBUG 1
*/
#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;

extern void impl_memmove(void * to, const void * from, size_t size);

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define TREEDB_ARRAY_MEMMOVE
#define ARRAY_NO_CLEARDEFS

#include <treedb/ns-types.h>

#define TREEDB_PAYLOAD \
        TREEDB_NS_CONCAT(uint,_t) a, b
#define ARRAY_MEMBER a

#include "handle-array.h"

#undef ARRAY
#undef PACKED_ARRAY
#undef PACKED2_ARRAY

#define PACKED_A (runtime->m_pTreedb->a)
#define PACKED2_A v3c_offsetof(TREEDB_NS(treedb_t), a)
#define A ((ARRAY_NS(ptr_t))ARRAY_unpack_pointer(PACKED_A))

#define PACKED_B (runtime->m_pTreedb->b)
#define PACKED2_B v3c_offsetof(TREEDB_NS(treedb_t), b)
#define B ((ARRAY_NS(ptr_t))ARRAY_unpack_pointer(PACKED_B))

#define TEST_A PACKED2_A
#define TEST_B PACKED2_B

// 1b637403-525b-4335-ae3a-6b80a2bbfb47
UUID_DEFINE
( UUID_treedb_handle_array_int_schema
, 0x1b, 0x63, 0x74, 0x03
, 0x52, 0x5b
, 0x43, 0x35
, 0xae, 0x3a
, 0x6b, 0x80, 0xa2, 0xbb, 0xfb, 0x47
);
const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;
TREEDB_NS(runtime_ptr_t) runtime;
int test_num = 0;

void impl_memmove(void * to, const void * from, size_t size)
{
	printf("### memmove(%zu, %zu, %zu) ###\n"
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, to)
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, from)
		, size
	);
	memmove(to, from, size);
}

void dumpA()
{
	cout << "A\n===\n";
	uint16_t n;
	TREEDB_PAYLOAD_NS(ptr_t) pa = A;
	for(n = 0; n < pa->m_count; ++n)
		cout << '[' << n << "] " << pa->m_items[n] << '\n';
	cout << flush;
}
void dumpB()
{
	cout << "B\n===\n";
	uint16_t n;
	TREEDB_PAYLOAD_NS(ptr_t) pa = B;
	for(n = 0; n < pa->m_count; ++n)
		cout << '[' << n << "] " << pa->m_items[n] << '\n';
	cout << flush;
}
void dump()
{
	dumpA();
	dumpB();
}

int check_expected(const char * s, TREEDB_PAYLOAD_NS(ptr_t) pa)
{
	uint16_t count = strlen(s);
	if(count != pa->m_count) {
		cerr << "check_expected failed(" << (pa == A ? 'A' : 'B' )
			<< ") count(" << pa->m_count << ") != expected("
			<< count << ")." << endl;
		return 1;
	}
	uint16_t n;
	int err = 0;
	for(n = 0; n < count; ++n) {
		int16_t u = s[n] - '0';
		if(pa->m_items[n] != u) {
			cerr << "check_expected failed(" << (pa == A ? 'A' : 'B' )
				<< ") entry[" << n << "] "
				<< pa->m_items[n]
				<< " != expected "
				<< u
				<< '.' << endl;
			err = 1;
		}
	}
	return err;
}

#include "two-test-common.h"

int setup_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	runtime = &han;
	int res1 = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_handle_array_int_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(res2 == eTreedbSearchEmpty)
		return 1;

	free_size = pFirstNode->m_uLength;

	if(TREEDB_PAYLOAD_NS(new)(runtime, PACKED2_A, 8, 4)) {
		fprintf(stderr, "Array A creation failed\n");
		return 1;
	}
	if(TREEDB_PAYLOAD_NS(new)(runtime, PACKED2_B, 4, 2)) {
		fprintf(stderr, "Array B creation failed\n");
		return 1;
	}

	unsigned int n;
	TREEDB_NS(dump_trees)(runtime);

	for(n = 0; n < 10; ++n)
		TREEDB_PAYLOAD_NS(push_back)(runtime, PACKED2_A, n);
	for(n = 0; n < 10; ++n)
		TREEDB_PAYLOAD_NS(push_back)(runtime, PACKED2_B, n);

	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}
int run_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_handle_array_int_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("run_test: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	runtime = &han;

	printf("Testing %s\n", tests[test_num].name);
	res = do_the_test(tests[test_num]);
	if(res) {
		printf("Test \"%s\" failed(%d).\n", tests[test_num].name, res);
		dump();
		return 3;
	}
	TREEDB_PAYLOAD_NS(free)(runtime, PACKED_A);
	TREEDB_PAYLOAD_NS(free)(runtime, PACKED_B);

	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_read_lock)(runtime);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		fprintf(stderr, "Free size(%zu) != %zu.\n", (size_t)free_size
			, (size_t)pFirstNode->m_uLength);
		return 4;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

int test()
{
	int res = setup_test();
	if(res) {
		printf("setup_test failed.\n");
		return res;
	}
	res = run_test();
	if(res) {
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res;
	for(test_num = 0; test_num < (int)(sizeof(tests)/sizeof(tests[0]));
			++test_num) {
		res = test();
		unlink(test_file);
		if(res)
			break;
	}
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
