/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef treedb_two_test_entries_h_included
#define treedb_two_test_entries_h_included

struct TestEntry {
	const char * name;
	bool copy;
	uintptr_t toWhere, toCount, fromWhere, fromCount;
	const char * expectedTo, * expectedFrom;
};
TestEntry tests[] =
{ { "copy_none_to_empty_start",  1,  0, 0,   0,0, "0123456789",    "0123456789" }
, { "copy_none_to_empty_middle", 1,  4, 0,   0,0, "0123456789",    "0123456789" }
, { "copy_none_to_empty_end",    1, 10, 0,  10,0, "0123456789",    "0123456789" }
, { "copy_some_to_empty_start",  1,  0, 0,   0,3, "0120123456789", "0123456789" }
, { "copy_some_to_empty_middle", 1,  4, 0,   0,3, "0123012456789", "0123456789" }
, { "copy_some_to_empty_end",    1, 10, 0,   0,3, "0123456789012", "0123456789" }
, { "copy_none_to_some_start",   1,  0, 3,   0,0, "3456789",       "0123456789" }
, { "copy_none_to_some_middle",  1,  4, 3,   0,0, "0123789",       "0123456789" }
, { "copy_none_to_some_end",     1,  7, 3,   0,0, "0123456",       "0123456789" }
, { "copy_some_to_some_start1",  1,  0, 3,   4,3, "4563456789",    "0123456789" }
, { "copy_some_to_some_start2",  1,  0, 3,   4,4, "45673456789",   "0123456789" }
, { "copy_some_to_some_start3",  1,  0, 3,   4,2, "453456789",     "0123456789" }
, { "copy_some_to_some_middle1", 1,  4, 3,   0,3, "0123012789",    "0123456789" }
, { "copy_some_to_some_middle2", 1,  4, 3,   0,4, "01230123789",   "0123456789" }
, { "copy_some_to_some_middle3", 1,  4, 3,   0,2, "012301789",     "0123456789" }
, { "copy_some_to_some_end1",    1,  7, 3,   0,3, "0123456012",    "0123456789" }
, { "copy_some_to_some_end2",    1,  7, 3,   0,4, "01234560123",   "0123456789" }
, { "copy_some_to_some_end3",    1,  7, 3,   0,2, "012345601",     "0123456789" }

, { "move_none_to_empty_start",  0,  0, 0,   0,0, "0123456789",    "0123456789" }
, { "move_none_to_empty_middle", 0,  4, 0,   0,0, "0123456789",    "0123456789" }
, { "move_none_to_empty_end",    0, 10, 0,  10,0, "0123456789",    "0123456789" }
, { "move_some_to_empty_start",  0,  0, 0,   0,3, "0120123456789", "3456789"    }
, { "move_some_to_empty_middle", 0,  4, 0,   0,3, "0123012456789", "3456789"    }
, { "move_some_to_empty_end",    0, 10, 0,   0,3, "0123456789012", "3456789"    }
, { "move_none_to_some_start",   0,  0, 3,   0,0, "3456789",       "0123456789" }
, { "move_none_to_some_middle",  0,  4, 3,   0,0, "0123789",       "0123456789" }
, { "move_none_to_some_end",     0,  7, 3,   0,0, "0123456",       "0123456789" }
, { "move_some_to_some_start1",  0,  0, 3,   4,3, "4563456789",    "0123789"    }
, { "move_some_to_some_start2",  0,  0, 3,   4,4, "45673456789",   "012389"     }
, { "move_some_to_some_start3",  0,  0, 3,   4,2, "453456789",     "01236789"   }
, { "move_some_to_some_middle1", 0,  4, 3,   0,3, "0123012789",    "3456789"    }
, { "move_some_to_some_middle2", 0,  4, 3,   0,4, "01230123789",   "456789"     }
, { "move_some_to_some_middle3", 0,  4, 3,   0,2, "012301789",     "23456789"   }
, { "move_some_to_some_end1",    0,  7, 3,   0,3, "0123456012",    "3456789"    }
, { "move_some_to_some_end2",    0,  7, 3,   0,4, "01234560123",   "456789"     }
, { "move_some_to_some_end3",    0,  7, 3,   0,2, "012345601",     "23456789"   }
};

int do_the_test(const TestEntry & e)
{
	int res;
	if(e.copy) {
		res = TREEDB_PAYLOAD_NS(copy_many)(runtime
				, TEST_A, e.toWhere, e.toCount
				, TEST_B, e.fromWhere, e.fromCount
				);
		if(res) {
			printf("copy_many() failed - %d.\n", res);
			return 1;
		}
	}else{
		res = TREEDB_PAYLOAD_NS(move_many)(runtime
				, TEST_A, e.toWhere, e.toCount
				, TEST_B, e.fromWhere, e.fromCount
				);
		if(res) {
			printf("move_many() failed - %d.\n", res);
			return 1;
		}
	}
	res = check_expected(e.expectedTo, A);
	if(res) {
		printf("check_expected(A) failed - %d.\n", res);
		return 2;
	}
	res = check_expected(e.expectedFrom, B);
	if(res) {
		printf("check_expected(B) failed - %d.\n", res);
		return 3;
	}
	return 0;
}

#endif /* #ifndef treedb_two_test_entries_h_included */
