/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */
/*
#undef TREEDB_DEBUG
#define TREEDB_DEBUG 1
*/
#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <iomanip>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;

extern void impl_memmove(void * to, const void * from, size_t size);

string to_string(const uint16_t * p, uintptr_t len)
{
	string res;
	uintptr_t n;
	char c;
	for(n = 0; n < len; ++n) {
		c = (char)p[n];
		res += c;
	}
	return res;
}

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define VARRAY_NO_CLEARDEFS
#define VARRAY_data_t uint16_t
#define TREEDB_PAYLOAD \
	TREEDB_PAYLOAD_NS(varray_t) a, b
#define VARRAY_MEMBER a

#include <treedb/varray.h>

#undef VARRAY
#undef PACKED_VARRAY
#define A (& runtime->m_pTreedb->a)
#define PACKED_A TREEDB_NS(impl_pack_pointer)(runtime, A)
#define B (& runtime->m_pTreedb->b)
#define PACKED_B TREEDB_NS(impl_pack_pointer)(runtime, B)

#define TEST_A PACKED_A
#define TEST_B PACKED_B

// 42821811-44ba-44b3-a1d5-5cc10b5f2088
UUID_DEFINE
( UUID_treedb_two_varray_uint_schema
, 0x42, 0x82, 0x18, 0x11
, 0x44, 0xba
, 0x44, 0xb3
, 0xa1, 0xd5
, 0x5c, 0xc1, 0x0b, 0x5f, 0x20, 0x88
);

#define ITEM_COUNT 10

// Note: no trailing '\0' terminators - the varray tracks their sizes.
const uint16_t items0[] = { 'b', 'l', 'u', 'e' };
const uint16_t items1[] = { 'g', 'r', 'e', 'e', 'n' };
const uint16_t items2[] = { 'r', 'e', 'd' };
const uint16_t items3[] = { 'y', 'e', 'l', 'l', 'o', 'w' };
const uint16_t items4[] = { 'm', 'a', 'g', 'e', 'n', 't', 'a' };
const uint16_t items5[] = { 'g', 'r', 'e', 'y' };
const uint16_t items6[] = { 's', 'i', 'l', 'v', 'e', 'r' };
const uint16_t items7[] = { 'g', 'o', 'l', 'd' };
const uint16_t items8[] = { 'o', 'r', 'a', 'n', 'g', 'e' };
const uint16_t items9[] = { 'b', 'l', 'a', 'c', 'k' };
const uint16_t * items[] =
{ items0, items1, items2, items3, items4, items5 , items6, items7, items8
, items9 };
VARRAY_count_t items_count[] =
{ COUNT_OF(items0)
, COUNT_OF(items1)
, COUNT_OF(items2)
, COUNT_OF(items3)
, COUNT_OF(items4)
, COUNT_OF(items5)
, COUNT_OF(items6)
, COUNT_OF(items7)
, COUNT_OF(items8)
, COUNT_OF(items9)
};
VARRAY_count_t item_totals[] =
{ VARRAY_count_t(0)
, VARRAY_count_t(item_totals[0] + items_count[0])
, VARRAY_count_t(item_totals[1] + items_count[1])
, VARRAY_count_t(item_totals[2] + items_count[2])
, VARRAY_count_t(item_totals[3] + items_count[3])
, VARRAY_count_t(item_totals[4] + items_count[4])
, VARRAY_count_t(item_totals[5] + items_count[5])
, VARRAY_count_t(item_totals[6] + items_count[6])
, VARRAY_count_t(item_totals[7] + items_count[7])
, VARRAY_count_t(item_totals[8] + items_count[8])
, VARRAY_count_t(item_totals[9] + items_count[9])
};

const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;
TREEDB_NS(runtime_ptr_t) runtime;
unsigned int test_num = 0;

void impl_memmove(void * to, const void * from, size_t size)
{
	printf("### memmove(%zu, %zu, %zu) ###\n"
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, to)
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, from)
		, size
	);
	memmove(to, from, size);
}

void dumpA()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = A;
	if(pa->m_offset.m_count <= 1)
		return;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	cout << "dump A " << (pa->m_offset.m_count-1) << '\n';
	for(n = 0; n < (unsigned int)(pa->m_offset.m_count-1); ++n)
		cout << "\t[" << n << "] offs " << setw(2) << po[n] <<", sz "
			<< setw(2) << (po[n+1] - po[n])
			<< " : " << to_string(pd + po[n], po[n+1] - po[n]) << '\n';
	cout << endl;
}
void dumpB()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = B;
	if(pa->m_offset.m_count <= 1)
		return;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	cout << "dump B " << (pa->m_offset.m_count-1) << '\n';
	for(n = 0; n < (unsigned int)(pa->m_offset.m_count-1); ++n)
		cout << "\t[" << n << "] offs " << setw(2) << po[n] <<", sz "
			<< setw(2) << (po[n+1] - po[n])
			<< " : " << to_string(pd + po[n], po[n+1] - po[n]) << '\n';
	cout << endl;
}
void dump()
{
	dumpA();
	dumpB();
}

int check_expected(const char * s, TREEDB_PAYLOAD_NS(varray_ptr_t) pa)
{
	uint16_t count = strlen(s);
	if(count != (pa->m_offset.m_count-1)) {
		cerr << "check_expected failed(" << (pa == A ? 'A' : 'B' )
			<< ") count(" << (pa->m_offset.m_count-1) << ") != expected("
			<< count << ")." << endl;
		return 1;
	}
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	uint16_t n;
	for(n = 0; n < count; ++n) {
		uint16_t u = s[n] - '0';
		if(items_count[u] != (po[n+1] - po[n])) {
			cerr << "check_expected failed(" << (pa == A ? 'A' : 'B' )
				<< ") entry[" << n << "] count(" << (items_count[u])
				<< ") != expected("
				<< (po[n+1] - po[n]) << ")." << endl;
			return 2;
		}
		if(memcmp(pd + po[n], items[u], po[n+1] - po[n])) {
			cerr << "check_expected failed(" << (pa == A ? 'A' : 'B' )
				<< ") entry[" << n << "] \""
				<< to_string(pd + po[n], po[n+1] - po[n])
				<< "\" != expected \""
				<< to_string(items[u], items_count[u])
				<< "\"." << endl;
			return 3;
		}
	}
	return 0;
}

#include "two-test-common.h"

int setup_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res1 = TREEDB_NS(heap_create)
		(& han
		, UUID_treedb_two_varray_uint_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}
	runtime = &han;

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(res2 == eTreedbSearchEmpty)
		return 1;

	free_size = pFirstNode->m_uLength;

	if(TREEDB_PAYLOAD_NS(init)(runtime, PACKED_A, 10, 5, 10)) {
		printf("Varray A initialisation failed\n");
		return 1;
	}
	if(TREEDB_PAYLOAD_NS(init)(runtime, PACKED_B, 10, 5, 10)) {
		printf("Varray B initialisation failed\n");
		return 1;
	}

	unsigned int n;
	TREEDB_NS(dump_trees)(runtime);
	for(n = 0; n < ITEM_COUNT; ++n) {
		res1 = TREEDB_PAYLOAD_NS(push_back)
				(runtime, PACKED_A, items[n], items_count[n]);
		if(res1) {
			cerr << '[' << n << "] push_back(A) failed(" << res1 << ")." << endl;
			return 1;
		}
	}
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = A;
	if((pa->m_offset.m_count-1) != ITEM_COUNT) {
		cerr << "pa->m_count A(" << (pa->m_offset.m_count-1) << ") != "
			<< ITEM_COUNT << '.' << endl;
		return 2;
	}

	for(n = 0; n < ITEM_COUNT; ++n) {
		res1 = TREEDB_PAYLOAD_NS(push_back)
				(runtime, PACKED_B, items[n], items_count[n]);
		if(res1) {
			cerr << '[' << n << "] push_back(B) failed(" << res1 << ")." << endl;
			return 1;
		}
	}
	pa = B;
	if((pa->m_offset.m_count-1) != ITEM_COUNT) {
		cerr << "pa->m_count B(" << (pa->m_offset.m_count-1) << ") != "
			<< ITEM_COUNT << '.' << endl;
		return 2;
	}

	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}
int run_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_two_varray_uint_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("run_test: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	runtime = &han;

	if(!TREEDB_PAYLOAD_NS(remove)(runtime, PACKED_A, ITEM_COUNT)) {
		printf("run_test: remove(%u) XPASS.\n", ITEM_COUNT);
		return 2;
	}
	printf("Testing %s\n", tests[test_num].name);
	res = do_the_test(tests[test_num]);
	if(res) {
		printf("Test \"%s\" failed(%d).\n", tests[test_num].name, res);
		dump();
		return 3;
	}
	//dump();
	TREEDB_PAYLOAD_NS(reset)(runtime, PACKED_A);
	TREEDB_PAYLOAD_NS(reset)(runtime, PACKED_B);

	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_read_lock)(runtime);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		printf("Free size(%zu) != %zu.\n", (size_t)free_size
			, (size_t)pFirstNode->m_uLength);
		return 4;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

int test()
{
	int res = setup_test();
	if(res) {
		printf("setup_test failed.\n");
		return res;
	}
	res = run_test();
	if(res) {
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res;
	for(test_num = 0; test_num < COUNT_OF(tests); ++test_num) {
		res = test();
		unlink(test_file);
		if(res)
			break;
	}
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
