/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */
/*
#undef TREEDB_DEBUG
#define TREEDB_DEBUG 1
*/
#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <iomanip>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;

extern void impl_memmove(void * to, const void * from, size_t size);
extern void varray_print_offsets(const char * msg);
#define VARRAY_print_offsets(x) varray_print_offsets(x)

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define VARRAY_NO_CLEARDEFS
#include <treedb/varray.h>

// a227db8f-b31a-4ad8-a95c-e099b327fe43
UUID_DEFINE
( UUID_treedb_varray_string_schema
, 0xa2, 0x27, 0xdb, 0x8f
, 0xb3, 0x1a
, 0x4a, 0xd8
, 0xa9, 0x5c
, 0xe0, 0x99, 0xb3, 0x27, 0xfe, 0x43
);

// Note: no trailing '\0' terminators - the varray tracks their sizes.
const char items0[] = { 'b', 'l', 'u', 'e' };
const char items1[] = { 'g', 'r', 'e', 'e', 'n' };
const char items2[] = { 'r', 'e', 'd' };
const char items3[] = { 'y', 'e', 'l', 'l', 'o', 'w' };
const char items4[] = { 'm', 'a', 'g', 'e', 'n', 't', 'a' };
const char items5[] = { 'g', 'r', 'e', 'y' };
const char items6[] = { 's', 'i', 'l', 'v', 'e', 'r' };
const char items7[] = { 'g', 'o', 'l', 'd' };
const char items8[] = { 'o', 'r', 'a', 'n', 'g', 'e' };
const char items9[] = { 'b', 'l', 'a', 'c', 'k' };
const char * items[] =
{ items0, items1, items2, items3, items4, items5 , items6, items7, items8
, items9 };
VARRAY_count_t items_len[] =
{ sizeof(items0)
, sizeof(items1)
, sizeof(items2)
, sizeof(items3)
, sizeof(items4)
, sizeof(items5)
, sizeof(items6)
, sizeof(items7)
, sizeof(items8)
, sizeof(items9)
};
VARRAY_size_t items_sz[] =
{ 0
, items_sz[0] + items_len[0]
, items_sz[1] + items_len[1]
, items_sz[2] + items_len[2]
, items_sz[3] + items_len[3]
, items_sz[4] + items_len[4]
, items_sz[5] + items_len[5]
, items_sz[6] + items_len[6]
, items_sz[7] + items_len[7]
, items_sz[8] + items_len[8]
, items_sz[9] + items_len[9]
};
#define ITEM_COUNT (sizeof(items) / sizeof(items[0]))
#define some_sz 3
#define some_offs 4

const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;
TREEDB_NS(runtime_ptr_t) runtime;
int test_num = 0;

void impl_memmove(void * to, const void * from, size_t size)
{
	printf("### memmove(%zu, %zu, %zu) ###\n"
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, to)
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, from)
		, size
	);
	memmove(to, from, size);
}

void dump()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(pa->m_offset.m_count <= 1)
		return;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	char * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	cout << "dump " << (pa->m_offset.m_count-1) << '\n';
	for(n = 0; n < (unsigned int)(pa->m_offset.m_count-1); ++n)
		cout << "\t[" << n << "] offs " << setw(2) << po[n] <<", sz "
			<< setw(2) << (po[n+1] - po[n])
			<< " : " << string(pd + po[n], po[n+1] - po[n]) << '\n';
	cout << endl;
}
void varray_print_offsets(const char * msg)
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	printf("%s (%zu) :", msg, (size_t)pa->m_offset.m_count);
	for(n = 0; n < pa->m_offset.m_count; ++n)
		printf(" [%u]%zu", n, (size_t)po[n]);
	printf("\n");
	fflush(stdout);
}

typedef int (*PFN)();

struct TestEntry
{
	PFN fn;
	const char * name;
};
int clear_fn()
{
	if(TREEDB_PAYLOAD_NS(clear)(runtime, PACKED_VARRAY))
		return 1;
	return 0;
}
int push_front_many()
{
	if(TREEDB_PAYLOAD_NS(push_front_many)(runtime, PACKED_VARRAY
			, some_sz, items_len))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	VARRAY_size_t expected = items_sz[some_sz] + items_sz[ITEM_COUNT];
	if(po[ITEM_COUNT + some_sz] != expected) {
		printf("po[(ITEM_COUNT + %u)(%zu)](%zu) !="
			" (items_sz[%u](%zu) + items_sz[ITEM_COUNT](%zu))(%zu).\n"
			, some_sz
			, (size_t)(ITEM_COUNT + some_sz)
			, (size_t)po[ITEM_COUNT + some_sz]
			, some_sz
			, (size_t)items_sz[some_sz]
			, (size_t)items_sz[ITEM_COUNT]
			, (size_t)(items_sz[some_sz] + items_sz[ITEM_COUNT])
		);
		return 2;
	}
	return 0;
}
int push_back_many()
{
	if(TREEDB_PAYLOAD_NS(push_back_many)(runtime, PACKED_VARRAY, some_sz
			, items_len))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	char * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	unsigned int n;
	/*printf("Offsets(%zu) :", (size_t)pa->m_offset.m_count);
	for(n = 0; n < pa->m_offset.m_count; ++n)
		printf(" [%u]%zu", n, (size_t)po[n]);
	printf("\n");
	fflush(stdout);*/
	if(pa->m_offset.m_count != (ITEM_COUNT + some_sz + 1)) {
		printf("pa->m_offset.m_count(%zu) != (ITEM_COUNT(%zu) + %u + 1)"
			"(%zu).\n"
			, (size_t)pa->m_offset.m_count
			, (size_t)ITEM_COUNT
			, some_sz
			, (size_t)(ITEM_COUNT + some_sz + 1)
		);
		return 2;
	}
	for(n = 0; n < some_sz; ++n)
		memcpy(pd + po[n + ITEM_COUNT], items[n], items_len[n]);
	VARRAY_count_t expected = items_sz[some_sz] + items_sz[ITEM_COUNT];
	if(po[ITEM_COUNT + some_sz] != expected) {
		printf("po[(ITEM_COUNT + %u)(%zu)](%zu) !="
			" (items_sz[%u](%zu) + items_sz[ITEM_COUNT](%zu))(%zu).\n"
			, some_sz
			, (size_t)(ITEM_COUNT + some_sz)
			, (size_t)po[ITEM_COUNT + some_sz]
			, some_sz
			, (size_t)items_sz[some_sz]
			, (size_t)items_sz[ITEM_COUNT]
			, (size_t)(items_sz[some_sz] + items_sz[ITEM_COUNT])
		);
		return 3;
	}
	return 0;
}
int insert_many()
{
	if(TREEDB_PAYLOAD_NS(insert_many)(runtime, PACKED_VARRAY, some_offs
			, some_sz, items_len))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(pa->m_offset.m_count != (ITEM_COUNT + some_sz + 1)) {
		printf("pa->m_offset.m_count(%zu) != (ITEM_COUNT(%zu) + %u)(%zu).\n"
			, (size_t)pa->m_offset.m_count
			, (size_t)ITEM_COUNT
			, some_sz
			, (size_t)(ITEM_COUNT + some_sz + 1)
		);
		return 2;
	}
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	char * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	unsigned int n;
	for(n = 0; n < some_sz; ++n)
		memcpy(pd + po[n + some_sz + 1], items[n], items_len[n]);
	VARRAY_count_t expected = items_sz[some_sz] + items_sz[ITEM_COUNT];
	if(po[ITEM_COUNT + some_sz] != expected) {
		printf("po[(ITEM_COUNT + %u)(%zu)](%zu) !="
			" (items_sz[%u](%zu) + items_sz[ITEM_COUNT](%zu))(%zu).\n"
			, some_sz
			, (size_t)(ITEM_COUNT + some_sz)
			, (size_t)po[ITEM_COUNT + some_sz]
			, some_sz
			, (size_t)items_sz[some_sz]
			, (size_t)items_sz[ITEM_COUNT]
			, (size_t)(items_sz[some_sz] + items_sz[ITEM_COUNT])
		);
		return 3;
	}
	return 0;
}
int pop_front_many()
{
	int n;
	if(TREEDB_PAYLOAD_NS(pop_front_many)
			(runtime, PACKED_VARRAY, some_sz))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	for(n = 0; n < (int)pa->m_offset.m_count; ++n) {
		VARRAY_size_t expected = items_sz[n + some_sz] - items_sz[some_sz];
		if(po[n] != expected) {
			printf("[%u] %zu != %zu\n"
				, n, (size_t)po[n], (size_t)expected);
			return 2;
		}
	}
	return 0;
}
int pop_back_many()
{
	int n;
	if(TREEDB_PAYLOAD_NS(pop_back_many)
			(runtime, PACKED_VARRAY, some_sz))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	for(n = 0; n < (int)pa->m_offset.m_count; ++n) {
		VARRAY_size_t expected = items_sz[n];
		if(po[n] != expected) {
			printf("[%u] %zu != %zu.\n"
				, n, (size_t)po[n], (size_t)expected);
			return 2;
		}
	}
	return 0;
}
int remove_many()
{
	int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa/* = VARRAY*/;
	/*printf("before size: %zu.\n", (size_t)pa->m_payload.m_count);*/
	if(TREEDB_PAYLOAD_NS(remove_many)
			(runtime, PACKED_VARRAY, some_offs, some_sz))
		return 1;
	pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	/*printf("after size: %zu.\n", (size_t)pa->m_payload.m_count);
	printf("items_sz[7] - items_sz[4] : %zu.\n"
		, (size_t)(items_sz[7] - items_sz[4]));*/
	for(n = 0; n < some_offs; ++n) {
		VARRAY_size_t expected = items_sz[n];
		if(po[n] != expected) {
			printf("below [%u] %zu != %zu.\n"
				, n, (size_t)po[n], (size_t)expected);
			return 2;
		}
	}
	for(; n < (int)pa->m_offset.m_count; ++n) {
		VARRAY_size_t expected
			= items_sz[n + some_sz]
				- (items_sz[some_offs + some_sz] - items_sz[some_offs]);
		if(po[n] != expected) {
			printf("above [%u] %zu != %zu.\n"
				, n, (size_t)po[n], (size_t)expected);
			return 3;
		}
	}
	return 0;
}
int remove_all()
{
	TREEDB_PAYLOAD_NS(clear)(runtime, PACKED_VARRAY);
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(pa->m_offset.m_count != 1)
		return 1;
	return 0;
}
int reverse_some_start()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(TREEDB_PAYLOAD_NS(reverse_some)(runtime, pa, 0, some_sz)) {
		printf("reverse_some() failed.\n");
		return 1;
	}
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	char * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	VARRAY_count_t size = 0;
	for(n = 0; n < some_sz; ++n) {
		if(po[n] != size) {
			printf("reversed [%u] po[n](%zu) != size(%zu).\n"
				, n, (size_t)po[n], (size_t)size);
			return 1;
		}
		size += items_len[some_sz - 1 - n];
		if((po[n+1] - po[n]) != items_len[some_sz - 1 - n]) {
			printf("reversed [%u] (po[n+1] - po[n])(%zu) "
				"!= items_len[some_sz - 1 - n](%zu)\n"
				, n, (size_t)(po[n+1] - po[n])
				, (size_t)items_len[some_sz - 1 - n]);
			return 2;
		}
		if(memcmp(items[some_sz - 1 - n], pd + po[n]
				, items_len[some_sz - 1 - n])) {
			printf("reversed [%u] memcmp() failed.\n", n);
			return 3;
		}
	}
	for(; n < ITEM_COUNT; ++n) {
		if(po[n] != size) {
			printf("above [%u] po[n](%zu) != size(%zu).\n"
				, n, (size_t)po[n], (size_t)size);
			return 4;
		}
		size += items_len[n];
		if((po[n+1] - po[n]) != items_len[n]) {
			printf("above [%u] (po[n+1] - po[n])(%zu) != items_len[n](%zu)\n"
				, n, (size_t)(po[n+1] - po[n]), (size_t)items_len[n]);
			return 5;
		}
		if(memcmp(items[n], pd + po[n], items_len[n])) {
			printf("above [%u] memcmp() failed.\n", n);
			return 6;
		}
	}
	return 0;
}
int reverse_some_middle()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(TREEDB_PAYLOAD_NS(reverse_some)(runtime, pa, some_offs, some_sz)) {
		printf("reverse_some() failed.\n");
		return 1;
	}
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	char * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	VARRAY_count_t size = 0;
	for(n = 0; n < some_offs; ++n) {
		if(po[n] != size)
			return 2;
		size += items_len[n];
		if((po[n+1] - po[n]) != items_len[n])
			return 3;
		if(memcmp(items[n], pd + po[n], po[n+1] - po[n]))
			return 4;
	}
	for(; n < (some_offs + some_sz); ++n) {
		unsigned int cmp = ITEM_COUNT - n;
		if(po[n] != size) {
			printf("reversed [%u] po[n](%zu) != size(%zu).\n"
				, n, (size_t)po[n], (size_t)size);
			return 5;
		}
		size += items_len[cmp];
		if((po[n+1] - po[n]) != items_len[cmp]) {
			printf("reversed [%u] (po[n+1] - po[n])(%zu) "
				"!= items_len[%u](%zu)\n"
				, n, (size_t)(po[n+1] - po[n]), cmp, (size_t)items_len[cmp]);
			return 6;
		}
		if(memcmp(items[cmp], pd + po[n], items_len[cmp])) {
			cout << "reversed [" << n << "] memcmp("
				<< string(items[cmp], items_len[cmp]) << ", "
				<< string(pd + po[n], items_len[cmp]) << ") failed." << endl;
			return 7;
		}
	}
	for(; n < ITEM_COUNT; ++n) {
		if(po[n] != size)
			return 8;
		size += items_len[n];
		if((po[n+1] - po[n]) != items_len[n])
			return 9;
		if(memcmp(items[n], pd + po[n], po[n+1] - po[n]))
			return 10;
	}
	return 0;
}
int reverse_some_end()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(TREEDB_PAYLOAD_NS(reverse_some)(runtime, pa, ITEM_COUNT - some_sz
			, some_sz)) {
		printf("reverse_some() failed.\n");
		return 1;
	}
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	char * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	VARRAY_size_t size = 0;
	for(n = 0; n < (ITEM_COUNT - some_sz); ++n) {
		if(po[n] != size)
			return 2;
		size += items_len[n];
		if((po[n+1] - po[n]) != items_len[n])
			return 3;
		if(memcmp(items[n], pd + po[n], po[n+1] - po[n]))
			return 4;
	}
	for(; n < ITEM_COUNT; ++n) {
		unsigned int cmp = ITEM_COUNT - n + (ITEM_COUNT - some_sz - 1);
		if(po[n] != size) {
			printf("reversed [%u] po[n](%zu) != size(%zu).\n"
				, n, (size_t)po[n], (size_t)size);
			return 5;
		}
		size += items_len[cmp];
		if((po[n+1] - po[n]) != items_len[cmp]) {
			printf("reversed [%u] (po[n+1] - po[n])(%zu) "
				"!= items_len[%u](%zu)\n"
				, n, (size_t)(po[n+1] - po[n]), cmp, (size_t)items_len[cmp]);
			return 6;
		}
		if(memcmp(items[cmp], pd + po[n], items_len[cmp])) {
			cout << "reversed [" << n << "] memcmp("
				<< string(items[cmp], items_len[cmp]) << ", "
				<< string(pd + po[n], items_len[cmp]) << ") failed." << endl;
			return 7;
		}
	}
	return 0;
}
int reverse_all()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(reverse)(runtime, VARRAY);
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	char * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	VARRAY_size_t size = 0;
	for(n = 0; n < (ITEM_COUNT - 1); ++n) {
		if(po[n] != size)
			return 4;
		size += items_len[ITEM_COUNT - 1 - n];
		if((po[n+1] - po[n]) != items_len[ITEM_COUNT - 1 - n])
			return 2;
		if(memcmp(items[ITEM_COUNT - 1 - n], pd + po[n], po[n+1] - po[n]))
			return 3;
	}
	return 0;
}
#define TEST_ENTRY(name) {& name, #name}
TestEntry tests[] =
{
	TEST_ENTRY(clear_fn),
	TEST_ENTRY(push_front_many),
	TEST_ENTRY(push_back_many),
	TEST_ENTRY(insert_many),
	TEST_ENTRY(pop_front_many),
	TEST_ENTRY(pop_back_many),
	TEST_ENTRY(remove_many),
	TEST_ENTRY(remove_all),
	TEST_ENTRY(reverse_some_start),
	TEST_ENTRY(reverse_some_middle),
	TEST_ENTRY(reverse_some_end),
	TEST_ENTRY(reverse_all)
};

int setup_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res1 = TREEDB_NS(heap_create)
		(& han
		, UUID_treedb_varray_string_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}
	runtime = &han;

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(res2 == eTreedbSearchEmpty)
		return 1;

	free_size = pFirstNode->m_uLength;

	if(TREEDB_PAYLOAD_NS(init)(runtime, PACKED_VARRAY, 10, 5, 10)) {
		printf("Array initialisation failed\n");
		return 1;
	}

	unsigned int n;
	TREEDB_NS(dump_trees)(runtime);
	for(n = 0; n < ITEM_COUNT; ++n) {
		res1 = TREEDB_PAYLOAD_NS(push_back)
				(runtime, PACKED_VARRAY, items[n], items_len[n]);
		if(res1) {
			cerr << '[' << n << "] push_back() failed(" << res1 << ")." << endl;
			return 1;
		}
	}
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if((pa->m_offset.m_count-1) != ITEM_COUNT) {
		cerr << "pa->m_count(" << (pa->m_offset.m_count-1) << ") != "
			<< ITEM_COUNT << '.' << endl;
		return 2;
	}

	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}
int run_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_varray_string_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("run_test: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	runtime = &han;

	if(!TREEDB_PAYLOAD_NS(remove)(runtime, PACKED_VARRAY, ITEM_COUNT)) {
		printf("run_test: remove(%zu) XPASS.\n", ITEM_COUNT);
		return 2;
	}
	printf("Testing %s\n", tests[test_num].name);
	res = tests[test_num].fn();
	if(res) {
		printf("Test \"%s\" failed(%d).\n", tests[test_num].name, res);
		dump();
		return 3;
	}
	//dump();
	TREEDB_PAYLOAD_NS(reset)(runtime, PACKED_VARRAY);

	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_read_lock)(runtime);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		printf("Free size(%zu) != %zu.\n", (size_t)free_size
			, (size_t)pFirstNode->m_uLength);
		return 4;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

int test()
{
	int res = setup_test();
	if(res) {
		printf("fill test failed.\n");
		return res;
	}
	res = run_test();
	if(res) {
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res;
	for(test_num = 0; test_num < (int)(sizeof(tests)/sizeof(tests[0]));
			++test_num) {
		res = test();
		unlink(test_file);
		if(res)
			break;
	}
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
