/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */
/*
#undef TREEDB_DEBUG
#define TREEDB_DEBUG 1
*/
#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <iomanip>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;

extern void impl_memmove(void * to, const void * from, size_t size);
extern void varray_print_offsets(const char * msg);
#define VARRAY_print_offsets(x) varray_print_offsets(x)

string to_string(const uint16_t * p, uintptr_t len)
{
	string res;
	uintptr_t n;
	char c;
	for(n = 0; n < len; ++n) {
		c = (char)p[n];
		res += c;
	}
	return res;
}

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define VARRAY_NO_CLEARDEFS
#define VARRAY_data_t uint16_t
#include <treedb/varray.h>

// 4686ba15-9b4c-46ff-a31c-4e3e23407695
UUID_DEFINE
( UUID_treedb_varray_uint_schema
, 0x46, 0x86, 0xba, 0x15
, 0x9b, 0x4c
, 0x46, 0xff
, 0xa3, 0x1c
, 0x4e, 0x3e, 0x23, 0x40, 0x76, 0x95
);
// Note: no trailing '\0' terminators - the varray tracks their sizes.
const uint16_t items0[] = { 'b', 'l', 'u', 'e' };
const uint16_t items1[] = { 'g', 'r', 'e', 'e', 'n' };
const uint16_t items2[] = { 'r', 'e', 'd' };
const uint16_t items3[] = { 'y', 'e', 'l', 'l', 'o', 'w' };
const uint16_t items4[] = { 'm', 'a', 'g', 'e', 'n', 't', 'a' };
const uint16_t items5[] = { 'g', 'r', 'e', 'y' };
const uint16_t items6[] = { 's', 'i', 'l', 'v', 'e', 'r' };
const uint16_t items7[] = { 'g', 'o', 'l', 'd' };
const uint16_t items8[] = { 'o', 'r', 'a', 'n', 'g', 'e' };
const uint16_t items9[] = { 'b', 'l', 'a', 'c', 'k' };
const uint16_t * items[] =
{ items0, items1, items2, items3, items4, items5 , items6, items7, items8
, items9 };
VARRAY_count_t items_count[] =
{ COUNT_OF(items0)
, COUNT_OF(items1)
, COUNT_OF(items2)
, COUNT_OF(items3)
, COUNT_OF(items4)
, COUNT_OF(items5)
, COUNT_OF(items6)
, COUNT_OF(items7)
, COUNT_OF(items8)
, COUNT_OF(items9)
};
VARRAY_count_t item_totals[] =
{ VARRAY_count_t(0)
, VARRAY_count_t(item_totals[0] + items_count[0])
, VARRAY_count_t(item_totals[1] + items_count[1])
, VARRAY_count_t(item_totals[2] + items_count[2])
, VARRAY_count_t(item_totals[3] + items_count[3])
, VARRAY_count_t(item_totals[4] + items_count[4])
, VARRAY_count_t(item_totals[5] + items_count[5])
, VARRAY_count_t(item_totals[6] + items_count[6])
, VARRAY_count_t(item_totals[7] + items_count[7])
, VARRAY_count_t(item_totals[8] + items_count[8])
, VARRAY_count_t(item_totals[9] + items_count[9])
};
#define ITEM_COUNT COUNT_OF(items)
#define some_count 3
#define some_offs 4

const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;
TREEDB_NS(runtime_ptr_t) runtime;
unsigned int test_num = 0;

void impl_memmove(void * to, const void * from, size_t size)
{
	printf("### memmove(%zu, %zu, %zu) ###\n"
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, to)
		, (size_t)TREEDB_NS(impl_pack_pointer)(runtime, from)
		, size
	);
	memmove(to, from, size);
}

void dump()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(pa->m_offset.m_count <= 1)
		return;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	cout << "dump " << (pa->m_offset.m_count-1) << '\n';
	for(n = 0; n < (unsigned int)(pa->m_offset.m_count-1); ++n)
		cout << "\t[" << n << "] offs " << setw(2) << po[n] <<", sz "
			<< setw(2) << (po[n+1] - po[n])
			<< " : " << to_string(pd + po[n], po[n+1] - po[n]) << '\n';
	cout << endl;
}
void varray_print_offsets(const char * msg)
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	printf("%s (%zu) :", msg, (size_t)pa->m_offset.m_count);
	for(n = 0; n < pa->m_offset.m_count; ++n)
		printf(" [%u]%zu", n, (size_t)po[n]);
	printf("\n");
	fflush(stdout);
}

typedef int (*PFN)();

struct TestEntry
{
	PFN fn;
	const char * name;
};
int clear_fn()
{
	if(TREEDB_PAYLOAD_NS(clear)(runtime, PACKED_VARRAY))
		return 1;
	return 0;
}
int push_front_many()
{
	if(TREEDB_PAYLOAD_NS(push_front_many)(runtime, PACKED_VARRAY
			, some_count, items_count))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	VARRAY_size_t expected = item_totals[some_count] + item_totals[ITEM_COUNT];
	if(po[ITEM_COUNT + some_count] != expected) {
		printf("po[(ITEM_COUNT + %u)(%zu)](%zu) !="
			" (item_totals[%u](%zu) + item_totals[ITEM_COUNT](%zu))(%zu).\n"
			, some_count
			, (size_t)(ITEM_COUNT + some_count)
			, (size_t)po[ITEM_COUNT + some_count]
			, some_count
			, (size_t)item_totals[some_count]
			, (size_t)item_totals[ITEM_COUNT]
			, (size_t)(item_totals[some_count] + item_totals[ITEM_COUNT])
		);
		return 2;
	}
	return 0;
}
int push_back_many()
{
	if(TREEDB_PAYLOAD_NS(push_back_many)(runtime, PACKED_VARRAY, some_count
			, items_count))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	unsigned int n;
	/*printf("Offsets(%zu) :", (size_t)pa->m_offset.m_count);
	for(n = 0; n < pa->m_offset.m_count; ++n)
		printf(" [%u]%zu", n, (size_t)po[n]);
	printf("\n");
	fflush(stdout);*/
	if(pa->m_offset.m_count != (ITEM_COUNT + some_count + 1)) {
		printf("pa->m_offset.m_count(%zu) != (ITEM_COUNT(%zu) + %u + 1)"
			"(%zu).\n"
			, (size_t)pa->m_offset.m_count
			, (size_t)ITEM_COUNT
			, some_count
			, (size_t)(ITEM_COUNT + some_count + 1)
		);
		return 2;
	}
	for(n = 0; n < some_count; ++n)
		memcpy(pd + po[n + ITEM_COUNT]
			, items[n]
			, items_count[n] * sizeof(uint16_t));
	VARRAY_size_t expected = item_totals[some_count] + item_totals[ITEM_COUNT];
	if(po[ITEM_COUNT + some_count] != expected) {
		printf("po[(ITEM_COUNT + %u)(%zu)](%zu) !="
			" (item_totals[%u](%zu) + item_totals[ITEM_COUNT](%zu))(%zu).\n"
			, some_count
			, (size_t)(ITEM_COUNT + some_count)
			, (size_t)po[ITEM_COUNT + some_count]
			, some_count
			, (size_t)item_totals[some_count]
			, (size_t)item_totals[ITEM_COUNT]
			, (size_t)(item_totals[some_count] + item_totals[ITEM_COUNT])
		);
		return 3;
	}
	return 0;
}
int insert_many()
{
	if(TREEDB_PAYLOAD_NS(insert_many)(runtime, PACKED_VARRAY, some_offs
			, some_count, items_count))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(pa->m_offset.m_count != (ITEM_COUNT + some_count + 1)) {
		printf("pa->m_offset.m_count(%zu) != (ITEM_COUNT(%zu) + %u)(%zu).\n"
			, (size_t)pa->m_offset.m_count
			, (size_t)ITEM_COUNT
			, some_count
			, (size_t)(ITEM_COUNT + some_count + 1)
		);
		return 2;
	}
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	unsigned int n;
	for(n = 0; n < some_count; ++n)
		memcpy(pd + po[n + some_count + 1]
			, items[n]
			, items_count[n] * sizeof(uint16_t));
	VARRAY_size_t expected = item_totals[some_count] + item_totals[ITEM_COUNT];
	if(po[ITEM_COUNT + some_count] != expected) {
		printf("po[(ITEM_COUNT + %u)(%zu)](%zu) !="
			" (item_totals[%u](%zu) + item_totals[ITEM_COUNT](%zu))(%zu).\n"
			, some_count
			, (size_t)(ITEM_COUNT + some_count)
			, (size_t)po[ITEM_COUNT + some_count]
			, some_count
			, (size_t)item_totals[some_count]
			, (size_t)item_totals[ITEM_COUNT]
			, (size_t)(item_totals[some_count] + item_totals[ITEM_COUNT])
		);
		return 3;
	}
	return 0;
}
int pop_front_many()
{
	int n;
	if(TREEDB_PAYLOAD_NS(pop_front_many)
			(runtime, PACKED_VARRAY, some_count))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	for(n = 0; n < (int)pa->m_offset.m_count; ++n) {
		VARRAY_size_t expected
			= item_totals[n + some_count] - item_totals[some_count];
		if(po[n] != expected) {
			printf("[%u] %zu != %zu\n"
				, n, (size_t)po[n], (size_t)expected);
			return 2;
		}
	}
	return 0;
}
int pop_back_many()
{
	int n;
	if(TREEDB_PAYLOAD_NS(pop_back_many)
			(runtime, PACKED_VARRAY, some_count))
		return 1;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	for(n = 0; n < (int)pa->m_offset.m_count; ++n) {
		VARRAY_size_t expected = item_totals[n];
		if(po[n] != expected) {
			printf("[%u] %zu != %zu.\n"
				, n, (size_t)po[n], (size_t)expected);
			return 2;
		}
	}
	return 0;
}
int remove_many()
{
	int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa/* = VARRAY*/;
	/*printf("before size: %zu.\n", (size_t)pa->m_payload.m_count);*/
	if(TREEDB_PAYLOAD_NS(remove_many)
			(runtime, PACKED_VARRAY, some_offs, some_count))
		return 1;
	pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	/*printf("after size: %zu.\n", (size_t)pa->m_payload.m_count);
	printf("item_totals[7] - item_totals[4] : %zu.\n"
		, (size_t)(item_totals[7] - item_totals[4]));*/
	for(n = 0; n < some_offs; ++n) {
		VARRAY_size_t expected = item_totals[n];
		if(po[n] != expected) {
			printf("below [%u] %zu != %zu.\n"
				, n, (size_t)po[n], (size_t)expected);
			return 2;
		}
	}
	for(; n < (int)pa->m_offset.m_count; ++n) {
		VARRAY_size_t expected
			= item_totals[n + some_count]
				- (item_totals[some_offs + some_count] - item_totals[some_offs]);
		if(po[n] != expected) {
			printf("above [%u] %zu != %zu.\n"
				, n, (size_t)po[n], (size_t)expected);
			return 3;
		}
	}
	return 0;
}
int remove_all()
{
	TREEDB_PAYLOAD_NS(clear)(runtime, PACKED_VARRAY);
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(pa->m_offset.m_count != 1)
		return 1;
	return 0;
}
int reverse_some_start()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(TREEDB_PAYLOAD_NS(reverse_some)(runtime, pa, 0, some_count)) {
		printf("reverse_some() failed.\n");
		return 1;
	}
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	VARRAY_size_t size = 0;
	for(n = 0; n < some_count; ++n) {
		if(po[n] != size) {
			printf("reversed [%u] po[n](%zu) != size(%zu).\n"
				, n, (size_t)po[n], (size_t)size);
			return 1;
		}
		size += items_count[some_count - 1 - n];
		if((po[n+1] - po[n]) != items_count[some_count - 1 - n]) {
			printf("reversed [%u] (po[n+1] - po[n])(%zu) "
				"!= items_count[some_count - 1 - n](%zu)\n"
				, n, (size_t)(po[n+1] - po[n])
				, (size_t)items_count[some_count - 1 - n]);
			return 2;
		}
		if(memcmp(items[some_count - 1 - n], pd + po[n]
				, items_count[some_count - 1 - n])) {
			printf("reversed [%u] memcmp() failed.\n", n);
			return 3;
		}
	}
	for(; n < ITEM_COUNT; ++n) {
		if(po[n] != size) {
			printf("above [%u] po[n](%zu) != size(%zu).\n"
				, n, (size_t)po[n], (size_t)size);
			return 4;
		}
		size += items_count[n];
		if((po[n+1] - po[n]) != items_count[n]) {
			printf("above [%u] (po[n+1] - po[n])(%zu) != items_count[n](%zu)\n"
				, n, (size_t)(po[n+1] - po[n]), (size_t)items_count[n]);
			return 5;
		}
		if(memcmp(items[n], pd + po[n], items_count[n])) {
			printf("above [%u] memcmp() failed.\n", n);
			return 6;
		}
	}
	return 0;
}
int reverse_some_middle()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(TREEDB_PAYLOAD_NS(reverse_some)(runtime, pa, some_offs, some_count)) {
		printf("reverse_some() failed.\n");
		return 1;
	}
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	VARRAY_size_t size = 0;
	for(n = 0; n < some_offs; ++n) {
		if(po[n] != size)
			return 2;
		size += items_count[n];
		if((po[n+1] - po[n]) != items_count[n])
			return 3;
		if(memcmp(items[n], pd + po[n], po[n+1] - po[n]))
			return 4;
	}
	for(; n < (some_offs + some_count); ++n) {
		unsigned int cmp = ITEM_COUNT - n;
		if(po[n] != size) {
			printf("reversed [%u] po[n](%zu) != size(%zu).\n"
				, n, (size_t)po[n], (size_t)size);
			return 5;
		}
		size += items_count[cmp];
		if((po[n+1] - po[n]) != items_count[cmp]) {
			printf("reversed [%u] (po[n+1] - po[n])(%zu) "
				"!= items_count[%u](%zu)\n"
				, n, (size_t)(po[n+1] - po[n]), cmp, (size_t)items_count[cmp]);
			return 6;
		}
		if(memcmp(items[cmp], pd + po[n], items_count[cmp])) {
			cout << "reversed [" << n << "] memcmp("
				<< to_string(items[cmp], items_count[cmp]) << ", "
				<< to_string(pd + po[n], items_count[cmp]) << ") failed." << endl;
			return 7;
		}
	}
	for(; n < ITEM_COUNT; ++n) {
		if(po[n] != size)
			return 8;
		size += items_count[n];
		if((po[n+1] - po[n]) != items_count[n])
			return 9;
		if(memcmp(items[n], pd + po[n], po[n+1] - po[n]))
			return 10;
	}
	return 0;
}
int reverse_some_end()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if(TREEDB_PAYLOAD_NS(reverse_some)(runtime, pa, ITEM_COUNT - some_count
			, some_count)) {
		printf("reverse_some() failed.\n");
		return 1;
	}
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	VARRAY_size_t size = 0;
	for(n = 0; n < (ITEM_COUNT - some_count); ++n) {
		if(po[n] != size)
			return 2;
		size += items_count[n];
		if((po[n+1] - po[n]) != items_count[n])
			return 3;
		if(memcmp(items[n], pd + po[n], po[n+1] - po[n]))
			return 4;
	}
	for(; n < ITEM_COUNT; ++n) {
		unsigned int cmp = ITEM_COUNT - n + (ITEM_COUNT - some_count - 1);
		if(po[n] != size) {
			printf("reversed [%u] po[n](%zu) != size(%zu).\n"
				, n, (size_t)po[n], (size_t)size);
			return 5;
		}
		size += items_count[cmp];
		if((po[n+1] - po[n]) != items_count[cmp]) {
			printf("reversed [%u] (po[n+1] - po[n])(%zu) "
				"!= items_count[%u](%zu)\n"
				, n, (size_t)(po[n+1] - po[n]), cmp, (size_t)items_count[cmp]);
			return 6;
		}
		if(memcmp(items[cmp], pd + po[n], items_count[cmp])) {
			cout << "reversed [" << n << "] memcmp("
				<< to_string(items[cmp], items_count[cmp]) << ", "
				<< to_string(pd + po[n], items_count[cmp]) << ") failed." << endl;
			return 7;
		}
	}
	return 0;
}
int reverse_all()
{
	unsigned int n;
	TREEDB_PAYLOAD_NS(reverse)(runtime, VARRAY);
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	VARRAY_count_t * po
		= TREEDB_PAYLOAD_NS(offset_unpack_data)(runtime, pa->m_offset.m_items);
	uint16_t * pd
		= TREEDB_PAYLOAD_NS(payload_unpack_data)(runtime, pa->m_payload.m_items);
	VARRAY_size_t size = 0;
	for(n = 0; n < (ITEM_COUNT - 1); ++n) {
		if(po[n] != size)
			return 4;
		size += items_count[ITEM_COUNT - 1 - n];
		if((po[n+1] - po[n]) != items_count[ITEM_COUNT - 1 - n])
			return 2;
		if(memcmp(items[ITEM_COUNT - 1 - n], pd + po[n], po[n+1] - po[n]))
			return 3;
	}
	return 0;
}
#define TEST_ENTRY(name) {& name, #name}
TestEntry tests[] =
{
	TEST_ENTRY(clear_fn),
	TEST_ENTRY(push_front_many),
	TEST_ENTRY(push_back_many),
	TEST_ENTRY(insert_many),
	TEST_ENTRY(pop_front_many),
	TEST_ENTRY(pop_back_many),
	TEST_ENTRY(remove_many),
	TEST_ENTRY(remove_all),
	TEST_ENTRY(reverse_some_start),
	TEST_ENTRY(reverse_some_middle),
	TEST_ENTRY(reverse_some_end),
	TEST_ENTRY(reverse_all)
};

int setup_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res1 = TREEDB_NS(heap_create)
		(& han
		, UUID_treedb_varray_uint_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}
	runtime = &han;

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(res2 == eTreedbSearchEmpty)
		return 1;

	free_size = pFirstNode->m_uLength;

	if(TREEDB_PAYLOAD_NS(init)(runtime, PACKED_VARRAY, 10, 5, 10)) {
		printf("Array initialisation failed\n");
		return 1;
	}

	unsigned int n;
	TREEDB_NS(dump_trees)(runtime);
	for(n = 0; n < ITEM_COUNT; ++n) {
		res1 = TREEDB_PAYLOAD_NS(push_back)
				(runtime, PACKED_VARRAY, items[n], items_count[n]);
		if(res1) {
			cerr << '[' << n << "] push_back() failed(" << res1 << ")." << endl;
			return 1;
		}
		//dump();
	}
	TREEDB_PAYLOAD_NS(varray_ptr_t) pa = VARRAY;
	if((pa->m_offset.m_count-1) != ITEM_COUNT) {
		cerr << "pa->m_count(" << (pa->m_offset.m_count-1) << ") != "
			<< ITEM_COUNT << '.' << endl;
		return 2;
	}

	TREEDB_NS(heap_unlock)(&han);
	TREEDB_NS(heap_close)(&han);
	return 0;
}
int run_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_varray_uint_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("run_test: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	runtime = &han;
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_write_lock)(runtime);

	if(!TREEDB_PAYLOAD_NS(remove)(runtime, PACKED_VARRAY, ITEM_COUNT)) {
		printf("run_test: remove(%zu) XPASS.\n", ITEM_COUNT);
		return 2;
	}
	printf("Testing %s\n", tests[test_num].name);
	res = tests[test_num].fn();
	if(res) {
		printf("Test \"%s\" failed(%d).\n", tests[test_num].name, res);
		dump();
		return 3;
	}
	//dump();
	TREEDB_PAYLOAD_NS(reset)(runtime, PACKED_VARRAY);

	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_read_lock)(runtime);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		printf("Free size(%zu) != %zu.\n", (size_t)free_size
			, (size_t)pFirstNode->m_uLength);
		return 4;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(&han);
	return 0;
}

int test()
{
	int res = setup_test();
	if(res) {
		printf("fill test failed.\n");
		return res;
	}
	res = run_test();
	if(res) {
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res;
	for(test_num = 0; test_num < COUNT_OF(tests); ++test_num) {
		res = test();
		unlink(test_file);
		if(res)
			break;
	}
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
