/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform - passing.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <string>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

//#define TREEDB_DEBUG 1

using namespace std;
using namespace v3c;

#define TREEDB_CHECK_HEAP(runtime) allocator_check_heap()
int heap_check_res = 0;
int allocator_check_heap();

#define TREEDB_have_dump_trees
#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define STRING_TREE_NO_CLEARDEFS
#define AVL_NO_CLEARDEFS
#include <treedb/avl-string.h>
#include "dump-string-trees.h"
#include <treedb/comet-guids.h>

TREEDB_NS(runtime_ptr_t) runtime = 0;

int allocator_check_heap()
{
	if(heap_check_res)
		return heap_check_res;
	heap_check_res = TREEDB_NS(check_heap)(runtime);
	return heap_check_res;
}

const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;

int fill_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	runtime = & han;
	int res1 = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_string_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("fill_test: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}
	STRING_TREE_NS(init)(runtime, STRING_TREE);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);

	if(res2 == eTreedbSearchEmpty)
		return 1;
	free_size = pFirstNode->m_uLength;

	string word;
	ifstream in("../../../v3c/3-comet/cpp_wordset_test.cpp");
	while(in.good()) {
		in >> word;
		STRING_TREE_NS(insert)(runtime, PACKED_STRING_TREE, word.c_str(), 0);
#if TREEDB_DEBUG != 1
		if(allocator_check_heap())
			return 1;
#endif
		if(heap_check_res) {
			printf("fill_test: heap check failed: %d.\n", heap_check_res);
			TREEDB_NS(heap_unlock)(runtime);
			TREEDB_NS(heap_close)(runtime);
			return 2;
		}
	}
	//TREEDB_NS(dump_trees)(runtime);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return 0;
}

void append_int
( STRING_TREE_NS(node_ptr_t) node
, unsigned int u)
{
	if(!node)
		return;
	unsigned int n;
	for(n = 0; node; ++n) {
#if TREEDB_DEBUG != 1
		if(allocator_check_heap())
			return;
#endif
		if(heap_check_res) {
			printf("append_int: heap check failed: %d.\n", heap_check_res);
			return;
		}
		char text[24];
		sprintf(text, "_%d", u);
		TREEDB_printf("append_int [%u] : node [%zu] %s -> %s%s\n", n
			, (size_t)TREEDB_pack_pointer(node)
			, node->m_item, node->m_item, text);

		/* The node's storage may be freed so unlink it just in case. */
		STRING_TREE_NS(unlink_node)(runtime, STRING_TREE, node);

		bool bTrim = false;
		size_t len = strlen(node->m_item);
		size_t space = len;
		if(len > 99) {
			len = space = len - 50;
			bTrim = true;
		}else
			space += strlen(text);
		space += 1 + offsetof(STRING_TREE_NS(node_t), m_item);
		STRING_TREE_NS(node_ptr_t) node2
		= (STRING_TREE_NS(node_ptr_t))TREEDB_NS(realloc)
			(runtime, node, space);

		/* Don't use node from here on, in case the address space moved. */
		if(!node2) {
			TREEDB_printf("append_int [%u] : realloc failed.\n", n);
			/* We now know that no address space move occurred. */
			STRING_TREE_NS(insert_node)(runtime, STRING_TREE, node);
			//TREEDB_NS(dump_trees)(runtime);
			node = STRING_TREE_NS_unpack_node(node->m_next);
			continue;
		}
		if(bTrim)
			node2->m_item[len] = '\0';
		else
			strcpy(node2->m_item + strlen(node2->m_item), text);
		if(STRING_TREE_NS(insert_node)(runtime, STRING_TREE, node2))
			node = STRING_TREE_NS_unpack_node(node2->m_next);
		else{
			/* A duplicate. */
			node = STRING_TREE_NS_unpack_node(node2->m_next);
			TREEDB_NS(free)(runtime, node2);
		}
	}
}
int empty_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	runtime = & han;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_string_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("main: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	TREEDB_NS(heap_unlock)(runtime);
	unsigned int n = 0, count = 0, total = 0;
	int v;
	TREEDB_NSA(node_ptr_t) pFirstNode = 0;

	STRING_TREE_NS(node_ptr_t) p;
	for(n = 0; true; ++n) {
		TREEDB_NS(heap_write_lock)(runtime);
		count = STRING_TREE_NS(count)(runtime, STRING_TREE);
		if(!total)
			total = count;
		if(!count) {
			TREEDB_NS(heap_unlock)(runtime);
			break;
		}
		v = rand();
		if(v < 0)
			v = -v;
		while((unsigned int)v >= count)
			v >>= 1;
		// Look up the item at index v.
		p = STRING_TREE_NS(index)(runtime, STRING_TREE, v);
		if(!p) {
			printf("get failed.\n");
			TREEDB_NS(heap_unlock)(runtime);
			break;
		}
		//printf("Remove : %u/%u %s\n", n + 1, total, p->m_item);
		STRING_TREE_NS(delete_node)(runtime, PACKED_STRING_TREE, p);
		//printf("Remove : after.\n");
		//TREEDB_NS(dump_trees)(runtime);
		append_int(STRING_TREE_NS_unpack_node(
			runtime->m_pTreedb->m_string_tree.m_head), n);
		if(heap_check_res) {
			printf("append_int: heap check failed: %d.\n", heap_check_res);
			TREEDB_NS(heap_unlock)(runtime);
			TREEDB_NS(heap_close)(runtime);
			return 2;
		}
		TREEDB_NS(heap_unlock)(runtime);
	}
	TREEDB_NS(heap_read_lock)(runtime);
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		printf("Error : wanted free_size == %zu, got %zu instead.\n"
			, (size_t)(free_size), (size_t)(pFirstNode->m_uLength));
		return 4;
	}
	if(pFirstNode->m_offset_next/*.node.value*/) {
		printf("Error : more than 1 free node.\n");
		return 5;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return 0;
}

int test()
{
	int res = fill_test();
	if(res) {
		printf("fill test failed.\n");
		return res;
	}
	res = empty_test();
	if(res) {
		printf("empty test failed.\n");
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res = test();
	unlink(test_file);
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
