/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#if TREEDB_NBITS==-1
#include <stdio.h>
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	printf("Not testing on a 32 bit platform - passing.\n");
	return 0;
}
#else

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <stdexcept>
#include <iostream>
#include <string>
#include <error.h>
#include <v3c/pretty_print_streambuf.hpp>
#include <v3c/string_utils.hpp>

using namespace std;
using namespace v3c;

#define TREEDB_SHARED
#define TREEDB_HAS_RESIZABLE_HEAP
#define TREEDB_have_dump_trees
#define STRING_TREE_NO_CLEARDEFS
#define AVL_NO_CLEARDEFS
#include "avl-string.h"
#include "dump-string-trees.h"
#include <treedb/comet-guids.h>

const char * test_file = "test.3db";
TREEDB_NS(count_t) free_size;

int fill_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	TREEDB_NS(runtime_ptr_t) runtime = &han;
	int res1 = TREEDB_NS(heap_create)
		( runtime
		, UUID_treedb_string_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		, TREEDB_PAGE_SIZE
		);
	if(res1) {
		printf("main: TREEDB_NS(heap_create)() failed: %d.\n", res1);
		return 1;
	}
	TREEDB_PAYLOAD_NS(init)(runtime, STRING_TREE);

	TREEDB_NSA(node_ptr_t) pFirstNode = 0;
	TreedbSearchResult res2 = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);

	if(res2 == eTreedbSearchEmpty)
		return 1;
	free_size = pFirstNode->m_uLength;

	string word;
	while(cin.good()) {
		cin >> word;
		TREEDB_PAYLOAD_NS(insert)(runtime, PACKED_STRING_TREE, word.c_str(), 0);
	}
	TREEDB_NS(dump_trees)(runtime);
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return 0;
}

int empty_test()
{
	TREEDB_NS(runtime_t) han = TREEDB_STORAGE_RUNTIME_INIT;
	int res = TREEDB_NS(heap_open)
		(& han
		, UUID_treedb_string_tree_schema
		, test_file
		, eHeapOpenRead | eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res) {
		printf("main: TREEDB_NS(heap_open)() failed: %d.\n", res);
		return 1;
	}
	TREEDB_NS(runtime_ptr_t) runtime = &han;
	TREEDB_NS(heap_unlock)(runtime);

	unsigned int n;
	int v;
	TREEDB_NSA(node_ptr_t) pFirstNode = 0;

	TREEDB_PAYLOAD_NS(node_ptr_t) p;
	while(true) {
		TREEDB_NS(heap_write_lock)(runtime);
		n = TREEDB_PAYLOAD_NS(count)(runtime, STRING_TREE);
		printf("[%u]\n", n);
		if(!n) {
			TREEDB_NS(dump_trees)(runtime);
			TREEDB_NS(heap_unlock)(runtime);
			break;
		}else if(n <= 3)
			TREEDB_NS(dump_trees)(runtime);
		v = rand();
		if(v < 0)
			v = -v;
		while((unsigned int)v >= n)
			v >>= 1;
		// Look up the item at index v.
		p = TREEDB_PAYLOAD_NS(index)(runtime, STRING_TREE, v);
		if(!p) {
			printf("get failed.\n");
			TREEDB_NS(heap_unlock)(runtime);
			break;
		}
		TREEDB_PAYLOAD_NS(delete_node)(runtime, PACKED_STRING_TREE, p);
		TREEDB_NS(heap_unlock)(runtime);
	}
	TREEDB_NS(heap_read_lock)(runtime);
	res = TREEDB_NSF(search_closest)
		(runtime, TREEDB_ALLOC_free_tree, 0, & pFirstNode);
	if(free_size != pFirstNode->m_uLength) {
		printf("Error : wanted free_size == %zu, got %zu instead.\n"
			, (size_t)(free_size), (size_t)(pFirstNode->m_uLength));
		return 4;
	}
	if(pFirstNode->m_offset_next/*.node.value*/) {
		printf("Error : more than 1 free node.\n");
		return 5;
	}
	TREEDB_NS(heap_unlock)(runtime);
	TREEDB_NS(heap_close)(runtime);
	return 0;
}

int test()
{
	int res = fill_test();
	if(res) {
		printf("fill test failed.\n");
		return res;
	}
	res = empty_test();
	if(res) {
		printf("empty test failed.\n");
		return res;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;

	int res = test();
	unlink(test_file);
	return res;
}
#endif /* #if TREEDB_NBITS==-1 */
