/* Copyright (C) 2011  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

//#define TREEDB_DEBUG_MMAP

#if !defined(TREEDB_DEBUG) && defined(TREEDB_DEBUG_MMAP)
  #define TREEDB_DEBUG 1
#endif

#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <mcheck.h>
#include <pthread.h>
#include <errno.h>
#include <v3c/atomic.h>
#include <sys/stat.h>
#include <sys/mman.h>

#if TREEDB_DEBUG
  #define TREEDB_DEBUG_RENDERERS
#endif

#ifndef TREEDB_PRIVATE_HEAP
static int g_control_fd = -1;
static int g_trace_fd = -1;
#else
  #define g_trace_fd STDERR_FILENO
#endif

#ifdef TREEDB_DEBUG_MMAP
#define mmap my_mmap
static void * my_mmap(void *__addr, size_t __len, int __prot,
		   int __flags, int __fd, __off_t __offset) __THROW;
#endif

static inline unsigned int my_utoa(char * buf, unsigned int v)
{
	int ndigits = 1;
	size_t max10 = 10;
	while(max10 < v) {
		max10 *= 10;
		++ndigits;
	}
	unsigned int n;
	for(n = 0; n < max10; ++n) {
		max10 /= 10;
		*buf++ = '0' + v / max10;
		v -= (v / max10) * max10;
	}
	*buf++ = '\0';
	return n;
}
static inline unsigned int my_itoa(char * buf, int v)
{
	if(v < 0) {
		*buf++ = '-';
		return 1 + my_utoa(buf, (unsigned int)(-v));
	}else
		return my_utoa(buf, (unsigned int)v);
}

#include <stdint.h>
#include <stdarg.h>

/* NOTE the 'c' param can have side effects. */
#define TREEDB_PCHAR(c) { \
	char ** pendp = (char **)arg; \
	*(*pendp)++ = c; \
	++retval; \
}

#include "printf.c"

static inline void
TREEDB_MALLOC_sprintf(char * buf, const char *fmt, ...)
{
	char * endp = buf;
	/* http://www.pagetable.com/?p=298 */
	va_list ap;

	va_start(ap, fmt);
	TREEDB_kvprintf(fmt, NULL, & endp, 10, ap);
	va_end(ap);
	*endp = '\0';
}
#if TREEDB_DEBUG
static char msgbuf[1024];

static inline void
TREEDB_MALLOC_printf(const char *fmt, ...)
{
	char * endp = msgbuf;
	/* http://www.pagetable.com/?p=298 */
	va_list ap;

	va_start(ap, fmt);
	TREEDB_kvprintf(fmt, NULL, & endp, 10, ap);
	va_end(ap);
	(void)write(g_trace_fd, msgbuf, endp - msgbuf);
}

  #define TREEDB_printf_DEFINED
  #define TREEDB_printf2_DEFINED
  #define TREEDB_printf4_DEFINED
  #define TREEDB_printf8_DEFINED
  #if TREEDB_DEBUG & 1
    #define TREEDB_printf(fmt, ...) \
      TREEDB_MALLOC_printf(msgbuf, fmt, ## __VA_ARGS__)
  #else
    #define TREEDB_printf(fmt, ...)
  #endif
  #if TREEDB_DEBUG & 2
    #define TREEDB_printf2(fmt, ...) \
      TREEDB_MALLOC_printf(msgbuf, fmt, ## __VA_ARGS__)
  #else
    #define TREEDB_printf2(fmt, ...)
  #endif
  #if TREEDB_DEBUG & 4
    #define TREEDB_printf4(fmt, ...) \
      TREEDB_MALLOC_printf(msgbuf, fmt, ## __VA_ARGS__)
  #else
    #define TREEDB_printf4(fmt, ...)
  #endif
  #if TREEDB_DEBUG & 8
    #define TREEDB_printf8(fmt, ...) \
      TREEDB_MALLOC_printf(msgbuf, fmt, ## __VA_ARGS__)
  #else
    #define TREEDB_printf8(fmt, ...)
  #endif

#endif

static int ok = 0;

static int g_dev_zero_fd = -1;

#define v3c_declare_prefix static
#define v3c_prefix static

/* We flush hard so the daemon knows about them - it still caches. */
#define TREEDB_MSYNC MS_SYNC

#define v3c_treedb_site_config_h_included
#define TREEDB_ALIGN TREEDB_NBITS
#define TREEDB_PAGE_SIZE 65536 /* 64K */

#if !(TREEDB_DEBUG & 1)
  #define TREEDB_OMIT_TreedbSearchResultDesc
  #define TREEDB_OMIT_TreedbInitResultDesc
  #define TREEDB_OMIT_v3c_c_strerror
#endif
#define TREEDB_OMIT_TreedbInsertResultDesc
#define TREEDB_OMIT_TreedbHeapStatusDesc
#define TREEDB_OMIT_TreedbFreeResultDesc
#define TREEDB_OMIT_TreedbKeyDistanceFlagsDesc
#define TREEDB_OMIT_TreedbNodeDistanceFlagsDesc
#define TREEDB_OMIT_TreedbHeapMismatched
#define TREEDB_OMIT_FILE_STORAGE_get_size
#define V3C_OMIT_file_lock_name
#define V3C_OMIT_file_lock_state_term
#define TREEDB_OMIT_TreedbAllocatorAbort

#if TREEDB_DEBUG
  #if (TREEDB_DEBUG & 1) /* also enable treedb-impl checks. */
    #define TREEDB_CHECK_HEAP(runtime) allocator_check_heap()
  #endif
  /* Otherwise only check on page allocs/frees. */
static int allocator_check_heap();
#else
  #define allocator_check_heap()
#endif

#define TREEDB_NS(x) TREEDB_NS_CONCAT(V3C_JOIN3(treedb_malloc_,x,_),)

/* We want the unused heap to mmap to /dev/zero */
#define TREEDB_HAS_RESIZABLE_HEAP
#define TREEDB_page_alloc(sz) TREEDB_NS(do_page_alloc)(runtime, sz)
#define TREEDB_page_free(sz) TREEDB_NS(do_page_free)(runtime, sz)
#define TREEDB_heap_init_size(offset, size) \
TREEDB_NS(do_init_size)(runtime, offset, size)

#include "treedb-malloc.h"

#define MALLOC_trace(fmt, ...) { \
	TREEDB_printf(fmt "\n", ## __VA_ARGS__); \
}
#define MALLOC_traceQ(fmt, ...) { \
	TREEDB_printf(fmt, ## __VA_ARGS__); \
}
#define MALLOC_traceX(fmt, ...)

#define Exit(fmt, ...) { \
	TREEDB_printf("TreedbMalloc: "fmt "\n", ## __VA_ARGS__); \
	treedb_malloc_term(); \
	exit(127); \
}
#define Exit2(fmt, ...) { \
	TREEDB_printf("TreedbMalloc: " fmt " : %s.\n", ## __VA_ARGS__, strerror(errno)); \
	treedb_malloc_term(); \
	exit(127); \
}

#define InitExit(msg) { \
	(void)write(STDERR_FILENO, msg "\n", strlen(msg "\n")); \
	pthread_mutex_unlock(& g_runtime.m_mx); \
	treedb_malloc_term(); \
	exit(127); \
}
#define InitExitErrno(msg) { \
	const char *str; \
	(void)write(STDERR_FILENO, msg, strlen(msg)); \
	str = " : "; \
	(void)write(STDERR_FILENO, str, strlen(str)); \
	str = strerror(errno); \
	(void)write(STDERR_FILENO, str, strlen(str)); \
	str = ".\n"; \
	(void)write(STDERR_FILENO, str, strlen(str)); \
	pthread_mutex_unlock(& g_runtime.m_mx); \
	treedb_malloc_term(); \
	exit(127); \
}

#define TREEDB_NS_prefix static inline
#define TREEDB_NS_declare_prefix static inline
#define TREEDB_SHARED

#define TREEDB_PAYLOAD_FIRST_MEMBER m_empty
#define TREEDB_PAYLOAD char m_empty[0]

#define TREEDB_pointer_to_offset(pv) \
	TREEDB_NS(impl_pointer_to_offset)(runtime, pv)
#define TREEDB_offset_to_pointer(u) \
	TREEDB_NS(impl_offset_to_pointer)(runtime, u)

#define TREEDB_pack_pointer(pv) ((uintptr_t)((char *)pv - (char *)0))
#define TREEDB_unpack_pointer(u) (u)

#define TREEDB_OMIT_pack_unpack_pointer_DECLARE

#include <treedb/ns-types.h>
#include <treedb/storage.h>
#include <v3c/file-lock-static.h>

#include <treedb/treedb-static.h>
#include <sys/mman.h>

// Very first and very last.
//static void treedb_malloc_init() __attribute__((constructor(101)));
static void treedb_malloc_term() __attribute__((destructor(101)));

static inline void * TREEDB_NS(do_page_alloc)
( TREEDB_NS(runtime_ptr_t) runtime
, size_t size
) {
	/* Map the extra memory onto our fd */
	size_t prev_size = runtime->m_uSize;
	if((size + prev_size) >= OVERLAY_PAGES_SIZE) {
		TREEDB_printf("do_page_alloc : 0x%zx exceeds the address range of "
			"this heap.\n"
			, size + prev_size);
		return 0;
	}
	TREEDB_printf("do_page_alloc : allocating %zu(0x%zx) bytes at %zu(0x%zx).\n"
		, size, size, prev_size, prev_size);
	allocator_check_heap();
	/* Convert the /dev/zero mapping section to r+w. */
	char * p = (char *)mmap(runtime->m_pBase + prev_size
		, size
		, runtime->m_map_protect
		, runtime->m_map_flags | MAP_FIXED
		, runtime->m_fls.m_fd
		, prev_size);
	if(p == (char *)-1) {
		Exit("do_page_alloc failed.");
		return 0;
	}
	allocator_check_heap();
	runtime->m_uSize += size;
	return runtime->m_pBase + prev_size;
}
static inline TreedbFreeResult TREEDB_NS(do_page_free)
( TREEDB_NS(runtime_ptr_t) runtime
, size_t new_size
) {
	
	TREEDB_printf("do_page_free : freeing %zu(0x%zx) bytes at %zu(0x%zx).\n"
		, runtime->m_uSize - new_size, runtime->m_uSize - new_size, new_size
		, new_size);

	/* Ensure updates are committed. */
	v3c_sync_synchronize();
	allocator_check_heap();

	/* Convert the r+w mapping section to /dev/zero. */
	char * res = (char *)mmap(runtime->m_pBase + new_size
		, runtime->m_uSize - new_size
		, PROT_READ
		, MAP_FIXED
		| MAP_SHARED
		, g_dev_zero_fd
		, new_size);
	if(res == (char *)-1) {
		Exit2("do_page_free failed.");
		return eTreedbFreeBadHeapFree;
	}
	allocator_check_heap();
	runtime->m_uSize = new_size;
	return eTreedbFreeSuccess;
}
static inline TreedbHeapStatus TREEDB_NS(do_init_size)
( TREEDB_NS(runtime_ptr_t) runtime
, size_t offset
, size_t size
) {
	(void)offset;
	if(runtime->m_fls.m_fd != -1) {
		int res = ftruncate(runtime->m_fls.m_fd, OVERLAY_PAGES_SIZE);
		/*check_return("heap_init_size", "ftruncate failed", res);*/
		if(res)
			return eTreedbHeapFileTruncateFailure;
	}
	runtime->m_uSize = size;
	return eTreedbHeapSuccess;
}
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_get_size)
( TREEDB_NS(runtime_ptr_t) runtime
) {
	(void)runtime;
	return eTreedbHeapSuccess;
}
/* TODO: We zero out freed memory in case it contains sensitive data */
//#define TREEDB_post_free(p, sz) TREEDB_NS(impl_post_free)(runtime, p, sz)

static inline void * TREEDB_NS(impl_offset_to_pointer)
( TREEDB_NS_runtime_ptr_t runtime, TREEDB_NS(count_t) u)
{
	if(!u)
		return 0;
	return (char *)(runtime->m_pBase) + u;
}
static inline TREEDB_NS(count_t) TREEDB_NS(impl_pointer_to_offset)
( TREEDB_NS_runtime_ptr_t runtime, const void * pv)
{
	if(!pv)
		return 0;
	return (char *)pv - (char *)(runtime->m_pBase);
}
static inline void * TREEDB_NS(impl_unpack_pointer)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS(count_t) u
) {
	(void)runtime;
	return (char *)0 + u;
}
static inline TREEDB_NS(count_t) TREEDB_NS(impl_pack_pointer)
( TREEDB_NS_runtime_ptr_t runtime
, const void * pv
) {
	(void)runtime;
	return (const char *)pv - (const char *)0;
}
static inline void render_string(const char * str)
{
	(void)write(g_trace_fd, str, strlen(str));
}
#ifdef TREEDB_DEBUG_RENDERERS
static inline void render_pointer(const void * v_)
{
	uintptr_t v = (uintptr_t)v_;
	strcpy(msgbuf, "0x");
	unsigned int n;
	char k, * p = msgbuf + sizeof(void *)*2+2;
	*p-- = '\0';
	for(n = 0; n < 2*sizeof(void *); ++n) {
		k = v & 0x0f;
		if(k < 10) k += '0'; else k += 'a' - 10;
		*p-- = k;
		v >>= 4;
	}
	render_string(msgbuf);
}
static inline void render_size_t(size_t v)
{
	int ndigits = 1;
	size_t max10 = 10;
	while(max10 < v) {
		max10 *= 10;
		++ndigits;
	}
	char * p = msgbuf;
	unsigned int n;
	for(n = 0; n < max10; ++n) {
		max10 /= 10;
		*p++ = '0' + v / max10;
		v -= (v / max10) * max10;
	}
	*p++ = '\0';
	render_string(msgbuf);
}
static inline void render_off_t(off_t v)
{
	if(v < 0) {
		render_string("-");
		render_size_t(-v);
	}else
		render_size_t(v);
}
#endif /* #if TREEDB_DEBUG_RENDERERS */

#ifdef TREEDB_DEBUG_MMAP
static inline void render_mmap_prot(int v)
{
	int next = 0;
	if(v & PROT_READ) {
		render_string("PROT_READ");
		next = 1;
	}
	if(v & PROT_WRITE) {
		if(next)
			render_string(" | ");
		render_string("PROT_WRITE");
		next = 1;
	}
	if(v & PROT_EXEC) {
		if(next)
			render_string(" | ");
		render_string("PROT_EXEC");
	}
}
static inline void render_mmap_flags(int v)
{
	int next = 0;
	if(v & MAP_PRIVATE) {
		render_string("MAP_PRIVATE");
		next = 1;
	}
	if(v & MAP_SHARED) {
		if(next)
			render_string(" | ");
		render_string("MAP_SHARED");
		next = 1;
	}
	if(v & MAP_FIXED) {
		if(next)
			render_string(" | ");
		render_string("MAP_FIXED");
		next = 1;
	}
	if(v & MAP_ANON) {
		if(next)
			render_string(" | ");
		render_string("MAP_ANON");
	}
}

#include <treedb/treedb-impl.h>

static void * my_mmap(void *__addr, size_t __len, int __prot,
		   int __flags, int __fd, __off_t __offset);
#else
#include <treedb/treedb-impl.h>
#endif

static TREEDB_NS(runtime_t) g_runtime = TREEDB_STORAGE_RUNTIME_INIT;
#ifdef TREEDB_PRIVATE_HEAP
  #define g_filename 0 // for mmap, although this isn't strictly needed.
#else
static char g_filename[4096];
static char g_trace_filename[4096];
static const char * g_fuse_dir = 0;
#endif

#define INITGUID
#include <uuid.h>
/* a6894af8-257b-4eef-9007-1a8465b986af */
UUID_DEFINE
( g_uSchema
, 0xa6, 0x89, 0x4a, 0xf8
, 0x25, 0x7b
, 0x4e, 0xef
, 0x90, 0x07
, 0x1a, 0x84, 0x65, 0xb9, 0x86, 0xaf
);

#ifndef TREEDB_PRIVATE_HEAP
static void atfork_before();
static void atfork_after_parent();
static void atfork_after_child();
#endif

static int in_treedb_malloc_init = 0;

#if TREEDB_DEBUG
static int allocator_check_heap()
{
	int res = TREEDB_NS(check_heap)(& g_runtime);
	if(res)
		Exit("allocator_check_heap failed:  %d.", res);
	return res;
}
#endif
v3c_prefix void v3c_treedb_allocator_abort(TreedbError e)
{
	render_string("Allocator Error: ");
	render_string(TreedbErrorDesc(e));
	render_string(" : aborting.\n");
	abort();
}

static void treedb_malloc_init()
{
	if(in_treedb_malloc_init)
		InitExit("treedb_malloc_init : recursing???");

	in_treedb_malloc_init = 1;

#ifndef TREEDB_PRIVATE_HEAP
	g_fuse_dir = getenv("V3C_TREEDB_MALLOC_FUSE_DIR");

	if(!g_fuse_dir) {
		system("set | grep TREEDB");
		InitExit("treedb_malloc_init : V3C_TREEDB_MALLOC_FUSE_DIR not set.");
	}
	{
		struct stat st;
		if(stat(g_fuse_dir, & st))
			InitExitErrno("Failed to stat fuse dir");
		if(st.st_size)
			InitExit("Directory isn't a fuse filesystem.");
	}
	MALLOC_trace("Opening fuse/control");
	TREEDB_MALLOC_sprintf(g_filename, "%s/control", g_fuse_dir);
	g_control_fd = open(g_filename, O_WRONLY | O_CLOEXEC);
	if(g_control_fd == -1)
		InitExitErrno("treedb_malloc_init : couldn't open control file");
	MALLOC_traceX("Got fuse/control, opening fuse/trace");
	TREEDB_MALLOC_sprintf(g_trace_filename, "%s/trace", g_fuse_dir);
	//mode_t old_mask = umask(0400);
	g_trace_fd = open(g_trace_filename, O_WRONLY | O_APPEND | O_CLOEXEC);
#endif // #ifndef TREEDB_PRIVATE_HEAP

	g_dev_zero_fd = open("/dev/zero", O_RDONLY | O_CLOEXEC);
	//umask(old_mask);
#ifndef TREEDB_PRIVATE_HEAP
	MALLOC_traceX("Got fuse/trace");
	if(pthread_atfork(& atfork_before, & atfork_after_parent
			, & atfork_after_child))
		InitExit("treedb_malloc_init : pthread_atfork() failed.");
	// Get the treedb-malloc-daemon to create a backing/overlay for us.
	TREEDB_MALLOC_sprintf(g_filename, "%s/%u", g_fuse_dir, getpid());
	MALLOC_trace("Creating %s", g_filename);
#else // #ifndef TREEDB_PRIVATE_HEAP
	MALLOC_trace("Creating mmap");
#endif // #ifndef TREEDB_PRIVATE_HEAP
	TreedbHeapStatus res;
#ifndef TREEDB_PRIVATE_HEAP
	res = TREEDB_NS(heap_create2)(& g_runtime, g_uSchema, g_filename
		, eHeapOpenRead | eHeapOpenWrite /*| eHeapOpenSync | eHeapOpenNoAtime */
			| eHeapOpenCloseOnExec
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite
		, eHeapAccessRead | eHeapAccessWrite
			| eHeapAccessShared
		, g_dev_zero_fd
		, 0
		, TREEDB_PAGE_SIZE, OVERLAY_PAGES_SIZE);
	if(res != eTreedbHeapSuccess) {
		InitExit("treedb_malloc_init : heap_create() failed.");
	}
	// This is a pointer heap, so an unmap/remap might make the address change.
	// Instead we unlock/relock the thread mutex.
	int ires = pthread_mutex_unlock(& g_runtime.m_mx);
	if(ires)
		InitExit("treedb_malloc_init : pthread_mutex_unlock() failed.");

#else // #ifndef TREEDB_PRIVATE_HEAP

	// We create an initial mapping of /dev/zero to cover our maximum memory
	// requirement.
	// We overlay writeable memory of TREEDB_PAGE_SIZE at the start.
	// Then page allocs and frees just slide the line between them.
	g_runtime.m_pBase = (char *)mmap(0, OVERLAY_PAGES_SIZE
		, PROT_READ
		, MAP_SHARED
		, g_dev_zero_fd
		, 0);
	if(g_runtime.m_pBase == (char *)-1)
		InitExit("treedb_malloc_init : mmap() for /dev/zero failed.");
	g_runtime.m_pBase = (char *)mmap(g_runtime.m_pBase, TREEDB_PAGE_SIZE
		, PROT_READ | PROT_WRITE
		, MAP_PRIVATE | MAP_ANON | MAP_FIXED
		, -1
		, 0);
	if(g_runtime.m_pBase == (char *)-1)
		InitExit("treedb_malloc_init : mmap() for data failed.");
	res = TREEDB_NS(init)(& g_runtime, g_uSchema
				, offsetof(TREEDB_NS(treedb_t), TREEDB_PAYLOAD_FIRST_MEMBER)
				, sizeof(TREEDB_NS(treedb_t))
				, TREEDB_PAGE_SIZE);
	if(res != eTreedbHeapSuccess) {
		InitExit("treedb_malloc_init : heap_create() failed.");
	}
	g_runtime.m_uSize = TREEDB_PAGE_SIZE;
	g_runtime.m_map_protect = PROT_READ | PROT_WRITE;
	g_runtime.m_map_flags = MAP_PRIVATE | MAP_ANON;
	g_runtime.m_fls.m_fd = -1;
#endif
	ok = 1;

	in_treedb_malloc_init = 0;
	MALLOC_trace("treedb_malloc_init complete.");
}
static void treedb_malloc_term()
{
	if(!ok) return;
	MALLOC_trace("treedb_malloc_term()");
	TREEDB_NS(heap_unlock)(& g_runtime);
	TREEDB_NS(heap_close)(& g_runtime);
#ifndef TREEDB_PRIVATE_HEAP
	if(g_trace_fd != -1)
		close(g_trace_fd);
#endif
	ok = 0;
}

#ifndef TREEDB_PRIVATE_HEAP
static void atfork_before()
{
	MALLOC_trace("*** atfork_before() {");
	if(ok) {
		MALLOC_trace("will overlay mapping");
		// Only init if memory was allocated.
		int ires = pthread_mutex_lock(& g_runtime.m_mx);
		if(ires)
			Exit("atfork_before : ires");
		msync((void *)g_runtime.m_pBase, g_runtime.m_uSize, TREEDB_MSYNC);

		/* Send the "overlay my mapping" control message. */
		char msg[20] = "O";
		size_t res, msglen = my_utoa(msg + 1, getpid()) + 1;

		res = write(g_control_fd, msg, msglen);

		if(res != msglen) {
			MALLOC_trace("*** write(g_control_fd) %zu != %zu.", res, msglen);
			pthread_mutex_unlock(& g_runtime.m_mx);
			exit(127);
		}
	}
	MALLOC_trace("*** atfork_before() }");
}
static void atfork_after_parent()
{
	MALLOC_trace("*** atfork_after_parent() {");
	if(ok) {
		MALLOC_trace("unlocking runtime");
		int ires = pthread_mutex_unlock(& g_runtime.m_mx);
		if(ires) {
			fprintf(stderr, "*** pthread_mutex_unlock() failed.\n");
			exit(127);
		}
	}
	MALLOC_trace("*** atfork_after_parent() }");
}
static void atfork_after_child()
{
	close(g_trace_fd);
	g_trace_fd = open(g_trace_filename, O_WRONLY | O_APPEND | O_CLOEXEC);
	MALLOC_trace("*** atfork_after_child() {");
	/* open the overlay file. */
	if(ok) {
		MALLOC_trace("will relock");
		TREEDB_MALLOC_sprintf(g_filename, "%s/%u-%u", g_fuse_dir, getppid(), getpid());
		int fd = open(g_filename, O_RDWR | O_CLOEXEC);
		if(fd == -1) {
			Exit2("atfork_after_child : open(%s) failed", g_filename);
		}
		if(TREEDB_NS(heap_relock)(& g_runtime, fd))
			Exit("atfork_after_child : heap_relock() failed");
	}
	MALLOC_trace("*** atfork_after_child() }");
}
#endif // #ifndef TREEDB_PRIVATE_HEAP

void *malloc(size_t __size)
{
	TREEDB_printf("malloc(%zu)\n", __size);
	if(!ok) treedb_malloc_init();
	if(!ok) return 0;

/*	const char msg[] = "malloc()\n";
	write(STDERR_FILENO, msg, sizeof(msg));*/
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return 0;
	void * ret = TREEDB_NS(malloc_aligned)(& g_runtime, __size
		, sizeof(void *), 0);
	pthread_mutex_unlock(& g_runtime.m_mx);
	TREEDB_printf("malloc returning (%p).\n", ret);
	return ret;
}
int didalloc(void *__ptr)
{
	if(!ok) treedb_malloc_init();
	if(!ok) return 1;

/*	const char msg[] = "malloc()\n";
	write(STDERR_FILENO, msg, sizeof(msg));*/
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return 0;
	int ret = TREEDB_NS(did_alloc)(& g_runtime, __ptr);
	pthread_mutex_unlock(& g_runtime.m_mx);
	return ret;
}
void *calloc(size_t __nmemb, size_t __size)
{
	if(!ok) treedb_malloc_init();
	if(!ok) return 0;
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return 0;
	void * res = TREEDB_NS(calloc_aligned)(& g_runtime, __nmemb, __size
		, sizeof(void *), 0);
	pthread_mutex_unlock(& g_runtime.m_mx);
	return res;
}
void *realloc(void *__ptr, size_t __size)
{
	if(!ok) treedb_malloc_init();
	if(!ok) return 0;
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return 0;
	void * res = TREEDB_NS(realloc_aligned)(& g_runtime, __ptr, __size
		, sizeof(void *), 0);
	pthread_mutex_unlock(& g_runtime.m_mx);
	return res;
}
/* We could fill the freed memory here to stress test apps. */
void free(void *__ptr)
{
	TREEDB_printf("free(%p).\n", __ptr);
	if(!ok) return;
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return;
	TREEDB_NS(free)(& g_runtime, __ptr);
	pthread_mutex_unlock(& g_runtime.m_mx);
}
void cfree(void * ptr)
{
	free(ptr);
}
void * memalign(size_t __alignment, size_t __size)
{
	if(!ok) treedb_malloc_init();
	if(!ok) return 0;
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return 0;
	void * res = TREEDB_NS(malloc_aligned)(& g_runtime, __size, __alignment, 0);
	pthread_mutex_unlock(& g_runtime.m_mx);
	return res;
}
/* Allocate SIZE bytes on a page boundary.  */
void *valloc(size_t __size)
{
	if(!ok) treedb_malloc_init();
	if(!ok) return 0;
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return 0;
	void * res = TREEDB_NS(malloc_aligned)(& g_runtime, __size, TREEDB_PAGE_SIZE
		, 0);
	pthread_mutex_unlock(& g_runtime.m_mx);
	return res;
}
/* Equivalent to valloc(minimum-page-that-holds(n)), that is, round up
   __size to nearest pagesize. */
void * pvalloc(size_t __size)
{
	if(!ok) treedb_malloc_init();
	if(!ok) return 0;
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return 0;
	void * res = TREEDB_NS(malloc_aligned)(& g_runtime
		, ((__size + TREEDB_PAGE_SIZE - 1) / TREEDB_PAGE_SIZE) *TREEDB_PAGE_SIZE
		, TREEDB_PAGE_SIZE, 0);
	pthread_mutex_unlock(& g_runtime.m_mx);
	return res;
}
/* Underlying allocation function; successive calls should return
   contiguous pieces of memory.  */
//extern void *(*__morecore) __MALLOC_PMT ((ptrdiff_t __size));

/* Default value of `__morecore'.  */
void *__default_morecore(ptrdiff_t __size)
{
	(void)__size;
	return 0;
}
int mallopt(int __param, int __val)
{
	(void)__param;
	(void)__val;
	return 1;
}
/* Release all but __pad bytes of freed top-most memory back to the
   system. Return 1 if successful, else 0. */
int malloc_trim(size_t __pad)
{
	(void)__pad;
	return 1;
}
/* Report the number of usable allocated bytes associated with allocated
   chunk __ptr. */
size_t malloc_usable_size(void *__ptr)
{
	if(!ok) return 0;
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return 0;
	size_t res = TREEDB_NS(usable_size)(& g_runtime, __ptr);
	pthread_mutex_unlock(& g_runtime.m_mx);
	return res;
}
/* Prints brief summary statistics on stderr. */
void malloc_stats(void)
{
}
/* Output information about state of allocator to stream FP.  */
int malloc_info(int __options, FILE *__fp)
{
	(void)__options;
	(void)__fp;
	return 1;
}
/* Record the state of all malloc variables in an opaque data structure. */
void * malloc_get_state()
{
	return 0;
}
/* Restore the state of all malloc variables from data obtained with
   malloc_get_state(). */
int malloc_set_state(void *__ptr)
{
	(void)__ptr;
	return 1;
}
/* Called once when malloc is initialized; redefining this variable in
   the application provides the preferred way to set up the hook
   pointers. */
// extern void (*__malloc_initialize_hook) __MALLOC_PMT ((void));
/* Hooks for debugging and user-defined versions. */
// extern void (*__free_hook) __MALLOC_PMT ((void *__ptr,
// 					__const __malloc_ptr_t));
// extern void *(*__malloc_hook) __MALLOC_PMT ((size_t __size,
// 					     __const __malloc_ptr_t));
// extern void *(*__realloc_hook) __MALLOC_PMT ((void *__ptr, size_t __size,
// 					      __const __malloc_ptr_t));
// extern void *(*__memalign_hook) __MALLOC_PMT ((size_t __alignment,
// 					       size_t __size,
// 					       __const __malloc_ptr_t));
// extern void (*__after_morecore_hook) __MALLOC_PMT ((void));

/* Activate a standard set of debugging hooks. */
void __malloc_check_init()
{
}

/* From stdlib.h */
/* Allocate memory of SIZE bytes with an alignment of ALIGNMENT.  */
int posix_memalign(void **__memptr, size_t __alignment, size_t __size)
{
	if(!ok) treedb_malloc_init();
	if(!ok) return 0;
	if(!__memptr)
		return 1;
	int ires = pthread_mutex_lock(& g_runtime.m_mx);
	if(ires) return 1;
	*__memptr = TREEDB_NS(malloc_aligned)(& g_runtime, __size, __alignment, 0);
	pthread_mutex_unlock(& g_runtime.m_mx);
	return 0;
}

/* From mcheck.h */
int mcheck(void (*ABORTFN) (enum mcheck_status))
{
	if(!ok) treedb_malloc_init();
	int ires, res;
	ires = pthread_mutex_lock(& g_runtime.m_mx);
	res = TREEDB_NS(check_heap)(& g_runtime);

	if(!res && !ires) {
		pthread_mutex_unlock(& g_runtime.m_mx);
		return 0;
	}
	if(ABORTFN) {
		if(!ires)
			pthread_mutex_unlock(& g_runtime.m_mx);
		ABORTFN(MCHECK_DISABLED);
		return 1;
	}
	static const char msg[] = "treedb_malloc heap check failure.\n";
	(void)write(STDERR_FILENO, msg, sizeof(msg) - 1);
	pthread_mutex_unlock(& g_runtime.m_mx);
	res /= ires;
	//sleep(60*30);
	abort();
	return 1;
}
#ifdef TREEDB_DEBUG_MMAP
static void * my_mmap(void *__addr, size_t __len, int __prot,
		   int __flags, int __fd, __off_t __offset)
{
	//TREEDB_MALLOC_sprintf(msgbuf, "cat /proc/%d/maps", getpid());
	//*endp++ = '\0';
	//system(msgbuf);
	render_string("mmap(");
	render_pointer(__addr);
	render_string(", size ");
	render_size_t(__len);
	render_string(", prot ");
	render_mmap_prot(__prot);
	render_string(", flags ");
	render_mmap_flags(__flags);
	render_string(", fd ");
	render_off_t(__fd);
	render_string(", offset ");
	render_off_t(__offset);
	render_string(")\n");
#undef mmap
	return mmap(__addr, __len, __prot,__flags, __fd, __offset);
}
#endif
