/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_treedb_malloc_defines_h_included
#define v3c_treedb_malloc_defines_h_included

#include <treedb/treedb.h>

/* THIS IS NOT A DISK BASED IMPLEMENTATION */
/* This isn't C++ either, so I can't just bundle the lot inside a C++ class and
let the compiler sort it out. So we need multiple passes for types to build on
other types until we're done.
*/

#define AVL2_NO_IMPL

#ifdef __cplusplus
extern "C" {
#endif

typedef struct treedb_malloc_allocator_runtime_t
treedb_malloc_allocator_runtime_t, * treedb_allocator_runtime;

typedef enum
{ eMallocHeapNodeTree16
, eMallocHeapNodeTree32
, eMallocHeapNodeLarge
} MallocHeapNodeType;

#define MallocHeapMembers \
	MallocHeapNodeType type

#define A16(x) treedb_malloc_allocator16_##x
#define AF16(x) treedb_malloc_allocator16_AvlAllocator_free_##x
#define AU16(x) treedb_malloc_allocator16_AvlAllocator_used_##x
#define A32(x) treedb_malloc_allocator32_##x
#define AF32(x) treedb_malloc_allocator32_AvlAllocator_free_##x
#define AU32(x) treedb_malloc_allocator32_AvlAllocator_used_##x
#define L2(x) treedb_malloc_list_##x
#define LG(x) treedb_malloc_heap_large_##x
#define HEAP(x) treedb_malloc_avl_tree_##x
//#define TREEDB_SHARED

#define OMIT_TREEDB_NS_SCHEMA_TYPEDEFS
/******************************************************************************/
/* Allocator16                                                                */
/******************************************************************************/
#define TREEDB_NS(x) A16(x)
#define TREEDB_NBITS 16

#ifdef DOXYGEN_processing
{
#endif

#include <treedb/ns-types.h>

#ifdef DOXYGEN_processing
}
#endif

#define treedb_malloc_allocator16_AvlAllocator_used_avl_runtime_ptr_t \
treedb_malloc_allocator16_runtime_ptr_t
#define treedb_malloc_allocator16_AvlAllocator_free_avl_runtime_ptr_t \
treedb_malloc_allocator16_runtime_ptr_t

#define treedb_malloc_allocator16_AvlAllocator_free_tree_t \
treedb_malloc_allocator16_free_tree_t
#define treedb_malloc_allocator16_AvlAllocator_used_tree_t \
treedb_malloc_allocator16_used_tree_t
#define OMIT_TREEDB_functions
#define OMIT_TREEDB_treedb_t

#define TREEDB_STORAGE_RUNTIME_CONDITION_MEMBERS
#define TREEDB_STORAGE_RUNTIME_CONDITION_MEMBERS_INIT
#define TREEDB_STORAGE_RUNTIME_OFFSET_MEMBERS
#define TREEDB_STORAGE_RUNTIME_OFFSET_MEMBERS_INIT

#ifdef TREEDB_SOMBRERO_ALLOC_OMIT_pack_unpack
  #define TREEDB_ALLOC_OMIT_used_pack_unpack_DECLARE
  #define TREEDB_ALLOC_OMIT_free_pack_unpack_DECLARE
#endif

#ifdef DOXYGEN_processing
{
#endif

#define AVL_ALLOCATOR_prefix extern
#include <treedb/storage.h>

#ifdef DOXYGEN_processing
}
#endif

#undef TREEDB_NS
#undef TREEDB_NBITS
/******************************************************************************/
/* Allocator32                                                                */
/******************************************************************************/
#define TREEDB_NS(x) treedb_malloc_allocator32_##x
#define TREEDB_NBITS 32

#ifdef DOXYGEN_processing
{
#endif

#include <treedb/ns-types.h>

#ifdef DOXYGEN_processing
}
#endif

#define treedb_malloc_allocator32_AvlAllocator_used_runtime_ptr_t \
treedb_malloc_allocator32_runtime_ptr_t
#define treedb_malloc_allocator32_AvlAllocator_free_runtime_ptr_t \
treedb_malloc_allocator32_runtime_ptr_t

#define treedb_malloc_allocator32_AvlAllocator_free_tree_t \
treedb_malloc_allocator32_free_tree_t
#define treedb_malloc_allocator32_AvlAllocator_used_tree_t \
treedb_malloc_allocator32_used_tree_t
#define OMIT_TREEDB_functions
#define OMIT_TREEDB_treedb_t

#ifdef DOXYGEN_processing
{
#endif

#define AVL_ALLOCATOR_prefix extern
#include <treedb/storage.h>
#include <treedb/avl-cleardefs.h>

#ifdef DOXYGEN_processing
}
#endif

#undef TREEDB_NS
#undef TREEDB_NBITS

#include <treedb/avl-allocator-cleardefs.h>

/******************************************************************************/
/* HEAP - treedb heap                                                         */
/******************************************************************************/
typedef struct MallocHeapNodeTree16_tag  MallocHeapNodeTree16;
typedef struct MallocHeapNodeTree32_tag  MallocHeapNodeTree32;
typedef struct MallocHeapNodeLarge_tag MallocHeapNodeLarge;

#define TREEDB_NS(x) HEAP(x)
#define AVL_TMPL_NS(x) HEAP(x)
#define treedb_malloc_avl_tree_count_t uint64_t
#define TREEDB_NBITS ptr
typedef struct AVL_TMPL_NS(node_t) AVL_TMPL_NS(node_t), * AVL_TMPL_NS(node_ptr_t);

typedef TREEDB_NS_CONCAT(uint,_t) AVL_TMPL_NS(size_t);
typedef AVL_TMPL_NS(size_t) AVL_TMPL_NS(packed_tree_ptr_t)
, AVL_TMPL_NS(packed_node_ptr_t);

#define treedb_malloc_avl_tree_treedb_packed_Heap_tree_ptr_t \
treedb_malloc_avl_tree_packed_tree_ptr_t
#define treedb_malloc_avl_tree_treedb_packed_Heap_node_ptr_t \
treedb_malloc_avl_tree_packed_node_ptr_t

#define TREEDB_STORAGE_size runtime->m_uSize

struct treedb_malloc_avl_tree_runtime_t;
typedef struct treedb_malloc_avl_tree_runtime_t treedb_malloc_avl_tree_runtime_t
, * treedb_malloc_avl_tree_runtime_ptr_t;
#define treedb_malloc_avl_tree_runtime_ptr_t treedb_malloc_avl_tree_tree_ptr_t
#define AVL_TMPL_key_t const char *

typedef union malloc_heap_node_t malloc_heap_node_t, * malloc_heap_node_ptr_t;

#define TREEDB_OMIT_typedef_runtime_t

#define AVL_TMPL_omit_node_t
#define AVL_TMPL_omit_node_t_typedefs
#define AVL_TMPL_omit_helper_alloc_node_DECLARE
#define AVL_TMPL_omit_helper_free_node_DECLARE
#define AVL_TMPL_omit_helper_pack_unpack_node_DECLARE
#define AVL_TMPL_omit_insert_DECLARE
#define AVL_TMPL_omit_remove_DECLARE
#define AVL_TMPL_omit_impl_pack_unpack_item_DECLARE
#define AVL_TMPL_omit_clear_DECLARE
#define AVL_TMPL_omit_delete_node_DECLARE

#define AVL_TMPL_omit_unpack_tree_DECLARE
#define AVL_TMPL_omit_pack_tree_DECLARE
#define AVL_TMPL_omit_unpack_node_DECLARE
#define AVL_TMPL_omit_pack_node_DECLARE

#ifdef DOXYGEN_processing
{
#endif

#define AVL_TMPL_prefix extern
#include <treedb/avl-template.h>
typedef const malloc_heap_node_ptr_t treedb_malloc_avl_tree_data_const_ref_t;
#include <treedb/avl-cleardefs.h>

#ifdef DOXYGEN_processing
}
#endif

#undef TREEDB_NS

struct A16(runtime_wrap_t);
typedef struct A16(runtime_wrap_t) A16(runtime_wrap_t), * A16(runtime_wrap_ptr_t);
struct A32(runtime_wrap_t);
typedef struct A32(runtime_wrap_t) A32(runtime_wrap_t), * A32(runtime_wrap_ptr_t);
struct LG(runtime_wrap_t);
typedef struct LG(runtime_wrap_t) LG(runtime_wrap_t), * LG(runtime_wrap_ptr_t);

#define HEAP_MEMBERS \
	HEAP(packed_node_ptr_t) m_heap_next, m_heap_prev, m_heap_parent \
	, m_heap_left, m_heap_right; \
	HEAP(count_t) m_count; \
	unsigned char m_depth
#define LIST_MEMBERS \
	L2(packed_node_ptr_t) m_next, m_prev
#define HEAP_HEAP_MEMBERS \
	MallocHeapNodeType type; \
	uint64_t m_uSize; \
	HEAP_MEMBERS
#define HEAP_NODE_MEMBERS \
	HEAP_HEAP_MEMBERS; \
	LIST_MEMBERS;

#undef TREEDB_OMIT_typedef_runtime_t

/******************************************************************************/
/* LIST                                                                       */
/******************************************************************************/
#define TREEDB_NS(x) L2(Treedb##x)
#define TREEDB_NBITS ptr
#define LIST_NS(x) L2(x)
#define LIST_MEMBER m_list

//#define LIST_count_t uint32_t
#define treedb_malloc_list_Treedbruntime_t treedb_malloc_list_list_t
#define treedb_malloc_list_Treedbruntime_ptr_t treedb_malloc_list_list_ptr_t
#define treedb_malloc_list_data_t malloc_heap_node_ptr_t


/* This doesn't work.
#define treedb_malloc_list_treedb_packed_List_list_ptr_t \
LIST_packed_list_ptr_t
*/
#define treedb_malloc_list_packed_list_ptr_t uintptr_t
#define treedb_malloc_list_packed_node_ptr_t uintptr_t
#define treedb_malloc_list_Treedbcount_t uintptr_t

#define treedb_malloc_list_treedb_packed_List_list_ptr_t \
	treedb_malloc_list_packed_list_ptr_t
/* This doesn't work.
#define treedb_malloc_list_treedb_packed_List_node_ptr_t \
LIST_packed_node_ptr_t
*/
#define treedb_malloc_list_treedb_packed_List_node_ptr_t \
	treedb_malloc_list_packed_node_ptr_t

typedef struct LIST_NS(node_t) LIST_NS(node_t), * LIST_NS(node_ptr_t);
struct LIST_NS(node_t) {
	HEAP_NODE_MEMBERS;
};
typedef malloc_heap_node_ptr_t * LIST_NS(data_ref_t);
typedef malloc_heap_node_ptr_t   LIST_NS(packed_data_ref_t);
#define LIST_TMPL_omit_node_t
#define LIST_TMPL_omit_node_t_typedefs
#define LIST_TMPL_omit_impl_free_node_DECLARE
#define LIST_TMPL_omit_delete_node_DECLARE
#define LIST_TMPL_omit_push_DECLARE
//#define LIST_TMPL_omit_list_t_typedefs
#define LIST_NO_IMPL
#define LIST_omit_runtime_t_typedef
//#define OMIT_TREEDB_runtime_t
#define LIST_omit_treedb_impl
#define LIST_omit_treedb_storage
#define LIST_prefix extern
#ifdef DOXYGEN_processing
{
#endif

#include <treedb/list.h>

#ifdef DOXYGEN_processing
}
#endif

#undef TREEDB_NS
#undef TREEDB_NBITS

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*+++++++++++++++++++++++++ and around we go again +++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/******************************************************************************/
/* Allocator16                                                                */
/******************************************************************************/
#ifdef DOXYGEN_processing
{
#endif

#include <treedb/avl-cleardefs.h>

#ifdef DOXYGEN_processing
}
#endif

#define TREEDB_NS(x) A16(x)
struct TREEDB_NS(treedb_t)
{
	A16(allocator_t) m_allocator;
	/* length-is 64K */
};
struct TREEDB_NS(runtime_wrap_t)
{
	HEAP_NODE_MEMBERS;
	TREEDB_NS(runtime_t) runtime;
};
extern void * TREEDB_NS(malloc)(TREEDB_NS(runtime_ptr_t) runtime, size_t size);
extern void * TREEDB_NS(calloc)(TREEDB_NS(runtime_ptr_t) runtime, size_t count
, size_t element_size);
extern void * TREEDB_NS(realloc)(TREEDB_NS(runtime_ptr_t) runtime, void * p
, size_t new_size);
extern TreedbFreeResult TREEDB_NS(free)(TREEDB_NS(runtime_ptr_t) runtime, void * p);

extern A16(runtime_wrap_ptr_t) A16(new_heap)(treedb_allocator_runtime h);
extern TreedbFreeResult A16(free_heap)(treedb_allocator_runtime h, A16(runtime_wrap_ptr_t) p);
extern TREEDB_NSA(node_ptr_t) A16(alloc_node)(A16(runtime_ptr_t) runtime
, size_t size);
#undef TREEDB_NS

#ifdef DOXYGEN_processing
{
#endif

#include <treedb/avl-cleardefs.h>

#ifdef DOXYGEN_processing
}
#endif

/******************************************************************************/
/* Allocator32                                                                */
/******************************************************************************/
#define TREEDB_NS(x) A32(x)
struct TREEDB_NS(treedb_t)
{
	A32(allocator_t) m_allocator;
	/* length-is 4G */
};
struct TREEDB_NS(runtime_wrap_t)
{
	HEAP_NODE_MEMBERS;
	TREEDB_NS(runtime_t) runtime;
};
extern void * TREEDB_NS(malloc)(TREEDB_NS(runtime_ptr_t) runtime, size_t size);
extern void * TREEDB_NS(calloc)(TREEDB_NS(runtime_ptr_t) runtime, size_t count
, size_t element_size);
extern void * TREEDB_NS(realloc)(TREEDB_NS(runtime_ptr_t) runtime, void * p
, size_t new_size);
extern TreedbFreeResult TREEDB_NS(free)(TREEDB_NS(runtime_ptr_t) runtime, void * p);

extern A32(runtime_wrap_ptr_t) A32(new_heap)(treedb_allocator_runtime h);
extern TreedbFreeResult A32(free_heap)(treedb_allocator_runtime h, A32(runtime_wrap_ptr_t) p);
extern TREEDB_NSA(node_ptr_t) A32(alloc_node)(A32(runtime_ptr_t) runtime
, size_t size);
#undef TREEDB_NS
#undef TREEDB_NBITS

#ifdef DOXYGEN_processing
{
#endif

#include <treedb/avl-cleardefs.h>

#ifdef DOXYGEN_processing
}
#endif

/******************************************************************************/
/* Large Allocation                                                           */
/******************************************************************************/
#define TREEDB_NS(x) LG(x)
#define TREEDB_OMIT_TREEDB_treedb_t
typedef struct TREEDB_NS(data_t) {
	uint64_t length;
	char data[];
} TREEDB_NS(data_t), * TREEDB_NS(data_ptr_t);

typedef struct TREEDB_NS(runtime_t) {
	/* Resource guard. */
	pthread_mutex_t m_mx;
	union {
		char * m_pBase;
		uintptr_t m_uSize;
		TREEDB_NS(data_ptr_t) m_pData;
	};
} TREEDB_NS(runtime_t), * TREEDB_NS(runtime_ptr_t);
struct TREEDB_NS(runtime_wrap_t)
{
	HEAP_NODE_MEMBERS;
	TREEDB_NS(runtime_t) runtime;
};
#undef TREEDB_NS
extern LG(runtime_wrap_ptr_t) LG(new_heap)(treedb_allocator_runtime h
, size_t size);
extern TreedbFreeResult LG(free_heap)(treedb_allocator_runtime h, LG(runtime_wrap_ptr_t) p);
extern void * LG(realloc)(treedb_allocator_runtime h, LG(runtime_wrap_ptr_t) p
	, size_t new_size);

/******************************************************************************/
/* HEAP - treedb heap                                                         */
/******************************************************************************/
typedef struct MallocHeapNodeHeader {
	HEAP_HEAP_MEMBERS;
} MallocHeapNodeHeader;

union malloc_heap_node_t {
	MallocHeapNodeType type; /*!< convenience */
	MallocHeapNodeHeader header; /*!< convenience */
	A16(runtime_wrap_t) a16;
	A32(runtime_wrap_t) a32;
	LG(runtime_wrap_t) large;
	L2(node_t) list_node;
	HEAP(node_ptr_t) heap_node;
};

typedef union treedb_malloc_heap_node_ref_ptr_t {
	MallocHeapNodeType * type; /*!< convenience */
	MallocHeapNodeHeader * header; /*!< convenience */
	A16(runtime_wrap_t) * a16;
	A32(runtime_wrap_t) * a32;
	LG(runtime_wrap_t) * large;
	L2(node_ptr_t) list_node;
	HEAP(node_ptr_t) heap_node;
} treedb_malloc_heap_node_ref_ptr_t;

struct HEAP(node_t) {
	HEAP_NODE_MEMBERS;
};

extern char * treedb_malloc_get_base(malloc_heap_node_ptr_t p);

#ifdef DOXYGEN_processing
{
#endif

#include <treedb/avl-cleardefs.h>

#ifdef DOXYGEN_processing
}
#endif

#undef TREEDB_NS

/******************************************************************************/
/* LIST                                                                       */
/******************************************************************************/
/******************************************************************************/
/* treedb_malloc_allocator                                                    */
/******************************************************************************/
struct treedb_malloc_allocator_runtime_t {
	/*! Resource guard. */
	pthread_mutex_t m_mx;
	/*! A list of malloc_heap_node_ptr_t's, for allocations < 4K */
	L2(list_t) a16l;
	/*! A list of malloc_heap_node_ptr_t's, for allocations >= 4K, < 128K */
	L2(list_t) a32l;
	/*! A tree of malloc_heap_node_ptr_t's, sorted by address. */
	HEAP(tree_t) heap;
	int free_emptied_heaps;
	uintptr_t small_allocation_threshold;  /*!< 4096 by default. */
	uintptr_t medium_allocation_threshold; /*!< 128 * 1024 by default. */
	uintptr_t small_allocation_heap_size;  /*!< 64K by default. */
	uintptr_t medium_allocation_heap_size; /*!< 4M by default. */
};

#ifdef __cplusplus
} // extern "C" {
#endif

#undef AVL2_NO_IMPL

#endif /* #ifndef v3c_treedb_malloc_defines_h_included */
