/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/* THIS IS NOT A DISK BASED IMPLEMENTATION */

#include "treedb.h"
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/mman.h>
#include <pthread.h>

#define TREEDB_SOMBRERO_ALLOC_OMIT_pack_unpack
#include "treedb-malloc-defines.h"

extern A16(runtime_t) A16(runtime_init);
extern A32(runtime_t) A32(runtime_init);
extern LG(runtime_t) LG(runtime_init);
extern treedb_malloc_allocator_runtime_t treedb_malloc_allocator_runtime_init;

treedb_allocator_runtime create_treedb_allocator(int free_emptied)
{
	treedb_allocator_runtime h =
		(treedb_allocator_runtime)malloc(sizeof(treedb_malloc_allocator_runtime_t));
	if(!h)
		return h;
	*h = treedb_malloc_allocator_runtime_init;
	h->free_emptied_heaps = free_emptied;
	return h;
}
TreedbFreeResult free_treedb_allocator(treedb_allocator_runtime h)
{
	if(!h)
		return eTreedbFreeBadHeap;
	/* The lists and the heap tree use the heap runtime wrappers to store their
	nodes so free the heaps and it all disappears. */
	treedb_malloc_heap_node_ref_ptr_t t;
	t.heap_node = (HEAP(node_ptr_t))(h->heap.m_head);
	TreedbFreeResult ret = eTreedbFreeSuccess, ret2;
	while(t.heap_node) {
		treedb_malloc_heap_node_ref_ptr_t t2;
		t2.heap_node = (HEAP(node_ptr_t))(t.heap_node->m_heap_next);
		switch(*t.type) {
		case eMallocHeapNodeTree16:
			ret2 = A16(free_heap)(h, t.a16);
			break;
		case eMallocHeapNodeTree32:
			ret2 = A32(free_heap)(h, t.a32);
			break;
		case eMallocHeapNodeLarge:
			ret2 = LG(free_heap)(h, t.large);
			break;
		default:
			ret2 = eTreedbFreeBadHeap;
			break;
		}
		t = t2;
		if(ret2 > ret)
			ret = ret2;
	}
	free(h);
	return ret;
}
/******************************************************************************/
/* Allocator16                                                                */
/******************************************************************************/
A16(runtime_wrap_ptr_t) A16(new_heap)(treedb_allocator_runtime h)
{
	(void)h;
	A16(runtime_wrap_ptr_t) p = (A16(runtime_wrap_ptr_t))
		malloc(sizeof(A16(runtime_wrap_t)));
	if(!p)
		return p;
	p->type = eMallocHeapNodeTree16;
	p->runtime = A16(runtime_init);
	p->runtime.m_pBase = mmap(0, 64 * 1024
			, PROT_READ | PROT_WRITE
			, MAP_PRIVATE | MAP_ANONYMOUS
			, 0
			, 0);
	if(p->runtime.m_pBase == (char *)-1) {
		p->runtime.m_pBase = 0;
		free(p);
		return 0;
	}
	TreedbInitResult res = A16(init_memory)(& p->runtime, sizeof(A16(treedb_t)));
	if(res != eTreedbInitSuccess) {
		munmap((char *)(p->runtime.m_pBase), 64 * 1024);
		free(p);
		return 0;
	}
	/* Resize the free node so it reaches the end of this heap. */
	A16(AvlAllocator_node_ptr_t) node
		= (A16(AvlAllocator_node_ptr_t))(p->runtime.m_pBase
			+ p->runtime.m_pTreedb->m_allocator.m_free.m_head);
	/* It's the only free node so we don't have to remove it and re-insert it
	because of the key change. */
	node->m_uLength += 64 * 1024 - TREEDB_PAGE_SIZE;
	p->m_uSize = 64 * 1024;
	return p;
}
TreedbFreeResult A16(free_heap)(treedb_allocator_runtime h, A16(runtime_wrap_ptr_t) p)
{
	(void)h;
	TreedbFreeResult ret = eTreedbFreeSuccess;
	if(p->runtime.m_pBase)
		munmap((char *)(p->runtime.m_pBase), 64 * 1024);
	else
		ret = eTreedbFreeBadHeap;
	free(p);
	return ret;
}
A16(AvlAllocator_node_ptr_t) A16(alloc_node)
( A16(runtime_ptr_t) runtime
, size_t size
) {
	char * p = (char *)treedb_malloc_allocator16_malloc(runtime, size);
	return (A16(AvlAllocator_node_ptr_t))
		(p - V3C_ROUND_UP_T2(uintptr_t, sizeof(A16(AvlAllocator_node_t)), 2));
}

/******************************************************************************/
/* Allocator32                                                                */
/******************************************************************************/
A32(runtime_wrap_ptr_t) A32(new_heap)(treedb_allocator_runtime h)
{
	(void)h;
	A32(runtime_wrap_ptr_t) p = (A32(runtime_wrap_ptr_t))
		malloc(sizeof(A32(runtime_wrap_t)));
	if(!p)
		return p;
	p->type = eMallocHeapNodeTree32;
	p->runtime = A32(runtime_init);
	p->runtime.m_pBase = mmap(0, 4UL * 1024UL * 1024UL
			, PROT_READ | PROT_WRITE
			, MAP_PRIVATE | MAP_ANONYMOUS
			, 0
			, 0);
	if(p->runtime.m_pBase == (char *)-1) {
		switch(errno) {
		case EINVAL:
			printf("A32(new_heap) : mmap failed : EINVAL.\n");
			break;
		case EACCES:
			printf("A32(new_heap) : mmap failed : EACCES.\n");
			break;
		case ENOMEM:
			printf("A32(new_heap) : mmap failed : ENOMEM.\n");
			break;
		case ENODEV:
			printf("A32(new_heap) : mmap failed : ENODEV.\n");
			break;
		case ENOEXEC:
			printf("A32(new_heap) : mmap failed : ENOEXEC.\n");
			break;
		default:
			printf("A32(new_heap) : mmap failed : <?>.\n");
			break;
		}
		p->runtime.m_pBase = 0;
		free(p);
		return 0;
	}
	TreedbInitResult res = A32(init_memory)(& p->runtime, sizeof(A32(treedb_t)));
	if(res != eTreedbInitSuccess) {
		munmap((char *)(p->runtime.m_pBase), 4UL * 1024UL * 1024UL);
		free(p);
		return 0;
	}
	/* Resize the free node so it reaches the end of this heap. */
	A32(AvlAllocator_node_ptr_t) node
		= (A32(AvlAllocator_node_ptr_t))(p->runtime.m_pBase
			+ p->runtime.m_pTreedb->m_allocator.m_free.m_head);
	/* It's the only free node so we don't have to remove it and re-insert it
	because of the key change. */
	node->m_uLength
			+= 4UL * 1024UL * 1024UL - TREEDB_PAGE_SIZE;
	p->m_uSize = 4UL * 1024UL * 1024UL;
	return p;
}
TreedbFreeResult A32(free_heap)(treedb_allocator_runtime h, A32(runtime_wrap_ptr_t) p)
{
	(void)h;
	TreedbFreeResult ret = eTreedbFreeSuccess;
	if(p->runtime.m_pBase)
		munmap((char *)(p->runtime.m_pBase), 4UL * 1024UL * 1024UL);
	else
		ret = eTreedbFreeBadHeap;
	free(p);
	return ret;
}
A32(AvlAllocator_node_ptr_t) A32(alloc_node)
( A32(runtime_ptr_t) runtime
, size_t size
) {
	char * p = (char *)treedb_malloc_allocator32_malloc(runtime, size);
	return (A32(AvlAllocator_node_ptr_t))
		(p - V3C_ROUND_UP_T2(uintptr_t, sizeof(A32(AvlAllocator_node_t)), 4));
}

/******************************************************************************/
/* Large Allocation                                                           */
/******************************************************************************/
LG(runtime_wrap_ptr_t) LG(new_heap)(treedb_allocator_runtime h
, size_t size)
{
	(void)h;
	LG(runtime_wrap_ptr_t) p = (LG(runtime_wrap_ptr_t))
		malloc(sizeof(LG(runtime_wrap_t)));
	if(!p)
		return p;
	p->type = eMallocHeapNodeLarge;
	p->runtime = LG(runtime_init);
	p->runtime.m_pBase = mmap(0, size + sizeof(uint64_t)
			, PROT_READ | PROT_WRITE
			, MAP_PRIVATE | MAP_ANONYMOUS
			, 0
			, 0);
	if(p->runtime.m_pBase == (char *)-1) {
		p->runtime.m_pBase = 0;
		free(p);
		return 0;
	}
	/* This demo get's a bit scraggy here :) */
	p->runtime.m_pData->length = p->m_uSize = size;
	return p;
}
TreedbFreeResult LG(free_heap)(treedb_allocator_runtime h, LG(runtime_wrap_ptr_t) p)
{
	(void)h;
	TreedbFreeResult ret = eTreedbFreeSuccess;
	if(p->runtime.m_pBase)
		munmap((char *)(p->runtime.m_pBase), p->runtime.m_pData->length + sizeof(uint64_t));
	else
		ret = eTreedbFreeBadHeap;
	free(p);
	return ret;
}
void * LG(realloc)(treedb_allocator_runtime h, LG(runtime_wrap_ptr_t) p, size_t new_size)
{
	(void)h;
	char * q = mremap((char *)(p->runtime.m_pBase), p->runtime.m_pData->length
		, new_size + sizeof(uint64_t)
		, MREMAP_MAYMOVE);
	if(q == (char *)-1)
		return 0;
	p->runtime.m_pData->length = p->m_uSize = new_size;
	return q;
}
