/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifdef DOXYGEN_processing
{
#endif

#include "array-template-defines.h"

#ifdef DOXYGEN_processing
}
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* 10. pointer pack / unpack */
#ifndef ARRAY_TMPL_omit_unpack_array
  #define ARRAY_TMPL_unpack_array(x) ARRAY_TMPL_NS(unpack_array)(runtime, x)
ARRAY_TMPL_prefix ARRAY_TMPL_array_ptr_t
ARRAY_TMPL_NS(unpack_array)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t p
) {
	ARRAY_TMPL_array_ptr_t ret
		= (ARRAY_TMPL_array_ptr_t)ARRAY_TMPL_unpack_pointer(p);
	TREEDB_printf8("array::unpack_array : %lu[0x%lx] -> %p\n"
		, (uintptr_t)p
		, (uintptr_t)p, (void *)(uintptr_t)ret);
	return ret;
}
#endif
#ifndef ARRAY_TMPL_omit_pack_array
  #define ARRAY_TMPL_pack_array(x) ARRAY_TMPL_NS(pack_array)(runtime, x)
ARRAY_TMPL_prefix ARRAY_TMPL_packed_array_ptr_t
ARRAY_TMPL_NS(pack_array)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_array_ptr_t p
) {
	ARRAY_TMPL_packed_array_ptr_t ret = ARRAY_TMPL_pack_pointer(p);
	TREEDB_printf8("array::pack_array : %p -> %lu[0x%lx]\n"
		, (void *)(uintptr_t)p, (uintptr_t)ret
		, (uintptr_t)ret);
	return ret;
}
#endif
#ifndef ARRAY_TMPL_omit_pack_data
  #define ARRAY_TMPL_pack_data(x) ARRAY_TMPL_NS(pack_data)(runtime, x)
ARRAY_TMPL_prefix ARRAY_TMPL_packed_data_ptr_t
ARRAY_TMPL_NS(pack_data)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_data_t * p
) {
	ARRAY_TMPL_packed_data_ptr_t ret = ARRAY_TMPL_pack_pointer(p);
	TREEDB_printf8("handle_array::pack_data : %p -> %lu[0x%lx]\n"
		, (void *)(uintptr_t)p, (uintptr_t)ret
		, (uintptr_t)ret);
	return ret;
}
#endif
#ifndef ARRAY_TMPL_omit_unpack_data
  #define ARRAY_TMPL_unpack_data(x) ARRAY_TMPL_NS(unpack_data)(runtime, x)
ARRAY_TMPL_prefix ARRAY_TMPL_data_t *
ARRAY_TMPL_NS(unpack_data)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_data_ptr_t p
) {
	ARRAY_TMPL_data_t * ret = (ARRAY_TMPL_data_t *)ARRAY_TMPL_unpack_pointer(p);
	TREEDB_printf8("handle_array::unpack_data : %lu[0x%lx] -> %p\n"
		, (uintptr_t)p
		, (uintptr_t)p, (void *)(uintptr_t)ret);
	return ret;
}
#endif

/**
Allocate array item storage.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(init)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t reserve
, ARRAY_TMPL_size_t increment
) {
	if(!ha)
		return 1;
	if(!increment)
		increment = 1;
	else if(ARRAY_TMPL_exceeds_count_limit(increment))
		return 2;
	if(ARRAY_TMPL_exceeds_count_limit(reserve))
		return 2;
	ARRAY_TMPL_count_t nitems
		= V3C_ROUND_UP_T(ARRAY_TMPL_count_t, reserve, increment);
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	ARRAY_TMPL_data_t * pd;
	if(nitems) {
		pd = (ARRAY_TMPL_data_t *)
			ARRAY_TMPL_alloc(nitems * ARRAY_TMPL_entry_size);
		pa = ARRAY_TMPL_unpack_array(ha);
	}else
		pd = 0;
	ARRAY_TMPL_count_MEMBER_SET(pa, 0);
	ARRAY_TMPL_reserve_MEMBER_SET(pa, nitems);
	ARRAY_TMPL_increment_MEMBER_SET(pa, increment);
	ARRAY_TMPL_items_MEMBER_SET(pa, ARRAY_TMPL_pack_data(pd));
	if(nitems)
		return pd ? 0 : 1;
	return 0;
}
#ifndef ARRAY_TMPL_omit_init_empty_ptr
/**
Set up an empty array.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(init_empty_ptr)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_array_ptr_t pa
, ARRAY_TMPL_size_t increment
) {
	if(!increment)
		increment = 1;
	else if(ARRAY_TMPL_exceeds_count_limit(increment))
		return 1;
	ARRAY_TMPL_count_MEMBER_SET(pa, 0);
	ARRAY_TMPL_reserve_MEMBER_SET(pa, 0);
	ARRAY_TMPL_increment_MEMBER_SET(pa, increment);
	ARRAY_TMPL_items_MEMBER_SET(pa, ARRAY_TMPL_pack_data(0));
	return 0;
}
#endif
#ifndef ARRAY_TMPL_omit_init_empty
/**
Set up an empty array.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(init_empty)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t increment
) {
	return ARRAY_TMPL_NS(init_empty_ptr)(runtime, ARRAY_TMPL_unpack_array(ha)
		, increment);
}
#endif

#ifndef ARRAY_TMPL_omit_new
/**
Allocate an array.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix ARRAY_TMPL_packed_array_ptr_t ARRAY_TMPL_NS(new)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_size_t reserve
, ARRAY_TMPL_size_t increment
) {
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_alloc_array;
	if(!pa)
		return 0;
	ARRAY_TMPL_packed_array_ptr_t ha = ARRAY_TMPL_pack_array(pa);
	if(ARRAY_TMPL_NS(init)(runtime, ha, reserve, increment)) {
		ARRAY_TMPL_free(ARRAY_TMPL_unpack_array(ha));
		return 0;
	}
	return ha;
}
#endif

/**
Free an array's entries.

The storage used must be write locked.
*/
static void ARRAY_TMPL_NS(reset)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
) {
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	(void)pa;
	if(ARRAY_TMPL_items_MEMBER(pa)) {
		if(ARRAY_TMPL_count_MEMBER(pa)) {
			if(ARRAY_TMPL_clear_entries(ha, 0, ARRAY_TMPL_count_MEMBER(pa)))
				pa = ARRAY_TMPL_unpack_array(ha);
			ARRAY_TMPL_count_MEMBER_SET(pa, 0);
		}
		ARRAY_TMPL_free(ARRAY_TMPL_unpack_data(ARRAY_TMPL_items_MEMBER(pa)));
		pa = ARRAY_TMPL_unpack_array(ha);
		ARRAY_TMPL_items_MEMBER(pa) = ARRAY_TMPL_pack_data(0);
	}
	ARRAY_TMPL_reserve_MEMBER_SET(pa, 0);
	ARRAY_TMPL_items_MEMBER_SET(pa, 0);
}

#ifndef ARRAY_TMPL_omit_free
/**
Free an array.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix void ARRAY_TMPL_NS(free)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
) {
	ARRAY_TMPL_NS(reset)(runtime, ha);
	ARRAY_TMPL_free(ARRAY_TMPL_unpack_array(ha));
}
#endif

/**
Clear (empty) the array.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix void ARRAY_TMPL_NS(clear)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
) {
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	ARRAY_TMPL_NS(remove_many)(runtime, ha, 0, ARRAY_TMPL_count_MEMBER(pa));
}
/** Make a copy (clone) of another array.

The storage used must be write locked.

The destination array must be initialised first.

If the destination array isn't empty, it is cleared.

If the allocator runs out of memory, the copy is cleared and clone() fails.

This is a "shallow copy" of the raw array data.

You may need to take extra steps to ensure that the copy is a consistant shallow
or deep copy.
*/
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(clone)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha_to
, ARRAY_TMPL_packed_array_ptr_t ha_from
) {
	ARRAY_TMPL_array_ptr_t pa_to = ARRAY_TMPL_unpack_array(ha_to);
	if(ARRAY_TMPL_items_MEMBER(pa_to))
		ARRAY_TMPL_NS(clear)(runtime, ha_to);
	ARRAY_TMPL_array_ptr_t pa_from = ARRAY_TMPL_unpack_array(ha_from);
	if(ARRAY_TMPL_NS(init)(runtime, ha_to, ARRAY_TMPL_count_MEMBER(pa_from)
			, ARRAY_TMPL_increment_MEMBER(pa_to)))
		return 1;
	pa_to = ARRAY_TMPL_unpack_array(ha_to);
	pa_from = ARRAY_TMPL_unpack_array(ha_from);
	ARRAY_TMPL_memcopy(ARRAY_TMPL_item_ptr(pa_to, 0)
		, ARRAY_TMPL_item_ptr(pa_from, 0)
		, ARRAY_TMPL_count_MEMBER(pa_from) * ARRAY_TMPL_entry_size);
	return 0;
}
#ifndef ARRAY_TMPL_omit_insert
/**
Insert an item.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(insert)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t where
, ARRAY_TMPL_input_t d
) {
	if(ARRAY_TMPL_exceeds_count_limit(where))
		return 1;
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	if(where > ARRAY_TMPL_count_MEMBER(pa))
		return 1;
	if(ARRAY_TMPL_count_MEMBER(pa) < ARRAY_TMPL_reserve_MEMBER(pa)) {
		; /* We have room to expand */
	}else{
		ARRAY_TMPL_count_t nitems
			= V3C_ROUND_UP_T(ARRAY_TMPL_count_t, ARRAY_TMPL_count_MEMBER(pa) + 1
				, ARRAY_TMPL_increment_MEMBER(pa));
		if(ARRAY_TMPL_realloc_packed(ha, nitems))
			return 1;
		pa = ARRAY_TMPL_unpack_array(ha);
		ARRAY_TMPL_reserve_MEMBER_SET(pa, nitems);
	}
	if(where != ARRAY_TMPL_count_MEMBER(pa))
		ARRAY_TMPL_memmove(ARRAY_TMPL_item_ptr(pa, where + 1)
			, ARRAY_TMPL_item_ptr(pa, where)
			, (ARRAY_TMPL_count_MEMBER(pa) - where) * ARRAY_TMPL_entry_size);
	ARRAY_TMPL_assign(ARRAY_TMPL_item(pa, where), d);
	ARRAY_TMPL_count_MEMBER_INCREMENT(pa);
	return 0;
}
#endif
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(insert_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t where
, ARRAY_TMPL_size_t count
) {
	if(!count)
		return 0;
	if(ARRAY_TMPL_exceeds_count_limit(
			(where + count - 1) * ARRAY_TMPL_entry_size))
		return 1;
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	if(where > ARRAY_TMPL_count_MEMBER(pa))
		return 2;
	if((ARRAY_TMPL_count_MEMBER(pa) + count) <= ARRAY_TMPL_reserve_MEMBER(pa)) {
		; /* We have room to expand */
	}else{
		ARRAY_TMPL_count_t nitems
			= V3C_ROUND_UP_T(ARRAY_TMPL_count_t
				, ARRAY_TMPL_count_MEMBER(pa) + count
				, ARRAY_TMPL_increment_MEMBER(pa));
		if(ARRAY_TMPL_realloc_packed(ha, nitems))
			return 3;
		pa = ARRAY_TMPL_unpack_array(ha);
		ARRAY_TMPL_reserve_MEMBER_SET(pa, nitems);
	}
	if(where != ARRAY_TMPL_count_MEMBER(pa))
		ARRAY_TMPL_memmove(ARRAY_TMPL_item_ptr(pa, where + count)
			, ARRAY_TMPL_item_ptr(pa, where)
			, (ARRAY_TMPL_count_MEMBER(pa) - where) * ARRAY_TMPL_entry_size);
	ARRAY_TMPL_count_MEMBER_ADD(pa, count);
	return 0;
}
#ifndef ARRAY_TMPL_omit_remove
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(remove)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t where
) {
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	if(where >= ARRAY_TMPL_count_MEMBER(pa))
		return 1;
	if(ARRAY_TMPL_clear_entries(ha, where, 1))
		pa = ARRAY_TMPL_unpack_array(ha); /* Just in case. */
	if((where + 1) != ARRAY_TMPL_count_MEMBER(pa))
		ARRAY_TMPL_memmove(ARRAY_TMPL_item_ptr(pa, where), ARRAY_TMPL_item_ptr(pa, where + 1)
			, (ARRAY_TMPL_count_MEMBER(pa) - where - 1) * ARRAY_TMPL_entry_size);
	if(ARRAY_TMPL_count_MEMBER(pa) % ARRAY_TMPL_increment_MEMBER(pa)) {
		; /* We have room to shrink */
	}else{
		ARRAY_TMPL_count_t nitems
			= V3C_ROUND_UP_T(ARRAY_TMPL_count_t, ARRAY_TMPL_count_MEMBER(pa) - 1
				, ARRAY_TMPL_increment_MEMBER(pa));
		if(nitems) {
			if(ARRAY_TMPL_realloc_packed(ha, nitems))
				return 1;
			pa = ARRAY_TMPL_unpack_array(ha);
		}else{
			ARRAY_TMPL_free(ARRAY_TMPL_unpack_data(ARRAY_TMPL_items_MEMBER(pa)));
			pa = ARRAY_TMPL_unpack_array(ha);
			ARRAY_TMPL_items_MEMBER(pa) = ARRAY_TMPL_pack_data(0);
		}
		ARRAY_TMPL_reserve_MEMBER_SET(pa, nitems);
	}
	ARRAY_TMPL_count_MEMBER_DECREMENT(pa);
	return 0;
}
#endif
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(remove_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t where
, ARRAY_TMPL_size_t count
) {
	if(!count)
		return 0;
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	if((where + count) > ARRAY_TMPL_count_MEMBER(pa))
		return 1;
	if(ARRAY_TMPL_clear_entries(ha, where, count))
		pa = ARRAY_TMPL_unpack_array(ha); /* Just in case. */
	if((where + count) != ARRAY_TMPL_count_MEMBER(pa))
		ARRAY_TMPL_memmove(ARRAY_TMPL_item_ptr(pa, where)
			, ARRAY_TMPL_item_ptr(pa, where + count)
			, (ARRAY_TMPL_count_MEMBER(pa) - where - count) * ARRAY_TMPL_entry_size);
	if(((ARRAY_TMPL_count_MEMBER(pa) - count) / ARRAY_TMPL_increment_MEMBER(pa))
			== (ARRAY_TMPL_count_MEMBER(pa) / ARRAY_TMPL_increment_MEMBER(pa))) {
		; /* We have room to shrink */
	}else{
		ARRAY_TMPL_count_t nitems = V3C_ROUND_UP_T(ARRAY_TMPL_count_t
			, ARRAY_TMPL_count_MEMBER(pa) - count, ARRAY_TMPL_increment_MEMBER(pa));
		if(nitems) {
			if(ARRAY_TMPL_realloc_packed(ha, nitems))
				return 1;
			pa = ARRAY_TMPL_unpack_array(ha);
		}else{
			ARRAY_TMPL_free(ARRAY_TMPL_unpack_data(ARRAY_TMPL_items_MEMBER(pa)));
			pa = ARRAY_TMPL_unpack_array(ha);
			ARRAY_TMPL_items_MEMBER(pa) = ARRAY_TMPL_pack_data(0);
		}
		ARRAY_TMPL_reserve_MEMBER_SET(pa, nitems);
	}
	ARRAY_TMPL_count_MEMBER_SUBTRACT(pa, count);
	return 0;
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(copy_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha_to
, ARRAY_TMPL_size_t to_where
, ARRAY_TMPL_size_t to_count
, ARRAY_TMPL_packed_array_ptr_t ha_from
, ARRAY_TMPL_size_t from_where
, ARRAY_TMPL_size_t from_count
) {
	ARRAY_TMPL_array_ptr_t pa_to;
	if(to_count == (ARRAY_TMPL_size_t)-1) {
		pa_to = ARRAY_TMPL_unpack_array(ha_to);
		to_count = ARRAY_TMPL_count_MEMBER(pa_to);
		if(to_where > to_count)
			return 1; // Specifying past-the-end is invalid.
		to_count -= to_where;
	}
	ARRAY_TMPL_array_ptr_t pa_from;
	if(from_count == (ARRAY_TMPL_size_t)-1) {
		pa_from = ARRAY_TMPL_unpack_array(ha_from);
		from_count = ARRAY_TMPL_count_MEMBER(pa_from);
		if(from_where > from_count)
			return 2; // Not enough entries.
		if(from_where == from_count) {
			if(!to_count)
				return 0;
			return ARRAY_TMPL_NS(remove_many)(runtime, ha_to, to_where
				, to_count);
		}
		from_count -= from_where;
	}
	// Fail if "to" and "from" overlap.
	if(ha_to == ha_from) {
		// Check that the source isn't within the destination.
		if(from_count && (from_where >= to_where)
				&& (from_where < (to_where + to_count)))
			return 3;
		// Check that the destination isn't within the source.
		if(to_count && (to_where >= from_where)
				&& (to_where < (from_where + from_count)))
			return 4;
	}
	if(from_count) {
		if(ARRAY_TMPL_NS(insert_many)(runtime, ha_to, to_where + to_count
				, from_count))
			return 1;
		pa_to = ARRAY_TMPL_unpack_array(ha_to);
		pa_from = ARRAY_TMPL_unpack_array(ha_from);
		ARRAY_TMPL_memcopy(ARRAY_TMPL_item_ptr(pa_to, to_where + to_count)
			, ARRAY_TMPL_item_ptr(pa_from, from_where)
			, from_count * ARRAY_TMPL_entry_size);
	}
	if(to_count)
		if(ARRAY_TMPL_NS(remove_many)(runtime, ha_to, to_where, to_count))
			return 2;
	return 0;
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(move_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha_to
, ARRAY_TMPL_size_t to_where
, ARRAY_TMPL_size_t to_count
, ARRAY_TMPL_packed_array_ptr_t ha_from
, ARRAY_TMPL_size_t from_where
, ARRAY_TMPL_size_t from_count
) {
	if(from_count) {
		if(ARRAY_TMPL_NS(insert_many)(runtime, ha_to, to_where + to_count
				, from_count))
			return 1;
		ARRAY_TMPL_array_ptr_t pa_to = ARRAY_TMPL_unpack_array(ha_to);
		ARRAY_TMPL_array_ptr_t pa_from = ARRAY_TMPL_unpack_array(ha_from);
		ARRAY_TMPL_memcopy(ARRAY_TMPL_item_ptr(pa_to, to_where + to_count)
			, ARRAY_TMPL_item_ptr(pa_from, from_where)
			, from_count * ARRAY_TMPL_entry_size);
		if(ARRAY_TMPL_NS(remove_many)(runtime, ha_from, from_where, from_count))
			return 2;
	}
	if(to_count)
		if(ARRAY_TMPL_NS(remove_many)(runtime, ha_to, to_where, to_count))
			return 3;
	return 0;
}
#ifndef ARRAY_TMPL_omit_push_back
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(push_back)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_input_t d
) {
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	return ARRAY_TMPL_NS(insert)(runtime, ha, ARRAY_TMPL_count_MEMBER(pa), d);
}
#endif
#ifndef ARRAY_TMPL_omit_push_back_many
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(push_back_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t count
) {
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	return ARRAY_TMPL_NS(insert_many)(runtime, ha, ARRAY_TMPL_count_MEMBER(pa), count);
}
#endif
#ifndef ARRAY_TMPL_omit_push_front
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(push_front)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_input_t d
) {
	return ARRAY_TMPL_NS(insert)(runtime, ha, 0, d);
}
#endif
#ifndef ARRAY_TMPL_omit_push_front_many
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(push_front_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t count
) {
	return ARRAY_TMPL_NS(insert_many)(runtime, ha, 0, count);
}
#endif
#ifndef ARRAY_TMPL_omit_pop_back
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(pop_back)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
) {
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	return ARRAY_TMPL_NS(remove)(runtime, ha, ARRAY_TMPL_count_MEMBER(pa) - 1);
}
#endif
#ifndef ARRAY_TMPL_omit_pop_back_many
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(pop_back_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t count
) {
	ARRAY_TMPL_array_ptr_t pa = ARRAY_TMPL_unpack_array(ha);
	return ARRAY_TMPL_NS(remove_many)(runtime, ha, ARRAY_TMPL_count_MEMBER(pa) - count
		, count);
}
#endif
#ifndef ARRAY_TMPL_omit_pop_front
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(pop_front)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
) {
	return ARRAY_TMPL_NS(remove)(runtime, ha, 0);
}
#endif
#ifndef ARRAY_TMPL_omit_pop_front_many
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(pop_front_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_array_ptr_t ha
, ARRAY_TMPL_size_t count
) {
	return ARRAY_TMPL_NS(remove_many)(runtime, ha, 0, count);
}
#endif
#ifndef ARRAY_TMPL_omit_reverse_some
/** ARRAY_TMPL_NS(reverse_some)

Unfortunately there's no equivalent for ARRAY_TMPL_clear_entries here.
*/
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(reverse_some)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_array_ptr_t pa
, ARRAY_TMPL_size_t lower
, ARRAY_TMPL_size_t count
) {
	(void)runtime;
	/* count includes the first item. */
	if((lower + count) > ARRAY_TMPL_count_MEMBER(pa))
		return 1;
	ARRAY_TMPL_count_t n, num = count/2, upper = lower + count - 1;
	for(n = 0; n < num; ++n) {
		ARRAY_TMPL_auto_data(item1);
		ARRAY_TMPL_auto_data(item2);
		ARRAY_TMPL_assign_to_auto(item1, ARRAY_TMPL_item(pa, lower));
		ARRAY_TMPL_assign_to_auto(item2, ARRAY_TMPL_item(pa, upper));
		/*printf("swapping %lu[%u] and %lu[%u].\n", (size_t)(lower)
			, item1, (size_t)(upper), item2);*/
		ARRAY_TMPL_assign_from_auto(ARRAY_TMPL_item(pa, lower), item2);
		ARRAY_TMPL_assign_from_auto(ARRAY_TMPL_item(pa, upper), item1);
		++lower, --upper;
	}
	/*fflush(stdout);*/
	return 0;
}
#endif
#ifndef ARRAY_TMPL_omit_reverse
ARRAY_TMPL_prefix void ARRAY_TMPL_NS(reverse)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_array_ptr_t pa
) {
	(void)ARRAY_TMPL_NS(reverse_some)(runtime, pa, 0, ARRAY_TMPL_count_MEMBER(pa));
}
#endif

#ifdef __cplusplus
} /* extern "C" { */
#endif
