/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifdef __cplusplus
extern "C" {
#endif

typedef TREEDB_NS(count_t) TREEDB_NSA(count_t);
typedef TREEDB_NSA(count_t) TREEDB_NSA(packed_node_ptr_t);
typedef TREEDB_NSA(count_t) TREEDB_NSA(packed_tree_ptr_t);

struct TREEDB_NSA(node_t) {
	TREEDB_NSA(packed_node_ptr_t) m_next, m_prev, m_parent, m_left, m_right;
	TREEDB_NS(count_t) m_count;
	unsigned char m_depth;
	/* bit 0 set means that node is free. */
	TREEDB_NSA(packed_node_ptr_t) m_offset_prev, m_offset_next;
	/* length-is */
	TREEDB_NS(count_t) m_uLength;
	char m_data[];
};

/* There are no free and used node types, only allocator node types. */
typedef struct TREEDB_NSA(node_t) TREEDB_NSA(node_t)
, * TREEDB_NSA(node_ptr_t);
typedef struct TREEDB_NSA(node_t) TREEDB_NSU(node_t)
, * TREEDB_NSU(node_ptr_t);
typedef struct TREEDB_NSA(node_t) TREEDB_NSF(node_t)
, * TREEDB_NSF(node_ptr_t);

#ifndef TREEDB_ALLOC_node_size
  #define TREEDB_ALLOC_node_size \
    V3C_ROUND_UP_T2(TREEDB_NS(count_t), sizeof(TREEDB_NSA(node_t)), \
      sizeof(TREEDB_NS(count_t)))
#endif

/* TODO pad checking. */
#ifndef TREEDB_trailing_pad_size
#define TREEDB_trailing_pad_size 0
/* #define TREEDB_trailing_pad_size sizeof(TREEDB_NS(count_t)) */
#endif

#define TREEDB_ALLOC_addr_to_nodeptr(p) \
  ((TREEDB_NSA(node_ptr_t)) (((char *)(p)) - TREEDB_ALLOC_node_size))
#define TREEDB_ALLOC_nodeptr_to_addr(p) \
  ((char *)(p) + TREEDB_ALLOC_node_size)

#define AVL2_TMPL_omit_node_t
#define AVL2_TMPL_omit_node_t_typedefs

#ifndef TREEDB_pack_allocator_node
  #define TREEDB_pack_allocator_node(a) TREEDB_NSF(pack_node)(runtime, a)
#endif

#ifndef TREEDB_pack_allocator_tree
  #define TREEDB_pack_allocator_tree(a) TREEDB_NSF(pack_tree)(runtime, a)
#endif

/******************************************************************************/
/* The used tree is sorted by address (offset).                               */
/******************************************************************************/
#define AVL2_omit_alloc_node

#define TREEDB_ALLOC_used_tree & TREEDB_ALLOCATOR.m_used
#define AVL2_NS(x) TREEDB_NSU(x)
#define AVL2_TREE_MEMBER m_uLength

#define AVL2_TMPL_count_t TREEDB_NSA(count_t)
#define AVL2_key_t TREEDB_NSA(count_t)

/* Given a node pointer, return the packed node value */
#define AVL2_unpack_key(pNode) TREEDB_pack_pointer(pNode)

#ifdef TREEDB_ALLOC_OMIT_used_pack_unpack_DECLARE
  #define AVL2_TMPL_omit_unpack_tree_DECLARE
  #define AVL2_TMPL_omit_pack_tree_DECLARE
  #define AVL2_TMPL_omit_unpack_node_DECLARE
  #define AVL2_TMPL_omit_pack_node_DECLARE
#endif

#ifdef AVL_ALLOCATOR_prefix
  #define AVL2_prefix AVL_ALLOCATOR_prefix
#endif

#ifdef DOXYGEN_processing
{
#endif

#define AVL2_NO_IMPL
#define AVL2_omit_treedb_storage
#define AVL2_omit_treedb_impl
#include <treedb/avl2.h>

#ifdef DOXYGEN_processing
}
#endif

/******************************************************************************/
/* The free tree is sorted by size and allows duplicates                      */
/******************************************************************************/
#define AVL2_omit_alloc_node

#define TREEDB_ALLOC_free_tree & TREEDB_ALLOCATOR.m_free
#define TREEDB_ALLOCATOR_packed_free_tree AVL2_TMPL_pack_tree(& TREEDB_ALLOCATOR.m_free)
#define AVL2_NS(x) TREEDB_NSF(x)
#define AVL2_TREE_MEMBER m_uLength

#define AVL2_TMPL_count_t TREEDB_NSA(count_t)
#define AVL2_key_t TREEDB_NSA(count_t)
/* Given a reference to the node, return the length */
#define AVL2_unpack_key(pNode) ((pNode)->AVL2_TMPL_item)
#define AVL2_TMPL_allow_duplicates 1

#define AVL2_TMPL_omit_node_t
#define AVL2_TMPL_omit_node_t_typedefs

#ifdef TREEDB_ALLOC_OMIT_free_pack_unpack_DECLARE
  #define AVL2_TMPL_omit_unpack_tree_DECLARE
  #define AVL2_TMPL_omit_pack_tree_DECLARE
  #define AVL2_TMPL_omit_unpack_node_DECLARE
  #define AVL2_TMPL_omit_pack_node_DECLARE
#endif

#ifdef AVL_ALLOCATOR_prefix
  #define AVL2_prefix AVL_ALLOCATOR_prefix
#endif

#ifdef DOXYGEN_processing
{
#endif

#define AVL2_NO_IMPL
#define AVL2_omit_treedb_storage
#define AVL2_omit_treedb_impl
#include <treedb/avl2.h>

#ifdef DOXYGEN_processing
}
#endif

#undef AVL2_TMPL_omit_delete_node_DECLARE

#undef AVL2_TMPL_omit_helper_alloc_node_DECLARE
#undef AVL2_TMPL_omit_helper_free_node_DECLARE

#undef AVL2_TMPL_omit_insert_DECLARE
#undef AVL2_TMPL_omit_remove_DECLARE
#undef AVL2_TMPL_omit_clear_DECLARE

#undef AVL_ALLOCATOR_prefix

#ifndef OMIT_Allocator_allocator_t_declare
typedef struct TREEDB_NS(allocator_t)
{
	TREEDB_NSF(tree_t) m_free;
	TREEDB_NSU(tree_t) m_used;
} TREEDB_NS(allocator_t), * TREEDB_NS(allocator_ptr_t);
#endif /* #ifndef OMIT_Allocator_allocator_t_declare */

#ifdef __cplusplus
} // extern "C" {
#endif
