/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <sys/mman.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <v3c/atomic.h>

#include <sys/stat.h>

#ifndef TREEDB_STORAGE_header
  #define TREEDB_STORAGE_header runtime->m_pTreedb->m_header
#endif
#ifndef TREEDB_STORAGE_size
  #define TREEDB_STORAGE_size runtime->m_uSize
#endif
#ifndef TREEDB_STORAGE_set_size
  #define TREEDB_STORAGE_set_size(sz) TREEDB_STORAGE_size = (sz)
#endif
#ifndef TREEDB_STORAGE_max_size
  #define TREEDB_STORAGE_max_size ((TREEDB_NS(count_t))-1)
#endif

#ifndef TREEDB_STORAGE_base
  #define TREEDB_STORAGE_base (runtime->m_pBase)
#endif

#ifndef TREEDB_NS_prefix
  #define TREEDB_NS_prefix
#endif

#ifndef TREEDB_NS_declare_prefix
  #define TREEDB_NS_declare_prefix extern TREEDB_NS_prefix
#endif

#ifndef TREEDB_heap_read_lock
  #define TREEDB_heap_read_lock() TREEDB_NS(heap_read_lock)(runtime)
#endif
#ifndef TREEDB_heap_write_lock
  #define TREEDB_heap_write_lock() TREEDB_NS(heap_write_lock)(runtime)
#endif
#ifndef TREEDB_heap_write_lock_at
  #define TREEDB_heap_write_lock_at(pv) \
    TREEDB_NS(heap_write_lock_at)(runtime, pv)
#endif
#ifndef TREEDB_heap_unlock
  #define TREEDB_heap_unlock() TREEDB_NS(heap_unlock)(runtime)
#endif
#ifndef TREEDB_heap_close
  #define TREEDB_heap_close() TREEDB_NS(heap_close)(runtime)
#endif
#ifndef TREEDB_heap_init
  #define TREEDB_heap_init(name, open_flags, create_permissions, access_flags \
    , offset, page_sz) TREEDB_NS(heap_init)(runtime, name, open_flags \
        , create_permissions, access_flags, offset, page_sz)
#endif
#ifndef TREEDB_heap_init_size
  #define TREEDB_heap_init_size(offset, size) \
    TREEDB_NS(heap_init_size)(runtime, offset, size)
#endif

#ifndef TREEDB_init
  #define TREEDB_init(uSchema, offs, hsz, sz) TREEDB_NS(init)(runtime, uSchema \
	, offs, hsz, sz)
#endif
#ifndef TREEDB_mmap
  #define TREEDB_mmap(addr, len, prot, flags, fd, offset) \
    mmap(addr, len, prot, flags, fd, offset)
#endif

#ifdef __cplusplus
extern "C" {
#endif

#if TREEDB_DEBUG & 1
extern char * v3c_c_strerror(int errnum);

#define TREEDB_error_wrap(res, text) \
	char * msg = v3c_c_strerror(res); \
	if(msg) { \
		TREEDB_printf(text " : %s.\n", msg); \
		free(msg); \
	}else{ \
		TREEDB_printf(text " : %d.\n", res); \
	}
#else
#define TREEDB_error_wrap(res, text)
#endif

#ifndef TREEDB_OMIT_FILE_STORAGE_get_size
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_get_size)
( TREEDB_NS(runtime_ptr_t) runtime
) {
#if HAVE_FSTAT
	struct stat s;
	if(fstat(runtime->m_fls.m_fd, &s))
		return eTreedbHeapBad;
	size_t top = (size_t)s.st_size;
#else
	size_t cur = lseek(runtime->m_fls.m_fd, 0, SEEK_CUR);
	size_t top = lseek(runtime->m_fls.m_fd, 0, SEEK_END);
	lseek(runtime->m_fls.m_fd, cur, SEEK_SET);
#endif
	TREEDB_printf("heap_get_size : will map %zu bytes.\n", top);
	TREEDB_STORAGE_set_size(top);
	return eTreedbHeapSuccess;
}
#endif

TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_read_lock_at)
( TREEDB_NS(runtime_ptr_t) runtime
, void * vp
) {
	TREEDB_printf("heap_read_lock_at\n");
	int res = pthread_mutex_lock(& runtime->m_mx);
	if(res) {
		TREEDB_error_wrap(res, "heap_read_lock_at : error : pthread_mutex_lock "
			"failed");
		return eTreedbHeapMutexLockError;
	}
	if(runtime->m_active_writer_thread == pthread_self()) {
		TREEDB_printf("heap_read_lock_at : error : heap already write locked.\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapAlreadyWriteLocked;
	}
	if((!(runtime->m_map_flags & MAP_ANON)) && (runtime->m_fls.m_fd == -1)) {
		TREEDB_printf("heap_read_lock_at : error : heap not open.\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapNotOpen;
	}
	if((TREEDB_STORAGE_base) != (void *)0) {
		TREEDB_printf("heap_read_lock_at : error : heap already read locked.\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapAlreadyReadLocked;
	}
	while(runtime->m_uWritersWaiting) {
		res = pthread_cond_wait(& runtime->m_cond, & runtime->m_mx);
		if(res) {
			TREEDB_error_wrap(res, "heap_read_lock_at : error : condition "
				"wait error");
			pthread_mutex_unlock(& runtime->m_mx);
			return eTreedbHeapCondWaitError;
		}
	}
	if(!runtime->m_uReadersActive) {
		// The first reader locks the file.
		if(!(runtime->m_map_flags & MAP_ANON)) {
			res = file_lock_state_update(& runtime->m_fls, eFileLockReadLocked);
			if(res) {
				TREEDB_printf("heap_read_lock_at : error : "
					"read lock update error.\n");
				pthread_mutex_unlock(& runtime->m_mx);
				return eTreedbHeapFileLockError;
			}
		}
		/* We need to re-fetch the file size as another process may have changed
		it. */
		TreedbHeapStatus hres = TREEDB_NS(heap_get_size)(runtime);
		if(hres != eTreedbHeapSuccess)
			return hres;
		TREEDB_STORAGE_base = (char *)TREEDB_mmap(vp
			, TREEDB_STORAGE_size
			, runtime->m_map_protect & (~ PROT_WRITE)
			, runtime->m_map_flags | (vp ? MAP_FIXED : 0)
			, runtime->m_fls.m_fd
			, runtime->m_offset);
		if((TREEDB_STORAGE_base) == (void *)-1) {
			TREEDB_printf("heap_read_lock_at : error : mmap failed!\n");
			TREEDB_STORAGE_base = 0;
			if(!(runtime->m_map_flags & MAP_ANON))
				file_lock_state_update(& runtime->m_fls, eFileLockUnlocked);
			pthread_mutex_unlock(& runtime->m_mx);
			return eTreedbHeapMmapError;
		}
	}
	++runtime->m_uReadersActive;

	res = pthread_mutex_unlock(& runtime->m_mx);
	if(res) {
		TREEDB_error_wrap(res, "heap_read_lock_at : pthread_mutex_unlock failed");
		return eTreedbHeapMutexUnlockError;
	}
	// Multiple readers are go!
	TREEDB_printf("heap_read_lock_at : success.\n");
	return eTreedbHeapSuccess;
}
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_read_lock)
( TREEDB_NS(runtime_ptr_t) runtime
) {
	TREEDB_printf("heap_read_lock\n");
	return TREEDB_NS(heap_read_lock_at)(runtime, 0);
}

TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_write_lock_at)
( TREEDB_NS(runtime_ptr_t) runtime
, void * vp
) {
	TREEDB_printf("heap_write_lock_at\n");
	int res = pthread_mutex_lock(& runtime->m_mx);
	if(res) {
		TREEDB_error_wrap(res, "heap_write_lock_at : "
			"pthread_mutex_lock() failed");
		return eTreedbHeapMutexLockError;
	}
	if(! (runtime->m_open_flags & eHeapOpenWrite)) {
		TREEDB_error_wrap(res, "heap_write_lock_at : error : "
			"read-only heap.\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapReadonly;
	}
	if(runtime->m_active_writer_thread) {
		TREEDB_error_wrap(res, "heap_write_lock_at : error : already "
			"write-locked.\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapAlreadyWriteLocked;
	}
	if((TREEDB_STORAGE_base) != (void *)0) {
		TREEDB_error_wrap(res, "heap_write_lock_at : error : memory already "
			"mapped.\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapAlreadyReadLocked;
	}
	if(!(runtime->m_map_flags & MAP_ANON) && (runtime->m_fls.m_fd == -1)) {
		TREEDB_error_wrap(res, "heap_write_lock_at : error : heap not open.\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapNotOpen;
	}
	++runtime->m_uWritersWaiting;

	while(runtime->m_uReadersActive && runtime->m_active_writer_thread) {
		res = pthread_cond_wait(& runtime->m_cond, & runtime->m_mx);
		if(res) {
			TREEDB_error_wrap(res, "heap_write_lock_at : pthread_cond_wait() "
				"failed");
			if(!(runtime->m_map_flags & MAP_ANON))
				pthread_mutex_unlock(& runtime->m_mx);
			return eTreedbHeapCondWaitError;
		}
	}
	runtime->m_active_writer_thread = pthread_self();

	--runtime->m_uWritersWaiting;

	if(!(runtime->m_map_flags & MAP_ANON)) {
		res = file_lock_state_update(& runtime->m_fls, eFileLockWriteLocked);
		if(res) {
			TREEDB_printf("heap_write_lock_at : file_lock_state_update failed");
			pthread_mutex_unlock(& runtime->m_mx);
			return eTreedbHeapFileLockError;
		}
	}
	/* We need to re-fetch the file size as another process may have changed
	it. */
	TreedbHeapStatus hres = TREEDB_NS(heap_get_size)(runtime);
	if(hres != eTreedbHeapSuccess)
		return hres;
	TREEDB_STORAGE_base = (char *)TREEDB_mmap(vp
		, TREEDB_STORAGE_size
		, runtime->m_map_protect
		, runtime->m_map_flags | (vp ? MAP_FIXED : 0)
		, runtime->m_fls.m_fd
		, runtime->m_offset);
	if((TREEDB_STORAGE_base) == (void *)-1) {
			TREEDB_printf("heap_write_lock_at : error : mmap failed!\n");
		TREEDB_STORAGE_base = 0;
		if(!(runtime->m_map_flags & MAP_ANON))
			file_lock_state_update(& runtime->m_fls, eFileLockUnlocked);
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapMmapError;
	}

	res = pthread_mutex_unlock(& runtime->m_mx);
	if(res) {
		TREEDB_error_wrap(res, "heap_write_lock_at : "
			"pthread_mutex_unlock failed");
		return eTreedbHeapMutexUnlockError;
	}
	return eTreedbHeapSuccess;
}
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_write_lock)
( TREEDB_NS(runtime_ptr_t) runtime
) {
	TREEDB_printf("heap_write_lock\n");
	return TREEDB_NS(heap_write_lock_at)(runtime, 0);
}
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_unlock)
( TREEDB_NS(runtime_ptr_t) runtime
) {
	TREEDB_printf("heap_unlock\n");
	int res = pthread_mutex_lock(& runtime->m_mx);
	if(res) {
		TREEDB_error_wrap(res, "heap_unlock : error : pthread_mutex_lock() "
			"failed");
		return eTreedbHeapMutexLockError;
	}
	if((TREEDB_STORAGE_base) == (void *)0) {
		TREEDB_printf("heap_unlock : error : memory not mapped!\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapAlreadyUnlocked;
	}
	if(pthread_equal(runtime->m_active_writer_thread, pthread_self())) {
		runtime->m_active_writer_thread = 0;
		if(!(runtime->m_map_flags & MAP_ANON)) {
			res = file_lock_state_update(& runtime->m_fls, eFileLockUnlocked);
			if(res) {
				TREEDB_printf("heap_unlock : error : "
					"file_lock_state_update failed!\n");
				pthread_mutex_unlock(& runtime->m_mx);
				return eTreedbHeapFileLockError;
			}
		}
		if(runtime->m_uReadersActive)
			pthread_cond_broadcast(& runtime->m_cond);
	}else{
		--runtime->m_uReadersActive;
		if(!(runtime->m_map_flags & MAP_ANON)) {
			if(!runtime->m_uReadersActive) {
				res = file_lock_state_update(& runtime->m_fls
					, eFileLockUnlocked);
				if(res) {
					TREEDB_printf("heap_unlock : error : file_lock_state_update"
						" failed!\n");
					pthread_mutex_unlock(& runtime->m_mx);
					return eTreedbHeapFileLockError;
				}
			}
		}
	}

	msync(TREEDB_STORAGE_base
		, TREEDB_STORAGE_size
		, TREEDB_MSYNC);

	res = munmap(TREEDB_STORAGE_base, TREEDB_STORAGE_size);
	if(res) {
		TREEDB_error_wrap(res, "heap_unlock : error : munmap() failed");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapUnMmapError;
	}
	TREEDB_STORAGE_base = 0;
	res = pthread_cond_broadcast(& runtime->m_cond);
	if(res) {
		TREEDB_error_wrap(res, "heap_unlock : error : pthread_cond_broadcast() "
			"failed");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapCondBroadcastError;
	}
	res = pthread_mutex_unlock(& runtime->m_mx);
	if(res) {
		TREEDB_error_wrap(res, "heap_read_lock : error : pthread_mutex_unlock "
			"failed");
		return eTreedbHeapMutexUnlockError;
	}
	return eTreedbHeapSuccess;
}
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_relock)
( TREEDB_NS(runtime_ptr_t) runtime
, int fd
) {
	TREEDB_printf("heap_relock\n");
	/* Done by the caller before fork().
	int res = pthread_mutex_lock(& runtime->m_mx);
	if(res) {
		TREEDB_error_wrap(res, "heap_relock : error : pthread_mutex_lock() "
			"failed");
		return eTreedbHeapMutexLockError;
	}*/
	if((TREEDB_STORAGE_base) == (void *)0) {
		TREEDB_printf("heap_relock : error : memory not mapped!\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapAlreadyUnlocked;
	}
	int was_write_locked = runtime->m_fls.m_cur == eFileLockWriteLocked;
	int res = file_lock_state_atfork_after_child(& runtime->m_fls);
	if(res) {
		TREEDB_printf("heap_relock : error : "
			"file_lock_state_atfork_after_child() failed.\n");
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapFileLockError;
	}
	// We can't msync the old heap - it's the parents heap.
	//msync(TREEDB_STORAGE_base, TREEDB_STORAGE_size, TREEDB_MSYNC);
	void * base = TREEDB_STORAGE_base;
	madvise(base, TREEDB_STORAGE_size, MADV_DONTNEED);
	int old_fd = runtime->m_fls.m_fd;
	runtime->m_fls.m_fd = fd;
	if(!(runtime->m_map_flags & MAP_ANON)) {
		res = file_lock_state_update(& runtime->m_fls
			, was_write_locked ? eFileLockWriteLocked : eFileLockReadLocked);
		if(res) {
			TREEDB_printf("heap_relock : file_lock_state_update failed.\n");
			pthread_mutex_unlock(& runtime->m_mx);
			return eTreedbHeapFileLockError;
		}
	}
	/* We need to overlay the new mmap over the old mmap, otherwise the os
	could sneak one into our range somehow. */
	TREEDB_STORAGE_base = (char *)mmap(base
		, TREEDB_STORAGE_size
		, runtime->m_map_protect & (~ (was_write_locked ? 0 : PROT_WRITE))
		, runtime->m_map_flags | MAP_FIXED
		, runtime->m_fls.m_fd
		, runtime->m_offset);
	if((TREEDB_STORAGE_base) == (void *)-1) {
		TREEDB_printf("heap_relock : error : mmap failed!\n");
		TREEDB_STORAGE_base = 0;
		if(!(runtime->m_map_flags & MAP_ANON))
			file_lock_state_update(& runtime->m_fls, eFileLockUnlocked);
		pthread_mutex_unlock(& runtime->m_mx);
		return eTreedbHeapMmapError;
	}
	if(old_fd != -1)
		close(old_fd);

	res = pthread_mutex_unlock(& runtime->m_mx);
	if(res) {
		TREEDB_error_wrap(res, "heap_relock : error : pthread_mutex_unlock "
			"failed");
		return eTreedbHeapMutexUnlockError;
	}
	TREEDB_printf("heap_relock : success.\n");
	return eTreedbHeapSuccess;
}
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_close)
( TREEDB_NS(runtime_ptr_t) runtime
) {
	TREEDB_printf("heap_close\n");
	if(TREEDB_STORAGE_base) {
		TREEDB_printf("heap_close : error : heap is locked!\n");
		return eTreedbHeapLockedError;
	}
	if(runtime->m_fls.m_fd != -1) {
		close(runtime->m_fls.m_fd);
		runtime->m_fls.m_fd = -1;
	}
	return eTreedbHeapSuccess;
}
#ifndef TREEDB_OMIT_FILE_STORAGE_header_check
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_header_check)
( TREEDB_NS(runtime_ptr_t) runtime
, const uuid_t uSchema
) {
	TREEDB_printf("heap_header_check\n");
	TreedbHeapStatus ret = eTreedbHeapSuccess;
	do {
		// We don't use the seek offset so be nice.
		size_t cur = lseek(runtime->m_fls.m_fd, 0, SEEK_CUR);
		size_t top = lseek(runtime->m_fls.m_fd, 0, SEEK_END);
		lseek(runtime->m_fls.m_fd, cur, SEEK_SET);
		TREEDB_printf("heap_header_check : %zu bytes.\n", top);

		if(top < sizeof(TREEDB_NS(treedb_t))) {
			ret = eTreedbHeapBadTop;
			break;
		}
		if(strcmp(TREEDB_STORAGE_header.m_szMagic, "3db")) {
			ret = eTreedbHeapOpenBadMagic;
			break;
		}
		if(top % TREEDB_PAGE_SIZE) {
			ret = eTreedbHeapBadTop;
			break;
		}
		if(uuid_compare(TREEDB_STORAGE_header.m_uAllocatorAlgorithm
				, UUID_treedb_avl_two_tree_allocator_algorithm)) {
			ret = eTreedbHeapOpenUnknownAllocatorAlgorithm;
			break;
		}
		if(TREEDB_STORAGE_header.m_uBytesPerOffset
				!= (sizeof(TREEDB_NS(count_t)))) {
			ret = eTreedbHeapOpenUnhandledOffsetSize;
			break;
		}
		if(TREEDB_STORAGE_header.m_uEndianIndex
				!= v3c_native_endian_index()) {
			ret = eTreedbHeapOpenUnhandledEndianIndex;
			break;
		}
		if(TREEDB_STORAGE_header.m_uFeatures != 3) {
			ret = eTreedbHeapOpenUnhandledFeatures; // We need count and depth.
			break;
		}
		if(uSchema && uuid_compare(TREEDB_STORAGE_header.m_uSchema, uSchema)) {
			ret = eTreedbHeapOpenUnknownSchema;
			break;
		}
	} while(0);
	return ret;
}
#endif
static void TREEDB_NS(set_mmap_flags)
( TREEDB_NS(runtime_ptr_t) runtime
, int access_flags
) {
	runtime->m_map_protect = 0;
	runtime->m_map_flags = 0;
	if(access_flags & eHeapAccessRead)
		runtime->m_map_protect |= PROT_READ;
	if(access_flags & eHeapAccessWrite)
		runtime->m_map_protect |= PROT_WRITE;
	if(access_flags & eHeapAccessExec)
		runtime->m_map_protect |= PROT_EXEC;
	if(access_flags & eHeapAccessPrivate)
		runtime->m_map_flags |= MAP_PRIVATE;
	if(access_flags & eHeapAccessShared)
		runtime->m_map_flags |= MAP_SHARED;
	if(access_flags & eHeapAccessAnonymous)
		runtime->m_map_flags |= MAP_ANON;
}

 /* heap_just_open_fd
   =================
1. file_lock_state_init
2. set_mmap_flags
3. set_size
4. heap_?_lock
*/
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_just_open_fd)
( TREEDB_NS(runtime_ptr_t) runtime
, int fd
, unsigned int open_flags
, unsigned int access_flags
, size_t offset
) {
	TREEDB_printf("heap_just_open_fd\n");
	if(runtime->m_fls.m_fd != -1)
		TREEDB_heap_close();
	runtime->m_fls.m_fd = fd;

	if(runtime->m_fls.m_fd == -1)
		return eTreedbHeapFileOpenFailure;

	if(file_lock_state_init(runtime->m_fls.m_fd, & runtime->m_fls, offset, 1
			, (open_flags & eHeapOpenWrite) ?
			eFileLockWriteLocked : eFileLockReadLocked))
		return eTreedbHeapFileLockError;
	TREEDB_NS(set_mmap_flags)(runtime, access_flags);
	runtime->m_offset = offset;
	runtime->m_open_flags = open_flags;

	TreedbHeapStatus res = TREEDB_NS(heap_get_size)(runtime);
	if(res != eTreedbHeapSuccess) {
		TREEDB_heap_close();
		return res;
	}

	if(open_flags & eHeapOpenWrite)
		res = TREEDB_heap_write_lock();
	else
		res = TREEDB_heap_read_lock();
	if(res != eTreedbHeapSuccess)
		TREEDB_heap_close();

	return res;
	/* IMPORTANT: the caller must unlock the heap!!! */
}

/* heap_just_open
   ==============
1. heap_just_open_fd(open(name))
*/
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_just_open)
( TREEDB_NS(runtime_ptr_t) runtime
, const char * name
, unsigned int open_flags
, unsigned int access_flags
, size_t offset
) {
	TREEDB_printf("heap_just_open %s\n", name);
	if(runtime->m_fls.m_fd != -1)
		TREEDB_heap_close();
	int fd = open(name
		, ((open_flags & eHeapOpenWrite) ? O_RDWR : O_RDONLY)
			| (open_flags & eHeapOpenSync ? O_FSYNC : 0)
			| (open_flags & eHeapOpenNoAtime ? O_NOATIME : 0)
			| (open_flags & eHeapOpenCloseOnExec ? O_CLOEXEC : 0)
		);
	if(fd == -1) {
		TREEDB_printf("heap_just_open : couldn't open fd \"%s\"!\n", name);
		return eTreedbHeapFileOpenFailure;
	}
	TreedbHeapStatus ret =
		TREEDB_NS(heap_just_open_fd)(runtime
			, fd
			, open_flags, access_flags, offset);
	if(ret != eTreedbHeapSuccess) {
		TREEDB_printf("heap_just_open : couldn't open \"%s\"!\n", name);
	}
	return ret;
}

#ifndef TREEDB_OMIT_FILE_STORAGE_heap_open

/* heap_open
   =========
1. heap_just_open
2. heap_header_check
*/
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_open)
( TREEDB_NS(runtime_ptr_t) runtime
, const uuid_t uSchema
, const char * name
, unsigned int open_flags
, unsigned int access_flags
, size_t offset
) {
	TREEDB_printf("heap_open %s\n", name);
	TreedbHeapStatus ret = TREEDB_NS(heap_just_open)
		(runtime, name, open_flags, access_flags, offset);
	if(ret != eTreedbHeapSuccess) {
		TREEDB_printf("heap_open: couldn't open \"%s\".\n", name);
		return ret;
	}
	ret = TREEDB_NS(heap_header_check)(runtime, uSchema);
	if(ret != eTreedbHeapSuccess) {
		TREEDB_heap_unlock();
		TREEDB_heap_close();
	}
	return ret;
	/* IMPORTANT: the caller must unlock the heap!!! */
}

/* heap_open_fd
   ============
1. heap_just_open_fd
2. heap_header_check
*/
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_open_fd)
( TREEDB_NS(runtime_ptr_t) runtime
, const uuid_t uSchema
, int fd
, unsigned int open_flags
, unsigned int access_flags
, size_t offset
) {
	TREEDB_printf("heap_open_fd\n");
	TreedbHeapStatus ret = TREEDB_NS(heap_just_open_fd)
		(runtime, fd, open_flags, access_flags, offset);
	if(ret != eTreedbHeapSuccess)
		return ret;
	ret = TREEDB_NS(heap_header_check)(runtime, uSchema);
	if(ret != eTreedbHeapSuccess) {
		TREEDB_heap_unlock();
		TREEDB_heap_close();
	}
	return ret;
	/* IMPORTANT: the caller must unlock the heap!!! */
}
#endif

#ifndef TREEDB_OMIT_FILE_STORAGE_heap_init_size
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_init_size)
( TREEDB_NS(runtime_ptr_t) runtime
, size_t offset
, size_t size
) {
	int res = ftruncate(runtime->m_fls.m_fd, offset + size);
	/*check_return("heap_init_size", "ftruncate failed", res);*/
	if(res)
		return eTreedbHeapFileTruncateFailure;

	TREEDB_STORAGE_set_size(size);
	return eTreedbHeapSuccess;
}
#endif

/* heap_init_fd
   ============
1. close the heap if it's already open(and != fd), otherwise unlock
2. file_lock_state_init
3. set_mmap_flags
4. ftruncate
5. sets the size field in the header
6. heap_write_lock
*/
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_init_fd)
( TREEDB_NS(runtime_ptr_t) runtime
, int fd
, unsigned int open_flags
, unsigned int access_flags
, size_t offset
, size_t size
) {
	TreedbHeapStatus res;
	TREEDB_printf("heap_init_fd\n");
	if(runtime->m_fls.m_fd != -1) {
		if(runtime->m_fls.m_fd != fd) {
			TREEDB_heap_close();
		}else{
			res = TREEDB_heap_unlock();
			if(res != eTreedbHeapSuccess)
				return res;
		}
	}
	runtime->m_fls.m_fd = fd;
	if(!(access_flags & eHeapAccessAnonymous)) {
		if(runtime->m_fls.m_fd == -1)
			return eTreedbHeapFileOpenFailure;

		if(file_lock_state_init(runtime->m_fls.m_fd, & runtime->m_fls, offset, 1
				, (open_flags & eHeapOpenWrite) ?
					eFileLockWriteLocked : eFileLockReadLocked))
			return eTreedbHeapFileLockError;
	}
	runtime->m_offset = offset;
	runtime->m_open_flags = open_flags;
	TREEDB_NS(set_mmap_flags)(runtime, access_flags);

	res = TREEDB_heap_init_size(offset, size);
	if(res != eTreedbHeapSuccess)
		return res;

	/* IMPORTANT: the caller must unlock the heap!!! */
	return TREEDB_heap_write_lock();
}

/* heap_init
   =========
1. heap_init_fd(open(name))
*/
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_init)
( TREEDB_NS(runtime_ptr_t) runtime
, const char * name
, unsigned int open_flags
, unsigned int create_permissions
, unsigned int access_flags
, size_t offset
, size_t size
) {
	TREEDB_printf("heap_init %s\n", name ? name : "(null)");
	int fd;
	if(!name)
		fd = -1;
	else{
		fd = open(name
		, ((open_flags & eHeapOpenWrite) ? O_RDWR : O_RDONLY)
			| (open_flags & eHeapOpenSync ? O_FSYNC : 0)
			| (open_flags & eHeapOpenNoAtime ? O_NOATIME : 0)
			| O_CREAT
		, treedb_mode_flags(create_permissions)
		);
		if(fd == -1) {
			TREEDB_printf("heap_init: couldn't create fd for \"%s\".\n", name);
			return eTreedbHeapFileOpenFailure;
		}
	}
	TreedbHeapStatus ret = TREEDB_NS(heap_init_fd)(runtime, fd
		, open_flags, access_flags, offset, size);
	if(ret != eTreedbHeapSuccess) {
		TREEDB_printf("heap_init: couldn't create \"%s\".\n", name);
	}
	return ret;
	/* IMPORTANT: the caller must unlock the heap!!! */
}

#ifndef TREEDB_OMIT_FILE_STORAGE_heap_create

/* heap_create
   ===========
1. heap_init
2. treedb_init
*/
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_create)
( TREEDB_NS(runtime_ptr_t) runtime
, const uuid_t uSchema
, const char * name
, unsigned int open_flags
, unsigned int create_permissions
, unsigned int access_flags
, size_t offset
, size_t size
) {
	TREEDB_printf("heap_create %s\n", name);
	TreedbHeapStatus ret = TREEDB_heap_init(name, open_flags
		, create_permissions, access_flags, offset, size);
	if(ret != eTreedbHeapSuccess)
		return ret;
	TreedbInitResult res2 = TREEDB_init(uSchema
				, offsetof(TREEDB_NS(treedb_t), TREEDB_PAYLOAD_FIRST_MEMBER)
				, sizeof(TREEDB_NS(treedb_t))
				, size);
	if(res2 != eTreedbInitSuccess) {
		TREEDB_printf("heap_create : bad code! bad code!\n");
		return eTreedbHeapCreateInitFailure;
	}
	return eTreedbHeapSuccess;
	/* IMPORTANT: the caller must unlock the heap!!! */
}
TREEDB_NS_prefix TreedbHeapStatus TREEDB_NS(heap_create2)
( TREEDB_NS(runtime_ptr_t) runtime
, const uuid_t uSchema
, const char * name
, unsigned int open_flags
, unsigned int create_permissions
, unsigned int access_flags
, int dev_zero_fd
, size_t offset
, size_t size
, size_t dev_zero_size
) {
	TREEDB_printf("heap_create2 %s\n", name);

	TreedbHeapStatus ret = TREEDB_heap_init(name, open_flags
		, create_permissions, access_flags, offset, size);
	if(ret != eTreedbHeapSuccess)
		return ret;
	// Forget the old mapping - we use the /dev/zero mapping address.
	TREEDB_heap_unlock();
	// Now create the /dev/zero mapping and overlay the writeable mapping.
	char * pv = (char *)mmap(0, dev_zero_size
		, PROT_READ
		, MAP_SHARED
		, dev_zero_fd
		, 0);
	if(pv == (char *)-1)
		return eTreedbHeapMmapError;

	TREEDB_heap_write_lock_at(pv);
	// This sets up the allocator.
	TreedbInitResult res2 = TREEDB_init(uSchema
				, offsetof(TREEDB_NS(treedb_t), TREEDB_PAYLOAD_FIRST_MEMBER)
				, sizeof(TREEDB_NS(treedb_t))
				, size);
	if(res2 != eTreedbInitSuccess) {
		TREEDB_printf("heap_create : bad code! bad code!\n");
		return eTreedbHeapCreateInitFailure;
	}
	return eTreedbHeapSuccess;
	/* IMPORTANT: the caller must unlock the heap!!! */
}
#endif
#ifdef TREEDB_HAS_RESIZABLE_HEAP
TREEDB_NS_prefix int TREEDB_NS(heap_resize)
( TREEDB_NS(runtime_ptr_t) runtime
, size_t size
) {
	TREEDB_printf("heap_resize %zu\n", size);
	if(!runtime->m_active_writer_thread) {
		TREEDB_printf("heap_resize : heap not write-locked!\n");
		return 1;
	}
	size = V3C_ROUND_UP_T2(size_t, size, TREEDB_PAGE_SIZE);
	size_t prev_size = TREEDB_STORAGE_size;
	if(prev_size == size)
		return 0;
	TREEDB_printf("heap_resize : %zu -> %zu.\n", prev_size, size);
	if(size >= TREEDB_STORAGE_max_size) {
		TREEDB_printf("heap_resize : %zu exceeds the address range of "
			"this heap.\n"
			, size);
		return 2;
	}
	msync(TREEDB_STORAGE_base, prev_size, TREEDB_MSYNC);
	char * p;
	if(size > prev_size) {
		int status = ftruncate(runtime->m_fls.m_fd, size + runtime->m_offset);
		if(status) {
			/*heap_warning("impl_extend_heap", "ftruncate failed");*/
			return 3; /* Out of disk space? */
		}
		p = (char *)mremap(TREEDB_STORAGE_base
			, prev_size
			, size
			, MREMAP_MAYMOVE
			);
		if(p == (char *)-1) {
			TREEDB_printf("impl_page_alloc : mremap failed!\n");
			return 4;
		}
	}else{
		p = (char *)mremap(TREEDB_STORAGE_base
			, prev_size
			, size
			, MREMAP_MAYMOVE
			);
		if(p == (char *)-1) {
			TREEDB_printf("impl_page_alloc : ftruncate failed!\n");
			return 4;
		}
		int status = ftruncate(runtime->m_fls.m_fd, size + runtime->m_offset);
		if(status) {
			/*heap_warning("impl_extend_heap", "ftruncate failed");*/
			return 3; /* Out of disk space? */
		}
	}
	TREEDB_STORAGE_base = p;

	/* Ensure the old address isn't used. */
	v3c_sync_synchronize();

	TREEDB_STORAGE_set_size(size);
	return 0;
}
/*! Extend the heap

A write lock is REQUIRED before calling this function.
*/
TREEDB_NS_prefix void * TREEDB_NS(impl_page_alloc)
( TREEDB_NS(runtime_ptr_t) runtime
, size_t size
) {
	TREEDB_printf("impl_page_alloc %zu\n", size);
	size_t prev_size = TREEDB_STORAGE_size;
	if(!size)
		return TREEDB_STORAGE_base + prev_size;
	int res = TREEDB_NS(heap_resize)(runtime, size + prev_size);
	if(res) {
		TREEDB_printf("impl_page_alloc : resize failed: %d.\n", res);
		return 0;
	}
	return TREEDB_STORAGE_base + prev_size;
}

/*! Shrink the heap

A write lock is REQUIRED before calling this function.
*/
TREEDB_NS_prefix TreedbFreeResult TREEDB_NS(impl_page_free)
( TREEDB_NS(runtime_ptr_t) runtime
, size_t new_size
) {
	int res = TREEDB_NS(heap_resize)(runtime, new_size);
	if(res) {
		TREEDB_printf("impl_page_free : resize failed: %d.\n", res);
		return eTreedbFreeBadHeap;
	}
	return eTreedbFreeSuccess;
}
#endif /* #ifdef TREEDB_HAS_RESIZABLE_HEAP */

#ifdef __cplusplus
} // extern "C" {
#endif
