/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifdef DOXYGEN_processing
{
#endif

#include "handle-array-template-defines.h"

#ifdef DOXYGEN_processing
}
#endif

/* 11. pointer pack / unpack */
#ifndef ARRAY_TMPL_omit_unpack
  #define ARRAY_TMPL_unpack(x) ARRAY_TMPL_NS(unpack)(runtime, x)
ARRAY_TMPL_prefix ARRAY_TMPL_ptr_t
ARRAY_TMPL_NS(unpack)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_ptr_t p
) {
	ARRAY_TMPL_ptr_t ret = (ARRAY_TMPL_ptr_t)ARRAY_TMPL_unpack_pointer(p);
	TREEDB_printf8("array::unpack : %lu[0x%lx] -> %p\n"
		, (uintptr_t)p
		, (uintptr_t)p, (void *)(uintptr_t)ret);
	return ret;
}
#endif
#ifndef ARRAY_TMPL_omit_pack
  #define ARRAY_TMPL_pack(x) ARRAY_TMPL_NS(pack)(runtime, x)
ARRAY_TMPL_prefix ARRAY_TMPL_packed_ptr_t
ARRAY_TMPL_NS(pack)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_ptr_t p
) {
	ARRAY_TMPL_packed_ptr_t ret = ARRAY_TMPL_pack_pointer(p);
	TREEDB_printf8("array::pack : %p -> %lu[0x%lx]\n"
		, (void *)(uintptr_t)p, (uintptr_t)ret
		, (uintptr_t)ret);
	return ret;
}
#endif
#ifndef ARRAY_TMPL_omit_unpack2
  #define ARRAY_TMPL_unpack2(x) ARRAY_TMPL_NS(unpack2)(runtime, x)
ARRAY_TMPL_prefix ARRAY_TMPL_ptr_t
ARRAY_TMPL_NS(unpack2)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t p
) {
	ARRAY_TMPL_ptr_t ret = (ARRAY_TMPL_ptr_t)ARRAY_TMPL_unpack_pointer2(p);
	TREEDB_printf8("array::unpack2 : %lu[0x%lx] -> %p\n"
		, (uintptr_t)p
		, (uintptr_t)p, (void *)(uintptr_t)ret);
	return ret;
}
#endif

#ifndef ARRAY_TMPL_omit_new
/**
Allocate an array.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(new)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_size_t reserve
, ARRAY_TMPL_size_t increment
) {
	if(ARRAY_TMPL_exceeds_count_limit(reserve))
		return 1;
	if(!increment)
		increment = 1;
	else if(ARRAY_TMPL_exceeds_count_limit(increment))
		return 2;
	ARRAY_TMPL_count_t nitems = V3C_ROUND_UP_T(ARRAY_TMPL_count_t, reserve, increment);
	ARRAY_TMPL_ptr_t pa = (ARRAY_TMPL_ptr_t)ARRAY_TMPL_alloc(nitems);
	if(!pa)
		return 3;
	pa->ARRAY_TMPL_count = 0;
	pa->ARRAY_TMPL_reserve = nitems;
	pa->ARRAY_TMPL_increment = increment;
	ARRAY_TMPL_assign_array(hha, pa);
	return 0;
}
#endif

#ifndef ARRAY_TMPL_omit_free
/**
Free an array.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix void ARRAY_TMPL_NS(free)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed_ptr_t ha
) {
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack(ha);
	(void)pa;
	if(ARRAY_TMPL_clear_entries(ha, 0, pa->ARRAY_TMPL_count)) {
		;
	}
	ARRAY_TMPL_free_packed(ha);
}
#endif

/**
Clear (empty) the array.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix void ARRAY_TMPL_NS(clear)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
) {
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack2(hha);
	ARRAY_TMPL_NS(remove_many)(runtime, hha, 0, pa->ARRAY_TMPL_count);
}
/** Make a copy (clone) of another array.

The storage used must be write locked.

The destination array must be initialised first.

If the destination array isn't empty, it is cleared.

If the allocator runs out of memory, the copy is cleared and clone() fails.

This is a "shallow copy" of the raw array data.

You may need to take extra steps to ensure that the copy is a consistant shallow
or deep copy.
*/
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(clone)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha_to
, ARRAY_TMPL_packed2_ptr_t hha_from
) {
	ARRAY_TMPL_packed_ptr_t ha_to
		= *(ARRAY_TMPL_packed_ptr_t*)ARRAY_TMPL_unpack_pointer(hha_to);
	if(ha_to)
		ARRAY_TMPL_NS(free)(runtime, hha_to);
	ARRAY_TMPL_ptr_t pa_to = ARRAY_TMPL_unpack2(hha_to);
	ARRAY_TMPL_ptr_t pa_from = ARRAY_TMPL_unpack2(hha_from);
	if(ARRAY_TMPL_NS(new)(runtime, hha_to, pa_from->ARRAY_TMPL_count
			, pa_to->ARRAY_TMPL_increment))
		return 1;
	pa_to = ARRAY_TMPL_unpack2(hha_to);
	pa_from = ARRAY_TMPL_unpack2(hha_from);
	ARRAY_TMPL_memcopy(ARRAY_TMPL_item_ptr(pa_to, 0), ARRAY_TMPL_item_ptr(pa_from, 0)
		, pa_from->ARRAY_TMPL_count * ARRAY_TMPL_entry_size);
	return 0;
}
/**
Insert an item.

The storage used must be write locked.
*/
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(insert)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_size_t where
, ARRAY_TMPL_input_t d
) {
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack2(hha);
	if(where > pa->ARRAY_TMPL_count)
		return 1;
	if(pa->ARRAY_TMPL_count < pa->ARRAY_TMPL_reserve) {
		; /* We have room to expand */
	}else{
		ARRAY_TMPL_count_t nitems = V3C_ROUND_UP_T(ARRAY_TMPL_count_t, pa->ARRAY_TMPL_count + 1
			, pa->ARRAY_TMPL_increment);
		if(ARRAY_TMPL_realloc_packed(hha, nitems))
			return 1;
		pa = ARRAY_TMPL_unpack2(hha);
		pa->ARRAY_TMPL_reserve = nitems;
	}
	if(where != pa->ARRAY_TMPL_count)
		ARRAY_TMPL_memmove(ARRAY_TMPL_item_ptr(pa, where + 1), ARRAY_TMPL_item_ptr(pa, where)
			, (pa->ARRAY_TMPL_count - where) * ARRAY_TMPL_entry_size);
	ARRAY_TMPL_assign(ARRAY_TMPL_item(pa, where), d);
	++pa->ARRAY_TMPL_count;
	return 0;
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(insert_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_size_t where
, ARRAY_TMPL_size_t count
) {
	if(!count)
		return 0;
	if(ARRAY_TMPL_exceeds_count_limit(where + count - 1))
		return 1;
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack2(hha);
	if(where > pa->ARRAY_TMPL_count)
		return 1;
	if((pa->ARRAY_TMPL_count + count) <= pa->ARRAY_TMPL_reserve) {
		; /* We have room to expand */
	}else{
		ARRAY_TMPL_count_t nitems = V3C_ROUND_UP_T(ARRAY_TMPL_count_t, pa->ARRAY_TMPL_count + count
			, pa->ARRAY_TMPL_increment);
		if(ARRAY_TMPL_realloc_packed(hha, nitems))
			return 2;
		pa = ARRAY_TMPL_unpack2(hha);
		pa->ARRAY_TMPL_reserve = nitems;
	}
	if(where != pa->ARRAY_TMPL_count)
		ARRAY_TMPL_memmove(ARRAY_TMPL_item_ptr(pa, where + count)
			, ARRAY_TMPL_item_ptr(pa, where)
			, (pa->ARRAY_TMPL_count - where) * ARRAY_TMPL_entry_size);
	pa->ARRAY_TMPL_count += count;
	return 0;
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(remove)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_size_t where
) {
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack2(hha);
	if(where >= pa->ARRAY_TMPL_count)
		return 1;
	if(ARRAY_TMPL_clear_entries(hha, where, 1))
		pa = ARRAY_TMPL_unpack2(hha); /* Just in case the memory moved. */
	if((where + 1) != pa->ARRAY_TMPL_count)
		ARRAY_TMPL_memmove(ARRAY_TMPL_item_ptr(pa, where), ARRAY_TMPL_item_ptr(pa, where + 1)
			, (pa->ARRAY_TMPL_count - where - 1) * ARRAY_TMPL_entry_size);
	if(pa->ARRAY_TMPL_count % pa->ARRAY_TMPL_increment) {
		; /* We have room to shrink */
	}else{
		ARRAY_TMPL_count_t nitems = V3C_ROUND_UP_T(ARRAY_TMPL_count_t, pa->ARRAY_TMPL_count - 1
			, pa->ARRAY_TMPL_increment);
		if(ARRAY_TMPL_realloc_packed(hha, nitems))
			return 1;
		pa = ARRAY_TMPL_unpack2(hha);
		pa->ARRAY_TMPL_reserve = nitems;
	}
	--pa->ARRAY_TMPL_count;
	return 0;
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(remove_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_size_t where
, ARRAY_TMPL_size_t count
) {
	if(!count)
		return 0;
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack2(hha);
	if((where + count) > pa->ARRAY_TMPL_count)
		return 1;
	if(ARRAY_TMPL_clear_entries(hha, where, count))
		pa = ARRAY_TMPL_unpack2(hha); /* Just in case the memory moved. */
	if((where + count) != pa->ARRAY_TMPL_count)
		ARRAY_TMPL_memmove(ARRAY_TMPL_item_ptr(pa, where)
			, ARRAY_TMPL_item_ptr(pa, where + count)
			, (pa->ARRAY_TMPL_count - where - count) * ARRAY_TMPL_entry_size);
	if(((pa->ARRAY_TMPL_count - count) / pa->ARRAY_TMPL_increment)
			== (pa->ARRAY_TMPL_count / pa->ARRAY_TMPL_increment)) {
		; /* We have room to shrink */
	}else{
		ARRAY_TMPL_count_t nitems = V3C_ROUND_UP_T(ARRAY_TMPL_count_t
			, pa->ARRAY_TMPL_count - count, pa->ARRAY_TMPL_increment);
		if(ARRAY_TMPL_realloc_packed(hha, nitems))
			return 1;
		pa = ARRAY_TMPL_unpack2(hha);
		pa->ARRAY_TMPL_reserve = nitems;
	}
	pa->ARRAY_TMPL_count -= count;
	return 0;
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(copy_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha_to
, ARRAY_TMPL_size_t to_where
, ARRAY_TMPL_size_t to_count
, ARRAY_TMPL_packed2_ptr_t hha_from
, ARRAY_TMPL_size_t from_where
, ARRAY_TMPL_size_t from_count
) {
	ARRAY_TMPL_ptr_t pa_to;
	if(to_count == (ARRAY_TMPL_size_t)-1) {
		pa_to = ARRAY_TMPL_unpack2(hha_to);
		to_count = pa_to->ARRAY_TMPL_count;
		if(to_where > to_count)
			return 1; // Specifying past-the-end is invalid.
		to_count -= to_where;
	}
	ARRAY_TMPL_ptr_t pa_from;
	if(from_count == (ARRAY_TMPL_size_t)-1) {
		pa_from = ARRAY_TMPL_unpack2(hha_from);
		from_count = pa_from->ARRAY_TMPL_count;
		if(from_where > from_count)
			return 2; // Not enough entries.
		if(from_where == from_count) {
			if(!to_count)
				return 0;
			return ARRAY_TMPL_NS(remove_many)(runtime, hha_to, to_where
				, to_count);
		}
		from_count -= from_where;
	}
	// Fail if "to" and "from" overlap.
	if(hha_to == hha_from) {
		// Check that the source isn't within the destination.
		if(from_count && (from_where >= to_where)
				&& (from_where < (to_where + to_count)))
			return 3;
		// Check that the destination isn't within the source.
		if(to_count && (to_where >= from_where)
				&& (to_where < (from_where + from_count)))
			return 4;
	}
	if(from_count) {
		if(ARRAY_TMPL_NS(insert_many)(runtime, hha_to, to_where + to_count
				, from_count))
			return 1;
		pa_to = ARRAY_TMPL_unpack2(hha_to);
		pa_from = ARRAY_TMPL_unpack2(hha_from);
		ARRAY_TMPL_memcopy(ARRAY_TMPL_item_ptr(pa_to, to_where + to_count)
			, ARRAY_TMPL_item_ptr(pa_from, from_where)
			, from_count * ARRAY_TMPL_entry_size);
	}
	if(to_count)
		if(ARRAY_TMPL_NS(remove_many)(runtime, hha_to, to_where, to_count))
			return 2;
	return 0;
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(move_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha_to
, ARRAY_TMPL_size_t to_where
, ARRAY_TMPL_size_t to_count
, ARRAY_TMPL_packed2_ptr_t hha_from
, ARRAY_TMPL_size_t from_where
, ARRAY_TMPL_size_t from_count
) {
	if(from_count) {
		if(ARRAY_TMPL_NS(insert_many)(runtime, hha_to, to_where + to_count
				, from_count))
			return 1;
		ARRAY_TMPL_ptr_t pa_to = ARRAY_TMPL_unpack2(hha_to);
		ARRAY_TMPL_ptr_t pa_from = ARRAY_TMPL_unpack2(hha_from);
		ARRAY_TMPL_memcopy(ARRAY_TMPL_item_ptr(pa_to, to_where + to_count)
			, ARRAY_TMPL_item_ptr(pa_from, from_where)
			, from_count * ARRAY_TMPL_entry_size);
		if(ARRAY_TMPL_NS(remove_many)(runtime, hha_from, from_where, from_count))
			return 2;
	}
	if(to_count)
		if(ARRAY_TMPL_NS(remove_many)(runtime, hha_to, to_where, to_count))
			return 3;
	return 0;
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(push_back)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_input_t d
) {
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack2(hha);
	return ARRAY_TMPL_NS(insert)(runtime, hha, pa->ARRAY_TMPL_count, d);
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(push_back_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_size_t count
) {
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack2(hha);
	return ARRAY_TMPL_NS(insert_many)(runtime, hha, pa->ARRAY_TMPL_count, count);
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(push_front)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_input_t d
) {
	return ARRAY_TMPL_NS(insert)(runtime, hha, 0, d);
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(push_front_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_size_t count
) {
	return ARRAY_TMPL_NS(insert_many)(runtime, hha, 0, count);
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(pop_back)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
) {
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack2(hha);
	return ARRAY_TMPL_NS(remove)(runtime, hha, pa->ARRAY_TMPL_count - 1);
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(pop_back_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_size_t count
) {
	ARRAY_TMPL_ptr_t pa = ARRAY_TMPL_unpack2(hha);
	return ARRAY_TMPL_NS(remove_many)(runtime, hha, pa->ARRAY_TMPL_count - count
		, count);
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(pop_front)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
) {
	return ARRAY_TMPL_NS(remove)(runtime, hha, 0);
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(pop_front_many)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_packed2_ptr_t hha
, ARRAY_TMPL_size_t count
) {
	return ARRAY_TMPL_NS(remove_many)(runtime, hha, 0, count);
}
ARRAY_TMPL_prefix int ARRAY_TMPL_NS(reverse_some)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_ptr_t pa
, ARRAY_TMPL_size_t lower
, ARRAY_TMPL_size_t count
) {
	(void)runtime;
	/* count includes the first item. */
	if((lower + count) > pa->ARRAY_TMPL_count)
		return 1;
	ARRAY_TMPL_count_t n, num = count/2, upper = lower + count - 1;
	for(n = 0; n < num; ++n) {
		ARRAY_TMPL_auto_data(item1);
		ARRAY_TMPL_auto_data(item2);
		ARRAY_TMPL_assign_to_auto(item1, ARRAY_TMPL_item(pa, lower));
		ARRAY_TMPL_assign_to_auto(item2, ARRAY_TMPL_item(pa, upper));
		/*printf("swapping %lu[%u] and %lu[%u].\n", (size_t)(lower)
			, item1, (size_t)(upper), item2);*/
		ARRAY_TMPL_assign_from_auto(ARRAY_TMPL_item(pa, lower), item2);
		ARRAY_TMPL_assign_from_auto(ARRAY_TMPL_item(pa, upper), item1);
		++lower, --upper;
	}
	/*fflush(stdout);*/
	return 0;
}
ARRAY_TMPL_prefix void ARRAY_TMPL_NS(reverse)
( ARRAY_TMPL_runtime_ptr_t runtime
, ARRAY_TMPL_ptr_t pa
) {
	(void)ARRAY_TMPL_NS(reverse_some)(runtime, pa, 0, pa->ARRAY_TMPL_count);
}
