/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdint.h>
#include "treedb.h"

#if TREEDB_NBITS==8
#elif TREEDB_NBITS==16
	#define TREEDB_need_exceed_check
#elif TREEDB_NBITS==32
  #if __WORDSIZE == 64
	#define TREEDB_need_exceed_check
  #endif
#elif TREEDB_NBITS==64
#elif TREEDB_NBITS==ptr
#else
  #error TREEDB_NBITS is not defined or invalid.
#endif

#ifndef TREEDB_NS
  #define TREEDB_NS(x) V3C_INVOKE3(V3C_JOIN3,x,_,TREEDB_NBITS)
#endif
#ifndef TREEDB_NS_CONCAT
  #define TREEDB_NS_CONCAT(a,b) V3C_INVOKE3(V3C_JOIN3,a,TREEDB_NBITS,b)
#endif

#ifndef TREEDB_OMIT_typedef_count_t
typedef TREEDB_NS_CONCAT(uint,_t) TREEDB_NS(count_t);
#endif

#ifndef TREEDB_NS_runtime_ptr_t
  #define TREEDB_NS_runtime_ptr_t TREEDB_NS(runtime_ptr_t)
#endif

/* runtime_t usage
Call heap_read_lock or heap_write_lock before accessing the heap and heap_unlock
afterwards.
This is typically done by an implementor wrapper.

We must have read->unlocked->write and write->unlocked->read because two threads
that initially have a read lock cannot both try to promote their lock to a write
lock as they would block each other forever.

Readers wait on the condition that there are no waiting or active writer
threads.
Writers wait on the condition that there are no readers and
m_active_writer_thread == 0 (no active writer).

This gives writers priority and allows readers to operate in parallel.

This also forces m_fls to transition to unlocked between read and write locks.
*/

#ifndef TREEDB_OMIT_typedef_runtime_t
typedef struct TREEDB_NS(runtime_t) TREEDB_NS(runtime_t)
, * TREEDB_NS(runtime_ptr_t);
#endif

/******************************************************************************/
/* AvlAllocator                                                               */
/*                                                                            */
/* The used and free trees are of different types. Even though they contain   */
/* the same node type, the free tree allows duplicates and sorts by size,     */
/* while the used tree sorts by address.                                      */
/*                                                                            */
/******************************************************************************/
#ifndef TREEDB_NSA
  #define TREEDB_NSA(x) TREEDB_NS(AvlAllocator_##x)
#endif
#ifndef TREEDB_NSF
  #define TREEDB_NSF(x) TREEDB_NSA(free_##x)
#endif
#ifndef TREEDB_NSU
  #define TREEDB_NSU(x) TREEDB_NSA(used_##x)
#endif

/* We can't rely on macros to stay around as long as the code, so we're forced
to use typedefs. */
#ifndef TREEDB_OMIT_treedb_ns_types_typedefs1
typedef TREEDB_NS(count_t) TREEDB_NSF(count_t), TREEDB_NSU(count_t);
typedef TREEDB_NS(runtime_ptr_t) TREEDB_NSF(runtime_ptr_t)
, TREEDB_NSU(runtime_ptr_t), TREEDB_NSA(runtime_ptr_t);
#endif

/******************************************************************************/
/* For the schema. */
#ifndef OMIT_TREEDB_NS_SCHEMA_TYPEDEFS
typedef struct TREEDB_NSF(tree_t) TREEDB_NS(free_tree_t);
typedef struct TREEDB_NSU(tree_t) TREEDB_NS(used_tree_t);
#endif
/******************************************************************************/

#ifndef TREEDB_malloc
#define TREEDB_malloc(sz) TREEDB_NS(malloc)(runtime, sz)
#define TREEDB_realloc(p, newsz) TREEDB_NS(realloc)(runtime, p, newsz)
#define TREEDB_free(p) TREEDB_NS(free)(runtime, p)
#define TREEDB_in_heap(p) TREEDB_NS(in_heap)(runtime, p)
#define TREEDB_did_alloc(p) TREEDB_NS(did_alloc)(runtime, p)
#endif
#ifndef TREEDB_unpack_pointer
  #define TREEDB_unpack_pointer(u) \
	TREEDB_NS(impl_unpack_pointer)(runtime, u)
#endif
#ifndef TREEDB_pack_pointer
  #define TREEDB_pack_pointer(pv) \
	TREEDB_NS(impl_pack_pointer)(runtime, pv)
#endif

#ifndef TREEDB_pointer_to_offset
  #define TREEDB_pointer_to_offset(pv) TREEDB_pack_pointer(pv)
#endif
#ifndef TREEDB_offset_to_pointer
  #define TREEDB_offset_to_pointer(u) TREEDB_unpack_pointer(u)
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifndef TREEDB_OMIT_typedef_treedb_t
typedef struct TREEDB_NS(treedb_t) TREEDB_NS(treedb_t)
, * TREEDB_NS(treedb_ptr_t);
#endif

#ifdef TREEDB_SHARED
  #include <v3c/file-lock.h>
#endif

#ifndef TREEDB_NS_declare_prefix
  #define TREEDB_NS_declare_prefix extern
#endif

TREEDB_NS_declare_prefix TreedbInitResult TREEDB_NS(init)
( TREEDB_NS_runtime_ptr_t runtime
, const uuid_t uSchema
, uint64_t uPayloadOffset
, uint64_t uHeaderSize
, uint64_t uSize
);

TREEDB_NS_declare_prefix TreedbInitResult TREEDB_NS(init_memory)
( TREEDB_NS_runtime_ptr_t runtime
, uint16_t uHeaderSize
);

TREEDB_NS_declare_prefix int TREEDB_NS(check_heap)(TREEDB_NS_runtime_ptr_t runtime);
TREEDB_NS_declare_prefix void * TREEDB_NS(malloc)(TREEDB_NS_runtime_ptr_t runtime, size_t size);
TREEDB_NS_declare_prefix void * TREEDB_NS(calloc)(TREEDB_NS_runtime_ptr_t runtime, size_t count
	, size_t element_size);
TREEDB_NS_declare_prefix void * TREEDB_NS(realloc)(TREEDB_NS_runtime_ptr_t runtime, void * p
	, size_t new_size);
TREEDB_NS_declare_prefix TreedbFreeResult TREEDB_NS(free)(TREEDB_NS_runtime_ptr_t runtime
	, void * p);
TREEDB_NS_declare_prefix int TREEDB_NS(in_heap)(TREEDB_NS_runtime_ptr_t runtime
	, const void * p);
TREEDB_NS_declare_prefix int TREEDB_NS(did_alloc)(TREEDB_NS_runtime_ptr_t runtime
	, const void * p);
TREEDB_NS_declare_prefix size_t TREEDB_NS(usable_size)(
	TREEDB_NS_runtime_ptr_t runtime, const void * p);

TREEDB_NS_declare_prefix void TREEDB_NS(impl_post_free)(TREEDB_NS_runtime_ptr_t runtime, void * p
, size_t size);

#ifndef TREEDB_OMIT_pack_unpack_pointer_DECLARE
TREEDB_NS_declare_prefix void * TREEDB_NS(impl_unpack_pointer)
( TREEDB_NS_runtime_ptr_t runtime, TREEDB_NS(count_t) u);
TREEDB_NS_declare_prefix TREEDB_NS(count_t) TREEDB_NS(impl_pack_pointer)
( TREEDB_NS_runtime_ptr_t runtime, const void * u);
#endif

#ifdef __cplusplus
} // extern "C" {
#endif
