/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <string.h>
#include <stdio.h>
#include <v3c/bits.h>

#ifndef TREEDB_NBITS
  #error TREEDB_NBITS is undefined.
#endif
#ifdef TREEDB_SHARED
  #ifdef TREEDB_HAS_HEAPS
    #error TREEDB_HAS_HEAPS define is defunct.
  #endif
  #ifndef TREEDB_OMIT_treedb_file_storage_impl

#ifdef DOXYGEN_processing
{
#endif

    #include <treedb/file-storage-impl.h>

#ifdef DOXYGEN_processing
}
#endif

  #endif
#endif

#include "treedb-defines.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifndef TREEDB_HAS_RESIZABLE_HEAP
TREEDB_NS_prefix void * TREEDB_NS(impl_page_alloc)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS_size_t size
) {
	(void)runtime;
	(void)size;
	return 0;
}
TREEDB_NS_prefix TreedbFreeResult TREEDB_NS(impl_page_free)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS_size_t new_size
) {
	(void)runtime;
	(void)new_size;
	return eTreedbFreeHeapIsFixed;
}
#endif

#ifndef TREEDB_OMIT_pack_unpack_pointer_DECLARE
TREEDB_NS_prefix void * TREEDB_NS(impl_unpack_pointer)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS(count_t) u
) {
	if(!u)
		return 0;
	return TREEDB_STORAGE_base + u;
}
TREEDB_NS_prefix TREEDB_NS(count_t) TREEDB_NS(impl_pack_pointer)
( TREEDB_NS_runtime_ptr_t runtime
, const void * pv
) {
	if(!pv)
		return 0;
	else
		return ((const char *)pv) - TREEDB_STORAGE_base;
}
#endif

#define TREEDB_NSA_unpack_node(x) \
	((TREEDB_NSA(node_ptr_t))TREEDB_unpack_pointer(x))
#define TREEDB_NS_unpack_free_node(x) \
	((TREEDB_NSA(node_ptr_t)) \
	TREEDB_unpack_pointer((x) & ~((TREEDB_NS(count_t))1)))
#define TREEDB_NSA_pack_node(x) \
	(TREEDB_pack_pointer(x))

extern TreedbFreeResult TREEDB_NS(free)
( TREEDB_NS_runtime_ptr_t runtime
, void * p);
extern void TREEDB_NS(impl_post_free)
( TREEDB_NS_runtime_ptr_t runtime
, void * p
, TREEDB_NS_size_t size);

#ifndef TREEDB_OMIT_treedb_init_size
TREEDB_NS_prefix void TREEDB_NS(init_size)
( TREEDB_NS_runtime_ptr_t runtime
, uint64_t uSize
) {
	TREEDB_STORAGE_size = uSize;
}
#endif

#ifndef TREEDB_OMIT_treedb_init_header

/* init_header
   ===========
1. check parameters
2. write header
3. set size
*/
TREEDB_NS_prefix TreedbInitResult TREEDB_NS(init_header)
( TREEDB_NS_runtime_ptr_t runtime
, const uuid_t uAllocatorAlgorithm
, const uuid_t uSchema
, unsigned int uBytesPerOffset
, uint64_t uPayloadOffset
, uint64_t uSize
) {
	TREEDB_printf("init_header : uBytesPerOffset %u, uPayloadOffset %zu\n"
		, uBytesPerOffset, (size_t)uPayloadOffset);
	if(!uBytesPerOffset)
		return eTreedbInitOffsetSizeTooSmall;
	if((uBytesPerOffset != 1) && (uBytesPerOffset & 1))
		return eTreedbInitOffsetSizeOdd; /* Not tested. */
	if(uBytesPerOffset > sizeof(void *))
		return eTreedbInitOffsetSizeTooBig;
	if(uPayloadOffset < sizeof(TreedbHeader))
		return eTreedbInitPayloadOffsetTooSmall;
	if(uPayloadOffset >= TREEDB_PAGE_SIZE)
		return eTreedbInitPayloadOffsetTooBig;

	// TreedbHeader.
	strcpy(runtime->m_pTreedb->m_header.m_szMagic, "3db");
	uuid_copy(runtime->m_pTreedb->m_header.m_uAllocatorAlgorithm
		, uAllocatorAlgorithm);
	uuid_copy(runtime->m_pTreedb->m_header.m_uSchema, uSchema);
	runtime->m_pTreedb->m_header.m_uBytesPerOffset = uBytesPerOffset;
	runtime->m_pTreedb->m_header.m_uEndianIndex = v3c_native_endian_index();
	runtime->m_pTreedb->m_header.m_uFeatures = 3; // count and depth.
	runtime->m_pTreedb->m_header.m_ulPayloadOffset = uPayloadOffset;

	TREEDB_NS(init_size)(runtime, uSize);

	return eTreedbInitSuccess;
}
#endif

#ifndef TREEDB_OMIT_treedb_init_memory

/* init_memory
   ===========
1. setup used + free trees
2. create + insert one free node
*/
TREEDB_NS_prefix TreedbInitResult TREEDB_NS(init_memory)
( TREEDB_NS_runtime_ptr_t runtime
, uint16_t uHeaderSize
) {
	TREEDB_printf("init_memory [%u]\n", uHeaderSize);
	if(uHeaderSize < TREEDB_MINSIZE) {
		TREEDB_printf("init_memory [%u] : error header too small.\n"
			, uHeaderSize);
		return eTreedbInitHeaderSizeTooSmall;
	}
	/* Initialise the allocator trees. */
	TREEDB_NSF(init)(runtime, TREEDB_ALLOC_free_tree);
	TREEDB_NSU(init)(runtime, TREEDB_ALLOC_used_tree);

	TREEDB_NS(count_t) used = TREEDB_NS_ALIGNED_OFFSET(uHeaderSize);

	/* Manually insert a free node that describes the available storage space in
	the first page.
	*/
	TREEDB_NSA(node_ptr_t) pFirstNode
		= (TREEDB_NSA(node_ptr_t))(TREEDB_STORAGE_base + used);
	used += TREEDB_NS_ALIGNED_OFFSET(sizeof(TREEDB_NSA(node_t)));

	TREEDB_ALLOC_node_set_length(pFirstNode, TREEDB_STORAGE_size - used);
	TREEDB_ALLOC_node_set_prev(pFirstNode, TREEDB_pack_allocator_node(0));
	TREEDB_ALLOC_node_set_next(pFirstNode, TREEDB_pack_allocator_node(0));
	TREEDB_NSF(insert_node)(runtime, TREEDB_ALLOC_free_tree, pFirstNode);

	return eTreedbInitSuccess;
}
#endif

#ifndef TREEDB_OMIT_treedb_init

/* init
   ====
1. init_header
2. init_memory
*/
TREEDB_NS_prefix TreedbInitResult TREEDB_NS(init)
( TREEDB_NS_runtime_ptr_t runtime
, const uuid_t uSchema
, uint64_t uPayloadOffset
, uint64_t uHeaderSize
, uint64_t uSize
) {
	TREEDB_printf("init : uPayloadOffset %zu, uHeaderSize %zu\n"
		, (size_t)uPayloadOffset, (size_t)uHeaderSize);
	TreedbInitResult res = TREEDB_NS(init_header)(runtime
		, UUID_treedb_avl_two_tree_allocator_algorithm, uSchema
		, sizeof(TREEDB_NS(count_t)), uPayloadOffset, uSize);
	if(res != eTreedbInitSuccess) {
		TREEDB_printf("init : error : init_header failed : %s.\n"
			, TreedbInitResultDesc(res));
		return res;
	}
	res = TREEDB_NS(init_memory)(runtime, uHeaderSize);
	if(res != eTreedbInitSuccess) {
		TREEDB_printf("init : error : init_memory failed : %s.\n"
			, TreedbInitResultDesc(res));
	}
	return res;
}
#endif

#ifndef TREEDB_have_dump_trees
TREEDB_NS_prefix void TREEDB_NS(dump_trees)
( TREEDB_NS_runtime_ptr_t runtime
) {
	(void)runtime;
}
#else
extern void TREEDB_NS(dump_trees)
( TREEDB_NS_runtime_ptr_t runtime
);
#endif

static void TREEDB_NS(dump_allocator_trees)
( TREEDB_NS_runtime_ptr_t runtime
) {
	(void)runtime;
	TREEDB_printf("[%zu] free : head(%zu), top(%zu), tail(%zu).\n"
		, (size_t)TREEDB_pointer_to_offset(TREEDB_ALLOC_free_tree)
		, (size_t)TREEDB_pointer_to_offset(TREEDB_ALLOC_free_head)
		, (size_t)TREEDB_pointer_to_offset(TREEDB_ALLOC_free_top)
		, (size_t)TREEDB_pointer_to_offset(TREEDB_ALLOC_free_tail)
	);
	TREEDB_printf("[%zu] used : head(%zu), top(%zu), tail(%zu).\n"
		, (size_t)TREEDB_pointer_to_offset(TREEDB_ALLOC_used_tree)
		, (size_t)TREEDB_pointer_to_offset(TREEDB_ALLOC_used_head)
		, (size_t)TREEDB_pointer_to_offset(TREEDB_ALLOC_used_top)
		, (size_t)TREEDB_pointer_to_offset(TREEDB_ALLOC_used_tail)
	);
}
#ifndef TREEDB_OMIT_treedb_check_heap
/******************************************************************************/
/* check_heap                                                                 */
/*                                                                            */
/* Returns 0 if the heap is OK, non-zero otherwise.                           */
/*                                                                            */
/******************************************************************************/
TREEDB_NS_prefix int TREEDB_NS(check_heap)
( TREEDB_NS_runtime_ptr_t runtime
) {
	TREEDB_printf("check_heap %p [%zu]\n"
		, TREEDB_STORAGE_base, (size_t)TREEDB_STORAGE_size);
	if(!TREEDB_STORAGE_size) {
		TREEDB_printf("\tError : Zero-length heap.\n");
		TREEDB_ALLOC_set_errno(eTreedbZeroLengthHeap);
		return eTreedbZeroLengthHeap;
	}
	TREEDB_NS(dump_allocator_trees)(runtime);
	TREEDB_NSA(node_ptr_t) p = 0, p2, p3;
	int saw_entries;
	TREEDB_NS(count_t) zero = 0;
	TreedbSearchResult res = TREEDB_NSU(search_ge)(runtime
		, TREEDB_ALLOC_used_tree, zero, &p);
	int is_free = 0;
	switch(res) {
	case eTreedbSearchEmpty:
		/* There should be just one free node. */
		is_free = 1;
		p = TREEDB_ALLOC_free_head;
	case eTreedbSearchGreater: {
		if(!p) {
			TREEDB_printf("\tError : Got a null node.\n");
			TREEDB_ALLOC_set_errno(eTreedbNullNode);
			return eTreedbNullNode;
		}
		/* Find the start. */
		TREEDB_NSA(packed_node_ptr_t) opp = TREEDB_ALLOC_node_prev(p);
		while(opp) {
			if(TREEDB_packed_node_isfree(opp)) {
				if(is_free) {
					TREEDB_printf("\tError : [%zu] m_offset_prev "
						"is free.\n"
						, (size_t)TREEDB_pointer_to_offset(p)
					);
					TREEDB_NS(dump_trees)(runtime);
					TREEDB_ALLOC_set_errno(eTreedbPrevNodeIsFree);
					return eTreedbPrevNodeIsFree;
				}
				is_free = 1;
			}else
				is_free = 0;
			p2 = TREEDB_NS_unpack_free_node(opp);
			if(p == p2) {
				TREEDB_printf("\tError : [%zu] m_offset_prev "
					"is ME.\n", (size_t)TREEDB_pointer_to_offset(p)
				);
				TREEDB_ALLOC_set_errno(eTreedbPrevNodeIsThis);
				return eTreedbPrevNodeIsThis;
			}
			if(!p2) {
				TREEDB_printf("\tError : TREEDB_ALLOC_node_prev(p) returned"
					" %zu but we have a null node.\n"
					, (size_t)opp
				);
				TREEDB_ALLOC_set_errno(eTreedbPrevNodeIsBad);
				return eTreedbPrevNodeIsBad;
			}
			p = p2;
			opp = TREEDB_ALLOC_node_prev(p);
		}
		if(is_free) {
			if(!TREEDB_NSF(search)(runtime
					, TREEDB_ALLOC_free_tree, TREEDB_ALLOC_node_length(p))
			) {
				TREEDB_printf("\tError : [%zu : %zu] node not "
					"in free heap.\n"
					, (size_t)TREEDB_pointer_to_offset(p)
					, (size_t)(TREEDB_ALLOC_node_length(p))
				);
				TREEDB_NS(dump_trees)(runtime);
				TREEDB_ALLOC_set_errno(eTreedbNodeNotInFreeHeap);
				return eTreedbNodeNotInFreeHeap;
			}
		}else{
			TREEDB_NS(count_t) offs = TREEDB_pack_pointer(p);
			if(!TREEDB_NSU(search)(runtime, TREEDB_ALLOC_used_tree, offs)) {
				TREEDB_printf("\tError : [%zu : %zu] node not "
					"in used heap.\n"
					, (size_t)TREEDB_pointer_to_offset(p)
						+ TREEDB_ALLOC_node_size
					, (size_t)(TREEDB_ALLOC_node_length(p)));
					TREEDB_NS(dump_trees)(runtime);
				TREEDB_ALLOC_set_errno(eTreedbNodeNotInUsedHeap);
				return eTreedbNodeNotInUsedHeap;
			}
		}
		saw_entries = TREEDB_ALLOC_node_next(p) != 0;
		if(saw_entries) {
			/* Walk the linked list forwards. */
			TREEDB_printf("\tBy offset %zu [%c:%zu:%zu] -> "
				, (size_t)(
					TREEDB_NSU(count)(runtime, TREEDB_ALLOC_used_tree)
					+ TREEDB_NSF(count)(runtime, TREEDB_ALLOC_free_tree)
					)
				, is_free ? 'F' : 'U'
				, (size_t)TREEDB_pointer_to_offset(p)
					+ (is_free ? 0 : TREEDB_ALLOC_node_size)
				, (size_t)TREEDB_ALLOC_node_length(p)
				);
		}
		while(TREEDB_ALLOC_node_next(p)) {
			if(TREEDB_pack_pointer(TREEDB_ALLOC_nodeptr_to_addr(p)
						+ TREEDB_ALLOC_node_length(p))
					!= TREEDB_item_offset(TREEDB_ALLOC_node_next(p))) {
				TREEDB_printf("\n\tError : [%zu] addr + length "
					"[%zu] != p->m_offset_next[%zu].\n"
					, (size_t)TREEDB_pointer_to_offset(p)
					, (size_t)TREEDB_pointer_to_offset(
							TREEDB_ALLOC_nodeptr_to_addr(p)
							+ (size_t)(TREEDB_ALLOC_node_length(p)))
					, (size_t)TREEDB_item_offset(
						TREEDB_ALLOC_node_next(p)));
				TREEDB_ALLOC_set_errno(eTreedbNodeLengthBad);
				return eTreedbNodeLengthBad;
			}
			if(TREEDB_packed_node_isfree(TREEDB_ALLOC_node_next(p))) {
				if(is_free) {
					TREEDB_printf("\n\tError : [%zu] m_offset_next is FREE.\n"
						, (size_t)TREEDB_pointer_to_offset(p)
					);
					TREEDB_NS(dump_trees)(runtime);
					TREEDB_ALLOC_set_errno(eTreedbUnexpectedFreeNode);
					return eTreedbUnexpectedFreeNode;
				}
				is_free = 1;
			}else
				is_free = 0;
			p2 = TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_next(p));
			if(p == p2) {
				TREEDB_printf("\n\tError : [%zu] m_offset_next is ME.\n"
					, (size_t)TREEDB_pointer_to_offset(p)
						+ TREEDB_ALLOC_node_size
				);
				TREEDB_ALLOC_set_errno(eTreedbNextNodeIsThis);
				return eTreedbNextNodeIsThis;
			}
			if(p > p2) {
				TREEDB_printf("\n\tError : [%zu] m_offset_next "
					"is LESS [%zu].\n"
					, (size_t)TREEDB_pointer_to_offset(p)
					, (size_t)TREEDB_pointer_to_offset(p2)
				);
				TREEDB_ALLOC_set_errno(eTreedbNextNodeIsLess);
				return eTreedbNextNodeIsLess;
			}
			if(p2) {
				p3 = TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_prev(p2));
				if(p3 != p) {
					TREEDB_printf("\n\tError : [%zu] "
						"m_offset_next->m_offset_prev[%zu] != This.\n"
						, (size_t)TREEDB_pointer_to_offset(p)
						, (size_t)TREEDB_pointer_to_offset(p3)
					);
					TREEDB_ALLOC_set_errno(eTreedbNodeLengthBad);
					return eTreedbNodeLengthBad;
				}
			}
			p3 = p;
			p = p2;
			if(is_free) {
				if(!TREEDB_NSF(search)(runtime, TREEDB_ALLOC_free_tree
					, TREEDB_ALLOC_node_length(p))) {
					TREEDB_printf("\n\tError 13 : [%zu : %zu] node not "
					"in free heap.\n"
					, (size_t)TREEDB_pointer_to_offset(p)
					, (size_t)(TREEDB_ALLOC_node_length(p)));
					TREEDB_NS(dump_trees)(runtime);
					TREEDB_ALLOC_set_errno(eTreedbNodeNotInFreeHeap);
					return eTreedbNodeNotInFreeHeap;
				}
			}else{
				TREEDB_NS(count_t) offs = TREEDB_pack_pointer(p);
				if(!TREEDB_NSU(search)(runtime, TREEDB_ALLOC_used_tree
						, offs)) {
					TREEDB_printf("\n\tError : [%zu : %zu] node not "
						"in used heap.\n"
						, (size_t)TREEDB_pointer_to_offset(p)
						, (size_t)(TREEDB_ALLOC_node_length(p)));
					TREEDB_NS(dump_trees)(runtime);
					TREEDB_ALLOC_set_errno(eTreedbNodeNotInUsedHeap);
					return eTreedbNodeNotInUsedHeap;
				}
			}
			TREEDB_printf("[%c:%zu:%zu] -> "
				, is_free ? 'F' : 'U'
				, (size_t)TREEDB_pointer_to_offset(p)
						+ (is_free ? 0 : TREEDB_ALLOC_node_size)
				, (size_t)TREEDB_ALLOC_node_length(p)
				);
		}
		if(saw_entries) {
			TREEDB_printf("NULL\n");
			size_t top = (size_t)TREEDB_pointer_to_offset(p)
				+ TREEDB_ALLOC_node_size + (size_t)TREEDB_ALLOC_node_length(p);
			if(top > TREEDB_STORAGE_size) {
				TREEDB_printf("\tError : last node end [%zu] past end of "
					"heap.\n", top);
				TREEDB_NS(dump_trees)(runtime);
				TREEDB_ALLOC_set_errno(eTreedbLastNodePastEndOfHeap);
				return eTreedbLastNodePastEndOfHeap;
			}
		}else if(res == eTreedbSearchEmpty) {
			p = TREEDB_ALLOC_free_head;
			size_t top = (size_t)TREEDB_pointer_to_offset(p)
				+ TREEDB_ALLOC_node_size + (size_t)TREEDB_ALLOC_node_length(p);
			if(top > TREEDB_STORAGE_size) {
				TREEDB_printf("\tError : last node end [%zu] past end of "
					"heap.\n", top);
				TREEDB_NS(dump_trees)(runtime);
				TREEDB_ALLOC_set_errno(eTreedbLastNodePastEndOfHeap);
				return eTreedbLastNodePastEndOfHeap;
			}
		}
		break; }
	default:
		TREEDB_printf("\tError : [%zu] invalid used heap (%s).\n"
			, (size_t)TREEDB_pointer_to_offset(p)
				+ TREEDB_ALLOC_node_size
			, TreedbSearchResultDesc(res));
		TREEDB_NS(dump_trees)(runtime);
		TREEDB_ALLOC_set_errno(eTreedbInvalidUsedHeap);
		return eTreedbInvalidUsedHeap;
	}
	p = TREEDB_ALLOC_used_head;
	saw_entries = p != 0;
	if(saw_entries) {
		TREEDB_printf("\tUsed %zu "
			, (size_t)TREEDB_NSU(count)(runtime, TREEDB_ALLOC_used_tree)
		);
	}
	for(; p; ) {
		TREEDB_NS(count_t) offs = TREEDB_pointer_to_offset(p);
		TREEDB_NS(count_t) offsp = TREEDB_pack_pointer(p);
		TREEDB_printf("[%zu:%zu] -> "
			, (size_t)offs
				+ TREEDB_ALLOC_node_size
			, (size_t)TREEDB_ALLOC_node_length(p)
			);
		if(!TREEDB_NSU(search)(runtime, TREEDB_ALLOC_used_tree, offsp)) {
				TREEDB_printf("\n\tError : [%zu : %zu] node not "
					"in the used heap.\n"
					, (size_t)TREEDB_pointer_to_offset(p)
						+ TREEDB_ALLOC_node_size
					, (size_t)(TREEDB_ALLOC_node_length(p)));
			TREEDB_NS(dump_trees)(runtime);
			TREEDB_ALLOC_set_errno(eTreedbNodeNotInUsedHeap);
			return eTreedbNodeNotInUsedHeap;
		}
		p2 = TREEDB_node_next(p);
		if(p == p2) {
			TREEDB_printf("\tError : [%zu] This->next == This.\n"
				, (size_t)offs);
			TREEDB_ALLOC_set_errno(eTreedbNextNodeIsThis);
			return eTreedbNextNodeIsThis;
		}
		if(p2) {
			/* Yes, this is an invariant - we're checking for corruption. */
			TREEDB_NS(count_t) offs2 = TREEDB_pointer_to_offset(p2);
			if(offs > offs2) {
				TREEDB_printf("\tError : This [%zu] "
					"> This->next [%zu].\n"
					, (size_t)offs + TREEDB_ALLOC_node_size
					, (size_t)offs2 + TREEDB_ALLOC_node_size
				);
				TREEDB_NS(dump_trees)(runtime);
				TREEDB_ALLOC_set_errno(eTreedbNextNodeIsLess);
				return eTreedbNextNodeIsLess;
			}
		}
		p = p2;
	}
	if(saw_entries) {
		TREEDB_printf("NULL\n");
	}
	p = TREEDB_ALLOC_free_head;
	saw_entries = p != 0;
	if(saw_entries) {
		TREEDB_printf("\tFree %zu "
			, (size_t)TREEDB_NSF(count)(runtime, TREEDB_ALLOC_free_tree));
	}
	for(; p; ) {
		TREEDB_printf("[%zu:%zu] -> "
			, (size_t)TREEDB_pointer_to_offset(p)
			, (size_t)TREEDB_ALLOC_node_length(p)
			);
		TREEDB_NS(count_t) len = TREEDB_ALLOC_node_length(p);
		if(!TREEDB_NSF(search)(runtime, TREEDB_ALLOC_free_tree, len)) {
				TREEDB_printf("\n\tError : [%zu : %zu] node not "
					"in the free heap.\n"
					, (size_t)TREEDB_pointer_to_offset(p)
					, (size_t)len
				);
			TREEDB_NS(dump_trees)(runtime);
			TREEDB_ALLOC_set_errno(eTreedbNodeNotInFreeHeap);
			return eTreedbNodeNotInFreeHeap;
		}
		p2 = TREEDB_node_next(p);
		if(p == p2) {
			TREEDB_printf("\tError 21 : [%zu] This->next "
				"== This.\n"
				, (size_t)TREEDB_pointer_to_offset(p)
			);
			TREEDB_ALLOC_set_errno(eTreedbNextNodeIsThis);
			return eTreedbNextNodeIsThis;
		}
		if(p2) {
			/* Yes, this is an invariant - we're checking for corruption. */
			TREEDB_NS(count_t) len2 = TREEDB_ALLOC_node_length(p2);
			if(len > len2) {
				TREEDB_printf("\tError 22 : length(This) [%zu:%zu] "
					"> length(This->next) [%zu:%zu].\n"
					, (size_t)TREEDB_pointer_to_offset(p)
					, (size_t)len
					, (size_t)TREEDB_pointer_to_offset(p2)
					, (size_t)len2
				);
				TREEDB_NS(dump_trees)(runtime);
				TREEDB_ALLOC_set_errno(eTreedbNextNodeIsLess);
				return eTreedbNextNodeIsLess;
			}
		}
		p = p2;
	}
	if(saw_entries) {
		TREEDB_printf("NULL\n");
	}
	return 0;
}
#endif /* #ifndef TREEDB_OMIT_treedb_check_heap */

#ifndef TREEDB_OMIT_treedb_impl_DECLARE
void * TREEDB_NS(impl_page_alloc)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS_size_t size);
TreedbFreeResult TREEDB_NS(impl_page_free)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS_size_t new_size);
#endif

#ifndef TREEDB_OMIT_treedb_fns_DECLARE
TREEDB_NS_prefix void TREEDB_NS(malloc_lower_split)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NSA(node_ptr_t) pn
, TREEDB_NS_size_t size
) {
	TREEDB_NS(count_t) chunk = TREEDB_ALLOC_node_size + size;
	TREEDB_NS(count_t) freesize = TREEDB_ALLOC_node_length(pn) - chunk;
	TREEDB_NSA(node_ptr_t) pn_free = (TREEDB_NSA(node_ptr_t))
		(TREEDB_ALLOC_nodeptr_to_addr(pn) + size);
	TREEDB_NSA(packed_node_ptr_t) packed_pn = TREEDB_NSA_pack_node(pn);
	TREEDB_NSA(packed_node_ptr_t) packed_pn_free
		= TREEDB_NSA_pack_node(pn_free);
	TREEDB_ALLOC_node_set_length(pn_free, freesize);
	TREEDB_printf("[malloc <] lower_split : "
		" free node %zu, node size %zu"
		, (size_t)TREEDB_pointer_to_offset(pn_free)
		, (size_t)TREEDB_ALLOC_node_size
	);
	TREEDB_NSF(unlink_node)(runtime, TREEDB_ALLOC_free_tree, pn);
	TREEDB_ALLOC_node_set_length(pn, size);
	TREEDB_NSU(insert_node)(runtime, TREEDB_ALLOC_used_tree, pn);
	/* Update free node linked list. */
	TREEDB_ALLOC_node_set_prev(pn_free, packed_pn);
	TREEDB_ALLOC_node_set_next(pn_free, TREEDB_ALLOC_node_next(pn));

	/* Let pn know that m_offset_next is free. */
	TREEDB_ALLOC_node_set_next(pn, TREEDB_packed_node_setfree(packed_pn_free));
	TREEDB_NSF(insert_node)(runtime, TREEDB_ALLOC_free_tree, pn_free);

	/* Update next -> next-> prev with pn_free. */
	if(TREEDB_ALLOC_node_next(pn_free))
		TREEDB_ALLOC_node_set_prev(
			TREEDB_NSA_unpack_node(TREEDB_ALLOC_node_next(pn_free))
			, TREEDB_packed_node_setfree(packed_pn_free)
		);

	/* Let the previous node know that pn is no longer free. */
	if(TREEDB_ALLOC_node_prev(pn))
		TREEDB_ALLOC_node_and_next(
			TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_prev(pn))
			, ~((TREEDB_NS(count_t))1)
		);
}
TREEDB_NS_prefix TREEDB_NSA(node_ptr_t) TREEDB_NS(malloc_upper_split)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NSA(node_ptr_t) pn
, TREEDB_NS_size_t size
, TREEDB_NS(count_t) align
, TREEDB_NS(count_t) offs
) {
	TREEDB_NS(count_t) chunk = TREEDB_ALLOC_node_size + size;
	TREEDB_NS(count_t) freesize = TREEDB_ALLOC_node_length(pn) - chunk;
	TREEDB_NSA(node_ptr_t) pn_new = (TREEDB_NSA(node_ptr_t))
		(V3C_ROUND_DOWN_PTR2(TREEDB_ALLOC_nodeptr_to_addr(pn) + freesize
				+ (TREEDB_ALLOC_node_size + offs), align)
			- (TREEDB_ALLOC_node_size + offs));
	TREEDB_NS(count_t) misalign
		= (TREEDB_ALLOC_nodeptr_to_addr(pn) + freesize) - (char *)pn_new;
	chunk += misalign;
	freesize -= misalign;
	TREEDB_NSA(packed_node_ptr_t) packed_pn_new = TREEDB_NSA_pack_node(pn_new);
	TREEDB_ALLOC_node_set_length(pn_new, chunk - TREEDB_ALLOC_node_size);
	TREEDB_printf("[malloc <] upper_split : free node %zu, node size %zu, "
		"align %zu, offs %zu"
		, (size_t)TREEDB_pointer_to_offset(pn_new) + TREEDB_ALLOC_node_size
		, (size_t)TREEDB_ALLOC_node_size
		, (size_t)align, (size_t)offs
	);
	TREEDB_NSU(insert_node)(runtime, TREEDB_ALLOC_used_tree, pn_new);
	TREEDB_NSF(unlink_node)(runtime, TREEDB_ALLOC_free_tree, pn);
	TREEDB_ALLOC_node_set_length(pn, freesize);
	TREEDB_NSF(insert_node)(runtime, TREEDB_ALLOC_free_tree, pn);

	TREEDB_ALLOC_node_set_prev(pn_new
		, TREEDB_packed_node_setfree(TREEDB_NSA_pack_node(pn)));
	TREEDB_ALLOC_node_set_next(pn_new, TREEDB_ALLOC_node_next(pn));

	TREEDB_ALLOC_node_set_next(pn, packed_pn_new);

	if(TREEDB_ALLOC_node_next(pn_new))
		TREEDB_ALLOC_node_set_prev(
			TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_next(pn_new))
			, packed_pn_new
		);

	return pn_new;
}
/**
The allocator follows these steps to get the memory requested:
 -# find a block of the exact size requested.

 -# if no such block exists, it looks for a larger block that can't be split
	because not enough space remains in the block for another node header.

 -# if no such block exists, use a larger block and split it.

 -# get more memory from the heap.

Notes:
 - every heap has an "offset" alignment that all nodes must be aligned to.

 - the node to data offset is the rounded-up-by-offset size of the allocator
	node.

 - the data pointer returned to the user is at a fixed offset from the node\n
	pointer, so to align the data, we must align the node.

 - preferably when a block is split, the lower part is returned.
	The size requested is rounded up using the offset alignment, so that the
	next node is suitably aligned.

 - sometimes the alignment restrictions imposed by the caller prevent the\n
	lower block from being considered, so the block is split such that the\n
	upper block matches the alignment.\n
	In this case the size may not be rounded up.

 - the free heap allows duplicates, so the "next" node may have the same size\n
	as the current node, and there may be a lot more like it to follow.

 - free nodes can be moved by extending the size used by a preceeding used node.
*/
TREEDB_NS_prefix void * TREEDB_NS(malloc_aligned)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS_size_t size
, TREEDB_NS(count_t) align
, TREEDB_NS(count_t) offs
) {
	TREEDB_printf("malloc_aligned : size %zu, align %zu, offs %zu.\n"
		, (size_t)size, (size_t)align, (size_t)offs);
	if(TREEDB_NS_CONCAT(v3c_nbits_u,)(align) != 1) {
		TREEDB_printf("Bad align %zu.\n", (size_t)align);
		return 0;
	}
	if(offs & ((TREEDB_NS(count_t))sizeof(TREEDB_NS(count_t)) - 1)) {
		TREEDB_printf("Bad offset %zu.\n", (size_t)offs);
		return 0;
	}
	if(TREEDB_DEBUG)
		if(TREEDB_CHECK_HEAP(runtime))
			return 0;

	if(TREEDB_NS_offset_out_of_range(size)) {
		TREEDB_printf("malloc: size too big!\n");
		return 0;
	}
	void * p;
	TREEDB_NSA(node_ptr_t) pn = 0;

	TreedbSearchResult res = TREEDB_NSF(search_ge)(runtime
		, TREEDB_ALLOC_free_tree, size, & pn);
	TREEDB_NS_ssize_t diff;
	TREEDB_NS_size_t aligned_size
		= V3C_ROUND_UP_T2(TREEDB_NS_size_t, size, sizeof(TREEDB_NS(count_t)));
	TREEDB_NS(count_t) offs_offset, aligned_offs_offset;
redo:
	switch(res) {
	case eTreedbSearchGreater:
		offs_offset = TREEDB_pointer_to_offset(pn) + TREEDB_ALLOC_node_size
			+ offs;
		TREEDB_printf("offs_offset %zu, offs_offset & (align-1) %zu.\n"
			, (size_t)offs_offset
			, (size_t)(offs_offset & (align-1))
		);
		if(!(offs_offset & (align-1))) {
			diff = (TREEDB_NS_ssize_t)(TREEDB_ALLOC_node_length(pn))
					- aligned_size;
			/* This node meets the alignment requirement. */
			if(diff >= (TREEDB_NS_ssize_t)TREEDB_ALLOC_node_size) {
				/* We can split the node. */
				TREEDB_printf(" : diff %zu > %zu"
					, (size_t)diff
					, (size_t)TREEDB_ALLOC_node_size
				);
				TREEDB_NS(malloc_lower_split)(runtime, pn, aligned_size);
				p = TREEDB_ALLOC_nodeptr_to_addr(pn);
				TREEDB_printf(" : size %zu, node %zu, "
					"returning %zu.\n"
					, (size_t)TREEDB_ALLOC_node_length(pn)
					, (size_t)TREEDB_pointer_to_offset(pn)
					, (size_t)TREEDB_pointer_to_offset(p)
				);

				if(TREEDB_DEBUG)
					if(TREEDB_CHECK_HEAP(runtime))
						return 0;

				return p;
			}else{
				TREEDB_printf("diff %zu < %zu : using the entire node.\n"
					, (size_t)diff
					, (size_t)TREEDB_ALLOC_node_size
				);
				/* We can't split the node. */
				goto use_aligned_node;
			}
		}
		/* This node doesn't meet the alignment requirement.*/
		/* Work down from the data end to get the new node offset. */
		offs_offset += TREEDB_ALLOC_node_length(pn) - size;
		aligned_offs_offset
			= V3C_ROUND_DOWN_T2(TREEDB_NS(count_t), offs_offset, align);

		diff = aligned_offs_offset - TREEDB_ALLOC_node_size
			- TREEDB_pointer_to_offset(pn)
			- offs
			- TREEDB_ALLOC_node_size; /* for the new node. */

		TREEDB_printf("malloc > : node [%zu:%zu] node size %zu, diff %zd.\n"
			, (size_t)TREEDB_pointer_to_offset(pn)
				+ TREEDB_ALLOC_node_size
			, (size_t)TREEDB_ALLOC_node_length(pn)
			, (size_t)TREEDB_ALLOC_node_size
			, (ssize_t)diff
		);
		if(diff >= 0) {
			/* We can split the node. */
			pn = TREEDB_NS(malloc_upper_split)(runtime, pn, aligned_size
				, align, offs);
			p = TREEDB_ALLOC_nodeptr_to_addr(pn);
			TREEDB_printf(" : size %zu, node %zu, returning %zu.\n"
				, (size_t)TREEDB_ALLOC_node_length(pn)
				, (size_t)TREEDB_pointer_to_offset(pn)
				, (size_t)TREEDB_pointer_to_offset(p));

			if(TREEDB_DEBUG)
				if(TREEDB_CHECK_HEAP(runtime))
					return 0;

			return p;
		}
		/* Fall through. */
	case eTreedbSearchEqual:
		offs_offset = TREEDB_pointer_to_offset(pn)
			+ TREEDB_ALLOC_node_size + offs;
		aligned_offs_offset
			= V3C_ROUND_UP_T2(TREEDB_NS(count_t), offs_offset, align);
		TREEDB_printf("[malloc = : offs_offset %zu, "
			"aligned_offs_offset %zu.\n"
			, (size_t)offs_offset
			, (size_t)aligned_offs_offset
		);
		if(aligned_offs_offset == offs_offset) {
use_aligned_node:
			TREEDB_printf("[malloc = : using aligned node.\n");
			TREEDB_NSF(unlink_node)(runtime, TREEDB_ALLOC_free_tree, pn);
			TREEDB_NSU(insert_node)(runtime, TREEDB_ALLOC_used_tree, pn);

			if(TREEDB_ALLOC_node_next(pn))
				TREEDB_ALLOC_node_and_prev(
					TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_next(pn))
					, (~(TREEDB_NS(count_t))1)
				);
			if(TREEDB_ALLOC_node_prev(pn))
				TREEDB_ALLOC_node_and_next(
					TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_prev(pn))
					, (~(TREEDB_NS(count_t))1)
				);
			p = TREEDB_ALLOC_nodeptr_to_addr(pn);
			TREEDB_printf("[malloc =] size %u, node %zu, returning %zu\n"
				, (unsigned int)size
				, (size_t)TREEDB_pointer_to_offset(pn)
				, (size_t)TREEDB_pointer_to_offset(p)
			);
			if(TREEDB_DEBUG)
				if(TREEDB_CHECK_HEAP(runtime))
					return 0;

			return p;
		}
		if(TREEDB_ALLOC_node_prev(pn)) {
			/* We may be able to move this node. */
			diff = (TREEDB_NS_ssize_t)(TREEDB_ALLOC_node_length(pn))
					- aligned_size;
			TREEDB_NS(count_t) odiff
				= aligned_offs_offset - offs_offset;
			if(diff >= (TREEDB_NS_ssize_t)odiff) {
				TREEDB_printf("[malloc =] size %u, node %zu(%zu) moving node to"
					" %zu.\n"
					, (unsigned int)size
					, (size_t)TREEDB_pointer_to_offset(pn)
						+ TREEDB_ALLOC_node_size
					, (size_t)TREEDB_ALLOC_node_length(pn)
					, (size_t)TREEDB_pointer_to_offset(pn)
						+ TREEDB_ALLOC_node_size
						+ odiff
				);
				/* We can move this node. Remember, free nodes can't be
				adjacent. */
				TREEDB_NSF(unlink_node)(runtime, TREEDB_ALLOC_free_tree, pn);
				TREEDB_ALLOC_node_length_sub(pn, odiff);
				memmove((char *)pn + odiff, pn, TREEDB_ALLOC_node_size);
				pn = (TREEDB_NSA(node_ptr_t))((char *)pn + odiff);
				TREEDB_NSA(packed_node_ptr_t) packed_pn
					= TREEDB_NSA_pack_node(pn);
				TREEDB_NSA(node_ptr_t) pn2
					= TREEDB_NSA_unpack_node(TREEDB_ALLOC_node_next(pn));
				if(pn2)
					TREEDB_ALLOC_node_set_prev(pn2, packed_pn);
				pn2 = TREEDB_NSA_unpack_node(TREEDB_ALLOC_node_prev(pn));
				TREEDB_ALLOC_node_set_next(pn2, packed_pn);
				TREEDB_ALLOC_node_length_add(pn2, odiff);
				TREEDB_NSU(insert_node)(runtime, TREEDB_ALLOC_used_tree, pn);

				p = TREEDB_ALLOC_nodeptr_to_addr(pn);
				TREEDB_printf("[malloc =] size %u, node %zu returning %zu\n"
					, (unsigned int)size
					, (size_t)TREEDB_pointer_to_offset(pn)
					, (size_t)TREEDB_pointer_to_offset(p));

				if(TREEDB_DEBUG)
					if(TREEDB_CHECK_HEAP(runtime))
						return 0;

				return p;
			}
		}
		/* This node didn't meet our alignment constraint. Try the next one. */
		/* The free list is sorted by size, so next's length is >= */
		TREEDB_printf("[malloc =] size %u, node %zu data offs %zu "
			"failed alignment %u, offset %u, will try another node.\n"
			, (unsigned int)size
			, (size_t)TREEDB_pointer_to_offset(pn)
				+ TREEDB_ALLOC_node_size
			, (size_t)TREEDB_pointer_to_offset(pn)
				+ TREEDB_ALLOC_node_size
				+ offs
			, (unsigned int)align, (unsigned int)offs
		);
		res = eTreedbSearchGreater;
		for(;;) {
			pn = TREEDB_node_next(pn);
			if(!pn)
				break;
			if(TREEDB_ALLOC_node_length(pn) >= size)
				goto redo;
		}
		/* Fall through */
	case eTreedbSearchEmpty:
	case eTreedbSearchLess: {
		if(pn) {
			TREEDB_printf("malloc > : got %zu, need %zu.\n"
				, (size_t)TREEDB_ALLOC_node_length(pn)
				, (size_t)size
			);
		}else{
			TREEDB_printf("malloc > : need %zu.\n", (size_t)size);
		}
		TREEDB_NS(count_t) page_size;
		if(sizeof(TREEDB_NS(count_t)) > 1) {
			/* Get the memory from the heap, rounding up to the next page. */
			page_size = V3C_ROUND_UP_T2(TREEDB_NS(count_t), size
				+ TREEDB_ALLOC_node_size, TREEDB_PAGE_SIZE);
			pn = (TREEDB_NSA(node_ptr_t))TREEDB_page_alloc(page_size);
			if(!pn) {
				TREEDB_printf("malloc : page_alloc failed.\n");
				return 0;
			}
		}else
			return 0;
		TREEDB_printf("malloc : page_alloc returned %zu.\n"
			, (size_t)TREEDB_pointer_to_offset(pn));
		int plast_is_free = 0;
		// TREEDB_ALLOC_free_tail isn't (necessarily) the last node in the heap,
		// it's the largest node.
		TREEDB_NSA(node_ptr_t) plast = TREEDB_ALLOC_used_tail;
		if(plast) {
			if(TREEDB_ALLOC_node_next(plast)) {
				plast
					= TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_next(plast));
				plast_is_free = 1;
			}
			if(TREEDB_ALLOC_node_next(plast)) {
				TREEDB_printf("malloc: two adjacent free nodes!\n");
				TREEDB_ALLOC_set_errno(eTreedbTwoAdjacentFreeNodes);
				return 0;
			}
			TREEDB_printf("[%zu]plast->m_uLength[%zu]\n"
				, (size_t)TREEDB_pointer_to_offset(plast)
				, (size_t)(TREEDB_ALLOC_node_length(plast)));
		}else{
			/* No used nodes means that there is but one free node. */
			plast = TREEDB_ALLOC_free_head;
			if(!plast) {
				TREEDB_printf("malloc: no used or free nodes!\n");
				TREEDB_ALLOC_set_errno(eTreedbNoNodes);
				return 0;
			}
			plast_is_free = 1;
		}
		char * plast_data = TREEDB_ALLOC_nodeptr_to_addr(plast);

		TREEDB_printf("malloc : plast_is_free : %d, last-data(%zu) "
			"+ last-len(%zu) == %zu, pn == %zu\n"
			, plast_is_free
			, (size_t)TREEDB_pointer_to_offset(plast_data)
			, (size_t)TREEDB_ALLOC_node_length(plast)
			, (size_t)(
				TREEDB_pointer_to_offset(plast_data)
				+ TREEDB_ALLOC_node_length(plast))
			, (size_t)TREEDB_pointer_to_offset(pn)
		);

		if(plast_is_free
				&& ((plast_data + TREEDB_ALLOC_node_length(plast))
					== (char *)pn)) {
			TREEDB_printf("malloc: Extend the last free node and do over.\n");
			TREEDB_NSF(unlink_node)(runtime, TREEDB_ALLOC_free_tree, plast);
			TREEDB_ALLOC_node_length_add(plast, page_size);
			TREEDB_NSF(insert_node)(runtime, TREEDB_ALLOC_free_tree, plast);
			res = eTreedbSearchGreater;
			pn = plast;
			goto redo;
		}
		TREEDB_printf("malloc: (plast=%zu) Create a new free node with data "
			"size %zu and do over.\n"
			, (size_t)TREEDB_pointer_to_offset(plast)
			, (size_t)(page_size - TREEDB_ALLOC_node_size)
		);

		TREEDB_printf("malloc: pn = %zu, plast = %zu\n"
			, (size_t)TREEDB_pointer_to_offset(pn)
			, (size_t)TREEDB_pointer_to_offset(plast)
			);
		TREEDB_NS(dump_allocator_trees)(runtime);

		TREEDB_ALLOC_node_length(pn) = page_size - TREEDB_ALLOC_node_size;
		TREEDB_ALLOC_node_set_prev(pn, TREEDB_NSA_pack_node(plast));
		TREEDB_ALLOC_node_set_next(pn, TREEDB_NSA_pack_node(0));
		TREEDB_ALLOC_node_set_next(plast
			, TREEDB_packed_node_setfree(TREEDB_NSA_pack_node(pn)));

		TREEDB_NSF(insert_node)(runtime, TREEDB_ALLOC_free_tree, pn);
		res = eTreedbSearchGreater;
		goto redo;
		break; }
	}
	return 0;
}
TREEDB_NS_prefix void * TREEDB_NS(malloc)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS_size_t size
) {
	return TREEDB_NS(malloc_aligned)(runtime, size, TREEDB_ALIGN, 0);
}
TREEDB_NS_prefix void * TREEDB_NS(calloc_aligned)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS_size_t count
, TREEDB_NS_size_t element_size
, TREEDB_NS(count_t) align
, TREEDB_NS(count_t) offs
) {
	TREEDB_NS_size_t n = 0, c = count * element_size;
	char * p, * ret = (char *)TREEDB_NS(malloc_aligned)(runtime, c, align
		, offs);
	if(!ret)
		return 0;
	for(p = ret, n = 0; n < c; ++n)
		*p++ = '\0';
	return ret;
}
TREEDB_NS_prefix void * TREEDB_NS(calloc)
( TREEDB_NS_runtime_ptr_t runtime
, TREEDB_NS_size_t count
, TREEDB_NS_size_t element_size
) {
	return TREEDB_NS(calloc_aligned)(runtime, count, element_size
		, TREEDB_ALIGN, 0);
}
TREEDB_NS_prefix void * TREEDB_NS(realloc_aligned)
( TREEDB_NS_runtime_ptr_t runtime
, void * p
, TREEDB_NS_size_t size
, TREEDB_NS(count_t) align
, TREEDB_NS(count_t) offs
) {
	if(!p)
		return TREEDB_NS(malloc_aligned)(runtime, size, align, offs);
	if(TREEDB_NS_CONCAT(v3c_nbits_u,)(align) != 1) {
		TREEDB_printf("Bad align %zu.\n", (size_t)align);
		return 0;
	}
	if(offs & ((TREEDB_NS(count_t))sizeof(TREEDB_NS(count_t)) - 1)) {
		TREEDB_printf("Bad offset %zu.\n", (size_t)offs);
		return 0;
	}
	if(p < (void *)(uintptr_t)TREEDB_ALLOC_node_size) {
		TREEDB_printf("realloc_aligned called with null-ish item!\n");
		TREEDB_ALLOC_set_errno(eTreedbInvalidParam);
		return 0;
	}
	TREEDB_NSA(node_ptr_t) pn = TREEDB_ALLOC_addr_to_nodeptr(p);
	TREEDB_NS(count_t) node_offset = TREEDB_pack_pointer(pn);
	if(!TREEDB_NSU(search)(runtime, TREEDB_ALLOC_used_tree, node_offset)) {
		TREEDB_printf("realloc called with item [%zu] not allocated from this "
			"heap.\n", (size_t)(node_offset + TREEDB_ALLOC_node_size));
		TREEDB_ALLOC_set_errno(eTreedbInvalidParam);
		return 0;
	}
	TREEDB_printf("[realloc_aligned] %zu %zu -> %zu\n"
		, (size_t)TREEDB_pointer_to_offset(pn)
			+ TREEDB_ALLOC_node_size
		, (size_t)(TREEDB_ALLOC_node_length(pn)), size);
	if(size > (TREEDB_NS_size_t)(~((TREEDB_NS(count_t))0))) {
		TREEDB_printf("realloc_aligned called with too large a value for this "
			"heap!\n");
		TREEDB_ALLOC_set_errno(eTreedbInvalidParam);
		return 0;
	}
	TREEDB_NS(count_t) offs_offset = TREEDB_pointer_to_offset(p);
	TREEDB_NS(count_t) misalign = (offs_offset - offs) & (align - 1);
	TREEDB_NS(count_t) osize = TREEDB_ALLOC_node_length(pn);
	if((size == osize) && !misalign)
		return p;
	if(size < osize) {
		if((TREEDB_ALLOC_node_length(pn) - size)
				< (TREEDB_ALLOC_node_size)) {
			if(!misalign)
				return p; /* The change is too small to consider. */
		}
	}
	// TODO for shrinkage, try to make a free node out of the freed space or
	// join the freed space with the next free node if the next node is free.
	// If this isn't possible then the allocator should leave the node as is.

	// TODO for enlargement, see if there's a free node after this one we can
	// use.

	/* Need to malloc a new node and copy. */
	/* The address range may change so save the offset. */
	TREEDB_NS(count_t) pn_offs = TREEDB_pack_pointer(pn);
	void * p2
		= (TREEDB_NSA(node_ptr_t))TREEDB_NS(malloc_aligned)
			(runtime, size, align, offs);
	if(!p2) {
		if(size < osize)
			return p; // realloc(smaller) cannot fail, but this wastes space.
		return 0;
	}
	TREEDB_NSA(node_ptr_t) pn2 = TREEDB_ALLOC_addr_to_nodeptr(p2);
	(void)pn2;
	/* Restore the pointer from the offset. */
	pn = (TREEDB_NSA(node_ptr_t))TREEDB_unpack_pointer(pn_offs);
	p = TREEDB_ALLOC_nodeptr_to_addr(pn);
	if(osize > size)
		osize = size;
	memcpy(p2, p, osize);
	TREEDB_NS(free)(runtime, p);
	return p2;
}
TREEDB_NS_prefix void * TREEDB_NS(realloc)
( TREEDB_NS_runtime_ptr_t runtime
, void * p
, TREEDB_NS_size_t size
) {
	return TREEDB_NS(realloc_aligned)(runtime, p, size, TREEDB_ALIGN, 0);
}
TREEDB_NS_prefix int TREEDB_NS(in_heap)
( TREEDB_NS_runtime_ptr_t runtime
, const void * p
) {
	return ((p > (void *)TREEDB_STORAGE_base)
			&& (p < (void *)(TREEDB_STORAGE_base + TREEDB_STORAGE_size)))
		? 1 : 0;
}
TREEDB_NS_prefix int TREEDB_NS(did_alloc)
( TREEDB_NS_runtime_ptr_t runtime
, const void * p
) {
	TREEDB_NSA(node_ptr_t) pn = TREEDB_ALLOC_addr_to_nodeptr(p);
	TREEDB_NS(count_t) offs = TREEDB_pack_pointer(pn);
	return TREEDB_NSU(search)
			(runtime, TREEDB_ALLOC_used_tree, offs)
		? 1 : 0;
}
TREEDB_NS_prefix size_t TREEDB_NS(usable_size)
( TREEDB_NS_runtime_ptr_t runtime
, const void * p
) {
	TREEDB_NSA(node_ptr_t) pn = TREEDB_ALLOC_addr_to_nodeptr(p);
	TREEDB_NS(count_t) offs = TREEDB_pack_pointer(pn);
	if(!TREEDB_NSU(search)(runtime, TREEDB_ALLOC_used_tree, offs))
		return 0;
	return TREEDB_ALLOC_node_length(pn);
}
TREEDB_NS_prefix TreedbFreeResult TREEDB_NS(free)
( TREEDB_NS_runtime_ptr_t runtime
, void * p
) {
	if(TREEDB_DEBUG)
		if(TREEDB_CHECK_HEAP(runtime))
			return eTreedbFreeBadHeap;

	if(p < (void *)(uintptr_t)TREEDB_ALLOC_node_size) {
		if(!p) {
			TREEDB_printf("free called with null item.\n");
			return eTreedbFreeUnknownPointer;
		}
		TREEDB_printf("free called with null-ish item.\n");
		TREEDB_ALLOC_set_errno(eTreedbInvalidParam);
		return eTreedbFreeUnknownPointer;
	}
	TREEDB_NSA(node_ptr_t) pn = TREEDB_ALLOC_addr_to_nodeptr(p);
	TREEDB_printf("[free] %zu\n"
		, (size_t)TREEDB_pointer_to_offset(pn)
			+ TREEDB_ALLOC_node_size
	);
	TREEDB_NS(count_t) offs = TREEDB_pack_pointer(pn);
	if(TREEDB_NSU(search_closest)
				(runtime, TREEDB_ALLOC_used_tree, offs, & pn)
			!= eTreedbSearchEqual) {
		TREEDB_printf("free called with item not allocated from this heap.\n");
		TREEDB_ALLOC_set_errno(eTreedbInvalidParam);
		return eTreedbFreeUnknownPointer;
	}
	TREEDB_NSU(unlink_node)(runtime, TREEDB_ALLOC_used_tree, pn);
	/* If next is free, join up. */
	TREEDB_NSA(node_ptr_t) pn2
		= TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_next(pn));
	if(TREEDB_packed_node_isfree(TREEDB_ALLOC_node_next(pn))) {
		TREEDB_printf("Joining with next [%zu:%zu]\n"
			, (size_t)TREEDB_pointer_to_offset(pn2)
				+ TREEDB_ALLOC_node_size
			, (size_t)TREEDB_ALLOC_node_length(pn2)
		);
		TREEDB_NSF(unlink_node)(runtime, TREEDB_ALLOC_free_tree, pn2);
		TREEDB_ALLOC_node_length_add(pn,
			TREEDB_ALLOC_node_size + TREEDB_ALLOC_node_length(pn2));
		TREEDB_ALLOC_node_set_next(pn, TREEDB_ALLOC_node_next(pn2));
	}
	/* If prev is free, join up. */
	if(TREEDB_packed_node_isfree(TREEDB_ALLOC_node_prev(pn))) {
		TREEDB_printf("Joining with prev %zu node size %zu, total"
			" node size %zu.\n"
			, (size_t)TREEDB_pointer_to_offset(
				TREEDB_NSA_unpack_node(
					TREEDB_packed_item_offset(TREEDB_ALLOC_node_prev(pn))))
				+ TREEDB_ALLOC_node_size
			, (size_t)TREEDB_ALLOC_node_length(pn)
			, (size_t)(TREEDB_ALLOC_node_size +
				TREEDB_ALLOC_node_length(pn))
		);
		TREEDB_NSA(node_ptr_t) pn2
			= TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_prev(pn));
		TREEDB_NSF(unlink_node)(runtime, TREEDB_ALLOC_free_tree, pn2);
		TREEDB_ALLOC_node_length_add(pn2,
			TREEDB_ALLOC_node_size + TREEDB_ALLOC_node_length(pn));
		TREEDB_ALLOC_node_set_next(pn2, TREEDB_ALLOC_node_next(pn));
		pn = pn2;
	}
	/* Let the previous node know that pn is free. */
	if(TREEDB_ALLOC_node_prev(pn))
		TREEDB_ALLOC_node_set_next(
			TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_prev(pn))
			, TREEDB_packed_node_setfree(TREEDB_NSA_pack_node(pn))
		);
	/* Let the next node know that pn is free. */
	if(TREEDB_ALLOC_node_next(pn))
		TREEDB_ALLOC_node_set_prev(
			TREEDB_NS_unpack_free_node(TREEDB_ALLOC_node_next(pn))
			, TREEDB_packed_node_setfree(TREEDB_NSA_pack_node(pn))
		);
	TREEDB_NSF(insert_node)(runtime, TREEDB_ALLOC_free_tree, pn);

	TREEDB_post_free(TREEDB_ALLOC_nodeptr_to_addr(pn)
		, TREEDB_ALLOC_node_length(pn));

	TreedbFreeResult res = eTreedbFreeSuccess;
#ifdef TREEDB_HAS_RESIZABLE_HEAP
	do {
		size_t prev_size = TREEDB_STORAGE_size;
		if(prev_size == TREEDB_PAGE_SIZE)
			break; /* Keep the first page. */

		char * p = TREEDB_ALLOC_nodeptr_to_addr(pn);
		size_t size = TREEDB_ALLOC_node_length(pn);
		if((p + size) < (TREEDB_STORAGE_base + prev_size))
			break; /* Not at the end. */

		TREEDB_NS(count_t) page_leader = TREEDB_PAGE_SIZE
			- ((uintptr_t)(p - TREEDB_STORAGE_base)
				& (TREEDB_PAGE_SIZE - 1));
		TREEDB_NS(count_t) page_count = (size - page_leader) / TREEDB_PAGE_SIZE;
		if(!page_count)
			break; /* No pages spanned. */
		/* Store the offset as the heap may be remapped. */
		TREEDB_NS(count_t) offs = TREEDB_pack_pointer(pn);

		res = TREEDB_page_free(prev_size - page_count * TREEDB_PAGE_SIZE);

		/* Restore the pointer from the offset. */
		pn = (TREEDB_NSA(node_ptr_t))TREEDB_unpack_pointer(offs);
		TREEDB_NSF(unlink_node)(runtime, TREEDB_ALLOC_free_tree, pn);
		TREEDB_ALLOC_node_length_sub(pn, page_count * TREEDB_PAGE_SIZE);
		TREEDB_NSF(insert_node)(runtime, TREEDB_ALLOC_free_tree, pn);
	} while(0);
#endif
	if(TREEDB_DEBUG)
		if(TREEDB_CHECK_HEAP(runtime))
			return eTreedbFreeBadHeap;

	return res;
}
#endif /* #ifndef TREEDB_OMIT_treedb_fns_DECLARE */

#ifndef TREEDB_OMIT_post_free_impl
TREEDB_NS_prefix void TREEDB_NS(impl_post_free)
( TREEDB_NS_runtime_ptr_t runtime
, void * p
, TREEDB_NS_size_t size
) {
	(void)runtime;
	(void)p;
	(void)size;
}
#endif

#ifndef TREEDB_OMIT_avl_allocator_impl

#ifdef DOXYGEN_processing
{
#endif

  #include "avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#endif
#undef TREEDB_NS_size_t
#undef TREEDB_NS_prefix
#undef TREEDB_NS_offset_out_of_range

#ifdef __cplusplus
} // extern "C" {
#endif
