/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdlib.h>
#include <string.h>
#include "treedb.h"

#include <sys/stat.h>

#ifndef v3c_prefix
#define v3c_prefix
#endif /* #ifndef v3c_prefix */

#ifdef __cplusplus
extern "C" {
#endif /* #ifdef __cplusplus */

#ifndef TREEDB_OMIT_v3c_c_strerror
v3c_prefix char * v3c_c_strerror(int errnum)
{
	char * msg = (char *)malloc(100);
	if(!msg)
		return 0; /* "<no memory for error message>"; */
  #if _GNU_SOURCE > 0
	//char * version
	char * ret = strerror_r(errnum, msg, 100);
  #else
	// int version
	const char * ret = msg;
	if(strerror_r(errnum, msg, N))
		ret = 0;
  #endif
	return ret;
}
#endif

#ifndef TREEDB_OMIT_TreedbSearchResultDesc
v3c_prefix const char * TreedbSearchResultDesc(TreedbSearchResult res)
{
	switch(res) {
	case eTreedbSearchEmpty:
		return "\"<>\"";
	case eTreedbSearchLess:
		return "\"<\"";
	case eTreedbSearchEqual:
		return "\"=\"";
	case eTreedbSearchGreater:
		return "\">\"";
	default:
		return "<?>";
	}
}
#endif
#ifndef TREEDB_OMIT_TreedbInsertResultDesc
v3c_prefix const char * TreedbInsertResultDesc(TreedbInsertResult res)
{
	switch(res) {
	case eTreedbInsertSuccess:
		return "Success";
	case eTreedbInsertExists:
		return "Exists";
	case eTreedbInsertNoMem:
		return "NoMem";
	default:
		return "<?>";
	}
}
#endif
#ifndef TREEDB_OMIT_TreedbInitResultDesc
v3c_prefix const char * TreedbInitResultDesc(TreedbInitResult res)
{
	switch(res) {
	case eTreedbInitSuccess:
		return "Success";
	case eTreedbInitOffsetSizeTooSmall:
		return "Offset size too small";
	case eTreedbInitOffsetSizeTooBig:
		return "Offset size too big";
	case eTreedbInitOffsetSizeOdd:
		return "Offset size odd (and not 1)";
	case eTreedbInitPayloadOffsetTooSmall:
		return "Payload offset too small";
	case eTreedbInitPayloadOffsetTooBig:
		return "Payload offset too big";
	case eTreedbInitHeaderSizeTooSmall:
		return "Header size too small";
	default:
		return "<?>";
	}
}
#endif
#ifndef TREEDB_OMIT_TreedbHeapStatusDesc
v3c_prefix const char * TreedbHeapStatusDesc(TreedbHeapStatus res)
{
	switch(res) {
	case eTreedbHeapSuccess:
		return "Success";
	case eTreedbHeapFileOpenFailure:
		return "File open failure";
	case eTreedbHeapBadTop:
		return "Bad top";
	case eTreedbHeapOpenBadMagic:
		return "Bad magic bytes";
	case eTreedbHeapOpenUnknownAllocatorAlgorithm:
		return "Unknown allocator algorithm";
	case eTreedbHeapOpenUnknownSchema:
		return "Unknown schema";
	case eTreedbHeapOpenUnhandledOffsetSize:
		return "Unhandled offset size";
	case eTreedbHeapOpenUnhandledEndianIndex:
		return "Unhandled endian index";
	case eTreedbHeapOpenUnhandledFeatures:
		return "Unhandled features";
	case eTreedbHeapCreateInitFailure:
		return "Initialisation failure";
	case eTreedbHeapAlreadyWriteLocked:
		return "Heap already write locked";
	case eTreedbHeapAlreadyReadLocked:
		return "Heap already read locked";
	case eTreedbHeapAlreadyUnlocked:
		return "Heap already unlocked";
	case eTreedbHeapPayloadSizeTooLarge:
		return "Payload size too large";
	case eTreedbHeapPointerAllocator:
		return "Cannot use a pointer allocator in a persistent treedb";
	case eTreedbHeapNotOpen:
		return "Heap not open";
	case eTreedbHeapReadonly:
		return "Read only heap";
	case eTreedbHeapMutexLockError:
		return "Mutex lock error";
	case eTreedbHeapCondWaitError:
		return "Condition wait errror";
	case eTreedbHeapCondBroadcastError:
		return "Condition broadcast error";
	case eTreedbHeapFileLockError:
		return "File lock error";
	case eTreedbHeapMmapError:
		return "Mmap error";
	case eTreedbHeapUnMmapError:
		return "Munmap error";
	case eTreedbHeapMutexUnlockError:
		return "Mutex unlock error";
	case eTreedbHeapLockedError:
		return "Heap is locked";
	case eTreedbHeapFileTruncateFailure:
		return "File truncate failure";
	case eTreedbHeapBad:
		return "Heap bad";
	case eTreedbHeapMallocFailure:
		return "Malloc failure";
	case eTreedbHeapNoRuntime:
		return "No runtime";
	default:
		return "<?>";
	}
}
#endif

#ifndef TREEDB_OMIT_TreedbTreedbErrorDesc
v3c_prefix const char * TreedbErrorDesc(TreedbError res)
{
	switch(res) {
	case eTreedbOk:
		return "Treedb OK";
	case eTreedbZeroLengthHeap:
		return "Zero length heap";
	case eTreedbNullNode:
		return "Null node";
	case eTreedbPrevNodeIsFree:
		return "Previous node is free";
	case eTreedbPrevNodeIsThis:
		return "Previous node is this";
	case eTreedbPrevNodeIsBad:
		return "Previous node is bad";
	case eTreedbNodeNotInFreeHeap:
		return "Node not in free heap";
	case eTreedbNodeNotInUsedHeap:
		return "Node not in used heap";
	case eTreedbNodeLengthBad:
		return "Node length bad";
	case eTreedbUnexpectedFreeNode:
		return "Unexpected free node";
	case eTreedbNextNodeIsThis:
		return "Next node is this";
	case eTreedbNextNodeIsLess:
		return "Next node is less";
	case eTreedbLastNodePastEndOfHeap:
		return "Last node past end of heap";
	case eTreedbInvalidUsedHeap:
		return "Invalid used heap";
	case eTreedbTwoAdjacentFreeNodes:
		return "Two adjacent free nodes";
	case eTreedbNoNodes:
		return "No nodes";
	case eTreedbInvalidParam:
		return "Invalid parameter";
	default:
		return "<?>";
	}
}
#endif


#ifndef TREEDB_OMIT_TreedbFreeResultDesc
v3c_prefix const char * TreedbFreeResultDesc(TreedbFreeResult res)
{
	switch(res) {
	case eTreedbFreeSuccess:
		return "Success";
	case eTreedbFreeUnknownPointer:
		return "Unknown pointer";
	case eTreedbFreeBadHeap:
		return "Bad heap";
	case eTreedbFreeBadHeapLock:
		return "Heap lock failed.";
	case eTreedbFreeBadHeapFree:
		return "Heap page free failed.";
	case eTreedbFreeHeapIsFixed:
		return "Heap cannot be resized.";
	default:
		return "<?>";
	}
}
#endif
#ifndef TREEDB_OMIT_TreedbKeyDistanceFlagsDesc
v3c_prefix const char * TreedbKeyDistanceFlagsDesc(TreedbKeyDistanceFlags flags)
{
	flags.flags &= (eKeyDistanceIncludeLower | eKeyDistanceIncludeUpper
		| eKeyDistanceExactLower | eKeyDistanceExactUpper);
	switch(flags.flags) {
	case  0: /*0000*/
	default:
		return "";
	case  1: /*0001*/
		return "include_lower";
	case  2: /*0010*/
		return "include_upper";
	case  4: /*0100*/
		return "exact_lower";
	case  8: /*1000*/
		return "exact_upper";
	case  3: /*0011*/
		return "include_lower | include_upper";
	case  5: /*0101*/
		return "include_lower | exact_lower";
	case  9: /*1001*/
		return "include_lower | exact_upper";
	case  6: /*0110*/
		return "include_upper | exact_lower";
	case 10: /*1010*/
		return "include_upper | exact_upper";
	case 12: /*1100*/
		return "exact_lower | exact_upper";
	case  7: /*0111*/
		return "include_lower | include_upper | exact_lower";
	case 11: /*1011*/
		return "include_lower | include_upper | exact_upper";
	case 13: /*1101*/
		return "include_lower | exact_lower | exact_upper";
	case 14: /*1110*/
		return "include_upper | exact_lower | exact_upper";
	case 15: /*1111*/
		return "include_lower | include_upper | exact_lower | exact_upper";
	}
}
#endif
#ifndef TREEDB_OMIT_TreedbNodeDistanceFlagsDesc
v3c_prefix const char * TreedbNodeDistanceFlagsDesc(TreedbNodeDistanceFlags flags)
{
	flags.flags &= (eNodeDistanceIncludeLower | eNodeDistanceIncludeUpper);
	switch(flags.flags) {
	case 0:
		return "";
	case eNodeDistanceIncludeLower:
		return "include_lower";
	case eNodeDistanceIncludeUpper:
		return "include_upper";
	default:
		return "include_lower | include_upper";
	}
}
#endif
#ifndef TREEDB_OMIT_TreedbHeapMismatched
v3c_prefix int TreedbHeapMismatched(TreedbHeapStatus res)
{
	if((res >= eTreedbHeapBadTop) && (res <= eTreedbHeapOpenUnhandledFeatures))
		return 1;
	return 0;
}
#endif

#ifndef TREEDB_OMIT_TreedbAllocatorAbort
v3c_prefix void v3c_treedb_allocator_abort(TreedbError e)
{
	fprintf(stderr, "Allocator Error: %s : aborting.\n", TreedbErrorDesc(e));
	abort();
}
#endif

v3c_prefix uint8_t v3c_native_endian_index()
{
	uint8_t res = 0;
	{
		union {
			uint16_t v;
			uint8_t a[2];
		} u;
		u.a[0] = 0;
		u.a[1] = 1;
		if(u.v == 1)
			res |= 1; // Big-endian words.
	}
	{
		union {
			uint32_t v;
			uint16_t a[2];
		} u;
		u.a[0] = 0;
		u.a[1] = 1;
		if(u.v == 1)
			res |= 2; // Big-endian double-words.
	}
	{
		union {
			uint64_t v;
			uint32_t a[2];
		} u;
		u.a[0] = 0;
		u.a[1] = 1;
		if(u.v == 1)
			res |= 4; // Big-endian quad-words.
	}
	return res;
}

v3c_prefix int treedb_mode_flags
( int flags_in
) {
	int flags = 0;
	if(flags_in & eHeapCreateOwnerRead)
		flags |= S_IRUSR;
	if(flags_in & eHeapCreateOwnerWrite)
		flags |= S_IWUSR;
	if(flags_in & eHeapCreateOwnerExec)
		flags |= S_IXUSR;
	if(flags_in & eHeapCreateGroupRead)
		flags |= S_IRGRP;
	if(flags_in & eHeapCreateGroupWrite)
		flags |= S_IWGRP;
	if(flags_in & eHeapCreateGroupExec)
		flags |= S_IXGRP;
	if(flags_in & eHeapCreateOtherRead)
		flags |= S_IROTH;
	if(flags_in & eHeapCreateOtherWrite)
		flags |= S_IWOTH;
	if(flags_in & eHeapCreateOtherExec)
		flags |= S_IXOTH;
	return flags;
}

#ifdef __cplusplus
} // extern "C" {
#endif /* #ifdef __cplusplus */
