/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_treedb_h_included
  #define v3c_treedb_h_included

  #ifndef v3c_declare_prefix
	#define v3c_declare_prefix extern
  #endif

  #ifndef TREEDB_DEBUG
	#define TREEDB_DEBUG 0
  #endif

  #ifndef __cplusplus
	#include <stdio.h>
  #else
	#include <cstdio>
	#include <iosfwd>
  #endif

  #ifndef TREEDB_printf_DEFINED
    #if (TREEDB_DEBUG & 1)
      #define TREEDB_printf(fmt, ...) \
        fprintf(stderr, fmt, ## __VA_ARGS__); fflush(stderr)
      #define TREEDB_cout cerr
    #else
      #define TREEDB_printf(fmt, ...)
      #define TREEDB_cout if(0) cerr
    #endif /* #if (TREEDB_DEBUG & 1) */
    #define TREEDB_printf_DEFINED
  #endif /* #ifndef TREEDB_printf_DEFINED */

  #ifndef TREEDB_printf2_DEFINED
    #if (TREEDB_DEBUG & 2)
      #define TREEDB_printf2(fmt, ...) \
        fprintf(stderr, fmt, ## __VA_ARGS__); fflush(stderr)
      #define TREEDB_cout2 cout
    #else
      #define TREEDB_printf2(fmt, ...)
      #define TREEDB_cout2 if(0) cerr
    #endif /* #if (TREEDB_DEBUG & 2) */
    #define TREEDB_printf2_DEFINED
  #endif /* #ifndef TREEDB_printf2_DEFINED */

  #ifndef TREEDB_printf4_DEFINED
    #if (TREEDB_DEBUG & 4)
      #define TREEDB_printf4(fmt, ...) \
        fprintf(stderr, fmt, ## __VA_ARGS__); fflush(stderr)
      #define TREEDB_cout4 cout
    #else
      #define TREEDB_printf4(fmt, ...)
      #define TREEDB_cout4 if(0) cerr
    #endif /* #if (TREEDB_DEBUG & 4) */
    #define TREEDB_printf4_DEFINED
  #endif /* #ifndef TREEDB_printf4_DEFINED */

  #ifndef TREEDB_printf8_DEFINED
    #if (TREEDB_DEBUG & 8)
      #define TREEDB_printf8(fmt, ...) \
        fprintf(stderr, fmt, ## __VA_ARGS__); fflush(stderr)
      #define TREEDB_cout8 cout
    #else
      #define TREEDB_printf8(fmt, ...)
      #define TREEDB_cout8 if(0) cerr
    #endif /* #if (TREEDB_DEBUG & 8) */
    #define TREEDB_printf8_DEFINED
  #endif /* #ifndef TREEDB_printf8_DEFINED */

  #ifndef TREEDB_MSYNC
	/*#define TREEDB_MSYNC MS_SYNC*/
	#define TREEDB_MSYNC MS_ASYNC
  #endif

  #include <v3c/utils.h>
  #include <v3c/comdefs.h>
  #include <treedb/site-config.h>

  #define TREEDB_ALIGNED_POINTER(p) \
	V3C_ROUND_UP_PTR2(p, TREEDB_ALIGN)

  #define TREEDB_NS_ALIGNED_OFFSET(u) \
	V3C_ROUND_UP_T2(TREEDB_NS(count_t), u, sizeof(TREEDB_NS(count_t)))

  #define TREEDB_NS_CONCAT(a,b) V3C_INVOKE3(V3C_JOIN3,a,TREEDB_NBITS,b)

  #include <uuid.h>

  #include "treedb-header.h"

  #ifdef __cplusplus
extern "C" {
  #endif

  #ifndef TreedbSearchResult_DEFINED
typedef enum TreedbSearchResult
{ eTreedbSearchEmpty
, eTreedbSearchLess
, eTreedbSearchEqual
, eTreedbSearchGreater
} TreedbSearchResult;
	#define TreedbSearchResult_DEFINED
  #endif

  #ifndef TreedbInsertResult_DEFINED
typedef enum TreedbInsertResult
{ eTreedbInsertSuccess
, eTreedbInsertExists
, eTreedbInsertNoMem
} TreedbInsertResult;
	#define TreedbInsertResult_DEFINED
  #endif

  #ifndef TreedbInitResult_DEFINED
typedef enum TreedbInitResult
{ eTreedbInitSuccess
, eTreedbInitOffsetSizeTooSmall
, eTreedbInitOffsetSizeTooBig
, eTreedbInitOffsetSizeOdd
, eTreedbInitNotEnoughAddressSpace
, eTreedbInitPayloadOffsetTooSmall
, eTreedbInitPayloadOffsetTooBig
, eTreedbInitHeaderSizeTooSmall
} TreedbInitResult;
	#define TreedbInitResult_DEFINED
  #endif

  #ifndef TreedbHeapStatus_DEFINED
typedef enum TreedbHeapStatus
{ eTreedbHeapSuccess
, eTreedbHeapFileOpenFailure
/* These values indicate an invalid or incompatible treedb. */
, eTreedbHeapBadTop
, eTreedbHeapOpenBadMagic
, eTreedbHeapOpenUnknownAllocatorAlgorithm
, eTreedbHeapOpenUnknownSchema
, eTreedbHeapOpenUnhandledOffsetSize
, eTreedbHeapOpenUnhandledEndianIndex
, eTreedbHeapOpenUnhandledFeatures
/* Coding errors. */
, eTreedbHeapCreateInitFailure
, eTreedbHeapAlreadyWriteLocked
, eTreedbHeapAlreadyReadLocked
, eTreedbHeapAlreadyUnlocked
, eTreedbHeapPayloadSizeTooLarge
, eTreedbHeapPointerAllocator
/* Already closed. */
, eTreedbHeapNotOpen
/* Permissions. */
, eTreedbHeapReadonly
/* Values from here on indicate an os or fs problem. */
, eTreedbHeapMutexLockError
, eTreedbHeapCondWaitError
, eTreedbHeapCondBroadcastError
, eTreedbHeapFileLockError
, eTreedbHeapMmapError
, eTreedbHeapUnMmapError
, eTreedbHeapMutexUnlockError
, eTreedbHeapLockedError
, eTreedbHeapFileTruncateFailure
, eTreedbHeapBad
, eTreedbHeapMallocFailure
, eTreedbHeapNoRuntime
} TreedbHeapStatus;
	#define TreedbHeapStatus_DEFINED
  #endif

  #ifndef TreedbError_DEFINED
typedef enum TreedbError
{ eTreedbOk
, eTreedbZeroLengthHeap
, eTreedbNullNode
, eTreedbPrevNodeIsFree
, eTreedbPrevNodeIsThis
, eTreedbPrevNodeIsBad
, eTreedbNodeNotInFreeHeap
, eTreedbNodeNotInUsedHeap
, eTreedbNodeLengthBad
, eTreedbUnexpectedFreeNode
, eTreedbNextNodeIsThis
, eTreedbNextNodeIsLess
, eTreedbLastNodePastEndOfHeap
, eTreedbInvalidUsedHeap
, eTreedbTwoAdjacentFreeNodes
, eTreedbNoNodes
, eTreedbInvalidParam
} TreedbError;
	#define TreedbError_DEFINED
  #endif

  #ifndef TreedbFreeResult_DEFINED
typedef enum TreedbFreeResult
{ eTreedbFreeSuccess
, eTreedbFreeUnknownPointer
, eTreedbFreeBadHeap
, eTreedbFreeBadHeapLock
, eTreedbFreeBadHeapFree
, eTreedbFreeHeapIsFixed
} TreedbFreeResult;
	#define TreedbFreeResult_DEFINED
  #endif

  #ifndef TreedbHeapOpenFlags_DEFINED
typedef enum TreedbHeapOpenFlags
{ eHeapOpenRead = 1
, eHeapOpenWrite = 2
, eHeapOpenSync = 4
, eHeapOpenNoAtime = 8
, eHeapOpenCloseOnExec = 16
} TreedbHeapOpenFlags;
	#define TreedbHeapOpenFlags_DEFINED
  #endif
  #ifndef TreedbHeapCreatePermissions_DEFINED
typedef enum TreedbHeapCreatePermissions
{ eHeapCreateOwnerRead  = 0x400
, eHeapCreateOwnerWrite = 0x200
, eHeapCreateOwnerExec  = 0x100
, eHeapCreateGroupRead  = 0x040
, eHeapCreateGroupWrite = 0x020
, eHeapCreateGroupExec  = 0x010
, eHeapCreateOtherRead  = 0x004
, eHeapCreateOtherWrite = 0x002
, eHeapCreateOtherExec  = 0x001
} TreedbHeapCreatePermissions;
	#define TreedbHeapCreatePermissions_DEFINED
  #endif
  #ifndef TreedbHeapAccessFlags_DEFINED
typedef enum TreedbHeapAccessFlags
{ eHeapAccessRead = 1
, eHeapAccessWrite = 2
, eHeapAccessExec = 4
, eHeapAccessPrivate = 8
, eHeapAccessShared = 16
, eHeapAccessAnonymous = 32
} TreedbHeapAccessFlags;
	#define TreedbHeapAccessFlags
  #endif

typedef union TreedbNodeDistanceFlags
{
	short flags;
	struct {
		unsigned short include_lower:1;
		unsigned short include_upper:1;
	};
} TreedbNodeDistanceFlags;
  #ifdef __cplusplus
static inline TreedbNodeDistanceFlags MakeNodeDistanceFlags(int x)
{ TreedbNodeDistanceFlags flags; flags.flags = (short)x; return flags; }
  #else
	#define MakeNodeDistanceFlags(x) ((TreedbNodeDistanceFlags)(x))
  #endif
typedef enum ETreedbNodeDistanceFlags
{ eNodeDistanceIncludeLower = 1
, eNodeDistanceIncludeUpper = 2
} ETreedbNodeDistanceFlags;
typedef union TreedbKeyDistanceFlags
{
	short flags;
	struct {
		/* The first two need to be the same as in TreedbDistanceFlags */
		unsigned short include_lower:1;
		unsigned short include_upper:1;
		unsigned short exact_lower:1;
		unsigned short exact_upper:1;
	};
	TreedbNodeDistanceFlags node_distance;
} TreedbKeyDistanceFlags;
  #ifdef __cplusplus
static inline TreedbKeyDistanceFlags MakeKeyDistanceFlags(int x)
{ TreedbKeyDistanceFlags flags; flags.flags = (short)x; return flags; }
  #else
	#define MakeNodeDistanceFlags(x) ((TreedbNodeDistanceFlags)(x))
  #endif
typedef enum ETreedbKeyDistanceFlags
{ eKeyDistanceIncludeLower = 1
, eKeyDistanceIncludeUpper = 2
, eKeyDistanceExactLower   = 4
, eKeyDistanceExactUpper   = 8
} ETreedbKeyDistanceFlags;

#ifndef TREEDB_OMIT_TreedbSearchResultDesc
v3c_declare_prefix const char * TreedbSearchResultDesc(TreedbSearchResult res);
#endif

#ifndef TREEDB_OMIT_TreedbInsertResultDesc
v3c_declare_prefix const char * TreedbInsertResultDesc(TreedbInsertResult res);
#endif

#ifndef TREEDB_OMIT_TreedbInitResultDesc
v3c_declare_prefix const char * TreedbInitResultDesc(TreedbInitResult res);
#endif

#ifndef TREEDB_OMIT_TreedbHeapStatusDesc
v3c_declare_prefix const char * TreedbHeapStatusDesc(TreedbHeapStatus res);
#endif

#ifndef TREEDB_OMIT_TreedbTreedbErrorDesc
v3c_declare_prefix const char * TreedbErrorDesc(TreedbError res);
#endif

#ifndef TREEDB_OMIT_TreedbFreeResultDesc
v3c_declare_prefix const char * TreedbFreeResultDesc(TreedbFreeResult res);
#endif

#ifndef TREEDB_OMIT_TreedbKeyDistanceFlagsDesc
v3c_declare_prefix const char * TreedbKeyDistanceFlagsDesc
( TreedbKeyDistanceFlags flags
);
#endif

#ifndef TREEDB_OMIT_TreedbNodeDistanceFlagsDesc
v3c_declare_prefix const char * TreedbNodeDistanceFlagsDesc
( TreedbNodeDistanceFlags flags
);
#endif

#ifndef TREEDB_OMIT_TreedbHeapMismatched
v3c_declare_prefix int TreedbHeapMismatched(TreedbHeapStatus res);
#endif

#ifndef TREEDB_OMIT_v3c_treedb_allocator_abort
v3c_declare_prefix void v3c_treedb_allocator_abort(TreedbError e);
#endif

v3c_declare_prefix uint8_t v3c_native_endian_index();


v3c_declare_prefix int treedb_mode_flags( int flags_in);


  #ifdef __cplusplus
} /* extern "C" { */
  #endif

#endif /* #ifndef v3c_treedb_h_included */
